/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/

package at.jku.mevss.anttracks.callcontext.util;

import java.util.Iterator;
import java.util.Objects;
import java.util.function.Consumer;
import java.util.function.Function;

/**
 * A wrapper around an {@link Iterator} that maps elements using a specified mapping function.
 *
 * @param <T>    The type of element returned by the wrapped iterator.
 * @param <R>    The type of element returned by this iterator (which is the result type of the mapping function).
 * @param <Iter> The type of the wrapped iterator.
 * @author Peter Feichtinger
 */
public class IteratorMapper<T, R, Iter extends Iterator<T>> implements Iterator<R> {

    private final Iter mIt;
    private final Function<? super T, ? extends R> mMapper;

    /**
     * Create a new {@link IteratorMapper} for the specified iterator using the specified mapping function.
     *
     * @param it     The iterator to wrap.
     * @param mapper The mapping function used to map elements returned by {@code it} to the return type.
     */
    public IteratorMapper(Iter it, Function<? super T, ? extends R> mapper) {
        mIt = Objects.requireNonNull(it);
        mMapper = Objects.requireNonNull(mapper);
    }

    /**
     * Get the iterator wrapped by this iterator.
     *
     * @return The iterator wrapped by this iterator.
     */
    public Iter getIterator() {
        return mIt;
    }

    @Override
    public boolean hasNext() {
        return mIt.hasNext();
    }

    @Override
    public R next() {
        return mMapper.apply(mIt.next());
    }

    @Override
    public void remove() {
        mIt.remove();
    }

    @Override
    public void forEachRemaining(Consumer<? super R> action) {
        mIt.forEachRemaining(t -> action.accept(mMapper.apply(t)));
    }
}
