import at.jku.mevss.anttracks.callcontext.CallContextBuilder;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

/**
 * Test showing that type analysis really works. This is an entirely interactive test, you will have to modify source code.
 * <p>
 * First of all, uncomment the log message in {@code SymbolsParser::amendStackTracesStatic} so extended stack traces get logged. Then you
 * can run a parse with type analysis enabled, and one with it commented out (see {@link CallContextBuilder#build()
 * CallContextBuilder::build}, or disable type analysis via {@link CallContextBuilder#enableTypeAnalysis(boolean)} in
 * {@code Symbols::initCallContext}). You will see that with type analysis, the stack traces of both allocation sites in the
 * {@code ListGetter} implementations can be extended by one element, {@code main}. Without type analysis, both calls in {@code main}
 * statically resolve to {@code ListGetter::get}, so there is no unique call and nothing can be extended.
 * <p>
 * Of course instead of logging stack trace extensions you can also look at the trace in the GUI: without type analysis the
 * {@code LinkedList} and {@code ArrayList} objects alive after the second GC will only have their allocation sites for a stack trace, with
 * type analysis the stack trace will also contain {@code main}.
 *
 * @author Peter Feichtinger
 */
public class TypeAnalysisTest {

    static interface ListGetter {
        // This method is called non-uniquely, so stack traces cannot be extended with static analysis alone
        public List<String> get();
    }

    static class LinkedListGetter implements ListGetter {
        @Override
        public List<String> get() {
            final LinkedList<String> l = new LinkedList<>();
            l.add("L");
            return l;
        }
    }

    static class ArrayListGetter implements ListGetter {
        @Override
        public List<String> get() {
            final ArrayList<String> l = new ArrayList<>();
            l.add("A");
            return l;
        }
    }

    public static void main(String[] args) {
        // Load classes
        setup();

        final ListGetter l = new LinkedListGetter();
        final ListGetter a = new ArrayListGetter();

        // Statically, both these calls are to ListGetter::get, with type analysis we will be able to distinguish them,
        // and subsequently be able to extend the stack traces of allocation sites in LinkedListGetter::get and ArrayListGetter::get
        final List<String> linkedList = l.get();
        final List<String> arrayList = a.get();

        System.gc();

        System.out.println(linkedList);
        System.out.println(arrayList);
    }

    private static void setup() {
        new LinkedListGetter();
        new ArrayListGetter();
        new LinkedList<Object>();
        new ArrayList<Object>();
        System.gc();
    }
}
