/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/

package at.jku.ssw.decorator;

import at.jku.ssw.marker.FeatureMarkerFactory;
import org.eclipse.core.resources.IResource;
import org.eclipse.jface.viewers.ILabelDecorator;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.LabelProviderChangedEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IDecoratorManager;
import org.eclipse.ui.PlatformUI;

import java.util.ArrayList;
import java.util.List;

/**
 * This class handles the resource decoration in the project browser window. It
 * adds a short text which denotes how many feature markings are set in any
 * particular resource.
 *
 * @author David Pfeffer
 */
public class FileDecorator extends LabelProvider implements ILabelDecorator {

    public static final String DECORATOR_ID = "at.jku.ssw.decorator.filedecorator"; //$NON-NLS-1$

    /**
     * Gets the custom decorator object. This method should be called to get the
     * custom decorator object by all methods that try to decorate resources.
     *
     * @return Custom Decorator Instance if the custom decorator is enabled null
     * if the custom decorator is not enabled
     */
    public static FileDecorator getFileDecorator() {
        IDecoratorManager decoratorManager = PlatformUI.getWorkbench().getDecoratorManager();

        // If the decorator is disabled, a null value is returned
        return (FileDecorator) decoratorManager.getLabelDecorator(DECORATOR_ID);
    }

    /**
     * Refreshes the decorator of a given {@link IResource} with all its
     * parents.
     *
     * @param resourceToBeUpdated The {@link IResoruce} from which on the decorators should be
     *                            refreshed
     */
    public void refreshWithParents(IResource resourceToBeUpdated) {
        List<IResource> resourceList = new ArrayList<>();
        for (IResource r = resourceToBeUpdated; r != null; r = r.getParent()) {
            resourceList.add(r);
        }
        refresh(resourceList);
    }

    /**
     * Refreshes a list of {@link IResource} objects.
     *
     * @param resourcesToBeUpdated The list of {@link IResource} objects which should be
     *                             refreshed.
     */
    public void refresh(List<IResource> resourcesToBeUpdated) {
        // Check if decoration is enabled
        FileDecorator demoDecorator = getFileDecorator();
        if (demoDecorator == null) {
            return; // do not decorate anything if not enabled
        }

        // Fire a label provider changed event to decorate the resources who
        // needs to be refreshed
        fireLabelEvent(new LabelProviderChangedEvent(demoDecorator, resourcesToBeUpdated.toArray()));
    }

    /**
     * Fires a {@link LabelProviderChangedEvent} using the current UI thread.
     *
     * @param event The {@link LabelProviderChangedEvent} to be fired.
     */
    private void fireLabelEvent(final LabelProviderChangedEvent event) {
        Display.getDefault().asyncExec(new Runnable() {
            @Override
            public void run() {
                fireLabelProviderChanged(event);
            }
        });
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * org.eclipse.jface.viewers.ILabelDecorator#decorateImage(org.eclipse.swt
     * .graphics.Image, java.lang.Object)
     */
    @Override
    public Image decorateImage(Image image, Object resource) {
        return image;
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * org.eclipse.jface.viewers.ILabelDecorator#decorateText(java.lang.String,
     * java.lang.Object)
     */
    @Override
    public String decorateText(String text, Object resource) {
        // Check if it is a valid IResoruce where we can count markers on
        if (resource instanceof IResource) {
            int markers = FeatureMarkerFactory.findMarkers((IResource) resource).size();

            // Only add suffix when there are markers defined in this resource
            if (markers > 0) {
                return text + " [" + markers + " assignment" + ((markers != 1) ? "s" : "") + "]";
            }
        }
        return text; // fallback with no added suffix
    }
}
