/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/

package at.jku.ssw.ftcm;

import at.jku.ssw.ftcm.FeaturePropertyPage.FeatureListEditor;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.fieldassist.ControlDecoration;
import org.eclipse.jface.fieldassist.FieldDecorationRegistry;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.*;

/**
 * This class handles the user interaction when a new feature is added to the
 * project.
 *
 * @author David Pfeffer
 */
public class FeatureAddDialog extends Dialog {

    private RGB featureColor;
    private String featureName;
    private int featureId;

    private Text tid;
    private Text tname;
    private Label lcolor2;

    private ControlDecoration decorId;
    private ControlDecoration decorName;
    private ControlDecoration decorColor;

    private FeatureListEditor featureListEditore;

    /**
     * Creates a new dialog instance.
     *
     * @param parentShell The parent Shell
     * @param psm         The project settings manager
     */
    protected FeatureAddDialog(Shell parentShell, FeatureListEditor featureListEditor) {
        super(parentShell);
        this.featureListEditore = featureListEditor;
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * org.eclipse.jface.dialogs.Dialog#createDialogArea(org.eclipse.swt.widgets
     * .Composite)
     */
    @Override
    protected Control createDialogArea(Composite parent) {
        Composite container = (Composite) super.createDialogArea(parent);

        GridData gridData = new GridData(SWT.FILL, SWT.CENTER, true, false);
        gridData.horizontalSpan = 2;

        GridLayout containerLayout = new GridLayout(3, false);
        containerLayout.horizontalSpacing = 20;
        container.setLayout(containerLayout);

        new Label(container, SWT.NONE).setText("Color:");
        lcolor2 = new Label(container, SWT.BORDER);
        lcolor2.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));

        decorColor = new ControlDecoration(lcolor2, SWT.CENTER);
        decorColor.setDescriptionText("Not a valid color");
        decorColor.setImage(FieldDecorationRegistry.getDefault().getFieldDecoration(FieldDecorationRegistry.DEC_ERROR).getImage());
        decorColor.setMarginWidth(2);

        Button bcolor = new Button(container, SWT.PUSH);
        bcolor.setText("Select ...");
        bcolor.addSelectionListener(new SelectionAdapter() {

            @Override
            public void widgetSelected(SelectionEvent e) {
                super.widgetSelected(e);

                ColorDialog colorDialog = new ColorDialog(getShell());
                colorDialog.setText("Select feature color");

                featureColor = colorDialog.open();
                if (featureColor != null) {
                    lcolor2.setBackground(new Color(getShell().getDisplay(), featureColor));
                }
                validateInput();
            }
        });

        // ID row
        Label lid = new Label(container, SWT.NONE);
        lid.setText("ID:");
        tid = new Text(container, SWT.BORDER);
        tid.setMessage(Integer.toString(featureListEditore.nextFreeFeatureId()));
        tid.setLayoutData(gridData);

        decorId = new ControlDecoration(tid, SWT.CENTER);
        decorId.setDescriptionText("Not a valid id");
        decorId.setImage(FieldDecorationRegistry.getDefault().getFieldDecoration(FieldDecorationRegistry.DEC_ERROR).getImage());
        decorId.setMarginWidth(2);
        decorId.hide();

        tid.addModifyListener(validateInputListener);

        // Name row
        Label lname = new Label(container, SWT.NONE);
        lname.setText("Name:");
        tname = new Text(container, SWT.BORDER);
        tname.setLayoutData(gridData);

        decorName = new ControlDecoration(tname, SWT.CENTER);
        decorName.setDescriptionText("Not a valid name");
        decorName.setImage(FieldDecorationRegistry.getDefault().getFieldDecoration(FieldDecorationRegistry.DEC_ERROR).getImage());
        decorName.setMarginWidth(2);

        tname.addModifyListener(validateInputListener);

        return container;
    }

    /**
     * Listener for forcing the validation of the user input.
     */
    private ModifyListener validateInputListener = new ModifyListener() {
        @Override
        public void modifyText(ModifyEvent e) {
            validateInput();
        }
    };

    /**
     * Validates the user input and handles the status of the buttons.
     */
    private void validateInput() {
        boolean validText = false, validID = false;

        // Check if name is valid
        if (tname.getText().length() > 0) {
            validText = true;
            decorName.hide();
        } else {
            decorName.show();
        }

        // Check if color is valid
        if (featureColor != null) {
            decorColor.hide();
        } else {
            decorColor.show();
        }

        // Check if id is valid
        try {

            int id;
            if (tid.getText().length() == 0 && tid.getMessage() != null) {
                id = Integer.parseInt(tid.getMessage());
            } else {
                id = Integer.parseInt(tid.getText());
            }

            if (featureListEditore.isFeatureIDAvailable(id)) {
                validID = true;
                decorId.hide();
            } else {
                decorId.show();
            }
        } catch (NumberFormatException exc) {
            decorId.show();
        }

        getButton(IDialogConstants.OK_ID).setEnabled(validText && validID && featureColor != null);
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * org.eclipse.jface.dialogs.Dialog#createButtonsForButtonBar(org.eclipse
     * .swt.widgets.Composite)
     */
    @Override
    protected void createButtonsForButtonBar(Composite parent) {
        super.createButtonsForButtonBar(parent);
        getButton(IDialogConstants.OK_ID).setEnabled(false);
    }

    /*
     * (non-Javadoc)
     *
     * @see org.eclipse.jface.dialogs.Dialog#okPressed()
     */
    @Override
    protected void okPressed() {
        if (tid.getText().length() == 0 && tid.getMessage() != null) {
            featureId = Integer.parseInt(tid.getMessage());
        } else {
            featureId = Integer.parseInt(tid.getText());
        }
        featureName = tname.getText();
        super.okPressed();
    }

    /**
     * Returns the feature name.
     *
     * @return The feature name.
     */
    public String getFeatureName() {
        return featureName;
    }

    /**
     * Returns the feature id.
     *
     * @return The feature id.
     */
    public int getFeatureId() {
        return featureId;
    }

    /**
     * Returns a RGB representation of the feature color.
     *
     * @return A RGB representation of the feature color.
     */
    public RGB getFeatureColor() {
        return featureColor;
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * org.eclipse.jface.window.Window#configureShell(org.eclipse.swt.widgets
     * .Shell)
     */
    @Override
    protected void configureShell(Shell newShell) {
        super.configureShell(newShell);
        newShell.setText("Add new feature");
    }

    /*
     * (non-Javadoc)
     *
     * @see org.eclipse.jface.dialogs.Dialog#getInitialSize()
     */
    @Override
    protected Point getInitialSize() {
        return new Point(300, 250);
    }
}
