/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/

package at.jku.ssw.ftcm;

import at.jku.ssw.ftcm.preferences.Feature;
import at.jku.ssw.ftcm.preferences.ProjectSettingsManager;
import at.jku.ssw.marker.FeatureMarkerFactory;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.*;
import org.eclipse.ui.IWorkbenchPropertyPage;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.dialogs.PropertyPage;
import org.osgi.service.prefs.BackingStoreException;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

public class FeaturePropertyPage extends PropertyPage implements IWorkbenchPropertyPage {

    private ProjectSettingsManager psm;

    private Button buttonEnableFeatureProcessing;
    private Button buttonEnableAutoPack;

    private FeatureListEditor featureListEditor;

    class FeatureListEditor extends Composite {

        private Table table;

        public FeatureListEditor(Composite parent, int style) {
            super(parent, style);

            Group group = new Group(parent, SWT.NONE);
            group.setLayout(new GridLayout(2, false));
            group.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
            group.setText("Feature list:");

            table = new Table(group, SWT.MULTI | SWT.FULL_SELECTION);
            GridData tableGridData = new GridData(SWT.FILL, SWT.FILL, true, true);
            table.setLayoutData(tableGridData);
            table.setHeaderVisible(true);
            table.setLinesVisible(true);

            String[] columnNames = {"Color", "ID", "Name"};
            for (String columnName : columnNames) {
                TableColumn column = new TableColumn(table, SWT.NONE);
                column.setText(columnName);
                column.setMoveable(false);
                column.pack();
            }
            table.getColumn(0).setAlignment(SWT.CENTER);
            table.getColumn(1).setAlignment(SWT.CENTER);

            Composite buttonComposite = new Composite(group, SWT.NONE);
            buttonComposite.setLayout(new GridLayout(1, false));
            buttonComposite.setLayoutData(new GridData(SWT.FILL, SWT.FILL, false, false));

            Button b1 = new Button(buttonComposite, SWT.PUSH);
            b1.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
            b1.addListener(SWT.Selection, new Listener() {

                @Override
                public void handleEvent(Event event) {
                    if (event.type != SWT.Selection) {
                        return;
                    }

                    FeatureAddDialog dialog = new FeatureAddDialog(getShell(), featureListEditor);

                    if (dialog.open() == Dialog.OK) {
                        addFeatureItem(dialog.getFeatureId(), dialog.getFeatureName(), new Color(getDisplay(), dialog.getFeatureColor()));
                    }

                }

            });
            b1.setText("Add");

            Button b2 = new Button(buttonComposite, SWT.PUSH);
            b2.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
            b2.addListener(SWT.Selection, new Listener() {

                @Override
                public void handleEvent(Event event) {
                    if (event.type != SWT.Selection) {
                        return;
                    }

                    List<Integer> featureidsToDelete = new ArrayList<>();
                    for (TableItem ti : table.getSelection()) {
                        featureidsToDelete.add((Integer) ti.getData("id"));
                    }

                    List<IMarker> markers = FeatureMarkerFactory.findMarkers(psm.getProject());
                    List<IMarker> markersToDelete = new ArrayList<>();
                    try {
                        for (IMarker m : markers) {
                            if (featureidsToDelete.contains((Integer) m.getAttribute("feature"))) {
                                markersToDelete.add(m);
                            }
                        }
                    } catch (CoreException e) {
                        e.printStackTrace();
                    }

                    Shell shell = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
                    if (markersToDelete.isEmpty()) {
                        table.remove(table.getSelectionIndices());
                    } else if (new MessageDialog(shell,
                                                 "This action will remove markers!",
                                                 null,
                                                 "There are currently " + markersToDelete.size() + " marker(s) assigned to the selected feature(s) which will also be removed!",
                                                 MessageDialog.ERROR,
                                                 new String[]{"Ok", "Cancel"},
                                                 0).open() == Dialog.OK) {
                        for (IMarker m : markersToDelete) {
                            FeatureMarkerFactory.deleteMarker(m);
                        }
                        table.remove(table.getSelectionIndices());
                    }
                }
            });
            b2.setText("Remove");
        }

        public void addFeatureItem(int id, String description, Color color) {
            TableItem tableItem = new TableItem(table, SWT.NONE);
            tableItem.setText(new String[]{"", String.valueOf(id), description});
            tableItem.setBackground(0, color);
            tableItem.setData("id", id);
        }

        public List<Feature> getFeatureList() {
            List<Feature> featureList = new ArrayList<>();

            for (TableItem item : table.getItems()) {
                RGB rgb = item.getBackground(0).getRGB();
                featureList.add(new Feature(Integer.parseInt(item.getText(1)), item.getText(2), rgb.red, rgb.green, rgb.blue));
            }
            return featureList;
        }

        public void setFeatureList(List<Feature> featureList) {
            for (Feature feature : featureList) {
                addFeatureItem(feature.getId(), feature.getDescription(), new Color(getDisplay(), new RGB(feature.getR(), feature.getG(), feature.getB())));
            }
        }

        public boolean isFeatureIDAvailable(int id) {
            List<Feature> featureList = getFeatureList();
            for (Feature f : featureList) {
                if (f.getId() == id) {
                    return false;
                }
            }
            return true;
        }

        public int nextFreeFeatureId() {

            List<Feature> featureList = getFeatureList();
            Collections.sort(featureList, new Comparator<Feature>() {
                @Override
                public int compare(Feature o1, Feature o2) {
                    return o1.getId() - o2.getId();
                }
            });

            return featureList.isEmpty() ? 1 : featureList.get(featureList.size() - 1).getId() + 1;
        }

        public void removeAll() {
            table.removeAll();
        }
    }

    public FeaturePropertyPage() {
        super();
    }

    @Override
    protected Control createContents(Composite parent) {

        IAdaptable adaptable = getElement();

        if (adaptable instanceof IJavaProject) {

            psm = ProjectSettingsManager.getInstance(((IJavaProject) adaptable).getProject());

            parent.setLayout(new GridLayout(1, false));

            Group group = new Group(parent, SWT.SHADOW_ETCHED_IN);
            group.setText("Feature processing:");
            group.setLayout(new GridLayout(1, true));
            group.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));

            buttonEnableFeatureProcessing = new Button(group, SWT.CHECK);
            buttonEnableFeatureProcessing.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
            buttonEnableFeatureProcessing.setText("Enable");
            buttonEnableFeatureProcessing.setSelection(psm.isBuilderRegistered());
            buttonEnableFeatureProcessing.pack();

            buttonEnableAutoPack = new Button(group, SWT.CHECK);
            buttonEnableAutoPack.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
            buttonEnableAutoPack.setText("Full build (for complete mapping file)");
            buttonEnableAutoPack.setSelection(psm.getProjectNode().getBoolean(ProjectSettingsManager.FEATURE_AUTO_CREATE, false));
            buttonEnableAutoPack.pack();

            group.pack();

            featureListEditor = new FeatureListEditor(parent, SWT.NONE);
            featureListEditor.setFeatureList(psm.getFeatureList());

            return parent;
        } else {
            return null;
        }
    }

    @Override
    protected void performDefaults() {
        super.performDefaults();
        featureListEditor.removeAll();
        buttonEnableFeatureProcessing.setSelection(false);
        buttonEnableAutoPack.setSelection(true);
    }

    @Override
    public boolean performOk() {
        if (buttonEnableFeatureProcessing.getSelection()) {
            psm.addFeatureBuilder();
        } else {
            psm.removeFeatureBuilder();
        }

        psm.setFeatureList(featureListEditor.getFeatureList());
        psm.getProjectNode().putBoolean(ProjectSettingsManager.FEATURE_AUTO_CREATE, buttonEnableAutoPack.getSelection());

        try {
            psm.getProjectNode().flush();
        } catch (BackingStoreException e) {
            MessageDialog.openError(getShell(),
                                    "Error while saving feature settings",
                                    "There was an error while saving the feature settings file:\n\t- The feature settings file is out of sync" + ".\n\nWhy did this " +
                                        "happen?\n\t- Most likely " + "the feature settings file got modified outside of Eclpse.\n\nWhat can I do now?\n\t - Please refresh" + " " +
                                        "your project and " + "try again.");
        }

        return super.performOk();
    }
}
