/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/

package at.jku.ssw.ftcm.builder;

import at.jku.ssw.ftcm.export.ClassFeature;
import at.jku.ssw.ftcm.export.MethodFeature;
import at.jku.ssw.ftcm.export.StatementFeature;
import at.jku.ssw.ftcm.preferences.Feature;
import at.jku.ssw.ftcm.preferences.ProjectSettingsManager;
import org.apache.bcel.classfile.ClassFormatException;
import org.apache.bcel.classfile.ClassParser;
import org.apache.bcel.classfile.JavaClass;
import org.eclipse.core.resources.*;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.core.IRegion;
import org.eclipse.jdt.core.JavaCore;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Map;

//import org.eclipse.core.resources.IMarker;

//import at.jku.ssw.marker.FeatureMarkerFactory;

/**
 * This class handles the generation of the feature mapping files.
 *
 * @author David Pfeffer
 */
public class FeatureBuilder extends IncrementalProjectBuilder {

    // The builder id
    public static final String BUILDER_ID = "at.jku.ssw.ftcm.builder.FeatureBuilder"; //$NON-NLS-1$

    // The mapping file name
    private final static String FEATURE_FILE_NAME = "myproject.features"; //$NON-NLS-1$

    // The storage for already processed resources in an incremental build
    private Map<IResource, ClassFeature> classFeatureMap;

    /**
     * Creates a new {@link FeatureBuilder}.
     */
    public FeatureBuilder() {
        classFeatureMap = new HashMap<>();
    }

    @Override
    protected void startupOnInitialize() {
    }

    @Override
    protected void clean(IProgressMonitor monitor) {
    }

    @Override
    protected IProject[] build(int kind, Map<String, String> args, IProgressMonitor monitor) throws CoreException {

        // Full build when forced or when the map is empty
        if (kind == IncrementalProjectBuilder.FULL_BUILD || classFeatureMap.size() == 0) {
            fullBuild(monitor);
        } else {
            // Fallback to a full build if no delta is available
            IResourceDelta delta = getDelta(getProject());
            if (delta == null) {
                fullBuild(monitor);
            } else {
                incrementalBuild(delta, monitor);
            }
        }

        // Write the feature file
        writeFeatureFile();

        return null;
    }

    /**
     * Handles an incremental project build.
     *
     * @param delta   The delta containing the changes from the last build
     *                invocation.
     * @param monitor The monitor instanced used for showing reporting information.
     */
    private void incrementalBuild(IResourceDelta delta, final IProgressMonitor monitor) {
        try {
            delta.accept(new IResourceDeltaVisitor() {

                @Override
                public boolean visit(IResourceDelta delta) {
                    if (delta.getKind() == IResourceDelta.ADDED || delta.getKind() == IResourceDelta.CHANGED) {
                        parseResource(delta.getResource(), monitor);
                    } else if (delta.getKind() == IResourceDelta.REMOVED) {
                        removeFeatures(delta.getResource(), monitor);
                    }
                    return true; // visit children
                }
            });
        } catch (CoreException e) {
            e.printStackTrace();
        }
    }

    /**
     * Handles a full project build with no known changes from the last build
     * invocation.
     *
     * @param monitor The monitor instanced used for showing reporting information.
     */
    private void fullBuild(final IProgressMonitor monitor) {
        try {
            getProject().accept(new IResourceVisitor() {

                @Override
                public boolean visit(IResource resource) throws CoreException {
                    parseResource(resource, monitor);
                    return true; // visit children
                }
            });
        } catch (CoreException e) {
            e.printStackTrace();
        }
    }

    /**
     * Handles the removing of resources which were deleted since the last build
     * invocation.
     *
     * @param resource The deleted resource.
     * @param monitor  The monitor instanced used for showing reporting information.
     */
    private void removeFeatures(IResource resource, IProgressMonitor monitor) {
        if (resource instanceof IFile && "class".equals(resource.getFileExtension())) {
            classFeatureMap.remove(resource);
        }
    }

    /**
     * Returns a list of matching class files for a given java file.
     *
     * @param javaFile The java file for which the class files should be found.
     * @return A list of matching class files for the given java file.
     */
    private IResource[] findClassFiles(IResource javaFile) {
        IRegion r = JavaCore.newRegion();
        r.add(JavaCore.create(javaFile));
        return JavaCore.getGeneratedResources(r, false);
    }

    /**
     * Handles the parsing of a single resource.
     *
     * @param resource The resource which should be parsed.
     * @param monitor  The monitor instanced used for showing reporting information.
     */
    private void parseResource(IResource resource, IProgressMonitor monitor) {
        // We are only interested in java files
        if (resource instanceof IFile && "java".equals(resource.getFileExtension())) {

            // System.err.println("PARSING: " +
            // resource.getProjectRelativePath());

            try {
                IFile javaFileResource = (IFile) resource.getAdapter(IFile.class);

                // Print and handle matching class files
                // System.err.println("\tmachting class files: ");
                for (IResource classFileResource : findClassFiles(javaFileResource)) {
                    handleClassFile(javaFileResource, classFileResource);
                }

                // Parse feature markers
                /*
                 * System.err.println("\tfound markers: "); for (IMarker m :
				 * FeatureMarkerFactory.findMarkers(javaFileResource)) {
				 * System.err.println("\t\t- at line " +
				 * m.getAttribute(IMarker.LINE_NUMBER) + " spanning over " +
				 * m.getAttribute("lines") + " line(s) with feature id " +
				 * m.getAttribute("feature")); }
				 */
            } catch (ClassFormatException e) {
                e.printStackTrace();
            } /*
             * catch (CoreException e) { e.printStackTrace(); }
			 */
        }
    }

    /**
     * Handles the feature mapping for a class file.
     *
     * @param javaFile  The java file used to take the marker information.
     * @param classFile The class file used to find the byte offsets.
     */
    private void handleClassFile(IResource javaFile, IResource classFile) {
        try {
            // System.err.println("\t\t- " +
            // classFile.getRawLocationURI().getPath());

            JavaClass javaClass = new ClassParser(classFile.getRawLocationURI().getPath().toString()).parse();
            ClassFeature classFeature = new ClassFeature(javaClass.getClassName());

            FeatureVisitor featureVisitor = new FeatureVisitor(javaClass, javaFile);
            featureVisitor.visit(classFeature);

            classFeatureMap.put(classFile, classFeature);
        } catch (ClassFormatException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Writes the feature file.
     */
    private void writeFeatureFile() {
        classFeatureMap.values();
        try (PrintStream out = new PrintStream(new FileOutputStream(getFeatureFilePath()))) {
            out.print("### DO NOT EDIT ###\n");
            ProjectSettingsManager manager = ProjectSettingsManager.getInstance(getProject());
            for (Feature feature : manager.getFeatureList()) {
                out.printf("%s (%d %d %d) {\n", feature.getDescription(), feature.getR(), feature.getG(), feature.getB());
                for (ClassFeature clazz : classFeatureMap.values()) {
                    out.printf("\t%s\n", clazz.getName());
                    for (MethodFeature method : clazz.getMethodFeatures()) {
                        out.printf("\t%s.%s(%s)\n", clazz.getName(), method.getName(), method.getSignature());
                        for (StatementFeature statement : method.getStatementFeatures()) {
                            out.printf("\t%s.%s(%s) [%d %d]\n", clazz.getName(), method.getName(), method.getSignature(), statement.getB_start(), statement.getB_end());
                        }
                    }
                }
                out.print("}\n");
            }
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }
        try {
            this.getProject().refreshLocal(IResource.DEPTH_ONE, null);
        } catch (CoreException e) {
            e.printStackTrace();
        }
    }

    /**
     * Returns the path of the feature file.
     *
     * @return The path of the feature file.
     */
    private String getFeatureFilePath() {
        IFile contributorsFile = this.getProject().getFile(FEATURE_FILE_NAME);
        IPath path = contributorsFile.getRawLocation();
        return path.toOSString();
    }
}
