/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/

package at.jku.ssw.ftcm.builder;

import at.jku.ssw.ftcm.export.ClassFeature;
import at.jku.ssw.ftcm.export.GenericFeature;
import at.jku.ssw.ftcm.export.MethodFeature;
import at.jku.ssw.ftcm.export.StatementFeature;
import at.jku.ssw.marker.FeatureMarkerFactory;
import org.apache.bcel.classfile.*;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;

import java.util.Stack;

public class FeatureVisitor extends EmptyVisitor implements Visitor {

    private IMarker[] statementMarker;
    private IMarker[] classMarker;
    private IMarker[] methodMarker;

    private JavaClass jc;
    private Stack<Node> nodeStack = new Stack<>();
    private Stack<GenericFeature> featureStack = new Stack<>();

    public FeatureVisitor(JavaClass jc, IResource resource) {
        this.jc = jc;
        try {
            statementMarker = resource.findMarkers(FeatureMarkerFactory.FEATURE_STATEMENT_MARKER, false, 0);
            methodMarker = resource.findMarkers(FeatureMarkerFactory.FEATURE_METHOD_MARKER, false, 0);
            classMarker = resource.findMarkers(FeatureMarkerFactory.FEATURE_CLASS_MARKER, false, 0);
        } catch (CoreException e) {
            e.printStackTrace();
        }
    }

    public Object predecessor() {
        return predecessor(0);
    }

    public Object predecessor(int level) {
        int size = nodeStack.size();
        if ((size < 2) || (level < 0)) {
            return null;
        } else {
            return nodeStack.elementAt(size - (level + 2)); // size - 1 ==
            // current
        }
    }

    public Object current() {
        return nodeStack.peek();
    }

    public void visit(ClassFeature cf) {
        featureStack.push(cf);
        jc.accept(this);
        featureStack.pop();
    }

    @Override
    public void visitJavaClass(JavaClass javaClass) {
        // System.err.print("Parsing class: " + javaClass.getClassName());
        nodeStack.push(javaClass);

        ClassFeature cf = (ClassFeature) featureStack.peek();

        // look for matching markings
        for (IMarker m : classMarker) {
            try {
                if (javaClass.getClassName().equals(m.getAttribute("name"))) {
                    cf.addId((Integer) m.getAttribute("feature"));
                    m.setAttribute("built", true);
                }
            } catch (CoreException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        }
        // System.err.println();

        // parse further methods
        Method[] methods = jc.getMethods();
        for (int i = 0; i < methods.length; i++) {
            methods[i].accept(this);
        }

        nodeStack.pop();
    }

    @Override
    public void visitMethod(Method method) {
        // System.err.print("Parsing method: " + method.getName() + " [" +
        // method.getSignature() + "]");
        nodeStack.push(method);

        ClassFeature cf = (ClassFeature) featureStack.peek();
        MethodFeature mf = new MethodFeature(method.getName(), method.getSignature());
        featureStack.push(mf);

        // look for matching markings
        for (IMarker m : methodMarker) {
            try {
                if (method.getSignature().equals(m.getAttribute("signature")) && method.getName().equals(m.getAttribute("name"))) {
                    mf.addId((Integer) m.getAttribute("feature"));
                    m.setAttribute("built", true);
                }
            } catch (CoreException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        }
        cf.addMethodFeature(mf);

        // System.err.println();

        // parse line number table
        LineNumberTable lnt = method.getLineNumberTable();
        lnt.accept(this);
        featureStack.pop();
        nodeStack.pop();
    }

    @Override
    public void visitLineNumberTable(LineNumberTable table) {
        // System.err.println("Parsing Line Number Table: ");
        nodeStack.push(table);

        LineNumber[] numbers = table.getLineNumberTable();
        for (int i = 0; i < numbers.length; i++) {
            numbers[i].accept(this);
        }

        nodeStack.pop();
    }

    @Override
    public void visitLineNumber(LineNumber obj) {
        // System.err.print("pc=" + obj.getStartPC() + " matches line=" +
        // obj.getLineNumber());
        nodeStack.push(obj);

        MethodFeature mf = (MethodFeature) featureStack.peek();
        StatementFeature sf = new StatementFeature(obj.getStartPC(), obj.getStartPC());

        for (IMarker m : statementMarker) {
            try {
                int lineNumber = obj.getLineNumber();
                int lineNumberStart = (int) m.getAttribute(IMarker.LINE_NUMBER);
                int lineNumberEnd = lineNumberStart + (int) m.getAttribute("lines") - 1;

                if (lineNumber >= lineNumberStart && lineNumber <= lineNumberEnd) {
                    sf.addId((int) m.getAttribute("feature"));
                    m.setAttribute("built", true);
                }
            } catch (CoreException e) {
                e.printStackTrace();
            }
        }
        // System.err.println();

        mf.addStatementFeature(sf);

        nodeStack.pop();
    }
}
