/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/

package at.jku.ssw.ftcm.proposal;

import at.jku.ssw.ftcm.Activator;
import at.jku.ssw.ftcm.preferences.Feature;
import at.jku.ssw.ftcm.preferences.ProjectSettingsManager;
import at.jku.ssw.marker.FeatureMarkerFactory;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jface.text.*;
import org.eclipse.jface.text.contentassist.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.part.FileEditorInput;

import java.util.ArrayList;
import java.util.List;

/**
 * This class handles the procession of feature proposals.
 *
 * @author David Pfeffer
 */
public class FeatureProposalProcessor implements IContentAssistProcessor {

    // Template if there are no proposals available
    private static final ICompletionProposal[] NO_PROPOSALS = new ICompletionProposal[0];

    /**
     * This class is a temporary representation of a feature proposal.
     *
     * @author David Pfeffer
     */
    private static final class Proposal
        implements ICompletionProposal, ICompletionProposalExtension, ICompletionProposalExtension2, ICompletionProposalExtension3, ICompletionProposalExtension4 {

        private final Feature feature;
        private final String prefix;
        private final int offset;

        /**
         * Creates a new feature proposal
         *
         * @param feature The proposed feature
         * @param prefix  The prefix
         * @param offset  The offset
         */
        public Proposal(Feature feature, String prefix, int offset) {
            this.feature = feature;
            this.prefix = prefix;
            this.offset = offset;
        }

        /*
         * (non-Javadoc)
         *
         * @see
         * org.eclipse.jface.text.contentassist.ICompletionProposal#apply(org
         * .eclipse.jface.text.IDocument)
         */
        @Override
        public void apply(IDocument document) {
            apply(null, '\0', 0, this.offset);
        }

        /*
         * (non-Javadoc)
         *
         * @see
         * org.eclipse.jface.text.contentassist.ICompletionProposal#getSelection
         * (org.eclipse.jface.text.IDocument)
         */
        @Override
        public Point getSelection(IDocument document) {
            return new Point(this.offset + feature.getDescription().length(), 0);
        }

        /*
         * (non-Javadoc)
         *
         * @see org.eclipse.jface.text.contentassist.ICompletionProposal#
         * getAdditionalProposalInfo()
         */
        @Override
        public String getAdditionalProposalInfo() {
            return null;
        }

        /*
         * (non-Javadoc)
         *
         * @see
         * org.eclipse.jface.text.contentassist.ICompletionProposal#getDisplayString
         * ()
         */
        @Override
        public String getDisplayString() {
            return feature.getDescription();
        }

        /*
         * (non-Javadoc)
         *
         * @see
         * org.eclipse.jface.text.contentassist.ICompletionProposal#getImage()
         */
        @Override
        public Image getImage() {
            Image image = new Image(Activator.getShell().getDisplay(), new Rectangle(0, 0, 8, 8));
            GC gc = new GC(image);
            gc.setBackground(new Color(Display.getCurrent(), feature.getR(), feature.getG(), feature.getB()));
            gc.fillRoundRectangle(0, 0, 8, 8, 4, 4);
            gc.dispose();
            return image;
        }

        /*
         * (non-Javadoc)
         *
         * @see org.eclipse.jface.text.contentassist.ICompletionProposal#
         * getContextInformation()
         */
        @Override
        public IContextInformation getContextInformation() {
            return null;
        }

        /*
         * (non-Javadoc)
         *
         * @see
         * org.eclipse.jface.text.contentassist.ICompletionProposalExtension
         * #apply(org.eclipse.jface.text.IDocument, char, int)
         */
        @Override
        public void apply(IDocument document, char trigger, int offset) {

            // This apply method assumes that the insertion at this offset is
            // valid, if not we do not insert anything
            if (!isValidFor(document, offset)) {
                return;
            }

            try {
                // Get the resource where the marker should be placed
                IResource res = (IResource) ((FileEditorInput) (PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().getActiveEditor()).getEditorInput()).getFile()
                                                                                                                                                                         .getAdapter(
                                                                                                                                                                             IResource.class);

                // Determine the current line by the given offset
                int startLine = document.getLineOfOffset(offset);

                // Set the marker using the factory
                FeatureMarkerFactory.createStatementMarker(res, document, startLine, startLine, feature);

                // Remove the user input
                document.replace(offset - prefix.length(), prefix.length(), "");
            } catch (BadLocationException | CoreException x) {
                // TODO Auto-generated catch block
                x.printStackTrace();
            }
        }

        /*
         * (non-Javadoc)
         *
         * @see
         * org.eclipse.jface.text.contentassist.ICompletionProposalExtension
         * #isValidFor(org.eclipse.jface.text.IDocument, int)
         */
        @Override
        public boolean isValidFor(IDocument document, int offset) {
            return validate(document, offset, null);
        }

        /*
         * (non-Javadoc)
         *
         * @see
         * org.eclipse.jface.text.contentassist.ICompletionProposalExtension
         * #getTriggerCharacters()
         */
        @Override
        public char[] getTriggerCharacters() {
            return null;
        }

        /*
         * (non-Javadoc)
         *
         * @see
         * org.eclipse.jface.text.contentassist.ICompletionProposalExtension
         * #getContextInformationPosition()
         */
        @Override
        public int getContextInformationPosition() {
            return 0;
        }

        /*
         * (non-Javadoc)
         *
         * @see
         * org.eclipse.jface.text.contentassist.ICompletionProposalExtension2
         * #apply(org.eclipse.jface.text.ITextViewer, char, int, int)
         */
        @Override
        public void apply(ITextViewer viewer, char trigger, int stateMask, int offset) {
            apply(viewer.getDocument(), trigger, offset);
        }

        /*
         * (non-Javadoc)
         *
         * @see
         * org.eclipse.jface.text.contentassist.ICompletionProposalExtension2
         * #selected(org.eclipse.jface.text.ITextViewer, boolean)
         */
        @Override
        public void selected(ITextViewer viewer, boolean smartToggle) {
        }

        /*
         * (non-Javadoc)
         *
         * @see
         * org.eclipse.jface.text.contentassist.ICompletionProposalExtension2
         * #unselected(org.eclipse.jface.text.ITextViewer)
         */
        @Override
        public void unselected(ITextViewer viewer) {
        }

        /*
         * (non-Javadoc)
         *
         * @see
         * org.eclipse.jface.text.contentassist.ICompletionProposalExtension2
         * #validate(org.eclipse.jface.text.IDocument, int,
         * org.eclipse.jface.text.DocumentEvent)
         */
        @Override
        public boolean validate(IDocument document, int offset, DocumentEvent event) {
            try {
                int prefixStart = this.offset - prefix.length();
                return offset >= this.offset && offset < this.offset + feature.getDescription().length() && document.get(prefixStart, offset - (prefixStart))
                                                                                                                    .equals((prefix + feature.getDescription()).substring(0,
                                                                                                                                                                          offset
                                                                                                                                                                              -
                                                                                                                                                                              prefixStart));
            } catch (BadLocationException x) {
                return false;
            }
        }

        /*
         * (non-Javadoc)
         *
         * @see
         * org.eclipse.jface.text.contentassist.ICompletionProposalExtension3
         * #getInformationControlCreator()
         */
        @Override
        public IInformationControlCreator getInformationControlCreator() {
            return null;
        }

        /*
         * (non-Javadoc)
         *
         * @see
         * org.eclipse.jface.text.contentassist.ICompletionProposalExtension3
         * #getPrefixCompletionText(org.eclipse.jface.text.IDocument, int)
         */
        @Override
        public CharSequence getPrefixCompletionText(IDocument document, int completionOffset) {
            return prefix + feature.getDescription();
        }

        /*
         * (non-Javadoc)
         *
         * @see
         * org.eclipse.jface.text.contentassist.ICompletionProposalExtension3
         * #getPrefixCompletionStart(org.eclipse.jface.text.IDocument, int)
         */
        @Override
        public int getPrefixCompletionStart(IDocument document, int completionOffset) {
            return this.offset - prefix.length();
        }

        /*
         * (non-Javadoc)
         *
         * @see
         * org.eclipse.jface.text.contentassist.ICompletionProposalExtension4
         * #isAutoInsertable()
         */
        @Override
        public boolean isAutoInsertable() {
            return true;
        }

    }

    /*
     * (non-Javadoc)
     *
     * @see org.eclipse.jface.text.contentassist.IContentAssistProcessor#
     * computeCompletionProposals(org.eclipse.jface.text.ITextViewer, int)
     */
    @Override
    public ICompletionProposal[] computeCompletionProposals(ITextViewer viewer, int offset) {
        try {
            // If there is no prefix we do not provide any proposals
            String prefix = getPrefix(viewer, offset);
            if (prefix == null || prefix.length() == 0) {
                return NO_PROPOSALS;
            }

            // Otherwise we provide a list of all features which math the
            // current prefix
            List<ICompletionProposal> result = new ArrayList<>();
            IProject project = ((FileEditorInput) (PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().getActiveEditor()).getEditorInput()).getFile().getProject();
            for (Feature feature : ProjectSettingsManager.getInstance(project).getFeatureList()) {
                if (feature.getDescription().toLowerCase().startsWith(prefix.toLowerCase())) {
                    result.add(createProposal(feature, prefix, offset));
                }
            }

            return (ICompletionProposal[]) result.toArray(new ICompletionProposal[result.size()]);
        } catch (BadLocationException x) {
            return NO_PROPOSALS;
        }
    }

    /**
     * This is a helper method to create a new proposal
     *
     * @param feature The proposed feature
     * @param prefix  The prefix
     * @param offset  The offset
     * @return The newly created proposal
     */
    private ICompletionProposal createProposal(Feature feature, String prefix, int offset) {
        return new Proposal(feature, prefix, offset);
    }

    /**
     * Gets the current prefix
     *
     * @param viewer The viewer
     * @param offset The offset
     * @return The current prefix
     * @throws BadLocationException
     */
    private String getPrefix(ITextViewer viewer, int offset) throws BadLocationException {
        IDocument doc = viewer.getDocument();
        if (doc == null || offset > doc.getLength()) {
            return null;
        }

        int length = 0;
        while (--offset >= 0 && Character.isJavaIdentifierPart(doc.getChar(offset))) {
            length++;
        }
        return doc.get(offset + 1, length);
    }

    /*
     * (non-Javadoc)
     *
     * @see org.eclipse.jface.text.contentassist.IContentAssistProcessor#
     * computeContextInformation(org.eclipse.jface.text.ITextViewer, int)
     */
    @Override
    public IContextInformation[] computeContextInformation(ITextViewer viewer, int offset) {
        return null;
    }

    /*
     * (non-Javadoc)
     *
     * @see org.eclipse.jface.text.contentassist.IContentAssistProcessor#
     * getCompletionProposalAutoActivationCharacters()
     */
    @Override
    public char[] getCompletionProposalAutoActivationCharacters() {
        return null;
    }

    /*
     * (non-Javadoc)
     *
     * @see org.eclipse.jface.text.contentassist.IContentAssistProcessor#
     * getContextInformationAutoActivationCharacters()
     */
    @Override
    public char[] getContextInformationAutoActivationCharacters() {
        return null;
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * org.eclipse.jface.text.contentassist.IContentAssistProcessor#getErrorMessage
     * ()
     */
    @Override
    public String getErrorMessage() {
        return null;
    }

    /*
     * (non-Javadoc)
     *
     * @see org.eclipse.jface.text.contentassist.IContentAssistProcessor#
     * getContextInformationValidator()
     */
    @Override
    public IContextInformationValidator getContextInformationValidator() {
        return null;
    }

}
