/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/

package at.jku.ssw.marker;

import at.jku.ssw.decorator.FileDecorator;
import at.jku.ssw.ftcm.Activator;
import at.jku.ssw.ftcm.SignatureResolver;
import at.jku.ssw.ftcm.preferences.Feature;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IType;
import org.eclipse.jface.text.*;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.TreeSelection;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class FeatureMarkerFactory {

    // Marker ID
    public static final String FEATURE_MARKER = "at.jku.ssw.marker.featureMarker"; //$NON-NLS-1$
    public static final String FEATURE_CLASS_MARKER = "at.jku.ssw.marker.featureClassMarker"; //$NON-NLS-1$
    public static final String FEATURE_METHOD_MARKER = "at.jku.ssw.marker.featureMethodMarker"; //$NON-NLS-1$
    public static final String FEATURE_STATEMENT_MARKER = "at.jku.ssw.marker.featureStatementMarker"; //$NON-NLS-1$

    // Annotation ID
    public static final String ANNOTATION = "at.jku.ssw.ftcm.annotation.FeatureAnnotation"; //$NON-NLS-1$

    /**
     * Creates a new class marker at a given {@link IResource}.
     *
     * @param res      The {@link IResource} where the marker should be added.
     * @param position The position where the marker should be added.
     * @param type     The type class where the marker should be created.
     * @param feature  The feature assigned to that marker.
     * @return The newly created marker.
     * @throws CoreException
     */
    public static IMarker createClassMarker(IResource res, Position position, IType type, Feature feature) throws CoreException {
        IMarker m = createMarker("Class", res, position, FeatureMarkerFactory.FEATURE_CLASS_MARKER, feature);
        m.setAttribute("name", type.getFullyQualifiedName());
        return m;
    }

    /**
     * Creates a new method marker at a given {@link IResource}.
     *
     * @param res      The {@link IResource} where the marker should be added.
     * @param position The position where the marker should be added.
     * @param method   The method on which the marker should be created.
     * @param feature  The feature assigned to that marker.
     * @return The newly created marker.
     * @throws CoreException
     */
    public static IMarker createMethodMarker(IResource res, Position position, IMethod method, Feature feature) throws CoreException {
        IMarker m = createMarker("Method", res, position, FeatureMarkerFactory.FEATURE_METHOD_MARKER, feature);
        m.setAttribute("signature", SignatureResolver.getResolvedSignature(method));
        m.setAttribute("name", method.getElementName());
        return m;
    }

    /**
     * Creates a new statement marker at given {@link IResource}.
     *
     * @param res       The {@link IResource} where the marker should be added.
     * @param doc       The {@link IDcoument} in which the marker should be added.
     * @param startLine The start line where marking should be set from.
     * @param endLine   The end line where marking should be set to.
     * @param feature   The feature assigned to that marker.
     * @return The newly created marker.
     * @throws CoreException
     */
    public static IMarker createStatementMarker(IResource res, IDocument doc, int startLine, int endLine, Feature feature) throws CoreException {
        try {
            IRegion startLineRegion = doc.getLineInformation(startLine);
            IRegion endLineRegion = doc.getLineInformation(endLine);
            int startOffset = startLineRegion.getOffset();
            int lenth = endLineRegion.getOffset() - startLineRegion.getOffset() + endLineRegion.getLength();

            IMarker m = createMarker("Statement", res, new Position(startOffset, lenth), FeatureMarkerFactory.FEATURE_STATEMENT_MARKER, feature);
            m.setAttribute(IMarker.LINE_NUMBER, startLine + 1);
            m.setAttribute("lines", endLine - startLine + 1);
            return m;
        } catch (BadLocationException e) {
            return null;
        }
    }

    /**
     * Creates a new marker at a given {@link IResource} with special prefix.
     *
     * @param prefix     The special string prefix
     * @param res        The {@link IResource} where the marker should be added.
     * @param position   The position where the marker should be added.
     * @param identifier The identifier of the marker.
     * @param feature    The feature assigned to that marker.
     * @return The newly created marker.
     * @throws CoreException
     */
    private static IMarker createMarker(String prefix, IResource res, Position position, String identifier, Feature feature) throws CoreException {
        IMarker marker = res.createMarker(identifier);
        marker.setAttribute(IMarker.CHAR_START, position.getOffset());
        marker.setAttribute(IMarker.CHAR_END, position.getOffset() + position.getLength());
        marker.setAttribute(IMarker.MESSAGE, prefix + " Feature " + feature.getDescription() + " (" + feature.getId() + ")");
        marker.setAttribute("feature", feature.getId());

        FileDecorator.getFileDecorator().refreshWithParents(res);

        return marker;
    }

    /**
     * Deletes a marker.
     *
     * @param marker The marker to delete.
     */
    public static void deleteMarker(IMarker marker) {
        IResource resource = marker.getResource();

        try {
            marker.delete();
        } catch (CoreException e) {
            e.printStackTrace();
        }

        FileDecorator.getFileDecorator().refreshWithParents(resource);
    }

    /**
     * Returns a list of a markers for given resource.
     *
     * @param resource The resource.
     * @return The list of a markers for given resource.
     */
    public static List<IMarker> findMarkers(IResource resource) {
        try {
            return Arrays.asList(resource.findMarkers(FEATURE_MARKER, true, IResource.DEPTH_INFINITE));
        } catch (CoreException e) {
            return new ArrayList<IMarker>();
        }
    }

    /**
     * Returns the selection of the package explorer.
     *
     * @return The selection of the package explorer.
     */
    public static TreeSelection getTreeSelection() {

        ISelection selection = Activator.getActiveWorkbenchWindow().getSelectionService().getSelection();
        if (selection instanceof TreeSelection) {
            return (TreeSelection) selection;
        }
        return null;
    }

    /**
     * Returns the selection of the editor.
     *
     * @return The selection of the editor.
     */
    public static TextSelection getTextSelection() {

        ISelection selection = Activator.getActiveWorkbenchWindow().getSelectionService().getSelection();
        if (selection instanceof TextSelection) {
            return (TextSelection) selection;
        }
        return null;
    }

}
