package at.jku.mevss.anttracks.classification;

import at.jku.mevss.anttracks.classification.enumerations.ClassifierSourceCollection;
import at.jku.mevss.anttracks.heap.IndexBasedHeap;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.util.DirectoryClassLoader;
import at.jku.mevss.anttracks.util.FilterUtil;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.ServiceConfigurationError;
import java.util.ServiceLoader;
import java.util.function.Supplier;
import java.util.logging.Logger;

public class FilterFactoryList extends ArrayList<FilterFactory> {

    private static final long serialVersionUID = 3157139101821000231L;

    protected final Logger LOGGER = Logger.getLogger(this.getClass().getSimpleName());
    private final Supplier<Symbols> symbolsSupplier;
    private final Supplier<IndexBasedHeap> fastHeapSupplier;

    public void remove(Filter filter) {
        FilterFactory factory = stream().filter(f -> f.getName().equals(filter.getName())).findFirst().orElse(null);
        if (factory != null) {
            this.remove(factory);
        }
    }

    public FilterFactoryList(Supplier<Symbols> symbolsSupplier, Supplier<IndexBasedHeap> fastHeapSupplier, String directory) {
        this(symbolsSupplier, fastHeapSupplier, directory, null);
    }

    @SuppressWarnings("rawtypes")
    public FilterFactoryList(Supplier<Symbols> symbolsSupplier,
                             Supplier<IndexBasedHeap> fastHeapSupplier,
                             String directory,
                             ClassifierSourceCollection sourceCollection) {
        this.symbolsSupplier = symbolsSupplier;
        this.fastHeapSupplier = fastHeapSupplier;

        loadCompiledFilters(Thread.currentThread().getContextClassLoader(), sourceCollection);

        if (directory != null) {
            loadCompiledFilters(new DirectoryClassLoader(directory), sourceCollection);
            loadSourceCodeFilters(directory, sourceCollection);
        }
    }

    private void loadSourceCodeFilters(String directory, ClassifierSourceCollection sourceCollection) {
        try {
            Files.list(Paths.get(directory)).forEach(f -> {
                if (f.toString().endsWith(FilterUtil.CUSTOM_FILTER_FILE_EXTENSION)) {
                    FilterFactory filterFactory = null;
                    try {
                        filterFactory = FilterUtil.loadFilter(f.toFile(), fastHeapSupplier, symbolsSupplier);
                        if (filterFactory != null && (filterFactory.getSourceCollection() == sourceCollection || filterFactory.getSourceCollection() ==
                                ClassifierSourceCollection.ALL)) {
                            add(filterFactory.create());
                        }
                    } catch (IOException | InstantiationException | IllegalAccessException e) {
                        e.printStackTrace();
                    }
                }
            });
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private void loadCompiledFilters(ClassLoader cl, ClassifierSourceCollection sourceCollection) {
        @SuppressWarnings("rawtypes")
        ServiceLoader<Filter> serviceLoader = ServiceLoader.load(Filter.class, cl);
        try {
            for (Filter filter : serviceLoader) {
                if (sourceCollection == null || filter.getSourceCollection() == sourceCollection || filter.getSourceCollection() == ClassifierSourceCollection.ALL) {
                    add(filter);
                }
            }
        } catch (ServiceConfigurationError sce) {
            LOGGER.warning("Loading object classifiers failed!\n" + sce);
        }
    }

    public Filter get(String name) {
        for (FilterFactory oc : this) {
            Filter filter = oc.create();
            if (filter.getName().equals(name)) {
                return filter;
            }
        }
        return null;
    }

    public void add(Filter oc) {
        for (int i = size() - 1; i >= 0; i--) {
            if (get(i).getName().equals(oc.getName())) {
                remove(i);
            }
        }
        add(new FilterFactory(symbolsSupplier, fastHeapSupplier, oc));
    }

    public boolean contains(String name) {
        return get(name) != null;
    }
}

