package at.jku.mevss.anttracks.classification.nodes;

import at.jku.mevss.anttracks.classification.*;
import at.jku.mevss.anttracks.heap.IndexBasedHeap;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.roots.RootPtr;
import at.jku.mevss.anttracks.heap.space.SpaceInfo;
import at.jku.mevss.anttracks.heap.symbols.AllocatedType;
import at.jku.mevss.anttracks.heap.symbols.AllocationSite;
import at.jku.mevss.anttracks.parser.EventType;
import javafx.beans.property.LongProperty;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.atomic.AtomicBoolean;

public interface GroupingNode {

    public interface FastHeapGroupingNode extends GroupingNode {
        @Override
        default FastHeapGroupingNode addChildIfNeeded(Classifier<?> classifier, Object key) {
            return (FastHeapGroupingNode) GroupingNode.super.addChildIfNeeded(classifier, key);
        }

        @Override
        default FastHeapGroupingNode addChildIfNeeded(Classifier<?> classifier, Object key, int subTreeLevel) {
            return (FastHeapGroupingNode) GroupingNode.super.addChildIfNeeded(classifier, key, subTreeLevel);
        }

        @Override
        FastHeapGroupingNode getChild(Object key);

        default GroupingNode classify(IndexBasedHeap fastHeap, Integer objIndex, ClassifierChain classifiers, ConditionalClassifierList conditionalClassifier) {
            return classify(fastHeap, objIndex, classifiers, conditionalClassifier, 1);
        }

        default GroupingNode classify(IndexBasedHeap fastHeap, Integer objIndex, ClassifierChain classifiers, ConditionalClassifierList conditionalClassifier, long n) {
            return classify(fastHeap, objIndex, classifiers, conditionalClassifier, n, 0);
        }

        default GroupingNode classify(IndexBasedHeap fastHeap,
                                      Integer objIndex,
                                      ClassifierChain classifiers,
                                      ConditionalClassifierList conditionalClassifiers,
                                      long n,
                                      int continueWithSubTreeLevel) {
            ConditionalClassifier conditionalClassifierToUse = findApplicableConditionalClassifier(conditionalClassifiers);

            // Replace remaining classifiers if matching conditional classifier is
            // found
            if (conditionalClassifierToUse != null) {
                conditionalClassifiers = (ConditionalClassifierList) conditionalClassifiers.clone();
                conditionalClassifiers.remove(conditionalClassifierToUse);
                classifiers = conditionalClassifierToUse.getClassifiers();
            }

            // Recursion hook: No more classifiers to process
            if (classifiers.length() == 0) {
                addLeaf(objIndex, n);
                return this;
            }

            // Invariant: At least one classifier is left to process
            Classifier<?> classifierToUse = classifiers.get(0);
            ClassifierChain remainingClassifiers = classifiers.dropFirst();

            try {
                switch (classifierToUse.getType()) {
                    case ONE: {
                        Object key = classifierToUse.classify(objIndex);
                        if (key == null) {
                            break;
                        }

                        if (classifierToUse instanceof Transformer) {
                            Transformer transfromer = (Transformer) classifierToUse;
                            // TODO: Heap currently is provided as null
                            GroupingNode transformed = transfromer.classify(objIndex);

                            // Merge the transformer into this HeapObjectGrouping
                            GroupingNode mergedTransfromer = addSubTreeChildIfNeeded(classifierToUse, transformed.getKey());
                            mergedTransfromer.merge(transformed);

                            if (remainingClassifiers.length() > 0) {
                                ArrayList<GroupingNode> toProcess = new ArrayList<>();
                                toProcess.add(transformed);
                                continueClassificationAfterTransformer(fastHeap,
                                                                       this,
                                                                       toProcess,
                                                                       objIndex,
                                                                       remainingClassifiers,
                                                                       conditionalClassifiers,
                                                                       n,
                                                                       new AtomicBoolean(true));
                            } else { // Transformer was last classifier, classify the
                                // object "before" the transformer
                                this.classify(fastHeap, objIndex, remainingClassifiers, conditionalClassifiers, n);
                            }
                        } else {
                            FastHeapGroupingNode group = addChildIfNeeded(classifierToUse, key, continueWithSubTreeLevel);
                            group.classify(fastHeap, objIndex, remainingClassifiers, conditionalClassifiers, n);
                        }
                        break;
                    }

                    case MANY: {
                        Object[] keys = (Object[]) classifierToUse.classify(objIndex);
                        if (keys == null) {
                            break;
                        }
                        if (keys.length == 0) {
                            // 24 April 2018, mw: Let's try to not stop when empty key array is returned by to continue on same level?
                            classify(fastHeap, objIndex, remainingClassifiers, conditionalClassifiers, n);
                        }
                        for (int i = 0; i < keys.length; i++) {
                            if (i > 0) {
                                addDuplicate(objIndex, n);
                            }
                            FastHeapGroupingNode group = addChildIfNeeded(classifierToUse, keys[i], continueWithSubTreeLevel);
                            group.classify(fastHeap, objIndex, remainingClassifiers, conditionalClassifiers, n);
                        }
                        break;
                    }
                    case HIERARCHY: {
                        FastHeapGroupingNode currentGrouping = this;
                        Object[] keys = (Object[]) classifierToUse.classify(objIndex);
                        if (keys == null) {
                            break;
                        }
                        if (keys.length == 0) {
                            // 24 April 2018, mw: Let's try to not stop when empty key array is returned by to continue on same level?
                            classify(fastHeap, objIndex, remainingClassifiers, conditionalClassifiers, n);
                        }
                        for (int i = 0; i < keys.length; i++) {
                            currentGrouping = currentGrouping.addChildIfNeeded(classifierToUse, keys[i], continueWithSubTreeLevel);

                            if (conditionalClassifiers != null) {
                                for (int y = 0; y < conditionalClassifiers.size(); y++) {
                                    // Shortcut: if conditional classifier is found that
                                    // should be applied during tree building, don't
                                    // build the
                                    // rest of the tree
                                    if (Arrays.equals(currentGrouping.getFullKey(), conditionalClassifiers.get(y).getConditions())) {
                                        currentGrouping.classify(fastHeap, objIndex, remainingClassifiers, conditionalClassifiers, n);
                                        return this;
                                    }
                                }
                            }
                        }
                        currentGrouping.classify(fastHeap, objIndex, remainingClassifiers, conditionalClassifiers, n);
                        break;
                    }
                    case MANY_HIERARCHY: {
                        FastHeapGroupingNode currentGrouping = this;
                        Object[][] keys = (Object[][]) classifierToUse.classify(objIndex);
                        if (keys == null) {
                            break;
                        }
                        if (keys.length == 0) {
                            // 24 April 2018, mw: Let's try to not stop when empty key array is returned by to continue on same level?
                            classify(fastHeap, objIndex, remainingClassifiers, conditionalClassifiers, n);
                        }
                        for (int i = 0; i < keys.length; i++) {
                            currentGrouping = this;
                            for (int j = 0; j < keys[i].length; j++) {
                                currentGrouping = currentGrouping.addChildIfNeeded(classifierToUse, keys[i][j], continueWithSubTreeLevel);

                                if (conditionalClassifiers != null) {
                                    for (int y = 0; y < conditionalClassifiers.size(); y++) {
                                        // Shortcut: if conditional classifier is found that
                                        // should be applied during tree building, don't
                                        // build the
                                        // rest of the tree
                                        if (Arrays.equals(currentGrouping.getFullKey(), conditionalClassifiers.get(y).getConditions())) {
                                            currentGrouping.classify(fastHeap, objIndex, remainingClassifiers, conditionalClassifiers, n);
                                            return this;
                                        }
                                    }
                                }
                            }
                            if (i > 0) {
                                addDuplicate(objIndex, n);
                            }
                            currentGrouping.classify(fastHeap, objIndex, remainingClassifiers, conditionalClassifiers, n);
                        }

                        break;
                    }
                }
            } catch (Exception ex) {
                GroupingNode root = this;
                while (root.getParent() != null) {
                    root = root.getParent();
                }
                root.addException(fastHeap.getAddress(objIndex),
                                  fastHeap.getObjectInfo(objIndex),
                                  fastHeap.getSpace(objIndex),
                                  fastHeap.getType(objIndex),
                                  fastHeap.getSize(objIndex),
                                  fastHeap.isArray(objIndex),
                                  fastHeap.getArrayLength(objIndex),
                                  fastHeap.getAllocationSite(objIndex),
                                  fastHeap.getEventType(objIndex),
                                  fastHeap.getRoot(objIndex),
                                  ex);
            }
            return this;
        }

        default void continueClassificationAfterTransformer(IndexBasedHeap fastHeap,
                                                            FastHeapGroupingNode transformerParent,
                                                            List<? extends GroupingNode> toProcess,
                                                            Integer objIndex,
                                                            ClassifierChain classifiers,
                                                            ConditionalClassifierList conditionalClassifiers,
                                                            long n,
                                                            AtomicBoolean firstClassification) {
            for (GroupingNode processGrouping : toProcess) {
                FastHeapGroupingNode child = getChild(processGrouping.getKey());
                if (processGrouping.getChildren().size() == 0) {
                    // Leaf nodes in transformer -> Continue classification with
                    // remaining classifiers on all leaf nodes
                    if (!firstClassification.get()) {
                        transformerParent.addDuplicate(objIndex, n);
                    }
                    child.classify(fastHeap, objIndex, classifiers, conditionalClassifiers, n, transformerParent.getSubTreeLevel());
                    firstClassification.set(false);
                } else {
                    child.continueClassificationAfterTransformer(fastHeap,
                                                                 transformerParent,
                                                                 processGrouping.getChildren(),
                                                                 objIndex,
                                                                 classifiers,
                                                                 conditionalClassifiers,
                                                                 n,
                                                                 firstClassification);
                }
            }
        }

        void addLeaf(int objId, long n);

        void addDuplicate(int objId, long n);
    }

    public interface DetailedHeapGroupingNode extends GroupingNode {
        @Override
        default DetailedHeapGroupingNode addChildIfNeeded(Classifier<?> classifier, Object key) {
            return (DetailedHeapGroupingNode) GroupingNode.super.addChildIfNeeded(classifier, key);
        }

        @Override
        default DetailedHeapGroupingNode addChildIfNeeded(Classifier<?> classifier, Object key, int subTreeLevel) {
            return (DetailedHeapGroupingNode) GroupingNode.super.addChildIfNeeded(classifier, key, subTreeLevel);
        }

        @Override
        DetailedHeapGroupingNode getChild(Object key);

        @Override
        DetailedHeapGroupingNode getChildBasedOnStringKey(String key);

        default GroupingNode classify(long address,
                                      ObjectInfo objectInfo,
                                      SpaceInfo space,
                                      AllocatedType type,
                                      int size,
                                      boolean isArray,
                                      int arrayLength,
                                      AllocationSite allocationSite,
                                      long[] pointedFrom,
                                      long[] pointsTo,
                                      EventType eventType,
                                      List<RootPtr> rootPtrs,
                                      ClassifierChain classifiers,
                                      ConditionalClassifierList conditionalClassifier) {
            return classify(address,
                            objectInfo,
                            space,
                            type,
                            size,
                            isArray,
                            arrayLength,
                            allocationSite,
                            pointedFrom,
                            pointsTo,
                            eventType,
                            rootPtrs,
                            classifiers,
                            conditionalClassifier,
                            1);
        }

        default GroupingNode classify(long address,
                                      ObjectInfo objectInfo,
                                      SpaceInfo space,
                                      AllocatedType type,
                                      int size,
                                      boolean isArray,
                                      int arrayLength,
                                      AllocationSite allocationSite,
                                      long[] pointedFrom,
                                      long[] pointsTo,
                                      EventType eventType,
                                      List<RootPtr> rootPtrs,
                                      ClassifierChain classifiers,
                                      ConditionalClassifierList conditionalClassifier,
                                      long n) {
            return classify(address,
                            objectInfo,
                            space,
                            type,
                            size,
                            isArray,
                            arrayLength,
                            allocationSite,
                            pointedFrom,
                            pointsTo,
                            eventType,
                            rootPtrs,
                            classifiers,
                            conditionalClassifier,
                            n,
                            0);
        }

        default GroupingNode classify(long address,
                                      ObjectInfo objectInfo,
                                      SpaceInfo space,
                                      AllocatedType type,
                                      int size,
                                      boolean isArray,
                                      int arrayLength,
                                      AllocationSite allocationSite,
                                      long[] pointedFrom,
                                      long[] pointsTo,
                                      EventType eventType,
                                      List<RootPtr> rootPtrs,
                                      ClassifierChain classifiers,
                                      ConditionalClassifierList conditionalClassifiers,
                                      long n,
                                      int continueWithSubTreeLevel) {
            ConditionalClassifier conditionalClassifierToUse = findApplicableConditionalClassifier(conditionalClassifiers);

            // Replace remaining classifiers if matching conditional classifier is
            // found
            if (conditionalClassifierToUse != null) {
                conditionalClassifiers = (ConditionalClassifierList) conditionalClassifiers.clone();
                conditionalClassifiers.remove(conditionalClassifierToUse);
                classifiers = conditionalClassifierToUse.getClassifiers();
            }

            // Recursion hook: No more classifiers to process
            if (classifiers.length() == 0) {
                addLeaf(size, n);
                return this;
            }

            // Invariant: At least one classifier is left to process
            Classifier<?> classifierToUse = classifiers.get(0);
            ClassifierChain remainingClassifiers = classifiers.dropFirst();

            try {
                switch (classifierToUse.getType()) {
                    case ONE: {
                        Object key = classifierToUse.classify(address,
                                                              objectInfo,
                                                              space,
                                                              type,
                                                              size,
                                                              isArray,
                                                              arrayLength,
                                                              allocationSite,
                                                              pointedFrom,
                                                              pointsTo,
                                                              eventType,
                                                              rootPtrs);
                        if (key == null) {
                            break;
                        }

                        if (classifierToUse instanceof Transformer) {
                            Transformer transfromer = (Transformer) classifierToUse;
                            // TODO: Heap currently is provided as null
                            GroupingNode transformed = transfromer.classify(address,
                                                                            objectInfo,
                                                                            space,
                                                                            type,
                                                                            size,
                                                                            isArray,
                                                                            arrayLength,
                                                                            allocationSite,
                                                                            pointedFrom,
                                                                            pointsTo,
                                                                            eventType,
                                                                            rootPtrs);

                            // Merge the transformer into this HeapObjectGrouping
                            GroupingNode mergedTransfromer = addSubTreeChildIfNeeded(classifierToUse, key);
                            mergedTransfromer.merge(transformed);

                            if (remainingClassifiers.length() > 0) {
                                ArrayList<GroupingNode> toProcess = new ArrayList<>();
                                toProcess.add(transformed);
                                continueClassificationAfterTransformer(address,
                                                                       objectInfo,
                                                                       space,
                                                                       type,
                                                                       size,
                                                                       isArray,
                                                                       arrayLength,
                                                                       allocationSite,
                                                                       pointedFrom,
                                                                       pointsTo,
                                                                       eventType,
                                                                       rootPtrs,
                                                                       this,
                                                                       toProcess,
                                                                       remainingClassifiers,
                                                                       conditionalClassifiers,
                                                                       n,
                                                                       new AtomicBoolean(true));
                            } else { // Transformer was last classifier, classify the
                                // object "before" the transformer
                                this.classify(address,
                                              objectInfo,
                                              space,
                                              type,
                                              size,
                                              isArray,
                                              arrayLength,
                                              allocationSite,
                                              pointedFrom,
                                              pointsTo,
                                              eventType,
                                              rootPtrs,
                                              remainingClassifiers,
                                              conditionalClassifiers,
                                              n);
                            }
                        } else {
                            DetailedHeapGroupingNode group = addChildIfNeeded(classifierToUse, key, continueWithSubTreeLevel);
                            group.classify(address,
                                           objectInfo,
                                           space,
                                           type,
                                           size,
                                           isArray,
                                           arrayLength,
                                           allocationSite,
                                           pointedFrom,
                                           pointsTo,
                                           eventType,
                                           rootPtrs,
                                           remainingClassifiers,
                                           conditionalClassifiers,
                                           n);
                        }
                        break;
                    }

                    case MANY: {
                        Object[] keys = (Object[]) classifierToUse.classify(address,
                                                                            objectInfo,
                                                                            space,
                                                                            type,
                                                                            size,
                                                                            isArray,
                                                                            arrayLength,
                                                                            allocationSite,
                                                                            pointedFrom,
                                                                            pointsTo,
                                                                            eventType,
                                                                            rootPtrs);
                        if (keys == null) {
                            break;
                        }
                        if (keys.length == 0) {
                            // 24 April 2018, mw: Let's try to not stop when empty key array is returned by to continue on same level?
                            classify(address,
                                     objectInfo,
                                     space,
                                     type,
                                     size,
                                     isArray,
                                     arrayLength,
                                     allocationSite,
                                     pointedFrom,
                                     pointsTo,
                                     eventType,
                                     rootPtrs,
                                     remainingClassifiers,
                                     conditionalClassifiers,
                                     n);
                        }

                        for (int i = 0; i < keys.length; i++) {
                            if (i > 0) {
                                addDuplicate(size, n);
                            }
                            DetailedHeapGroupingNode group = addChildIfNeeded(classifierToUse, keys[i], continueWithSubTreeLevel);
                            group.classify(address,
                                           objectInfo,
                                           space,
                                           type,
                                           size,
                                           isArray,
                                           arrayLength,
                                           allocationSite,
                                           pointedFrom,
                                           pointsTo,
                                           eventType,
                                           rootPtrs,
                                           remainingClassifiers,
                                           conditionalClassifiers,
                                           n);
                        }
                        break;
                    }
                    case HIERARCHY: {
                        DetailedHeapGroupingNode currentGrouping = this;
                        Object[] keys = (Object[]) classifierToUse.classify(address,
                                                                            objectInfo,
                                                                            space,
                                                                            type,
                                                                            size,
                                                                            isArray,
                                                                            arrayLength,
                                                                            allocationSite,
                                                                            pointedFrom,
                                                                            pointsTo,
                                                                            eventType,
                                                                            rootPtrs);
                        if (keys == null) {
                            break;
                        }
                        if (keys.length == 0) {
                            // 24 April 2018, mw: Let's try to not stop when empty key array is returned by to continue on same level?
                            classify(address,
                                     objectInfo,
                                     space,
                                     type,
                                     size,
                                     isArray,
                                     arrayLength,
                                     allocationSite,
                                     pointedFrom,
                                     pointsTo,
                                     eventType,
                                     rootPtrs,
                                     remainingClassifiers,
                                     conditionalClassifiers,
                                     n);
                        }
                        for (int i = 0; i < keys.length; i++) {
                            currentGrouping = currentGrouping.addChildIfNeeded(classifierToUse, keys[i], continueWithSubTreeLevel);

                            if (conditionalClassifiers != null) {
                                for (int y = 0; y < conditionalClassifiers.size(); y++) {
                                    // Shortcut: if conditional classifier is found that
                                    // should be applied during tree building, don't
                                    // build the
                                    // rest of the tree
                                    if (Arrays.equals(currentGrouping.getFullKey(), conditionalClassifiers.get(y).getConditions())) {
                                        currentGrouping.classify(address,
                                                                 objectInfo,
                                                                 space,
                                                                 type,
                                                                 size,
                                                                 isArray,
                                                                 arrayLength,
                                                                 allocationSite,
                                                                 pointedFrom,
                                                                 pointsTo,
                                                                 eventType,
                                                                 rootPtrs,
                                                                 remainingClassifiers,
                                                                 conditionalClassifiers,
                                                                 n);
                                        return this;
                                    }
                                }
                            }
                        }
                        currentGrouping.classify(address,
                                                 objectInfo,
                                                 space,
                                                 type,
                                                 size,
                                                 isArray,
                                                 arrayLength,
                                                 allocationSite,
                                                 pointedFrom,
                                                 pointsTo,
                                                 eventType,
                                                 rootPtrs,
                                                 remainingClassifiers,
                                                 conditionalClassifiers,
                                                 n);
                        break;
                    }
                    case MANY_HIERARCHY: {
                        DetailedHeapGroupingNode currentGrouping = this;
                        Object[][] keys = (Object[][]) classifierToUse.classify(address,
                                                                                objectInfo,
                                                                                space,
                                                                                type,
                                                                                size,
                                                                                isArray,
                                                                                arrayLength,
                                                                                allocationSite,
                                                                                pointedFrom,
                                                                                pointsTo,
                                                                                eventType,
                                                                                rootPtrs);
                        if (keys == null) {
                            break;
                        }
                        if (keys.length == 0) {
                            // 24 April 2018, mw: Let's try to not stop when empty key array is returned by to continue on same level?
                            classify(address,
                                     objectInfo,
                                     space,
                                     type,
                                     size,
                                     isArray,
                                     arrayLength,
                                     allocationSite,
                                     pointedFrom,
                                     pointsTo,
                                     eventType,
                                     rootPtrs,
                                     remainingClassifiers,
                                     conditionalClassifiers,
                                     n);
                        }
                        for (int i = 0; i < keys.length; i++) {
                            currentGrouping = this;
                            for (int j = 0; j < keys[i].length; j++) {
                                currentGrouping = currentGrouping.addChildIfNeeded(classifierToUse, keys[i][j], continueWithSubTreeLevel);

                                if (conditionalClassifiers != null) {
                                    for (int y = 0; y < conditionalClassifiers.size(); y++) {
                                        // Shortcut: if conditional classifier is found that
                                        // should be applied during tree building, don't
                                        // build the
                                        // rest of the tree
                                        if (Arrays.equals(currentGrouping.getFullKey(), conditionalClassifiers.get(y).getConditions())) {
                                            currentGrouping.classify(address,
                                                                     objectInfo,
                                                                     space,
                                                                     type,
                                                                     size,
                                                                     isArray,
                                                                     arrayLength,
                                                                     allocationSite,
                                                                     pointedFrom,
                                                                     pointsTo,
                                                                     eventType,
                                                                     rootPtrs,
                                                                     remainingClassifiers,
                                                                     conditionalClassifiers,
                                                                     n);
                                            return this;
                                        }
                                    }
                                }
                            }
                            if (i > 0) {
                                addDuplicate(size, n);
                            }
                            currentGrouping.classify(address,
                                                     objectInfo,
                                                     space,
                                                     type,
                                                     size,
                                                     isArray,
                                                     arrayLength,
                                                     allocationSite,
                                                     pointedFrom,
                                                     pointsTo,
                                                     eventType,
                                                     rootPtrs,
                                                     remainingClassifiers,
                                                     conditionalClassifiers,
                                                     n);
                        }

                        break;
                    }
                }
            } catch (Exception ex) {
                GroupingNode root = this;
                while (root.getParent() != null) {
                    root = root.getParent();
                }
                root.addException(address, objectInfo, space, type, size, isArray, arrayLength, allocationSite, eventType, rootPtrs, ex);
            }
            return this;
        }

        default void continueClassificationAfterTransformer(long address,
                                                            ObjectInfo objectInfo,
                                                            SpaceInfo space,
                                                            AllocatedType type,
                                                            int size,
                                                            boolean isArray,
                                                            int arrayLength,
                                                            AllocationSite allocationSite,
                                                            long[] pointedFrom,
                                                            long[] pointsTo,
                                                            EventType eventType,
                                                            List<RootPtr> rootPtrs,
                                                            DetailedHeapGroupingNode transformerParent,
                                                            List<? extends GroupingNode> toProcess,
                                                            ClassifierChain classifiers,
                                                            ConditionalClassifierList conditionalClassifiers,
                                                            long n,
                                                            AtomicBoolean firstClassification) {
            for (GroupingNode processGrouping : toProcess) {
                DetailedHeapGroupingNode child = getChild(processGrouping.getKey());
                if (processGrouping.getChildren().size() == 0) {
                    // Leaf nodes in transformer -> Continue classification with
                    // remaining classifiers on all leaf nodes
                    if (!firstClassification.get()) {
                        transformerParent.addDuplicate(size, n);
                    }
                    child.classify(address,
                                   objectInfo,
                                   space,
                                   type,
                                   size,
                                   isArray,
                                   arrayLength,
                                   allocationSite,
                                   pointedFrom,
                                   pointsTo,
                                   eventType,
                                   rootPtrs,
                                   classifiers,
                                   conditionalClassifiers,
                                   n,
                                   transformerParent.getSubTreeLevel());
                    firstClassification.set(false);
                } else {
                    child.continueClassificationAfterTransformer(address,
                                                                 objectInfo,
                                                                 space,
                                                                 type,
                                                                 size,
                                                                 isArray,
                                                                 arrayLength,
                                                                 allocationSite,
                                                                 pointedFrom,
                                                                 pointsTo,
                                                                 eventType,
                                                                 rootPtrs,
                                                                 transformerParent,
                                                                 processGrouping.getChildren(),
                                                                 classifiers,
                                                                 conditionalClassifiers,
                                                                 n,
                                                                 firstClassification);
                }
            }
        }

        void addLeaf(int size, long n);

        void addDuplicate(int size, long n);
    }

    void setObjectCount(double count);

    void setByteCount(long bytes);

    interface GroupingDataCollection {
        /**
         * @return Return the number of objects in this collection.
         */
        int getObjectCount();

        long getBytes();

        void clear();
    }

    Object getKey();

    List<GroupingNode> getChildren();

    int getSubTreeLevel();

    /**
     * @return Returns only local data
     */
    GroupingDataCollection getData();

    /**
     * @param refSubTree
     * @return Returns data of children on given SubTreeLevel, including own data if this node is also on given subtreelevel.
     */
    GroupingDataCollection getData(int refSubTree);

    GroupingNode getParent();

    default GroupingNode getFirstParentOnSubTreeLevel(int subTreeLevel) {
        for (GroupingNode node = getParent(); node != null; node = node.getParent()) {
            if (node.getSubTreeLevel() == subTreeLevel) {
                return node;
            }
        }

        return null;
    }

    Classifier<?> getClassifier();

    boolean isCalculatedOnSampling();

    long getObjectCount();

    long getByteCount();

    void fillClosureSizes();

    LongProperty transitiveClosureSizeProperty();

    boolean isClosureSizeCalculated();

    LongProperty retainedSizeProperty();

    boolean isGCSizeCalculated();

    LongProperty dataStructureSizeProperty();

    boolean isDataStructureSizeCalculated();

    LongProperty deepDataStructureSizeProperty();

    boolean isDeepDataStructureSizeCalculated();

    boolean getAndSetIsClosuresBeingCalculated(boolean newValue);

    List<GroupingException> getExceptions();

    public Object[] getFullKey();

    void setClassifier(Classifier<?> classifier);

    GroupingNode merge(GroupingNode other, int refSubTreeLevel);

    boolean containsChild(Object key);

    int getLevel();

    void clear();

    void setCalculatedOnSampling(boolean b);

    void addException(long address,
                      ObjectInfo objectInfo,
                      SpaceInfo space,
                      AllocatedType type,
                      int size,
                      boolean isArray,
                      int arrayLength,
                      AllocationSite allocationSite,
                      EventType eventType,
                      List<RootPtr> rootPtrs,
                      Exception ex);

    GroupingNode sampleTopDown();

    GroupingNode getChild(Object key);

    GroupingNode getChildBasedOnStringKey(String key);

    void setSubTreeLevel(int subTreeLevel);

    default ConditionalClassifier findApplicableConditionalClassifier(ConditionalClassifierList conditionalClassifier) {
        if (conditionalClassifier == null) {
            return null;
        }

        for (int i = 0; i < conditionalClassifier.size(); i++) {
            if (Arrays.equals(getFullKey(), conditionalClassifier.get(i).getConditions())) {
                return conditionalClassifier.get(i);
            }
        }

        return null;
    }

    default GroupingNode addSubTreeChildIfNeeded(Classifier<?> classifier, Object key) {
        return addChildIfNeeded(classifier, key, this.getSubTreeLevel() + 1);
    }

    default GroupingNode addChildIfNeeded(Classifier<?> classifier, Object key) {
        return addChildIfNeeded(classifier, key, getSubTreeLevel());
    }

    default GroupingNode addChildIfNeeded(Classifier<?> classifier, Object key, int subTreeLevel) {
        GroupingNode child = getChild(key);
        if (child == null) {
            child = addChild(classifier, getLevel() + 1, subTreeLevel, key);
        } else {
            assert child.getClassifier()
                        .getName()
                        .equals(classifier.getName()) && child.getLevel() == getLevel() + 1 && child.getSubTreeLevel() == subTreeLevel : String.format(
                    "Found child must match the classifier (%s <-> %s) and level (%d <-> %d) and subtreelevel (%d <-> %d",
                    child.getClassifier(),
                    classifier,
                    child.getLevel(),
                    getLevel() + 1,
                    child.getSubTreeLevel(),
                    subTreeLevel);
        }
        return child;
    }

    /**
     * FOR INTERNAL USE ONLY
     *
     * @param key
     * @return
     */
    GroupingNode addChild(Classifier<?> classifier, int level, int subTreeLevel, Object key);

    default GroupingNode merge(GroupingNode other) {
        return merge(other, this.getSubTreeLevel() - other.getSubTreeLevel());
    }

    default List<GroupingNode> getFirstChildrenOnSameSubtreeLevel() {
        ArrayList<GroupingNode> children = new ArrayList<>();
        ArrayDeque<GroupingNode> toCheck = new ArrayDeque<>(getChildren());

        while (!toCheck.isEmpty()) {
            GroupingNode x = toCheck.poll();
            if (x.getSubTreeLevel() == getSubTreeLevel()) {
                children.add(x);
            } else {
                toCheck.addAll(x.getChildren());
            }
        }

        return children;
    }

    default List<GroupingNode> getFirstChildrenOnSubtreeLevel(int subTreeLevel) {
        ArrayList<GroupingNode> children = new ArrayList<>();
        ArrayDeque<GroupingNode> toCheck = new ArrayDeque<>(getChildren());

        while (!toCheck.isEmpty()) {
            GroupingNode x = toCheck.poll();
            if (x.getSubTreeLevel() == subTreeLevel) {
                children.add(x);
            } else {
                toCheck.addAll(x.getChildren());
            }
        }

        return children;
    }

    default GroupingDataCollection getData(boolean recursive) {
        if (recursive) {
            return getData(getSubTreeLevel());
        } else {
            return getData();
        }
    }

    default DataGroupingNode subtractRecursive(GroupingNode other) {
        assert getKey().equals(other.getKey()) : "Two nodes may only be subtracted if they have the same key";
        DataGroupingNode result = subtract(other);
        for (GroupingNode child : getChildren()) {
            GroupingNode otherChild = other.getChild(child.getKey());
            if (otherChild == null) {
                // add this child (and all successors) with unchanged values to the result grouping
                result.addChild(child.subtractRecursive(new DataGroupingNode(null,
                                                                             child.getFullKey(),
                                                                             child.getClassifier(),
                                                                             child.getLevel(),
                                                                             child.getSubTreeLevel(),
                                                                             0,
                                                                             0,
                                                                             0,
                                                                             0,
                                                                             0,
                                                                             0)));
            } else {
                // subtract recursively
                result.addChild(child.subtractRecursive(otherChild));
            }
        }

        // now all successors of this node are in the result tree
        // however, there might be successors of the other node which are not handled yet because there is no matching successor of this node
        // thus we also traverse the other node's successors and handle all those for which there is no matching successor of this node

        for (GroupingNode otherChild : other.getChildren()) {
            GroupingNode child = getChild(otherChild.getKey());
            if (child == null) {
                // add this child (and all successors) with negated values to the result grouping
                result.addChild(otherChild.subtractRecursive(new DataGroupingNode(null,
                                                                                  otherChild.getFullKey(),
                                                                                  otherChild.getClassifier(),
                                                                                  otherChild.getLevel(),
                                                                                  otherChild.getSubTreeLevel(),
                                                                                  otherChild.getObjectCount() * 2,
                                                                                  otherChild.getByteCount() * 2,
                                                                                  otherChild.transitiveClosureSizeProperty().get() * 2,
                                                                                  otherChild.retainedSizeProperty().get() * 2,
                                                                                  otherChild.dataStructureSizeProperty().get() * 2,
                                                                                  otherChild.deepDataStructureSizeProperty().get() * 2)));
            }
        }

        return result;
    }

    default DataGroupingNode subtract(GroupingNode other) {
        assert other == null || getKey().equals(other.getKey()) : "Nodes may only be subtracted if the share the same key";

        DataGroupingNode result;
        if (other != null) {
            result = new DataGroupingNode(null,
                                          getFullKey(),
                                          getClassifier(),
                                          getLevel(),
                                          getSubTreeLevel(),
                                          getObjectCount() - other.getObjectCount(),
                                          getByteCount() - other.getByteCount(),
                                          transitiveClosureSizeProperty().get() - other.transitiveClosureSizeProperty().get(),
                                          retainedSizeProperty().get() - other.retainedSizeProperty().get(),
                                          dataStructureSizeProperty().get() - other.dataStructureSizeProperty().get(),
                                          deepDataStructureSizeProperty().get() - other.deepDataStructureSizeProperty().get());
        } else {
            result = new DataGroupingNode(null,
                                          getFullKey(),
                                          getClassifier(),
                                          getLevel(),
                                          getSubTreeLevel(),
                                          getObjectCount(),
                                          getByteCount(),
                                          transitiveClosureSizeProperty().get(),
                                          retainedSizeProperty().get(),
                                          dataStructureSizeProperty().get(),
                                          deepDataStructureSizeProperty().get());
        }

        return result;
    }

    default GroupingNode getChildWithFullKey(Object[] key) {
        if (getLevel() + 1 == key.length) {
            if (getKey().equals(key[getLevel()])) {
                return this;
            } else {
                return null;
            }
        } else {
            GroupingNode child = getChildBasedOnStringKey(key[getLevel() + 1].toString());
            if (child == null) {
                return null;
            }
            return child.getChildWithFullKey(key);
        }
    }
}
