package at.jku.mevss.anttracks.classification.nodes;

import at.jku.mevss.anttracks.heap.Closures;
import at.jku.mevss.anttracks.heap.IndexBasedHeap;
import at.jku.mevss.anttracks.util.ApplicationStatistics;

import java.util.Arrays;
import java.util.BitSet;

public class IndexCollection implements GroupingNode.GroupingDataCollection {
    private int[] data;
    private int count;
    private boolean sorted;
    public final IndexBasedHeap heap;
    private Closures closures = null;

    public IndexCollection(IndexBasedHeap heap) {
        this(1, heap);
    }

    public IndexCollection(int initialSize, IndexBasedHeap heap) {
        if (initialSize < 1) {
            throw new IllegalArgumentException("Initial size must be at least 1!");
        }
        this.data = new int[initialSize];
        this.count = 0;
        this.sorted = false;
        this.heap = heap;
    }

    public IndexCollection(int[] data, IndexBasedHeap fastHeap) {
        this.data = data;
        this.count = data.length;
        this.sorted = false;
        this.heap = fastHeap;
    }

    public int get(int index) {
        if (index < 0 || index >= count) {
            throw new IllegalArgumentException("Invalid index given");
        }

        return data[index];
    }

    public int getCount() {
        return count;
    }

    IndexBasedHeap getHeap() { return heap; }

    public long getBytes() {
        long bytes = 0;
        for (int i = 0; i < count; i++) {
            bytes += heap.getSize(get(i));
        }
        return bytes;
    }

    public Closures calculateClosures() {
        return calculateClosures(null);
    }

    public Closures calculateClosures(BitSet assumeChildClosure) {
        if (!sorted) {
            sort();
        }
        return heap.getClosures(Arrays.copyOf(data, count), assumeChildClosure);
    }

    /*
    public Closures getClosures() {
        return getClosures(null);
    }

    public Closures getClosures(HashSet<Integer> assumeHelpClosure) {
        if (closures == null) {
            closures = heap.getClosures(Arrays.copyOf(data, count), assumeHelpClosure);
        }
        return closures;
    }
    */

    void setCount(int count) {
        this.count = count;
    }

    boolean isSorted() {
        return sorted;
    }

    void add(int index) {
        if (!isSorted()) {
            sort();
        }
        int insertionPoint = Arrays.binarySearch(data, 0, count, index);
        if (insertionPoint < 0) {
            // index not yet stored
            // enlarge if necessary
            if (count >= data.length) {
                data = Arrays.copyOfRange(data, 0, data.length * 2);
            }

            insertionPoint = Math.abs(insertionPoint) - 1;
            // shift data from insertion point to the right by 1
            System.arraycopy(data, insertionPoint, data, insertionPoint + 1, count - insertionPoint);
            // insert new index into the resulting gap
            data[insertionPoint] = index;
            count++;
        }
    }

    /**
     * Used by union - does not check uniqueness/sort (guaranteed by union algorithm)
     *
     * @param index
     */
    void fastAdd(int index) {
        if (count >= data.length) {
            data = Arrays.copyOfRange(data, 0, data.length * 2);
        }

        data[count++] = index;
        sorted = false;
    }

    @Override
    public void clear() {
        count = 0;
    }

    void sort() {
        Arrays.sort(data, 0, count);
        sorted = true;
    }

    public boolean contains(int x) {
        if (!sorted) {
            sort();
        }
        return Arrays.binarySearch(data, x) >= 0;
    }

    public synchronized void unionWith(IndexCollection[] childData) {
        if (childData == null) {
            return;
        }

        int finalSize = count;
        for (IndexCollection child : childData) {
            if (child != null) {
                finalSize += child.count;
            }
        }

        if (finalSize == this.count) {
            return;
        }

        int loc = this.count;

        // Increase data size to fit all children
        this.data = Arrays.copyOf(this.data, finalSize);
        this.count = finalSize;
        // Copy all children
        for (IndexCollection child : childData) {
            if (child != null) {
                System.arraycopy(child.data, 0, this.data, loc, child.count);
                loc += child.count;
            }
        }
        // Sort array
        sort();
        // Eliminate duplicates
        int[] newData = new int[this.data.length];
        this.count = 0;
        for (int fromIdx = 0; fromIdx < this.data.length; fromIdx++) {
            if (fromIdx + 1 < this.data.length && this.data[fromIdx] == this.data[fromIdx + 1]) {
                continue;
            }
            newData[this.count++] = this.data[fromIdx];
        }
        this.data = newData;
    }

    public synchronized void unionWith(IndexCollection other) {
        if (other == null) {
            return;
        }

        if (!this.isSorted()) {
            sort();
        }
        if (!other.isSorted()) {
            other.sort();
        }

        int initialSize = count + other.getCount();
        IndexCollection union = new IndexCollection(initialSize >= 1 ? initialSize : 1, heap);
        // other remains unchanged
        int i = 0, j = 0;
        // union while retaining element order
        while (i < count && j < other.getCount()) {
            if (this.get(i) < other.get(j)) {
                union.fastAdd(this.get(i++));

            } else if (this.get(i) == other.get(j)) {
                union.fastAdd(this.get(i));
                i++;
                j++;

            } else { // this.get(i) > other.get(j)
                union.fastAdd(other.get(j++));
            }
        }

        // add remaining elements
        while (i < count) {
            union.fastAdd(this.get(i++));
        }
        while (j < other.getCount()) {
            union.fastAdd(other.get(j++));
        }

        // save
        this.data = union.data;
        this.setCount(union.getCount());
    }

    public IndexCollection clone() {
        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("IndexCollection - Clone");
        IndexCollection dolly = new IndexCollection(heap);
        dolly.data = Arrays.copyOf(data, data.length);
        dolly.count = count;
        dolly.sorted = sorted;
        dolly.closures = closures;
        m.end();
        return dolly;
    }

    @Override
    public int getObjectCount() {
        return count;
    }

    public boolean isEmpty() {
        return count == 0;
    }
}
