/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.diff;

import at.jku.mevss.anttracks.classification.ClassifierChain;
import at.jku.mevss.anttracks.classification.ConditionalClassifierList;
import at.jku.mevss.anttracks.classification.Filter;
import at.jku.mevss.anttracks.classification.nodes.MapGroupingNode;
import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.ObjectStream;
import at.jku.mevss.anttracks.heap.ObjectVisitor;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.roots.RootPtr;
import at.jku.mevss.anttracks.heap.space.SpaceInfo;
import at.jku.mevss.anttracks.heap.symbols.AllocatedType;
import at.jku.mevss.anttracks.heap.symbols.AllocationSite;
import at.jku.mevss.anttracks.parser.EventType;
import at.jku.mevss.anttracks.util.*;
import at.jku.mevss.anttracks.util.ApplicationStatistics.Measurement;

import javax.naming.OperationNotSupportedException;
import java.io.DataOutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.logging.Logger;
import java.util.stream.LongStream;

public class ComplexAddressCollection implements AddressCollection {

    public class ComplexAddressCollectionIterator implements AddressCollectionIterator {

        private Part part;
        private int p;
        private int i;

        public ComplexAddressCollectionIterator() {
            removeEmptyParts();
            if (parts.size() > 0) {
                part = parts.get(0);
            } else {
                part = null;
            }
            p = 0;
            i = -1;
        }

        public boolean hasNext() {
            return part != null && (i + 1 < part.getSize() || p + 1 < parts.size());
        }

        public long next() {
            i++;
            if (i == part.getSize()) {
                p++;
                part = parts.get(p);
                i = 0;
            }
            return part.addresses[i];
        }

    }

    private class Part {

        private long[] addresses;
        private MutableLong startAddress;
        private MutableInt size;
        private final Lock lock;

        public Part() {
            long[] arr = new long[objectsPerPart];
            Arrays.fill(arr, NO_ELEMENT);
            addresses = arr;
            size = new MutableInt(0);
            startAddress = new MutableLong(0);
            lock = new ReentrantLock();
        }

        public Part(long[] arr, int size) {
            addresses = arr;
            this.size = new MutableInt(size);
            startAddress = new MutableLong(size > 0 ? addresses[0] : 0);
            lock = new ReentrantLock();
        }

        public long[] getAddresses() {
            return addresses;
        }

        public long getStartAddress() {
            return startAddress.get();
        }

        public int getSize() {
            return size.get();
        }

        public void lock() {
            lock.lock();
        }

        public void unlock() {
            lock.unlock();
        }

        public void resetStartAddress() {
            startAddress.set(addresses[0]);
        }

        public boolean hasToBeResized() {
            return hasToBeResized(1);
        }

        public boolean hasToBeResized(int n) {
            return getSize() + n > objectsPerPart;
        }

        @Override
        protected Part clone() {
            Part cloned = new Part();
            cloned.addresses = addresses.clone();
            cloned.size = new MutableInt(getSize());
            cloned.startAddress = new MutableLong(getStartAddress());
            return cloned;
        }
    }

    private static final Logger LOGGER = Logger.getLogger(ComplexAddressCollection.class.getSimpleName());
    private static final long NO_ELEMENT = -1;
    private static final boolean MULTI_THREADED = false;

    private int objectsPerPart = 30_000;
    protected ArrayList<Part> parts = new ArrayList<>();
    // This member is only set before classification, do not use in other
    // situations, because it may contain wrong information
    protected long[] partsStartAddresses;

    protected String name = null;

    protected DetailedHeap lastHeap;

    public ComplexAddressCollection() {
        this("Default");
    }

    public ComplexAddressCollection(String name) {
        this(name, 30_000);
    }

    public ComplexAddressCollection(String name, int objectsPerPart) {
        this.name = name;
        this.objectsPerPart = objectsPerPart;
        parts.add(new Part());
    }

    @Override
    public void clear() {
        parts.clear();
        parts.add(new Part());
    }

    @Override
    public AddressCollection sort() {
        return this;
    }

    @Override
    public boolean isSorted() {
        return true;
    }

    private int findPart(long addr) {
        // If the address is the very first address, return the first array
        if (addr < parts.get(0).getStartAddress()) {
            return 0;
        }

        for (int i = 0; i < parts.size() - 1; i++) {
            if (parts.get(i).getStartAddress() <= addr && parts.get(i + 1).getStartAddress() > addr) {
                return i;
            }
        }
        // Last object array
        return parts.size() - 1;
    }

    /*
    public void putSyncIfAbsent(long addr) {
        if (!contains(addr)) {
            put(addr);
        }
    }

    public void putSync(long addr) {
        Part part = null;
        int idx = 0;

        synchronized (parts) {
            idx = findPart(addr);
            part = parts.get(idx);
            part.lock();
            if (part.hasToBeResized()) {
                long[] next = Arrays.copyOfRange(parts.get(idx).getAddresses(), objectsPerPart / 2,
                                                 objectsPerPart + objectsPerPart / 2);
                part.size.set(objectsPerPart / 2);

                Part newPart = new Part(next, objectsPerPart / 2);
                parts.add(idx + 1, newPart);

                // Check if put should happen in newly created array
                if (addr > newPart.getStartAddress()) {
                    part.unlock();
                    part = newPart;
                    part.lock();
                }
            }
        }

        int curEle = part.getSize();
        part.getAddresses()[curEle] = addr;

        while (curEle > 0) {
            if (part.getAddresses()[curEle] < part.getAddresses()[curEle - 1]) {
                long temp = part.getAddresses()[curEle];
                part.getAddresses()[curEle] = part.getAddresses()[curEle - 1];
                part.getAddresses()[curEle - 1] = temp;
                curEle--;
                // TODO: Check equality --> exception throw
            } else {
                break;
            }
        }

        part.size.inc();

        if (curEle == 0) {
            part.resetStartAddress();
        }

        part.unlock();
    }
    */

    public void put(long addr) {
        Part part = null;
        int idx = 0;

        idx = findPart(addr);
        part = parts.get(idx);
        if (part.hasToBeResized()) {
            long[] next = Arrays.copyOfRange(parts.get(idx).getAddresses(), objectsPerPart / 2, objectsPerPart + objectsPerPart / 2);
            part.size.set(objectsPerPart / 2);

            Part newPart = new Part(next, objectsPerPart / 2);
            parts.add(idx + 1, newPart);

            // Check if put should happen in newly created array
            if (addr > newPart.getStartAddress()) {
                part = newPart;
            }
        }

        int curEle = part.getSize();
        part.getAddresses()[curEle] = addr;

        while (curEle > 0) {
            if (part.getAddresses()[curEle] < part.getAddresses()[curEle - 1]) {
                long temp = part.getAddresses()[curEle];
                part.getAddresses()[curEle] = part.getAddresses()[curEle - 1];
                part.getAddresses()[curEle - 1] = temp;
                curEle--;
                // TODO: Check equality --> exception throw
            } else {
                break;
            }
        }

        part.size.inc();

        if (curEle == 0) {
            part.resetStartAddress();
        }
    }

    public boolean removeSync(long addr) {
        int idx = 0;
        Part part = null;

        synchronized (parts) {
            idx = findPart(addr);
            part = parts.get(idx);
            part.lock();
        }

        if (part.getSize() > 0) {
            // Check if element is found
            int elePos = binarySearch(part.getAddresses(), 0, part.getSize(), addr);
            if (elePos < 0) {
                part.unlock();
                return false;
            }

            // Copy if not last element
            if (elePos + 1 != part.getSize()) {
                System.arraycopy(part.getAddresses(), elePos + 1, part.getAddresses(), elePos, part.getSize() - elePos - 1);
            }
            part.size.dec();

            if (elePos == 0 && part.getSize() > 0) {
                part.resetStartAddress();
            }

            part.unlock();
            return true;
        }
        part.unlock();
        return false;
    }

    /*
    public boolean remove(long addr) {
        int idx = 0;
        Part part = null;

        idx = findPart(addr);
        part = parts.get(idx);

        if (part.getSize() > 0) {
            // Check if element is found
            int elePos = binarySearch(part.getAddresses(), 0, part.getSize(), addr);
            if (elePos < 0) {
                return false;
            }

            // Copy if not last element
            if (elePos + 1 != part.getSize()) {
                System.arraycopy(part.getAddresses(), elePos + 1, part.getAddresses(), elePos,
                                 part.getSize() - elePos - 1);
            }
            part.size.dec();

            if (elePos == 0 && part.getSize() > 0) {
                part.resetStartAddress();
            }

            return true;
        }
        return false;
    }
    */

    /*
    public boolean containsWithSync(long addr) {
        Part part = null;
        int idx = 0;

        synchronized (parts) {
            idx = findPart(addr);
            part = parts.get(idx);
            part.lock();
        }

        boolean res = binarySearch(part.getAddresses(), 0, part.getSize(), addr) >= 0;
        // LOGGER.finest(Thread.currentThread().getName() + ":\tCONTAINS @ " +
        // idx + " (unsuccessful - not found)");
        part.unlock();
        return res;
    }
    */

    public boolean contains(long addr) {
        Part part = null;
        int idx = 0;

        idx = findPart(addr);
        part = parts.get(idx);
        boolean res = binarySearch(part.getAddresses(), 0, part.getSize(), addr) >= 0;
        // LOGGER.finest(Thread.currentThread().getName() + ":\tCONTAINS @ " +
        // idx + " (unsuccessful - not found)");
        return res;
    }

    public boolean containsWithoutSyncFullBinary(long addr) {
        Part part = null;
        int idx = 0;

        idx = binarySearch(partsStartAddresses, 0, partsStartAddresses.length, addr);
        if (idx < 0) {
            if (idx == -1) {
                // If address would be before first address, we know that the
                // collection does not contain it
                return false;
            } else {
                // -2 and smaller -> search for the address
                idx = -idx - 2;
            }
        }
        part = parts.get(idx);
        boolean res = binarySearch(part.getAddresses(), 0, part.getSize(), addr) >= 0;
        // LOGGER.finest(Thread.currentThread().getName() + ":\tCONTAINS @ " +
        // idx + " (unsuccessful - not found)");
        return res;
    }

    // Not thread-safe
    public int getObjectCount() {
        synchronized (parts) {
            return parts.stream().mapToInt(x -> x.getSize()).sum();
        }
    }

    // Not thread-safe
    public LongStream stream() {
        List<LongStream> streams = new ArrayList<>();
        for (int i = 0; i < parts.size(); i++) {
            streams.add(Arrays.stream(parts.get(i).getAddresses()).limit(parts.get(i).getSize()));
        }
        return streams.stream().flatMap(stream -> stream.boxed()).mapToLong(x -> x);
    }

    protected void cloneFrom(ComplexAddressCollection other) {
        parts.clear();
        setLastHeap(other.lastHeap);
        for (int i = 0; i < other.parts.size(); i++) {
            parts.add(other.parts.get(i).clone());
        }
    }

    public void visit(AddressVisitor visitor) {
        ArrayList<Thread> threads = new ArrayList<>();
        ParallelizationUtil.temporaryExecutorServiceBlocking((threadId, threadCount) -> {
            for (int partId = threadId; partId < parts.size(); partId += threadCount) {
                long[] arr = parts.get(partId).getAddresses();
                int size = parts.get(partId).getSize();
                for (int elementId = 0; elementId < size; elementId++) {
                    visitor.visit(arr[elementId]);
                }
            }
        });
    }

    public interface AddressVisitor {
        void visit(long address);
    }

    public AddressCollection without(AddressCollection other) {
        Measurement m = ApplicationStatistics.getInstance().createMeasurement("AddressCollection: without", Thread.currentThread().getName());

        assert other instanceof ComplexAddressCollection : "ComplexAddressCollections can only be set-operated with other ComplexAddressCollections";
        ComplexAddressCollection otherComplexAddressCollection = (ComplexAddressCollection) other;

        otherComplexAddressCollection.removeEmptyParts();

        ArrayList<Thread> threads = new ArrayList<>();

        // int nThreads = this.parts.size() / 20 + 1;
        int nThreads = MULTI_THREADED ? Consts.AVAILABLE_PROCESSORS() : 1;

        float partsPerThread = (1.0f * parts.size()) / nThreads;
        float currentIncrement = 0.0f;
        int to = 0;

        for (int threadNr = 0; threadNr < nThreads; threadNr++) {
            currentIncrement += partsPerThread;

            final int threadFrom = to;
            final int threadTo = threadFrom + (int) currentIncrement;

            to = threadTo;
            currentIncrement -= (int) currentIncrement;

            Thread t = new Thread(() -> {
                int upperBound = threadTo;
                int iPartNr = threadFrom;
                Part iPart = parts.get(iPartNr);

                // Thread is not assigned to any part (e.g. collection contains
                // less than Consts.AVAILABLE_PROCESSORS parts)
                if (iPartNr == upperBound) {
                    return;
                }

                int jPartNr = otherComplexAddressCollection.findPart(iPart.getStartAddress());
                Part jPart = otherComplexAddressCollection.parts.get(jPartNr);

                for (int i = 0, j = 0; ; ) {
                    if (i >= iPart.getSize()) {
                        if (iPartNr + 1 == upperBound) {
                            // Last i part finished, nothing more to remove
                            return;
                        }
                        // Get next i part
                        iPartNr++;
                        iPart = parts.get(iPartNr);
                        i = 0;
                        continue;
                    }
                    if (j == jPart.getSize()) {
                        if (jPartNr + 1 == otherComplexAddressCollection.parts.size()) {
                            // Last j part finished, nothing more to remove
                            return;
                        }
                        // Get next j part
                        jPartNr++;
                        jPart = otherComplexAddressCollection.parts.get(jPartNr);
                        j = 0;
                        continue;
                    }

                    int equals = 0;

                    while (i < iPart.getSize() && j < jPart.getSize() && jPart.addresses[j] == iPart.addresses[i]) {
                        i++;
                        j++;
                        equals++;
                    }

                    if (equals > 0) {
                        // If we had matches, move the index back and remove the
                        // equal parts
                        i -= equals;
                        removeAt(iPart, i, equals);
                    } else if (jPart.addresses[j] < iPart.addresses[i]) {
                        j++;
                    } else if (jPart.addresses[j] > iPart.addresses[i]) {
                        i++;
                    }
                }
            }, "Without Thread #" + threadNr);
            threads.add(t);
            t.start();
        }

        threads.forEach(x -> {
            try {
                x.join();
            } catch (InterruptedException ex) {
                ex.printStackTrace();
            }
        });

        removeEmptyParts();

        m.end();
        LOGGER.finest(m.toString());
        return this;
    }

    @Override
    public AddressCollection without(AddressCollection[] other) {
        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("ComplexAddressCollection - without - " + other.length);
        if (other.length == 1) {
            without(other[0]);
        }
        without(other[0].unionWith(Arrays.copyOfRange(other, 1, other.length)));
        m.end();
        return this;
    }

    private void removeAt(Part part, int pos, int n) {
        assert n > 0 : "At least one element has to be removed";
        assert pos + n <= part.getSize() : "Cannot exceed array length -> " + String.format("pos: %d, n: %d, from: %d, to: %d, length: %d, size: %d",
                                                                                            pos,
                                                                                            n,
                                                                                            pos + n,
                                                                                            pos,
                                                                                            part.getSize() - pos - n,
                                                                                            part.getSize());

        // Copy if not last element
        if (pos + n != part.getSize()) {
            System.arraycopy(part.getAddresses(), pos + n, part.getAddresses(), pos, part.getSize() - pos - n);
        }
        part.size.sub(n);

        if (pos == 0 && part.getSize() > 0) {
            part.resetStartAddress();
        }
    }

    public AddressCollection unionWith(AddressCollection other) {
        Measurement m = ApplicationStatistics.getInstance().createMeasurement("AddressCollection: UNION", Thread.currentThread().getName());

        assert other instanceof ComplexAddressCollection : "ComplexAddressCollections can only be set-operated with other ComplexAddressCollections";
        ComplexAddressCollection otherComplexAddressCollection = (ComplexAddressCollection) other;

        otherComplexAddressCollection.removeEmptyParts();

        int iPartNr = 0;
        int jPartNr = 0;
        Part iPart = parts.get(iPartNr);
        Part jPart = otherComplexAddressCollection.parts.get(jPartNr);
        boolean fixAdd = false;

        for (int i = 0, j = 0; ; ) {
            if (i >= iPart.getSize()) {
                if (iPartNr + 1 == parts.size()) {
                    // Last i part finished, add all remaining values from j
                    fixAdd = true;
                } else {
                    // Get next i part
                    iPartNr++;
                    iPart = parts.get(iPartNr);
                    i = 0;
                    continue;
                }
            }
            if (j >= jPart.getSize()) {
                if (jPartNr + 1 == otherComplexAddressCollection.parts.size()) {
                    // Last j part finished, nothing more to add
                    break;
                }
                // Get next j part
                jPartNr++;
                jPart = otherComplexAddressCollection.parts.get(jPartNr);
                j = 0;
                continue;
            }

            if (fixAdd) {
                // End of i arrays reached
                // Just add all remaining j arrays
                while (jPartNr < otherComplexAddressCollection.parts.size()) {
                    // Set jPart
                    jPart = otherComplexAddressCollection.parts.get(jPartNr);

                    // Clone non-empty j part and add it
                    assert jPart.getSize() - j > 0 : "At least one object has to be cloned! Part should also not be empty - is it: " + (jPart.getSize() == 0 ?
                                                                                                                                        "YES!" :
                                                                                                                                        ("No, has size " + jPart
                                                                                                                                                .getSize()));

                    long[] next = Arrays.copyOfRange(jPart.getAddresses(), j, objectsPerPart + j);
                    Part newPart = new Part(next, jPart.getSize() - j);
                    parts.add(newPart);

                    // Set index to next j part
                    jPartNr++;
                    j = 0;
                }
                // Everything added, end adding loop
                break;
            } else {
                int smaller = 0;

                while (i + smaller < iPart.getSize() && j + smaller < jPart.getSize() && iPart.getSize() + smaller <= objectsPerPart && jPart.addresses[j + smaller] <
                        iPart
                                .addresses[i]) {
                    smaller++;
                }

                if (smaller > 0) {
                    boolean previousPartUsed = putInPart(iPart, i, iPartNr, jPart, j, smaller);
                    if (!previousPartUsed) {
                        i++;
                    }
                    j += smaller;
                } else if (jPart.addresses[j] == iPart.addresses[i]) {
                    i++;
                    j++;
                } else if (jPart.addresses[j] > iPart.addresses[i]) {
                    i++;
                } else {
                    assert false : "At least one index has to be moved";
                }
            }
        }

        removeEmptyParts();

        m.end();
        LOGGER.finest(m.toString());
        return this;
    }

    public AddressCollection unionWith(AddressCollection[] other) {
        throw new UnsupportedOperationException("Not yet implemented");
    }

    private void removeEmptyParts() {
        ArrayList<Part> cleaned = new ArrayList<>();
        for (Part p : parts) {
            if (p.getSize() > 0) {
                cleaned.add(p);
            }
        }
        if (cleaned.size() == 0) {
            cleaned.add(new Part());
        }
        parts = cleaned;
        // printAverageFillRate();
    }

    private boolean putInPart(Part toPart, int toPos, int toPartNr, Part fromPart, int fromPos, int n) {
        boolean previousPartUsed = false;

        assert toPos < objectsPerPart : "Cannot write after array";

        // Expand current array if needed
        if (toPart.hasToBeResized(n)) {
            long[] next = Arrays.copyOfRange(toPart.getAddresses(), toPos, objectsPerPart + toPos);
            Part newPart = new Part(next, toPart.getSize() - toPos);
            parts.add(toPartNr + 1, newPart);

            toPart.size.set(toPos);
        }

        assert !toPart.hasToBeResized(n) : "At least after resizing data has to fit into array";

        // if (toPos == 0 && toPartNr != 0) { // Try to add in last part
        // Part lastPart = this.parts.get(toPartNr - 1);
        // if (!lastPart.hasToBeResized(n)) {
        // toPart = lastPart;
        // toPos = toPart.getSize();
        // previousPartUsed = true;
        // }
        // }

        // Shift to right only needed if
        if (toPart.getSize() != toPos) {
            System.arraycopy(toPart.addresses, toPos, toPart.addresses, toPos + n, toPart.getSize() - toPos);
        }
        for (int copyIdx = 0; copyIdx < n; copyIdx++) {
            toPart.addresses[toPos + copyIdx] = fromPart.addresses[fromPos + copyIdx];
        }

        toPart.size.add(n);

        if (toPos == 0) {
            toPart.resetStartAddress();
        }

        return previousPartUsed;
    }

    public AddressCollection intersectWith(AddressCollection other) {
        Measurement m = ApplicationStatistics.getInstance().createMeasurement("AddressCollection: INTERSECT", Thread.currentThread().getName());

        assert other instanceof ComplexAddressCollection : "ComplexAddressCollections can only be set-operated with other ComplexAddressCollections";
        ComplexAddressCollection otherComplexAddressCollection = (ComplexAddressCollection) other;

        otherComplexAddressCollection.removeEmptyParts();

        int iPartNr = 0;
        int jPartNr = 0;
        Part iPart = parts.get(iPartNr);
        Part jPart = otherComplexAddressCollection.parts.get(jPartNr);

        for (int i = 0, j = 0; ; ) {
            if (i >= iPart.getSize()) {
                if (iPartNr + 1 == parts.size()) {
                    // Last i part finished, nothing more to remove
                    break;
                } else {
                    // Get next i part
                    iPartNr++;
                    iPart = parts.get(iPartNr);
                    i = 0;
                    continue;
                }
            }
            if (j == jPart.getSize()) {
                if (jPartNr + 1 == otherComplexAddressCollection.parts.size()) {
                    // Last j part finished, remove all remaining elements
                    // Cut current part is current position
                    iPart.size.set(i);
                    // Remove remaining parts
                    int partsAfterClear = iPartNr + 1;
                    while (parts.size() > partsAfterClear) {
                        parts.remove(partsAfterClear);
                    }
                    break;
                }
                // Get next j part
                jPartNr++;
                jPart = otherComplexAddressCollection.parts.get(jPartNr);
                j = 0;
                continue;
            }

            int greater = 0;

            while (i < iPart.getSize() && j < jPart.getSize() && jPart.addresses[j] > iPart.addresses[i]) {
                greater++;
                i++;
            }

            if (greater > 0) {
                i -= greater;
                removeAt(iPart, i, greater);
            } else if (jPart.addresses[j] < iPart.addresses[i]) {
                j++;
            } else if (jPart.addresses[j] == iPart.addresses[i]) {
                i++;
                j++;
            }
        }
        removeEmptyParts();

        m.end();
        LOGGER.finest(m.toString());
        return this;
    }

    public ComplexAddressCollection cloneWithout(AddressCollection other) {
        ComplexAddressCollection cloned = clone();
        cloned.without(other);
        return cloned;
    }

    public ComplexAddressCollection cloneAndIntersectWith(AddressCollection other) {
        ComplexAddressCollection cloned = clone();
        cloned.intersectWith(other);
        return cloned;
    }

    public MapGroupingNode classify(ClassifierChain classifier, ConditionalClassifierList conditionalClassifier, List<Filter> filters) {
        partsStartAddresses = new long[parts.size()];
        initPartsStartAddresses();

        Measurement m = ApplicationStatistics.getInstance().createMeasurement("AddressCollection: Classify", Thread.currentThread().getName());
        MapGroupingNode group = lastHeap.toObjectStream().filter(new Filter() {
            @Override
            protected Boolean classify() throws Exception {
                return containsWithoutSyncFullBinary(address());
            }
        }).filter(filters == null ? new Filter[0] : filters.toArray(new Filter[0])).groupMapParallel(classifier, conditionalClassifier);
        m.end();
        return group;
    }

    public long getBytes() {
        partsStartAddresses = new long[parts.size()];
        initPartsStartAddresses();

        class CounterVisitor extends Counter implements ObjectVisitor {

            @Override
            public void visit(long address,
                              ObjectInfo objectInfo,
                              SpaceInfo space,
                              AllocatedType type,
                              int size,
                              boolean isArray,
                              int arrayLength,
                              AllocationSite allocationSite,
                              long[] pointedFrom,
                              long[] pointsTo,
                              EventType eventType,
                              List<RootPtr> rootPtrs) {
                add(size);
            }
        }
        ObjectStream.ThreadVisitorGenerator<CounterVisitor> gen = () -> new CounterVisitor();

        Measurement m = ApplicationStatistics.getInstance().createMeasurement("AddressCollection: Classify", Thread.currentThread().getName());
        Counter sum = new Counter();
        lastHeap.toObjectStream().filter(new Filter() {
            @Override
            protected Boolean classify() throws Exception {
                return containsWithoutSyncFullBinary(address());
            }
        }).forEachParallel(gen, ObjectVisitor.Settings.NO_INFOS).forEach(sum::add);
        m.end();
        return sum.get();
    }

    private void initPartsStartAddresses() {
        for (int i = 0; i < parts.size(); i++) {
            partsStartAddresses[i] = parts.get(i).getStartAddress();
        }
    }

    // Not thread-safe
    public void setLastHeap(DetailedHeap lastHeap) {
        this.lastHeap = lastHeap;
    }

    // Not thread-safe
    /// This method can be used during development to ensure data structure
    // consistency. Otherwise it is not used
    @SuppressWarnings("unused")
    private void verifyArrays() {
        synchronized (parts) {
            long last = -1;
            for (int i = 0; i < parts.size(); i++) {
                long[] arr = parts.get(i).getAddresses();
                int size = parts.get(i).getSize();
                long startAddr = parts.get(i).getStartAddress();

                for (int ele = 0; ele < size; ele++) {
                    assert last < arr[ele] : "Elements must be sorted ascending";
                    last = arr[ele];
                }

                if (size > 0) {
                    assert startAddr == arr[0] : "Start address must be up to date";
                }
            }

            last = -1;
            for (int i = 0; i < parts.size(); i++) {
                assert last < parts.get(i).getStartAddress() : "Start addresses must be sorted ascending! Problem @ " + i + " in [" + parts.stream()
                                                                                                                                           .map(part -> part.startAddress
                                                                                                                                                   .toString())
                                                                                                                                           .reduce((a, b) -> a + ", " + b)
                                                                                                                                           .get() + "]";
                last = parts.get(i).getStartAddress();
            }
        }
    }

    // This method may be used for evaluation
    @SuppressWarnings("unused")
    private void printAverageFillRate() {
        System.out.println(parts.stream().mapToInt(part -> part.getSize()).sum() / parts.size());
    }

    private static int binarySearch(long[] a, int fromIndex, int toIndex, long key) {
        int low = fromIndex;
        int high = toIndex - 1;

        while (low <= high) {
            int mid = (low + high) / 2;
            long midVal = a[mid];

            if (midVal < key) {
                low = mid + 1;
            } else if (midVal > key) {
                high = mid - 1;
            } else {
                return mid; // key found
            }
        }
        return -(low + 1); // key not found.
    }

    @Override
    public void writeToMetaData(DataOutputStream dos) throws OperationNotSupportedException {
        throw new OperationNotSupportedException("Not yet implemented");
    }

    public AddressCollectionIterator getIterator() {
        return new ComplexAddressCollectionIterator();
    }

    public DetailedHeap getLastHeap() {
        return lastHeap;
    }

    @Override
    public ComplexAddressCollection clone() {
        synchronized (parts) {
            ComplexAddressCollection dolly = new ComplexAddressCollection(name);
            dolly.cloneFrom(this);
            return dolly;
        }
    }

    public static ComplexAddressCollection from(long[] addresses, String name) {
        return from(addresses, name, 30_000);
    }

    public static ComplexAddressCollection from(long[] addresses, String name, int objectsPerPart) {
        ComplexAddressCollection collection = new ComplexAddressCollection(name);
        for (int i = 0; i < addresses.length / objectsPerPart + 1; i++) {
            int partSize = Math.min(addresses.length - i * objectsPerPart, objectsPerPart);
            Part p = collection.new Part(Arrays.copyOfRange(addresses, i * objectsPerPart, (i + 1) * objectsPerPart), partSize);
            collection.parts.add(p);
        }
        return collection;
    }
}
