/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.diff;

import at.jku.mevss.anttracks.classification.Classifier;
import at.jku.mevss.anttracks.classification.nodes.GroupingNode;
import at.jku.mevss.anttracks.classification.nodes.MapGroupingNode;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

public class DiffGrouping {
    private final GroupingNode perm;
    private final GroupingNode born;
    private final GroupingNode died;
    private final GroupingNode temp;
    private final Object key;
    private final int level;
    private final DiffGrouping parent;

    public DiffGrouping(GroupingNode perm, GroupingNode born, GroupingNode died, GroupingNode temp) {
        this(perm, born, died, temp, "Overall", 0, null);
    }

    public DiffGrouping(GroupingNode perm, GroupingNode born, GroupingNode died, GroupingNode temp, Object key, int level, DiffGrouping parent) {
        this.perm = perm;
        this.born = born;
        this.died = died;
        this.temp = temp;
        this.key = key;
        this.level = level;
        this.parent = parent;
    }

    public List<DiffGrouping> getChildren() {
        Map<String, Object[]> unionChildrenKeys = new HashMap<>();

        if (perm != null) {
            perm.getChildren()
                .stream()
                .filter(child -> !unionChildrenKeys.containsKey(Arrays.toString(child.getFullKey())))
                .forEach(child -> unionChildrenKeys.put(Arrays.toString(child.getFullKey()), child.getFullKey()));
        }
        if (born != null) {
            born.getChildren()
                .stream()
                .filter(child -> !unionChildrenKeys.containsKey(Arrays.toString(child.getFullKey())))
                .forEach(child -> unionChildrenKeys.put(Arrays.toString(child.getFullKey()), child.getFullKey()));
        }
        if (died != null) {
            died.getChildren()
                .stream()
                .filter(child -> !unionChildrenKeys.containsKey(Arrays.toString(child.getFullKey())))
                .forEach(child -> unionChildrenKeys.put(Arrays.toString(child.getFullKey()), child.getFullKey()));
        }
        if (temp != null) {
            temp.getChildren()
                .stream()
                .filter(child -> !unionChildrenKeys.containsKey(Arrays.toString(child.getFullKey())))
                .forEach(child -> unionChildrenKeys.put(Arrays.toString(child.getFullKey()), child.getFullKey()));
        }

        return unionChildrenKeys.values()
                                .stream()
                                .map(key -> new DiffGrouping(perm == null ? null : (MapGroupingNode) perm.getChild(key[level + 1]),
                                                             born == null ? null : (MapGroupingNode) born.getChild(key[level + 1]),
                                                             died == null ? null : (MapGroupingNode) died.getChild(key[level + 1]),
                                                             temp == null ? null : (MapGroupingNode) temp.getChild(key[level + 1]),
                                                             key[level + 1],
                                                             level + 1,
                                                             this))
                                .collect(Collectors.toList());
    }

    public GroupingNode getPerm() {
        return perm;
    }

    public GroupingNode getBorn() {
        return born;
    }

    public GroupingNode getDied() {
        return died;
    }

    public GroupingNode getTemp() {
        return temp;
    }

    public Object getKey() {
        return key;
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("Perm:\n");
        sb.append(perm);
        sb.append("\nBorn:\n");
        sb.append(born);
        sb.append("\nDied:\n");
        sb.append(died);
        sb.append("\nTemp:\n");
        sb.append(temp);
        return sb.toString();
    }

    public void clear() {
        if (perm != null) {
            perm.clear();
        }
        if (born != null) {
            born.clear();
        }
        if (died != null) {
            died.clear();
        }
        if (temp != null) {
            temp.clear();
        }
    }

    public int getSubTreeLevel() {
        if (perm != null) {
            return perm.getSubTreeLevel();
        }
        if (born != null) {
            return born.getSubTreeLevel();
        }
        if (temp != null) {
            return temp.getSubTreeLevel();
        }
        if (died != null) {
            return died.getSubTreeLevel();
        }
        assert false : "At least one object-set must be non-null";
        return -1;
    }

    public Classifier<?> getClassifier() {
        if (perm != null) {
            return perm.getClassifier();
        }
        if (born != null) {
            return born.getClassifier();
        }
        if (temp != null) {
            return temp.getClassifier();
        }
        if (died != null) {
            return died.getClassifier();
        }
        assert false : "At least one object-set must be non-null";
        return null;
    }

    public DiffGrouping getParent() {
        return parent;
    }

    public void sampleTopDown() {
        perm.sampleTopDown();
        born.sampleTopDown();
        died.sampleTopDown();
        temp.sampleTopDown();
    }
}
