/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.diff;

import at.jku.mevss.anttracks.diff.Interval.IntervalInfo;
import at.jku.mevss.anttracks.diff.Interval.IntervalListener;
import at.jku.mevss.anttracks.heap.*;
import at.jku.mevss.anttracks.heap.io.HeapIndexReader;
import at.jku.mevss.anttracks.heap.io.HeapPosition;
import at.jku.mevss.anttracks.heap.io.MetaDataReaderConfig;
import at.jku.mevss.anttracks.heap.io.MetaDataWriterListener;
import at.jku.mevss.anttracks.heap.statistics.MemoryConsumption;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.parser.ParsingInfo;
import at.jku.mevss.anttracks.parser.TraceParser;
import at.jku.mevss.anttracks.parser.TraceParsingEventHandler;
import at.jku.mevss.anttracks.parser.TraceScannerFactory;
import at.jku.mevss.anttracks.parser.heap.HeapBuilder;
import at.jku.mevss.anttracks.util.TraceException;

import java.io.IOException;
import java.util.concurrent.CopyOnWriteArrayList;

public class HeapDiffTraceParser extends TraceParser<HeapDiff> {

    private final static boolean VERBOSE = false;

    private DetailedHeap heap;

    private final CopyOnWriteArrayList<IntervalListener> intervalListener = new CopyOnWriteArrayList<>();

    private IntervalListener forwardingIntervalListener = new IntervalListener() {

        @Override
        public void gcStartHeapStateAdded(long time, MemoryConsumption perm, MemoryConsumption died, MemoryConsumption born, MemoryConsumption temp) {
            intervalListener.forEach(listener -> listener.gcStartHeapStateAdded(time, perm, died, born, temp));
        }

        @Override
        public void gcEndHeapStateAdded(long time,
                                        MemoryConsumption perm,
                                        MemoryConsumption died,
                                        MemoryConsumption born,
                                        MemoryConsumption temp,
                                        IntervalInfo intervalInfo) {
            intervalListener.forEach(listener -> listener.gcEndHeapStateAdded(time, perm, died, born, temp, intervalInfo));

        }

        @Override
        public void intervalEnded(long time,
                                  DetailedHeap heap,
                                  MemoryConsumption perm,
                                  MemoryConsumption died,
                                  MemoryConsumption born,
                                  MemoryConsumption temp,
                                  IntervalInfo intervalInfo) {
            intervalListener.forEach(listener -> listener.intervalEnded(time, heap, perm, died, born, temp, intervalInfo));
        }
    };

    public void addIntervalListener(IntervalListener listener) {
        intervalListener.add(listener);
    }

    private HeapListener heapListener = new HeapListener() {
        @Override
        public void phaseChanging(DetailedHeap heap,
                                  GarbageCollectionType from,
                                  GarbageCollectionType to,
                                  GarbageCollectionCause cause,
                                  boolean concurrent,
                                  boolean failed,
                                  long time,
                                  long position,
                                  long length,
                                  boolean inSelectedTimeWindow) {
            handleInterval(time, to != GarbageCollectionType.MUTATOR);
        }

        @Override
        public void phaseChanged(DetailedHeap heap,
                                 GarbageCollectionType from,
                                 GarbageCollectionType to,
                                 GarbageCollectionCause cause,
                                 boolean concurrent,
                                 boolean failed,
                                 long time,
                                 long position,
                                 long length,
                                 boolean inSelectedTimeWindow) {
            // Nothing to do
        }

        @Override
        public void close(DetailedHeap heap) {
            if (workspace != null) {
                workspace.close(heap);
            }
        }

        private void handleInterval(long time, boolean gcStart) {
            workspace.add(heap, time, gcStart);
            if (VERBOSE) {
                workspace.printIntervalState(symbols);
            }
        }
    };

    public HeapDiffTraceParser(Symbols symbols, MetaDataReaderConfig readerConfig, GarbageCollectionLookup gcFrom, GarbageCollectionLookup gcTo) throws Exception {
        super(symbols, readerConfig, gcFrom, gcTo);
    }

    public HeapDiffTraceParser(Symbols symbols, MetaDataReaderConfig readerConfig, long fromTime, long toTime) throws Exception {
        super(symbols, readerConfig, fromTime, toTime);
    }

    @Override
    protected void doWorkspaceCompletion(HeapDiff workspace) throws TraceException {
        heap.complete();
    }

    @Override
    protected void doRemoveListenersOnCompletion(HeapDiff workspace) {
        heap.removeAllListeners();
        heapListener = null;
        workspace.removeAllListener();
        forwardingIntervalListener = null;
        intervalListener.clear();
    }

    @Override
    protected HeapDiff generateWorkspaceFromMetaData(HeapIndexReader heapIndexReader, HeapPosition heapPosition) throws IOException {
        if (heapPosition.fileName != -1) {
            heap = heapIndexReader.getHeap(heapPosition.fileName, symbols);
        } else {
            heap = HeapBuilder.constructHeap(symbols);
        }

        heap.addListener(heapListener);

        HeapDiff diff = new HeapDiff(fromTime, toTime);
        diff.addIntervalListener(forwardingIntervalListener);

        // This is the case if the first selected GC point has exaclty been reconstructed from a heap dump file.
        // The GC-end event at which the heap dump file got created will not be read from the trace anymore, therefore
        // we have to start the diffing here manually.
        if (heapPosition.fromPosition == heapPosition.toPosition) {
            // We only dump heaps on GC-end, so this cannot be a GC start
            // Start the diffing with GC-end
            diff.add(heap, fromTime, false);
        }

        return diff;
    }

    @Override
    protected HeapDiff generatePlainWorkspace(TraceScannerFactory factory) throws IOException {
        heap = HeapBuilder.constructHeap(symbols);
        if (writerConfig != null) {
            heap.addListener(new MetaDataWriterListener(writerConfig, symbols));
        }

        heap.addListener(heapListener);

        HeapDiff diff = new HeapDiff(fromTime, toTime);
        diff.addIntervalListener(forwardingIntervalListener);
        return diff;
    }

    @Override
    protected void doParseCleanupAfterSuccessfulParse(HeapDiff workspace) throws TraceException {
        heap.assignLabsIncorrectlyAssumedToBeFillers(false);
    }

    @Override
    protected TraceParsingEventHandler createMainEventHandler(ParsingInfo parsingInfo) {
        return new HeapDiffEventHandler(heap, workspace, symbols, parsingInfo);
    }
}
