package at.jku.mevss.anttracks.diff;

import at.jku.mevss.anttracks.classification.ClassifierChain;
import at.jku.mevss.anttracks.classification.ConditionalClassifierList;
import at.jku.mevss.anttracks.classification.Filter;
import at.jku.mevss.anttracks.classification.nodes.GroupingNode;
import at.jku.mevss.anttracks.classification.nodes.MapGroupingNode;
import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.util.ApplicationStatistics;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.stream.LongStream;

public class SimpleAddressCollection implements AddressCollection {
    private String name;
    private long[] data;
    private int count;
    private boolean sorted;
    private DetailedHeap lastHeap;

    public SimpleAddressCollection(String name) {
        this.name = name;
        data = new long[1];
        count = 0;
        sorted = true;
        lastHeap = null;
    }

    @Override
    public AddressCollectionIterator getIterator() {
        return new AddressCollectionIterator() {

            int i = 0;

            @Override
            public boolean hasNext() {
                return i < count;
            }

            @Override
            public long next() {
                return data[i++];
            }
        };
    }

    @Override
    public void put(long address) {
        if (count == data.length) {
            int newSize = count + (count >> 1);
            if (newSize == count) {
                newSize++;
            }
            data = Arrays.copyOfRange(data, 0, newSize);
        }
        data[count] = address;
        count++;
        sorted = false;
    }

    @Override
    public boolean contains(long address) {
        return Arrays.binarySearch(data, 0, count, address) >= 0;
    }

    @Override
    public void clear() {
        for (int i = 0; i < count; i++) {
            data[i] = 0;
        }
        count = 0;
        sorted = true;
    }

    @Override
    public LongStream stream() {
        return Arrays.stream(data).limit(count);
    }

    @Override
    public AddressCollection unionWith(AddressCollection other) {
        // TODO: Check if simple solution is faster
        // return unionWith(new AddressCollection[]{other});

        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("SimpleAddressCollection - union");

        assert other instanceof SimpleAddressCollection : "SimpleAddressCollections can only be merged with SimpleAddressCollections";

        if (!this.sorted) {
            sort();
        }
        if (!other.isSorted()) {
            other.sort();
        }

        SimpleAddressCollection o = (SimpleAddressCollection) other;

        long[] newData = new long[this.count + o.count];
        int newCount = 0;

        int i = 0, j = 0;
        while (i < this.count && j < o.count) {
            if (this.data[i] < o.data[j]) {
                newData[newCount] = this.data[i];
                i++;
                newCount++;
            } else if (this.data[i] == o.data[j]) {
                newData[newCount] = this.data[i];
                newCount++;
                i++;
                j++;
            } else if (this.data[i] > o.data[j]) {
                newData[newCount] = o.data[j];
                j++;
                newCount++;
            }
        }
        while (i < this.count) {
            newData[newCount] = this.data[i];
            i++;
            newCount++;
        }
        while (j < o.count) {
            newData[newCount] = o.data[j];
            j++;
            newCount++;
        }
        this.data = newData;
        this.count = newCount;
        this.sorted = true;
        m.end();
        return this;
    }

    @Override
    public AddressCollection unionWith(AddressCollection[] other) {
        if (other.length == 0) {
            return this;
        }
        if (other.length == 1) {
            return unionWith(other[0]);
        }

        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("SimpleAddressCollection - union - " + other.length);
        for (int i = 0; i < other.length; i++) {
            assert other[i] instanceof SimpleAddressCollection : "SimpleAddressCollections can only be merged with SimpleAddressCollections";
        }

        int finalSize = this.getObjectCount();
        for (AddressCollection o : other) {
            finalSize += o.getObjectCount();
        }

        int insertionPoint = this.getObjectCount();

        // Increase data size to fit all children
        this.data = Arrays.copyOf(this.data, finalSize);
        this.count = finalSize;
        // Copy all children
        for (AddressCollection o : other) {
            System.arraycopy(((SimpleAddressCollection) o).data, 0, this.data, insertionPoint, o.getObjectCount());
            insertionPoint += o.getObjectCount();
        }
        // Sort array
        sort();
        // Eliminate duplicates
        long[] newData = new long[finalSize];
        this.count = 0;
        for (int fromIdx = 0; fromIdx < finalSize; fromIdx++) {
            if (fromIdx + 1 < finalSize && this.data[fromIdx] == this.data[fromIdx + 1]) {
                continue;
            }
            newData[this.count++] = this.data[fromIdx];
        }
        this.data = newData;
        this.sorted = true;
        m.end();
        return this;
    }

    @Override
    public AddressCollection intersectWith(AddressCollection other) {
        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("SimpleAddressCollection - intersect");
        assert other instanceof SimpleAddressCollection : "SimpleAddressCollections can only be merged with SimpleAddressCollections";
        SimpleAddressCollection o = (SimpleAddressCollection) other;

        if (!sorted) {
            sort();
        }
        if (!o.sorted) {
            o.sort();
        }

        int matches = 0;
        for (int i = 0, j = 0; i < this.count && j < o.count; ) {
            if (data[i] == o.data[j]) {
                data[matches] = data[i];
                matches++;
                i++;
                j++;
            } else if (data[i] < o.data[j]) {
                i++;
            } else if (data[i] > o.data[j]) {
                j++;
            }
        }
        this.count = matches;
        this.sorted = true;
        m.end();
        return this;
    }

    @Override
    public AddressCollection without(AddressCollection other) {
        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("SimpleAddressCollection - without");
        assert other instanceof SimpleAddressCollection : "SimpleAddressCollections can only be merged with SimpleAddressCollections";
        SimpleAddressCollection o = (SimpleAddressCollection) other;

        if (!sorted) {
            sort();
        }
        if (!o.sorted) {
            o.sort();
        }

        int i = 0, j = 0, matches = 0;
        for (; i < this.count && j < o.count; ) {
            if (data[i] == o.data[j]) {
                i++;
                j++;
            } else if (data[i] < o.data[j]) {
                data[matches] = data[i];
                matches++;
                i++;
            } else if (data[i] > o.data[j]) {
                j++;

            }
        }
        if (j == o.count) {
            // Case that other data has reached the end -> keep remaining this.data
            for (; i < this.count; i++) {
                data[matches] = data[i];
                matches++;
            }
        }
        this.count = matches;
        this.sorted = true;
        m.end();
        return this;
    }

    @Override
    public AddressCollection without(AddressCollection[] other) {
        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("SimpleAddressCollection - without - " + other.length);
        if (other.length == 0) {
            return this;
        } else if (other.length == 1) {
            without(other[0]);
        } else {
            without(other[0].unionWith(Arrays.copyOfRange(other, 1, other.length)));
        }
        m.end();
        return this;
    }

    @Override
    public AddressCollection cloneAndIntersectWith(AddressCollection other) {
        SimpleAddressCollection cloned = clone();
        cloned.intersectWith(other);
        return cloned;
    }

    @Override
    public AddressCollection cloneWithout(AddressCollection other) {
        SimpleAddressCollection cloned = clone();
        cloned.without(other);
        return cloned;
    }

    @Override
    public SimpleAddressCollection clone() {
        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("SimpleAddressCollection - clone");
        SimpleAddressCollection dolly = new SimpleAddressCollection("");
        dolly.cloneFrom(this);
        m.end();
        return dolly;
    }

    private void cloneFrom(SimpleAddressCollection parent) {
        this.name = parent.name;
        this.data = new long[parent.count];
        this.count = parent.count;
        this.sorted = parent.sorted;
        for (int i = 0; i < count; i++) {
            data[i] = parent.data[i];
        }
        this.lastHeap = parent.lastHeap;
    }

    @Override
    public void setLastHeap(DetailedHeap lastHeap) {
        this.lastHeap = lastHeap;
    }

    @Override
    public GroupingNode classify(ClassifierChain classifier, ConditionalClassifierList conditionalClassifier, List<Filter> filters) {
        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("SimpleAddressCollection - classify");

        MapGroupingNode result = lastHeap.toObjectStream().filter(new Filter() {
            @Override
            protected Boolean classify() throws Exception {
                return contains(address());
            }
        }).filter(filters != null ? filters.toArray(new Filter[filters.size()]) : null).groupMapParallel(classifier, conditionalClassifier);
        m.end();
        return result;
    }

    @Override
    public int getObjectCount() {
        return count;
    }

    @Override
    public long getBytes() {
        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("SimpleAddressCollection - getBytes");
        long bytes = classify(new ClassifierChain(), null, null).getByteCount();
        m.end();
        return bytes;
    }

    public AddressCollection sort() {
        if (!sorted) {
            ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("SimpleAddressCollection - sort");
            Arrays.sort(data, 0, count);
            sorted = true;
            m.end();
        }
        return this;
    }

    @Override
    public boolean isSorted() {
        return sorted;
    }

    public static SimpleAddressCollection from(long[] addresses, String name) {
        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("SimpleAddressCollection - from array");
        SimpleAddressCollection collection = new SimpleAddressCollection(name);
        collection.data = addresses;
        collection.count = addresses.length;
        collection.sorted = false;
        m.end();
        return collection;
    }

    @Override
    public void writeToMetaData(DataOutputStream dos) {
        try {
            dos.writeUTF(this.name);
            dos.writeInt(count);
            dos.writeBoolean(sorted);
            for (int i = 0; i < count; i++) {
                dos.writeLong(data[i]);
            }
        } catch (IOException e) {
            e.printStackTrace();
        }

    }

    public static AddressCollection readFromMetaData(DataInputStream dis) {
        try {
            SimpleAddressCollection sac = new SimpleAddressCollection(dis.readUTF());
            sac.count = dis.readInt();
            sac.sorted = dis.readBoolean();
            sac.data = new long[sac.count];
            for (int i = 0; i < sac.count; i++) {
                sac.data[i] = dis.readLong();
            }
            return sac;
        } catch (IOException e) {
            e.printStackTrace();
        }
        return null;
    }
}
