/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap;

import at.jku.mevss.anttracks.classification.Classifier;
import at.jku.mevss.anttracks.classification.ClassifierChain;
import at.jku.mevss.anttracks.classification.ConditionalClassifierList;
import at.jku.mevss.anttracks.classification.Filter;
import at.jku.mevss.anttracks.classification.nodes.MapGroupingNode;

import java.util.List;
import java.util.stream.Stream;

public interface ObjectStream {
    public interface ThreadVisitorGenerator<I extends ObjectVisitor> {
        public abstract I generate();
    }

    public interface IterationListener {
        public abstract void objectsIterated(long objectCount);
    }

    /**
     * Returns a collection containing all {@link HeapObjectFilter}s to apply
     *
     * @return A collection containing all {@link HeapObjectFilter}s to apply
     */
    //    Collection<HeapObjectFilter> getFilter();

    /**
     * Adds the given {@link IterationListener}.
     *
     * @param ll The {@link IterationListener} to add
     */
    public void addLabListener(IterationListener ll);

    /**
     * Removes the given {@link IterationListener}. If the {@link IterationListener} is not
     * present, nothing happens.
     *
     * @param ll The {@link IterationListener} to remove
     */
    public void removeLabListener(IterationListener ll);

    /**
     * Adds the given filters to the filter queue
     *
     * @param filter The filters to add
     * @return The {@link ObjectStream} with the given filters added
     */
    public ObjectStream filter(Filter... filter);

    /**
     * Sorts the stream by its default sorting parameter
     */
    public ObjectStream sorted();

    /**
     * Groups single-threaded according to the given classifiers, taking into
     * account all filters in the filter queue
     *
     * @param classifier            Classifiers applied to all objects
     * @param conditionalClassifier Classifiers applied to a certain subgroup of objects
     * @return The resulting grouping
     */
    public MapGroupingNode groupMap(ClassifierChain classifier, ConditionalClassifierList conditionalClassifier);

    /**
     * Groups multi-threaded according to the given classifiers, taking into
     * account all filters in the filter queue
     *
     * @param classifier            Classifiers applied to all objects
     * @param conditionalClassifier Classifiers applied to a certain subgroup of objects
     * @return The resulting grouping
     */
    public MapGroupingNode groupMapParallel(ClassifierChain classifier, ConditionalClassifierList conditionalClassifier);

    /**
     * Visits all objects which are not excluded by at least one filter in the
     * filter queue
     *
     * @param visitor The visitor to be executed on each visited object
     */
    public abstract void forEach(ObjectVisitor visitor, ObjectVisitor.Settings visitorSettings);

    public abstract <I extends ObjectVisitor> List<I> forEachParallel(ThreadVisitorGenerator<I> threadLocalDataGenerator, ObjectVisitor.Settings visitorSettings);

    public abstract <A> Stream<A> map(Classifier<A> classifier);
}
