package at.jku.mevss.anttracks.heap.datastructures;

import at.jku.mevss.anttracks.heap.datastructures.cocor.DataStructurePart;
import at.jku.mevss.anttracks.heap.datastructures.cocor.PointsToType;
import at.jku.mevss.anttracks.heap.symbols.AllocatedType;
import at.jku.mevss.anttracks.heap.symbols.AllocatedTypes;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import static at.jku.mevss.anttracks.util.DataStructureUtil.*;

public class DataStructureDefinitions {
    Logger LOGGER = Logger.getLogger(DataStructureDefinitions.class.getName());

    public final List<DataStructurePart> parts;
    public final Map<String, DataStructureDefinition> definitions;

    public DataStructureDefinitions() {
        parts = new ArrayList<>();
        definitions = new HashMap<>();
    }

    public void addDataStructurePart(DataStructurePart dsp) {
        parts.add(dsp);
    }

    public void resolveDataStructureParts(AllocatedTypes allocatedTypes) {
        // For each type that has no matching DataStructurePart, create it
        allocatedTypes.forEach(allocatedType -> {
            if (parts.stream().noneMatch(dsp -> dsp.matches(allocatedType))) {
                DataStructurePart part = new DataStructurePart(allocatedType.getFullyQualifiedName(true), false);
                if (allocatedType.isArray()) {
                    // Arrays follow
                    part.addPointsToType(new PointsToType("*", true));
                }
                parts.add(part);
                if (part.getPointsToTypes().isEmpty()) {
                    LOGGER.fine(String.format("Created empty DataStructurePart for type %s", allocatedType.getExternalName(false, true)));
                } else {
                    LOGGER.fine(String.format("Created 'follow-all' DataStructurePart for type %s", allocatedType.getExternalName(false, true)));
                }
            } else {
                LOGGER.fine(String.format("DataStructurePart already exists for type %s", allocatedType.getExternalName(false, true)));
            }
        });

        // Now, resolve the DataStructureParts
        // This matches all their PointsToTypes to DataStructureParts
        parts.forEach(dsp -> dsp.resolve(parts));

        // Currently, we may have more DataStructureParts than types
        // This may be due to the fact that we have data structure definition in files for data structures that do not exist in our application

        // As a next step, to reduce usage complexity, convert all DataStructureParts to DataStructureDefinitions
        // There is exactly one DataStructureDefinition per AllocatedType, and one AllocatedType per DataStructureDefinition
        allocatedTypes.forEach(allocatedType -> {
            DataStructurePart matchingPart = parts.stream()
                                                  .filter(dsp -> dsp.matches(allocatedType))
                                                  .findAny()
                                                  .orElse(null);
            assert matchingPart != null : "Every type must have a matching DataStructurePart, yet this was not the case for " + allocatedType;

            DataStructureDefinition definition =
                    new DataStructureDefinition(allocatedType,
                                                matchingPart.getDeepPointsToDataStructureParts().isEmpty()
                                                ? new AllocatedType[0]
                                                : matchingPart.getDeepPointsToDataStructureParts().stream().anyMatch(dsp -> dsp == starDummyDataStructurePart)
                                                  ? new AllocatedType[]{deepStarDummyAllocatedType}
                                                  : matchingPart.getDeepPointsToDataStructureParts()
                                                                .stream()
                                                                .map(dsp -> allocatedTypes.stream()
                                                                                          .filter(t -> dsp.matches(t))
                                                                                          .findAny()
                                                                                          // .orElse(new NonFoundType(dsp.getType()))) // it
                                                                                          // may happen that we do not find a matching type
                                                                                          // for a DataStructurePart since we may have
                                                                                          // definitions for never allocated types
                                                                                          .orElse(null))
                                                                .filter(type -> type != null)
                                                                .toArray(AllocatedType[]::new),
                                                matchingPart.getFlatDataStructureParts().isEmpty()
                                                ? new AllocatedType[0]
                                                : matchingPart.getFlatDataStructureParts().stream().anyMatch(dsp -> dsp == starDummyDataStructurePart)
                                                  ? new AllocatedType[]{flatStarDummyAllocatedType}
                                                  : matchingPart.getFlatDataStructureParts()
                                                                .stream()
                                                                .map(dsp -> allocatedTypes.stream()
                                                                                          .filter(t -> dsp.matches(t))
                                                                                          .findAny()
                                                                                          // .orElse(new NonFoundType(dsp.getType()))) // it
                                                                                          // may happen that we do not find a matching type
                                                                                          // for a DataStructurePart since we may have
                                                                                          // definitions for never allocated types
                                                                                          .orElse(null))
                                                                .filter(type -> type != null)
                                                                .toArray(AllocatedType[]::new),
                                                matchingPart.isHead());
            /*
            assert Arrays.stream(definition.followTypes).noneMatch(t -> t instanceof NonFoundType)
                    :
                    "There may be no null references in DataStructureDefinitions, something must have gone wrong in the conversion from DataStructureParts to " +
                            "DataStructureDefinitions for type " + definition.type.getFullyQualifiedName() + " and the following deep types: " +
                            Arrays.toString(Arrays.stream(definition.followTypes)
                                                  .filter(t -> t instanceof NonFoundType)
                                                  .map(t -> ((NonFoundType) t).expectedType)
                                                  .toArray(String[]::new));
            assert Arrays.stream(definition.flatTypes).noneMatch(t -> t instanceof NonFoundType)
                    :
                    "There may be no null references in DataStructureDefinitions, something must have gone wrong in the conversion from DataStructureParts to " +
                            "DataStructureDefinitions for type " + definition.type.getFullyQualifiedName() + " and the following flat types: " +
                            Arrays.toString(Arrays.stream(definition.flatTypes)
                                                  .filter(t -> t instanceof NonFoundType)
                                                  .map(t -> ((NonFoundType) t).expectedType)
                                                  .toArray(String[]::new));
            */
            definitions.put(allocatedType.getFullyQualifiedName(true), definition);
            allocatedType.dataStructureDefinition = definition;
        });
    }
}
