package at.jku.mevss.anttracks.heap.datastructures.cocor;

import at.jku.mevss.anttracks.heap.symbols.AllocatedType;

import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

public class DataStructurePart {
    private String type;
    private Set<PointsToType> pointsToTypes = new HashSet<>();
    private boolean isHead;

    private List<DataStructurePart> flatPointsToTypes;
    private List<DataStructurePart> deepArrayPointsToTypes;
    private List<DataStructurePart> deepPointsToDataStructureParts;

    /**
     * The default constructor used data structure parts
     *
     * @param type   The data structures part's fully qualified name (may contain "*" as placeholder)
     * @param isHead <code>true</code> if the given type is a data structure head, <code>false</code> otherwise
     */
    public DataStructurePart(String type, boolean isHead) {
        this.type = type;
        this.isHead = isHead;
    }

    public void resolve(List<DataStructurePart> definitions) {
        pointsToTypes.stream().forEach(ptt -> ptt.resolve(definitions));

        flatPointsToTypes = pointsToTypes.stream()
                                         .filter(ptt -> ptt.isFlat())
                                         .flatMap(ptt -> ptt.getMatchingDataStructureParts().stream())
                                         .collect(Collectors.toList());
        deepArrayPointsToTypes = pointsToTypes.stream()
                                              .filter(ptt -> ptt.isFollow())
                                              .filter(ptt -> ptt.mayBeArray())
                                              .flatMap(ptt -> ptt.getMatchingDataStructureParts().stream())
                                              .collect(Collectors.toList());
        deepPointsToDataStructureParts = pointsToTypes.stream()
                                                      .filter(ptt -> ptt.isFollow())
                                                      .flatMap(ptt -> ptt.getMatchingDataStructureParts().stream())
                                                      .collect(Collectors.toList());
    }

    public List<DataStructurePart> getDeepPointsToDataStructureParts() {
        return deepPointsToDataStructureParts;
    }

    public List<DataStructurePart> getDeepArrayDataStructureParts() {
        return deepArrayPointsToTypes;
    }

    public List<DataStructurePart> getFlatDataStructureParts() {
        return flatPointsToTypes;
    }

    @Override
    public String toString() {

        StringBuilder sb = new StringBuilder();
        sb.append(type);
        sb.append(" { \n");
        sb.append(pointsToTypes.stream()
                               .flatMap(ptt -> ptt.getMatchingDataStructureParts().stream())
                               .sorted(Comparator.comparing(dsPart -> dsPart.type))
                               .map(dsPart -> dsPart.type)
                               .collect(Collectors.joining(";\n")));
        sb.append(" }");
        return sb.toString();
    }

    public String getType() {
        return type;
    }

    public void setType(String type) {this.type = type;}

    public Set<PointsToType> getPointsToTypes() {
        return pointsToTypes;
    }

    public void addPointsToType(PointsToType subPart) {
        pointsToTypes.add(subPart);
    }

    public boolean isHead() {
        return isHead;
    }

    public boolean isInternal() { return !isHead; }

    /**
     * Data structure parts match an object of a certain type if they share the exact same name
     *
     * @param type The object's type to check against
     * @return <code>true</code> if the given type matched the data structure part
     */
    public boolean matches(AllocatedType type) {
        return this.type.equals(type.getFullyQualifiedName(true));
    }
}