/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap.io;

import at.jku.mevss.anttracks.heap.GarbageCollectionCause;
import at.jku.mevss.anttracks.heap.GarbageCollectionType;
import at.jku.mevss.anttracks.heap.StatisticGCInfo;
import at.jku.mevss.anttracks.heap.statistics.*;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.parser.EventType;
import at.jku.mevss.anttracks.parser.io.BaseFile;
import at.jku.mevss.anttracks.util.Consts;

import java.io.*;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

public class StatisticsReader implements AutoCloseable {
    private static Logger LOGGER = Logger.getLogger(StatisticsReader.class.getSimpleName());

    private final DataInputStream in;

    public StatisticsReader(String path) throws IOException {
        this.in = new DataInputStream(new BufferedInputStream(BaseFile.openR(path + File.separator + Consts.STATISTICS_META_FILE)));
    }

    @Override
    public void close() throws IOException {
        in.close();
    }

    public ArrayList<Statistics> read(Symbols symbols) throws IOException {
        ArrayList<Statistics> statistics = new ArrayList<>();

        int magic = in.readInt();
        assert magic == Consts.HEAP_FILES_MAGIC_PREFIX;
        try {
            while (true) {
                long time = in.readLong();
                EventType meta = EventType.parse(in.readInt());
                GarbageCollectionType type = GarbageCollectionType.fromID(in.readInt());
                GarbageCollectionCause cause = symbols.causes.get(in.readInt());
                boolean concurrent = in.readBoolean();
                boolean failed = in.readBoolean();
                StatisticGCInfo info = new StatisticGCInfo(type, cause, concurrent, failed, meta, time);

                SpaceStatistics eden = readSpaceStatistics();
                SpaceStatistics survivor = readSpaceStatistics();
                SpaceStatistics old = readSpaceStatistics();
                statistics.add(new Statistics(info, eden, survivor, old));
            }
        } catch (EOFException e) {
            in.close();
        }

        return statistics;
    }

    private SpaceStatistics readSpaceStatistics() throws IOException {
        Allocators allocator = readAllocators();
        MemoryConsumption memoryConsumption = readMemoryConsumption();
        ObjectTypes objectTypes = readObjectTypes();
        MemoryConsumption[] featureConsumption = readFeatureAllocation();

        return new SpaceStatistics(memoryConsumption, allocator, objectTypes, featureConsumption);
    }

    private Allocators readAllocators() throws IOException {
        long vm = in.readLong();
        long ir = in.readLong();
        long c1 = in.readLong();
        long c2 = in.readLong();
        return new Allocators(vm, ir, c1, c2);
    }

    private MemoryConsumption readMemoryConsumption() throws IOException {
        long objects = in.readLong();
        long bytes = in.readLong();
        return new MemoryConsumption(objects, bytes);
    }

    private ObjectTypes readObjectTypes() throws IOException {
        long instances = in.readLong();
        long smallArrays = in.readLong();
        long bigArrays = in.readLong();

        return new ObjectTypes(instances, smallArrays, bigArrays);
    }

    private MemoryConsumption[] readFeatureAllocation() throws IOException {
        if (in.readBoolean()) {
            int f = in.readInt();
            MemoryConsumption[] consumptions = new MemoryConsumption[f];
            for (int i = 0; i < f; i++) {
                long objects = in.readLong();
                long bytes = in.readLong();
                consumptions[i] = new MemoryConsumption(objects, bytes);
            }
            return consumptions;
        }
        return null;
    }

    public static List<Statistics> readStatisticsFromMetadata(String fullMetaDataPath, Symbols symbols) {
        List<Statistics> statistics = null;
        try {
            try (StatisticsReader statisticsReader = new StatisticsReader(fullMetaDataPath)) {
                statistics = statisticsReader.read(symbols);
            }
        } catch (Throwable t) {
            LOGGER.log(Level.WARNING, "reading meta data failed (this may be due to a change in the meta data format) -> reparse");
            statistics = null;
        }

        return statistics;
    }
}
