package at.jku.mevss.anttracks.heap.iteration;

import at.jku.mevss.anttracks.classification.Classifier;
import at.jku.mevss.anttracks.classification.ClassifierChain;
import at.jku.mevss.anttracks.classification.ConditionalClassifierList;
import at.jku.mevss.anttracks.classification.Filter;
import at.jku.mevss.anttracks.classification.nodes.MapGroupingNode;
import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.ObjectStream;
import at.jku.mevss.anttracks.heap.ObjectVisitor;
import at.jku.mevss.anttracks.heap.space.Space;
import at.jku.mevss.anttracks.util.Consts;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;
import java.util.concurrent.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class IteratingObjectStream implements ObjectStream {
    public IteratingObjectStream(DetailedHeap heap, boolean current) {
        this.heap = heap;
        this.spaces = heap.getSpacesCloned();
        this.current = current;
    }

    boolean current;

    private final DetailedHeap heap;
    private final Space[] spaces;

    private final List<Filter> filter = new ArrayList<>();
    private final List<IterationListener> listener = new ArrayList<>();

    @Override
    public ObjectStream filter(Filter... filter) {
        if (filter != null) {
            for (Filter of : filter) {
                this.filter.add(of);
            }
        }
        return this;
    }

    @Override
    public ObjectStream sorted() {
        Arrays.sort(spaces, Comparator.comparingLong(space -> space == null ? -1 : space.getAddress()));
        return this;
    }

    @Override
    public MapGroupingNode groupMap(ClassifierChain classifier, ConditionalClassifierList conditionalClassifier) {
        ThreadLocalHeapObjectMapGroupingVisitor grouping = new ThreadLocalHeapObjectMapGroupingVisitor(classifier, conditionalClassifier);
        forEach(grouping, ObjectVisitor.Settings.ALL_INFOS);
        grouping.sampleTopDown();
        return grouping;
    }

    @Override
    public MapGroupingNode groupMapParallel(ClassifierChain classifier, ConditionalClassifierList conditionalClassifier) {
        List<ThreadLocalHeapObjectMapGroupingVisitor> returns = forEachParallel(() -> new ThreadLocalHeapObjectMapGroupingVisitor(classifier, conditionalClassifier),
                                                                                ObjectVisitor.Settings.ALL_INFOS);
        MapGroupingNode finalGroup = returns.stream().map(x -> (MapGroupingNode) x).reduce((a, b) -> (MapGroupingNode) a.merge(b)).get();
        finalGroup.sampleTopDown();
        return finalGroup;
    }

    @Override
    public void forEach(ObjectVisitor visitor, ObjectVisitor.Settings visitorSettings) {
        for (Space space : spaces) {
            if (space != null) {
                space.iterate(heap, visitor, visitorSettings, filter, current, null, listener);
            }
        }
    }

    @Override
    public <I extends ObjectVisitor> List<I> forEachParallel(ObjectStream.ThreadVisitorGenerator<I> threadLocalVisitorGenerator,
                                                             ObjectVisitor.Settings visitorSettings) {
        List<I> returns;
        List<Future<I>> futures = new ArrayList<>();
        ExecutorService executorService = Executors.newFixedThreadPool(Consts.AVAILABLE_PROCESSORS() * 2);

        for (Space space : spaces) {
            if (space != null) {
                space.iterateAsync(heap, threadLocalVisitorGenerator, visitorSettings, filter, current, executorService, listener).forEach(futures::add);
            }
        }

        try {
            // Wait "forever" until executor service finished all tasks
            executorService.shutdown();
            executorService.awaitTermination(Long.MAX_VALUE, TimeUnit.MILLISECONDS);
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }

        returns = futures.stream().map(future -> {
            try {
                return future.get();
            } catch (InterruptedException e) {
                e.printStackTrace();
            } catch (ExecutionException e) {
                e.printStackTrace();
            }
            return null;
        }).filter(x -> x != null).collect(Collectors.toList());
        return returns;
    }

    @Override
    public <A> Stream<A> map(Classifier<A> classifier) {
        List<A> ret = new ArrayList<>();
        forEach((address, objectInfo, space, type, size, isArray, arrayLength, allocationSite, pointedFrom, pointsTo, eventType, rootPtrs) -> {
            try {
                ret.add(classifier.classify(address, objectInfo, space, type, size, isArray, arrayLength, allocationSite, pointedFrom, pointsTo, eventType, rootPtrs));
            } catch (Exception e) {
            }
        }, ObjectVisitor.Settings.ALL_INFOS);
        return ret.stream();
    }

    @Override
    public void addLabListener(IterationListener ll) {
        listener.add(ll);
    }

    @Override
    public void removeLabListener(IterationListener ll) {
        listener.remove(ll);
    }
}