/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap.labs;

import at.jku.mevss.anttracks.classification.Filter;
import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.ObjectAccess;
import at.jku.mevss.anttracks.heap.ObjectVisitor;
import at.jku.mevss.anttracks.heap.objects.ArrayObjectInfo;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.space.Space;
import at.jku.mevss.anttracks.heap.space.SpaceInfo;
import at.jku.mevss.anttracks.parser.heap.pointer.PtrUpdateVisitor;
import at.jku.mevss.anttracks.util.TraceException;

import java.util.List;

public final class SingleObjectLab extends Lab {

    private final ObjectInfo obj;
    private long[] pointers;
    /**
     * This is not updated during heap parsing, but only set once at the end of heap parsing
     */
    private long[] pointedFrom;
    private long forwardedAddr;
    //private int age;

    private boolean marking;

    public SingleObjectLab(String thread, int virtualOrFiller, long addr, ObjectInfo obj) {
        super(thread, virtualOrFiller, addr);
        this.obj = obj;
        pointers = null;
        pointedFrom = null;
        forwardedAddr = UNSET_FORWARDING_ADDR;
        //this.age = age;
    }

    @Override
    public long capacity() {
        return obj.size;
    }

    @Override
    public long position() {
        return obj.size;
    }

    @Override
    public int getObjectCount() {
        return 1;
    }

    @Override
    public void iterate(DetailedHeap heap,
                        SpaceInfo space,
                        List<Filter> filter,
                        ObjectVisitor visitor,
                        ObjectVisitor.Settings visitorSettings) {
        boolean accept = true;
        if (filter != null) {
            for (Filter f : filter) {
                try {
                    if (!f.classify(addr,
                                    obj,
                                    space,
                                    obj.type,
                                    obj.size,
                                    visitorSettings.arrayInfoNeeded ? obj instanceof ArrayObjectInfo : false,
                                    visitorSettings.arrayInfoNeeded ? (obj instanceof ArrayObjectInfo ? ((ArrayObjectInfo) obj).getLength() : -1) : -1,
                                    obj.allocationSite,
                                    visitorSettings.pointerInfoNeeded ? pointedFrom : null,
                                    visitorSettings.pointerInfoNeeded ? pointers : null,
                                    obj.eventType,
                                    visitorSettings.rootPointerInfoNeeded ? heap.rootPtrs.get(addr) : null)) {
                        accept = false;
                        break;
                    }
                } catch (Exception e) {
                    e.printStackTrace();
                    accept = false;
                    break;
                }
            }
        }
        if (accept) {
            visitor.visit(addr,
                          obj,
                          space,
                          obj.type,
                          obj.size,
                          obj instanceof ArrayObjectInfo,
                          (obj instanceof ArrayObjectInfo) ? ((ArrayObjectInfo) obj).getLength() : -1,
                          obj.allocationSite,
                          pointedFrom,
                          pointers,
                          obj.eventType,
                          heap == null ? null : heap.rootPtrs.get(addr));
        }
    }

    @Override
    public long tryAllocate(ObjectInfo obj, long objAddr) throws TraceException {
        return objAddr == addr ? addr : OBJECT_NOT_ASSIGNED;
    }

    @Override
    public ObjectInfo getObject(long objAddr) throws TraceException {
        return objAddr == addr ? obj : null;
    }

    @Override
    public void setObjectPointers(long objAddr, long[] newPtrs) throws TraceException {
        if (objAddr != addr) {
            throw new TraceException(String.format("Could not set new Pointers, object not found at address: %,d", objAddr));
        }
        pointers = newPtrs;
    }

    @Override
    public long getObjectForwardingAddressUnescaped(long objAddr) throws TraceException {
        if (objAddr != addr) {
            throw new TraceException(String.format("Cannot find object at address: %,d", objAddr));
        }
        return forwardedAddr;
    }

    @Override
    public long getObjectForwardingAddressEscaped(long objAddr) throws TraceException {
        if (objAddr != addr) {
            throw new TraceException(String.format("Cannot find object at address: %,d", objAddr));
        }
        return forwardedAddr & ObjectAccess.PTR_ADJUSTMENT_NEEDED_RESET_MASK;
    }

    @Override
    public long getObjectAddressAtIndex(int index) throws TraceException {
        if (index > 0) {
            throw new TraceException("SingleObjectLab may only contain a single object");
        }
        return addr;
    }

    @Override
    public void setObjectForwardingAddr(long objAddr, long forwAddr) throws TraceException {
        if (objAddr != addr) {
            throw new TraceException(String.format("Could not set new forwarding address, object not found at addres: %,d", objAddr));
        }
        forwardedAddr = forwAddr;

    }

    @Override
    public long[] getObjectPointers(long objAddr) throws TraceException {
        if (objAddr != addr) {
            throw new TraceException(String.format("Cannot find object at addres: %,d", objAddr));
        }
        return pointers;
    }

    @Override
    public void setObjectForwardingAddrAndPtrs(long objAddr, long forwAddr, long[] ptrs) throws TraceException {
        if (objAddr != addr) {
            throw new TraceException(String.format("Could not set new forwarding address, object not found at address: %,d", objAddr));
        }
        forwardedAddr = forwAddr;
        pointers = ptrs;
    }

    @Override
    public void resetForwardingAddresses() {
        forwardedAddr = UNSET_FORWARDING_ADDR;
    }

	/*
    @Override
	public void setAge(long objAddr, int age) throws TraceException {
		this.age = age + 1;
	}
	*/

	/*
    @Override
	public ObjectInfoAge getObjectAge(long objAddr) throws TraceException {
		ObjectInfo obj = getObject(objAddr);
		return new ObjectInfoAge(obj, age);
	}
	*/

    @Override
    public Lab clone() {
        return new SingleObjectLab(thread, kind, addr, obj);
    }

    @Override
    public ObjectInfo getObjectAtIndex(int index) throws TraceException {
        if (index > 0) {
            throw new TraceException("SingleObjectLab may only contain a single object");
        }
        return obj;
    }

    @Override
    public void setObjectPointersAtIndex(int index, long[] ptrs) throws TraceException {
        if (index != 0) {
            throw new TraceException(String.format("Could not set new pointers, object index in SingleObjectLab may not be large than 0"));
        }
        pointers = ptrs;
    }

    @Override
    public void setFromPointerCount(long objAddr, int n) throws TraceException {
        if (objAddr != addr) {
            throw new TraceException(String.format("Could not init from-pointer, object not found at address: %,d", objAddr));
        }
        pointedFrom = new long[n];
        for (int i = 0; i < pointedFrom.length; i++) {
            pointedFrom[i] = -1;
        }
    }

    @Override
    public long[] getObjectFromPointers(long objAddr) throws TraceException {
        if (objAddr != addr) {
            throw new TraceException(String.format("Could not set from-pointer, object not found at addres: %,d", objAddr));
        }

        return pointedFrom;
    }

    @Override
    public void reduceSize() {
        return;
    }

    @Override
    public boolean mayContainPointers() {
        return pointers != null;
    }

	/*
    @Override
	public ObjectInfoAge getObjectAgeAtIndex(int index) throws TraceException {
		if (index > 0) {
			throw new TraceException("SingleObjectLab may only contain a single object");
		}
		return new ObjectInfoAge(obj, age);
	}
	*/

    @Override
    public long[] getObjectPointersAtIndex(int index) throws TraceException {
        if (index > 0) {
            throw new TraceException("SingleObjectLab may only contain a single object");
        }
        return pointers;
    }

    @Override
    public long[] getObjectFromPointersAtIndex(int index) throws TraceException {
        if (index > 0) {
            throw new TraceException("SingleObjectLab may only contain a single object");
        }
        return pointedFrom;
    }

    @Override
    public long getObjectForwardingAddrUnescapedAtIndex(int index) throws TraceException {
        if (index > 0) {
            throw new TraceException("SingleObjectLab may only contain a single object");
        }
        return forwardedAddr;
    }

    @Override
    public long getObjectForwardingAddrEscapedAtIndex(int index) throws TraceException {
        if (index > 0) {
            throw new TraceException("SingleObjectLab may only contain a single object");
        }
        return forwardedAddr & ObjectAccess.PTR_ADJUSTMENT_NEEDED_RESET_MASK;
    }

    @Override
    public int getAddressIndex(long objAddr) throws TraceException {
        if (objAddr != addr) {
            throw new TraceException(String.format("Given address %,d does not point to the object of this single object lab!", objAddr));
        }
        return 0;
    }

    @Override
    public void mark(int index) throws TraceException {
        if (index != 0) {
            throw new TraceException("SingleObjectLab may only contain a single object/mark");
        }
        marking = true;
    }

    @Override
    public void mark(long objAddr) throws TraceException {
        if (objAddr != addr) {
            throw new TraceException("Given address does not match lab object address");
        }

        marking = true;
    }

    @Override
    public boolean isMarked(int index) throws TraceException {
        if (index != 0) {
            throw new TraceException("SingleObjectLab may only contain a single object/mark");
        }
        return marking;
    }

    @Override
    public boolean isMarked(long objAddr) throws TraceException {
        if (objAddr != addr) {
            throw new TraceException("Given address does not match lab object address");
        }

        return marking;
    }

    @Override
    public void clearMarkings() {
        marking = false;
    }

    @Override
    public void iterateUpdatePointer(PtrUpdateVisitor iterator, Space space) {
        iterator.visit(space, this, addr, pointers);
    }

    @Override
    public void variableCapacity() { }

    @Override
    public void setFromPointers(long address, long[] ptrs) throws TraceException {
        if (address != addr) {
            throw new TraceException(String.format("Could not set from-pointers, object not found at address: %,d", address));
        }
        pointedFrom = ptrs;
    }

    @Override
    public boolean ptrNeedAdjustment(long objAddr) throws TraceException {
        if (objAddr != addr) {
            throw new TraceException(String.format("Could not retrieve pointer-adjustment status, object not found at address: %,d", objAddr));
        }
        return (forwardedAddr & ObjectAccess.PTR_ADJUSTMENT_NEEDED_MASK) == 1 ? true : false;
    }

    @Override
    public void setPtrNeedAdjustment(long objAddr, boolean needAdjustment) throws TraceException {
        if (objAddr != addr) {
            throw new TraceException(String.format("Could not set pointer-adjustment status, object not found at address: %,d", objAddr));
        }
        if (needAdjustment) {
            forwardedAddr |= ObjectAccess.PTR_ADJUSTMENT_NEEDED_MASK;
        } else {
            forwardedAddr &= ObjectAccess.PTR_ADJUSTMENT_NEEDED_RESET_MASK;
        }
    }

    public MultiObjectLab toMultiObjectLab(boolean expectPointers) {
        MultiObjectLab lab = new MultiObjectLab(thread, Lab.KIND_REGION_VIRTUAL, addr, capacity(), expectPointers);
        try {
            lab.tryAllocate(obj, addr);
            if (expectPointers) {
                lab.setObjectForwardingAddrAndPtrs(addr, forwardedAddr, pointers);
            }
        } catch (TraceException e) {
            e.printStackTrace();
        }
        return lab;
    }
}
