/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap.listener;

import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.objects.ArrayObjectInfo;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLong;

public class ArrayLengthLogListener extends NewObjectLogListener {

    private final File file;

    private final Object lock;
    private final Map<Integer, AtomicLong> lengths;

    public ArrayLengthLogListener(File file) {
        this.file = file;
        if (file.exists()) {
            file.delete();
        }
        lock = new Object();
        lengths = new ConcurrentHashMap<>();
    }

    public static void init(DetailedHeap heap) {
        String log = System.getProperty("statistics.arraylength.log");
        if (log != null) {
            heap.addListener(new ArrayLengthLogListener(new File(log)));
        }
    }

    @Override
    protected void record(ObjectInfo obj) {
        if (!obj.isArray()) {
            return;
        }
        ArrayObjectInfo array = (ArrayObjectInfo) obj;
        record(array.getLength());
    }

    private void record(int length) {
        AtomicLong counter = lengths.get(length);
        if (counter == null) {
            synchronized (lock) {
                counter = lengths.get(length);
                if (counter == null) {
                    counter = new AtomicLong(0);
                    lengths.put(length, counter);
                }
            }
        }
        counter.incrementAndGet();
    }

    @Override
    public void close(DetailedHeap heap) {
        try (BufferedWriter out = new BufferedWriter(new FileWriter(file))) {
            for (Integer length : lengths.keySet()) {
                long count = lengths.get(length).get();
                if (count == 0) {
                    continue;
                }
                out.write(String.valueOf(length));
                out.write(" ");
                out.write(String.valueOf(count));
                out.write("\n");
            }
        } catch (IOException e) {
            e.printStackTrace(System.err);
        }
    }
}
