/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap.listener;

import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.symbols.AllocationSite;
import at.jku.mevss.anttracks.heap.symbols.Symbols;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLong;

public class StackTraceLogListener extends NewObjectLogListener {

    private final File file_static, file_dynamic;

    private final Object lock;
    private final Map<Integer, AtomicLong> depths;
    private final Symbols symbols;

    public StackTraceLogListener(String file, Symbols symbols) {
        file_static = new File(file.replace("{0}", "static"));
        file_dynamic = new File(file.replace("{0}", "dynamic"));
        if (file_static.exists()) {
            file_static.delete();
        }
        if (file_dynamic.exists()) {
            file_dynamic.delete();
        }
        this.symbols = symbols;
        lock = new Object();
        depths = new ConcurrentHashMap<>();
    }

    public static void init(DetailedHeap heap) {
        String log = System.getProperty("statistics.stacktrace.log");
        if (log != null) {
            heap.addListener(new StackTraceLogListener(log, heap.getSymbols()));
        }
    }

    @Override
    protected void record(ObjectInfo obj) {
        int depth = symbols.sites.getById(obj.allocationSite.id).callSites.length;
        AtomicLong counter = depths.get(depth);
        if (counter == null) {
            synchronized (lock) {
                counter = depths.get(depth);
                if (counter == null) {
                    counter = new AtomicLong(0);
                    depths.put(depth, counter);
                }
            }
        }
        counter.incrementAndGet();
    }

    @Override
    public void close(DetailedHeap heap) {
        try (BufferedWriter out = new BufferedWriter(new FileWriter(file_static))) {
            for (AllocationSite site : symbols.sites) {
                if (site == null) {
                    continue;
                }
                out.write(String.valueOf(site.callSites.length));
                out.write(" 1\n");
            }
        } catch (IOException e) {
            e.printStackTrace(System.err);
        }
        try (BufferedWriter out = new BufferedWriter(new FileWriter(file_dynamic))) {
            for (Integer depth : depths.keySet()) {
                long count = depths.get(depth).get();
                out.write(String.valueOf(depth));
                out.write(" ");
                out.write(String.valueOf(count));
                out.write("\n");
            }
        } catch (IOException e) {
            e.printStackTrace(System.err);
        }
    }

}
