/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap.listener;

import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.symbols.AllocatedType;
import at.jku.mevss.anttracks.heap.symbols.AllocationSite;
import at.jku.mevss.anttracks.heap.symbols.AllocationSites;
import at.jku.mevss.anttracks.heap.symbols.Symbols;

import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.atomic.AtomicLongArray;

public class TotalAllocationListener extends NewObjectLogListener {

    private final Symbols symbols;
    private AtomicLongArray allocationSiteCounts;

    public TotalAllocationListener(DetailedHeap heap) {
        this.symbols = heap.getSymbols();
        this.allocationSiteCounts = new AtomicLongArray(getAllocationSiteCountsArrayLength(symbols));
    }

    private int getAllocationSiteCountsArrayLength(Symbols symbols) {
        AllocationSites sites = symbols.sites;
        Integer index = -1;
        Iterator<AllocationSite> iter = sites.iterator();
        while (iter.hasNext()) {
            AllocationSite site = iter.next();
            if (site != null && index < site.id) {
                index = site.id;
            }
        }
        return index + 1;
    }

    @Override
    public void close(DetailedHeap heap) {
        try {
            writeAllocSiteDump(new OutputStreamWriter(System.out));
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    @Override
    protected void record(ObjectInfo obj) {
        int allocationSiteId = obj.allocationSite.id;
        allocationSiteCounts.incrementAndGet(allocationSiteId);

    }

    public static void init(DetailedHeap heap) {
        // heap.addListener(new TotalAllocationListener(heap));
    }

    private synchronized void writeAllocSiteDump(Writer writer) throws IOException {
        long total = 0;
        Iterator<AllocationSite> iter = symbols.sites.iterator();
        List<String> lines = new LinkedList<>();
        while (iter.hasNext()) {
            AllocationSite site = iter.next();
            if (site == null) {
                continue;
            }
            int allocationSiteId = site.id;
            long count = allocationSiteCounts.get(allocationSiteId);
            // writer.append(getAllocationSiteString(site, count));
            // writer.append("\n");
            lines.add(getAllocationSiteLine(site, count));
            total += count;
        }
        lines.sort((a, b) -> a.compareTo(b));
        for (String line : lines) {
            writer.append(line);
        }
        writer.append("\n");
        writer.append("Total allocations: " + total);
        writer.flush();
    }

    private String getAllocationSiteLine(AllocationSite site, long count) {
        StringBuilder builder = new StringBuilder();
        int typeId = site.allocatedTypeId;
        AllocatedType type = symbols.types.get(typeId);
        builder.append(type + "@");
        String duckTracksComparableAllocationSite = getDuckTracksComparableAllocationSite(site.toString());
        builder.append(duckTracksComparableAllocationSite);
        builder.append("=");
        builder.append(count);
        builder.append("\n");
        return builder.toString();
    }

    private String getDuckTracksComparableAllocationSite(String allocationSite) {
        int firstLineBreakIndex = allocationSite.indexOf("\n");
        if (firstLineBreakIndex < 0) {
            return allocationSite;
        } else {
            return allocationSite.substring(0, firstLineBreakIndex);
        }
    }

}
