/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap.objects;

import at.jku.mevss.anttracks.heap.symbols.AllocatedType;
import at.jku.mevss.anttracks.heap.symbols.AllocationSite;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.parser.EventType;

import java.util.Arrays;

public abstract class ObjectInfo {

    // Special thing: id is excluded from hash and equals
    public int id = -1;

    public EventType eventType;
    public AllocationSite allocationSite;
    public AllocatedType type;
    public int size;

    public ObjectInfo(EventType eventType, AllocationSite allocationSite, AllocatedType type, int size) {
        this.eventType = eventType;
        this.allocationSite = allocationSite;
        this.type = type;
        this.size = size;
    }

    protected ObjectInfo(EventType eventType, AllocationSite allocationSite, AllocatedType type, int size, int heapWordSize) {
        this.eventType = eventType;
        this.allocationSite = allocationSite;
        this.type = type;
        this.size = padObjectSize(size, heapWordSize);
    }

    protected ObjectInfo() {}

    protected static int padObjectSize(int unpadded, int heapWordSize) {
        int padding = unpadded % heapWordSize;
        int padded = unpadded + (padding == 0 ? 0 : heapWordSize - padding);
        return padded;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        ObjectInfo that = (ObjectInfo) o;

        if (size != that.size) {
            return false;
        }
        if (eventType != that.eventType) {
            return false;
        }
        if (!allocationSite.equals(that.allocationSite)) {
            return false;
        }
        return type.equals(that.type);
    }

    @Override
    public int hashCode() {
        int result = eventType.hashCode();
        result = 31 * result + allocationSite.hashCode();
        result = 31 * result + type.hashCode();
        result = 31 * result + size;
        return result;
    }

    @Override
    public String toString() {
        return String.format("Object of type %d (%s) @ %d %s (size %d, event type %d)",
                             type.id,
                             type.name,
                             allocationSite.id,
                             Arrays.stream(allocationSite.callSites)
                                   .map(location -> location.text)
                                   .reduce((location, location2) -> location + " <- " + location2)
                                   .orElse("???"),
                             size,
                             eventType.id);
    }

    public abstract boolean isArray();

    public static ObjectInfo get(EventType eventType,
                                 AllocationSite allocationSite,
                                 AllocatedType type,
                                 int size,
                                 boolean isArray,
                                 int arrayLength,
                                 InstanceObjectInfo instancePrototype,
                                 InstanceMirrorObjectInfo instanceMirrorPrototype,
                                 ArrayObjectInfo arrayPrototype,
                                 ObjectInfoCache cache,
                                 Symbols symbols) {
        ObjectInfo obj;
        if (type.name.equals(InstanceMirrorObjectInfo.MIRROR_CLASS)) {
            obj = InstanceMirrorObjectInfo.get(cache, instanceMirrorPrototype, eventType, allocationSite, type, size, symbols);
        } else {
            if (isArray) {
                obj = ArrayObjectInfo.get(cache, arrayPrototype, eventType, allocationSite, type, arrayLength, symbols);
            } else {
                obj = InstanceObjectInfo.get(cache, instancePrototype, eventType, allocationSite, type, symbols);
            }
        }
        return obj;
    }
}
