/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap.objects;

import at.jku.mevss.anttracks.util.CollectionsUtil;

import java.util.Collection;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Stream;

public class ObjectInfoCache {

    private final AtomicInteger currentId = new AtomicInteger();

    private ConcurrentHashMap<ObjectInfo, ObjectInfo> cache;
    private ConcurrentHashMap<Integer, ObjectInfo> idCache;

    public ObjectInfoCache() {
        cache = new ConcurrentHashMap<>();
        idCache = new ConcurrentHashMap<>();
    }

    public ObjectInfoCache(Collection<ObjectInfo> objects) {
        cache = CollectionsUtil.map(objects, o -> o, o -> o, () -> new ConcurrentHashMap<>());
        idCache = CollectionsUtil.map(objects, o -> o.id, o -> o, () -> new ConcurrentHashMap<>());
    }

    public ObjectInfo get(ObjectInfo info) {
        ObjectInfo cached = cache.get(info);
        if (cached != null) {
            return cached;
        }

        cached = cache.putIfAbsent(info, info);
        if (cached == null) {
            cached = info;
            if (cached.id < 0) {
                cached.id = currentId.getAndIncrement();
            }
            idCache.put(cached.id, cached);
        }
        return cached;
    }

    public ObjectInfo get(InstanceObjectInfo info) {
        ObjectInfo cached = cache.get(info);
        if (cached != null) {
            return cached;
        }

        ObjectInfo nu = new InstanceObjectInfo(info);
        cached = cache.putIfAbsent(nu, nu);
        if (cached == null) {
            cached = nu;
            if (cached.id < 0) {
                cached.id = currentId.getAndIncrement();
            }
            idCache.put(cached.id, cached);
        }
        return cached;
    }

    public ObjectInfo get(ArrayObjectInfo info) {
        ObjectInfo cached = cache.get(info);
        if (cached != null) {
            return cached;
        }

        ObjectInfo nu = new ArrayObjectInfo(info);
        cached = cache.putIfAbsent(nu, nu);
        if (cached == null) {
            cached = nu;
            if (cached.id < 0) {
                cached.id = currentId.getAndIncrement();
            }
            idCache.put(cached.id, cached);
        }
        return cached;
    }

    public ObjectInfo get(InstanceMirrorObjectInfo info) {
        ObjectInfo cached = cache.get(info);
        if (cached != null) {
            return cached;
        }

        ObjectInfo nu = new InstanceMirrorObjectInfo(info);
        cached = cache.putIfAbsent(nu, nu);
        if (cached == null) {
            cached = nu;
            if (cached.id < 0) {
                cached.id = currentId.getAndIncrement();
            }
            idCache.put(cached.id, cached);
        }
        return cached;
    }

    public ObjectInfo getById(int id) {
        return idCache.get(id);
    }

    public void clear() {
        cache.clear();
        idCache.clear();
    }

    public int getCount() {
        return cache.size();
    }

    public ObjectInfo[] toArray() {
        return cache.values().toArray(new ObjectInfo[cache.size()]);
    }

    public ObjectInfo getByHash(int objectInfoHash) {
        for (Map.Entry<ObjectInfo, ObjectInfo> objectInfoObjectInfoEntry : cache.entrySet()) {
            if (objectInfoObjectInfoEntry.getKey().hashCode() == objectInfoHash) {
                return objectInfoObjectInfoEntry.getValue();
            }
        }
        return null;
    }

    public int size() {
        return cache.size();
    }

    public Stream<ObjectInfo> stream() {
        return cache.values().stream();
    }

    @Override
    public ObjectInfoCache clone() {
        ObjectInfoCache dolly = new ObjectInfoCache();
        dolly.currentId.set(currentId.get());
        dolly.cache = new ConcurrentHashMap<>(cache);
        dolly.idCache = new ConcurrentHashMap<>(idCache);
        return dolly;
    }
}
