/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap.statistics;

import at.jku.mevss.anttracks.heap.*;
import at.jku.mevss.anttracks.heap.io.StatisticsReader;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.roots.RootPtr;
import at.jku.mevss.anttracks.heap.space.SpaceInfo;
import at.jku.mevss.anttracks.heap.symbols.AllocatedType;
import at.jku.mevss.anttracks.heap.symbols.AllocationSite;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.parser.EventType;
import at.jku.mevss.anttracks.util.ApplicationStatistics;

import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

public class Statistics {
    private static Logger LOGGER = Logger.getLogger(Statistics.class.getSimpleName());

    public final StatisticGCInfo info;
    public final SpaceStatistics eden;
    public final SpaceStatistics survivor;
    public final SpaceStatistics old;

    public Statistics(StatisticGCInfo info, SpaceStatistics eden, SpaceStatistics survivor, SpaceStatistics old) {
        this.info = info;
        this.eden = eden;
        this.survivor = survivor;
        this.old = old;
    }

    public static Statistics collect(DetailedHeap heap,
                                     GarbageCollectionType gcType,
                                     GarbageCollectionCause gcCause,
                                     boolean concurrent,
                                     boolean failed,
                                     EventType gcMeta,
                                     long time) {
        class ThreadLocalStatisticsVisitor implements ObjectVisitor {
            SpaceStatistics eden = new SpaceStatistics(heap.getSymbols());
            SpaceStatistics survivor = new SpaceStatistics(heap.getSymbols());
            SpaceStatistics old = new SpaceStatistics(heap.getSymbols());

            @Override
            public void visit(long address,
                              ObjectInfo objectInfo,
                              SpaceInfo space,
                              AllocatedType type,
                              int size,
                              boolean isArray,
                              int arrayLength,
                              AllocationSite allocationSite,
                              long[] pointedFrom,
                              long[] pointsTo,
                              EventType eventType,
                              List<RootPtr> rootPtrs) {
                switch (space.getType()) {
                    case EDEN:
                        eden.add(objectInfo);
                        break;

                    case SURVIVOR:
                        survivor.add(objectInfo);
                        break;

                    case OLD:
                        old.add(objectInfo);
                        break;
                }
            }
        }

        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("Collect statistics");
        StatisticGCInfo gcInfo = new StatisticGCInfo(gcType, gcCause, concurrent, failed, gcMeta, time);

        SpaceStatistics eden = new SpaceStatistics(heap.getSymbols());
        SpaceStatistics survivor = new SpaceStatistics(heap.getSymbols());
        SpaceStatistics old = new SpaceStatistics(heap.getSymbols());
        heap.toObjectStream().forEachParallel(() -> new ThreadLocalStatisticsVisitor(), ObjectVisitor.Settings.NO_INFOS).forEach(tlsv -> {
            eden.merge(tlsv.eden);
            survivor.merge(tlsv.survivor);
            old.merge(tlsv.old);
        });

        m.end();
        return new Statistics(gcInfo, eden, survivor, old);

    }

    public static List<Statistics> readStatisticsFromMetadata(PreprocessedHeap preprocessedHeap) {
        return readStatisticsFromMetadata(preprocessedHeap.metaDir.getAbsolutePath(), preprocessedHeap.symbols);
    }

    public static List<Statistics> readStatisticsFromMetadata(String fullMetaDataPath, Symbols symbols) {
        LOGGER.log(Level.INFO, "reading statistics from meta data");
        List<Statistics> statistics = null;
        try {
            try (StatisticsReader statisticsReader = new StatisticsReader(fullMetaDataPath)) {
                statistics = statisticsReader.read(symbols);
            }
        } catch (Throwable t) {
            LOGGER.log(Level.WARNING, "reading statistics from meta data failed (this may be due to a change in the meta data format) -> reparse");
            statistics = null;
        }

        return statistics;
    }

    public static Statistics findStatistics(List<Statistics> statistics, GarbageCollectionLookup gcLookup) {
        int id = 0;
        for (Statistics stat : statistics) {
            if (stat.info.matches(gcLookup)) {
                if (id == gcLookup.id) {
                    return stat;
                } else {
                    id++;
                }
            }
        }
        return null;
    }
}
