/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap.symbols;

import at.jku.mevss.anttracks.heap.datastructures.DataStructureDefinition;
import at.jku.mevss.anttracks.util.SignatureConverter;

import java.util.Arrays;
import java.util.Comparator;
import java.util.Map;
import java.util.regex.Pattern;

public class AllocatedType {
    public static class MirrorAllocatedType extends AllocatedType {
        public final AllocatedType mirrorKlass;

        public MirrorAllocatedType(AllocatedType classClass, AllocatedType mirrorKlass) {
            // Clone Class<> class
            super(mirrorKlass != null ? -mirrorKlass.id : AllocatedType.ALLOCATED_TYPE_IDENTIFIER_UNKNOWN_MIRROR, classClass.superId, classClass.name, classClass.size);
            this.fieldInfos = classClass.fieldInfos;
            this.staticFieldInfos = classClass.staticFieldInfos;
            this.nonStaticFieldInfos = classClass.nonStaticFieldInfos;
            this.refFields = classClass.refFields;
            this.nNonStaticReferenceFields = classClass.nNonStaticReferenceFields;
            this.fieldIndex = classClass.fieldIndex;
            this.methodInfos = classClass.methodInfos;
            this.methodIndex = classClass.methodIndex;

            // Set mirrored class
            this.mirrorKlass = mirrorKlass;
        }

        public String getExternalName(boolean omitPackage, boolean showMirrorType) {
            if (!showMirrorType) {
                return super.getExternalName(omitPackage, showMirrorType);
            } else {
                return super.getExternalName(omitPackage, showMirrorType) + "<" + (mirrorKlass == null ?
                                                                                   "?" :
                                                                                   mirrorKlass.getExternalName(omitPackage, showMirrorType)) + ">";
            }
        }
    }

    public static final int MAGIC_BYTE = 24;
    public static final int MAGIC_BYTE_TYPE_FIELD_INFO = 56;
    public static final int MAGIC_BYTE_TYPE_SUPER_FIELD_INFO = 57;
    public static final int MAGIC_BYTE_TYPE_METHOD_INFO = 58;
    public static final int ALLOCATED_TYPE_IDENTIFIER_UNKNOWN = 0;
    // Objects with ID 1 are class objects of which we do not know the mirror class
    public static final int ALLOCATED_TYPE_IDENTIFIER_UNKNOWN_MIRROR = 1;
    public static final char ALLOCATED_TYPE_IDENTIFIER_UNKNOWN_SIG = 'U';

    public final int id;
    public final int superId;
    public final String name;
    public int size;
    protected FieldInfo[] fieldInfos;
    protected FieldInfo[] staticFieldInfos;
    protected FieldInfo[] nonStaticFieldInfos;
    protected int refFields;
    protected int nNonStaticReferenceFields;
    protected int fieldIndex;
    protected final String externalName;
    public MethodInfo[] methodInfos;
    protected int methodIndex;

    private AllocatedType superType;

    public DataStructureDefinition dataStructureDefinition;

    public AllocatedType(int id, int superId, String name, int size) {
        this.id = id;
        this.superId = superId;
        this.name = name;
        this.size = size;
        this.externalName = SignatureConverter.convertToJavaType(name);
        createOrExpandFieldInfos(0);

        // By default has an empty data structure definition
        this.dataStructureDefinition = new DataStructureDefinition(this);
    }

    public class FieldInfo {
        public final int offset;
        public final String name;
        public final String signature;
        private int typeId;
        public final int flags;
        public final boolean supr;

        public FieldInfo(boolean supr, int offset, String signature, String name, int flags) {
            this.supr = supr;
            this.offset = offset;
            this.signature = signature;
            this.name = name;
            this.flags = flags;
        }

        public boolean isStatic() {
            return (flags & (1L << 3)) != 0;
        }

        public boolean isArray() {
            return signature.startsWith("[");
        }

        public boolean isInstance() {
            return signature.startsWith("L");
        }

        public int getSize() {
            switch (signature) {
                case "B":
                    return 1;
                case "C":
                    return 2;
                case "D":
                    return 8;
                case "F":
                    return 4;
                case "I":
                    return 4;
                case "J":
                    return 8;
                case "S":
                    return 2;
                case "Z":
                    return 1;
                default:
                    // TODO: Return pointer size depending on compressed oops
                    return 4;
            }
        }

        @Override
        public String toString() {
            return signature + " " + name + "(Offset: " + offset + ", Inherited: " + supr + ", Access: " + flags + "[Instance: " + isInstance() + ", Static:" +
                    isStatic() + "])";
        }

        public void setTypeId(int typeId) {
            this.typeId = typeId;
        }

        public int getTypeId() {
            return typeId;
        }
    }

    public class MethodInfo {
        public final int idnum;
        public final String name;
        public final Map<Integer, String> locals;
        public final AllocatedType type;

        public MethodInfo(int idnum, String name, Map<Integer, String> locals, AllocatedType type) {
            this.idnum = idnum;
            this.name = name;
            this.locals = locals;
            this.type = type;
        }

        public String toString() {
            return idnum + " : " + name;
        }
    }

    public void addFieldInfo(boolean supr, int offset, String signature, String name, int flags) {
        FieldInfo info = new FieldInfo(supr, offset, signature, name, flags);
        fieldInfos[fieldIndex++] = info;
    }

    public void addMethodInfo(int idnum, String name, Map<Integer, String> locals) {
        MethodInfo info = new MethodInfo(idnum, name, locals, this);
        methodInfos[methodIndex++] = info;
    }

    public FieldInfo[] getAllFieldInfos() {
        return fieldInfos;
    }

    public FieldInfo[] getNonStaticFieldInfos() {
        return nonStaticFieldInfos;
    }

    public FieldInfo[] getStaticFieldInfos() {
        return staticFieldInfos;
    }

    public void complete(Symbols symbols) {
        completeFieldInfo();

        // resolve superId to AllocatedType
        superType = superId > 0 ? symbols.types.get(superId) : null; // no super type for java.lang.Object
    }

    private void completeFieldInfo() {
        fieldInfos = Arrays.stream(fieldInfos).sorted(Comparator.comparingInt(o -> o.offset)).toArray(FieldInfo[]::new);
        nonStaticFieldInfos = Arrays.stream(fieldInfos).filter(x -> !x.isStatic()).sorted(Comparator.comparingInt(o -> o.offset)).toArray(FieldInfo[]::new);
        staticFieldInfos = Arrays.stream(fieldInfos).filter(FieldInfo::isStatic).sorted(Comparator.comparingInt(o -> o.offset)).toArray(FieldInfo[]::new);

        if (nNonStaticReferenceFields != 0 || fieldInfos == null) {
            return;
        }

        for (int i = 0; i < fieldInfos.length; i++) {
            FieldInfo field = fieldInfos[i];
            int bit = 1;

            if (!field.isInstance() && !field.isArray()) {
                bit = 0;
            }
            if (field.isStatic() && !name.equals(AllocatedTypes.MIRROR_CLASS_NAME)) {
                bit = 0;
            }

            nNonStaticReferenceFields += bit;
            refFields = (refFields << 1) | bit;
        }
    }

    public int nonStaticReferenceFieldsCount() {
        return nNonStaticReferenceFields;
    }

    public int referenceFieldsCount() {
        return Integer.bitCount(refFields);
    }

    public void createOrExpandFieldInfos(int length) {
        if (fieldInfos == null) {
            fieldInfos = new FieldInfo[length];
        } else {
            fieldInfos = Arrays.copyOf(fieldInfos, fieldInfos.length + length);
        }
    }

    public void createOrExandMethodInfos(int length) {
        if (methodInfos == null) {
            methodInfos = new MethodInfo[length];
        } else {
            methodInfos = Arrays.copyOf(methodInfos, methodInfos.length + length);
        }
    }

    @Override
    public String toString() {
        return name;
    }

    public boolean isKnownToHaveWrongPointerNumber(Symbols symbols) {
        AllocatedType cur = this;
        do {
            if (cur.name.equals("Ljava/lang/ref/Reference;")) {
                // Classes that inherit from Reference may have incomplete
                // pointer information, see "Efficient Memory Traces with Full
                // Pointer Information", see section 3.3
                return true;
            }
            if (cur.name.equals("Ljava/lang/invoke/MemberName;")) {
                // Don't know yet why they are wrong, issue on YouTack: AT-90
                return true;
            }
            if (cur.superId > ALLOCATED_TYPE_IDENTIFIER_UNKNOWN) {
                cur = symbols.types.get(cur.superId);
            } else {
                cur = null;
            }
        } while (cur != null);
        return false;
    }

    public String getFullyQualifiedName() {
        return getFullyQualifiedName(false);
    }

    public String getFullyQualifiedName(boolean showMirrorType) {
        return getExternalName(false, showMirrorType);
    }

    public String getSimpleName() {
        return getSimpleName(false);
    }

    public String getSimpleName(boolean showMirrorType) {
        return getExternalName(true, showMirrorType);
    }

    public String getExternalName(boolean omitPackage, boolean showMirrorType) {
        if (omitPackage && externalName.contains(".")) {
            return externalName.substring(externalName.lastIndexOf('.') + 1);
        }
        return externalName;
    }

    /**
     * Tells whether this AllocatedType is a (direct or indirect) subtype of an allocated type with the given name
     *
     * @param externalName Name of the allocated type including package and mirrortype (AllocatedType.getExternalName(false, true))
     * @return true if there is an AllocatedType with the given name that is a supertype of this AllocatedType, false otherwise
     */
    public boolean isSubtypeOf(String externalName) {
        for (AllocatedType cur = superType; cur != null; cur = cur.superType) {
            if (cur.getExternalName(false, true).equals(externalName)) {
                return true;
            }
        }

        return false;
    }

    public boolean isAssignableTo(AllocatedType fieldType) {
        AllocatedType currentType = this;
        do {
            // my type matches the given type
            if (fieldType.id == currentType.id) {
                return true;
            }
            // was no match, check if my parent type is matching
            currentType = currentType.superType;
        } while (currentType != null);
        return false;
    }

    /**
     * Tells whether this AllocatedType is a (direct or indirect) subtype of an allocated type whose name matches the given pattern
     *
     * @param pattern pattern that the supertype's name (including package and mirrortype) must match
     * @return true if there is an AllocatedType whose name matches the pattern and is a supertype of this AllocatedType, false otherwise
     */
    public boolean isSubtypeOf(Pattern pattern) {
        for (AllocatedType cur = superType; cur != null; cur = cur.superType) {
            if (pattern.matcher(cur.getFullyQualifiedName(true)).matches()) {
                return true;
            }
        }

        return false;
    }

    public boolean isPrimitiveArray() {
        // Basic data types have a name of length 1 (e.g., C)
        // Basic data type arrays have a name of length 2 (e.g., [C)
        // Reference types (and non-basic data type arrays) have a name length of at least 3 (e.g., Lmypackage/MyClassName; or [LX;)
        return name.length() == 2;
    }

    public boolean isArray() {
        return name.startsWith("[");
    }
}
