/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.heap.symbols;

import at.jku.mevss.anttracks.util.ArraysUtil;
import at.jku.mevss.anttracks.util.SignatureConverter;

public class AllocationSite {

    public static class Location {
        public final String signature;
        public final int bci;

        public final String text;

        public Location(String signature, int bci) {
            this.signature = signature;
            this.bci = bci;
            this.text = SignatureConverter.convertToJavaMethodSignature(signature, false) + " : " + bci;
        }

        @Override
        public int hashCode() {
            final int prime = 31;
            int result = 1;
            result = prime * result + bci;
            result = prime * result + ((signature == null) ? 0 : signature.hashCode());
            return result;
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj) {
                return true;
            }
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            Location other = (Location) obj;
            if (bci != other.bci) {
                return false;
            }
            if (signature == null) {
                if (other.signature != null) {
                    return false;
                }
            } else if (!signature.equals(other.signature)) {
                return false;
            }
            return true;
        }

        @Override
        public String toString() {
            return signature + ":" + bci;
        }
    }

    public static final int MAGIC_BYTE = 42;
    public static final int MAGIC_BYTE_SIMPLE = 84;
    public static final int ALLOCATION_SITE_IDENTIFIER_UNKNOWN = 0;
    public static final int UNKNOWN_BCI = -1;

    public final int id;
    public final Location[] callSites;
    public final int allocatedTypeId;

    /**
     * The original size of the call stack as parsed from the symbols file.
     */
    public final int stackSizeOriginal;
    /**
     * The size of the call stack after static extension.
     */
    public final int stackSizeExtendedStatic;

    public AllocationSite(int id, Location[] callSites, int allocatedTypeId) {
        this.id = id;
        this.callSites = callSites.clone();
        this.allocatedTypeId = allocatedTypeId;
        stackSizeOriginal = this.callSites.length;
        stackSizeExtendedStatic = stackSizeOriginal;
    }

    private AllocationSite(int id, Location[] callSites, int allocatedTypeId, Location[] extendedCallSites) {
        this.id = id;
        this.allocatedTypeId = allocatedTypeId;
        this.callSites = ArraysUtil.concat(Location.class, callSites, extendedCallSites);
        stackSizeOriginal = callSites.length;
        stackSizeExtendedStatic = callSites.length + extendedCallSites.length;
    }

    private AllocationSite(AllocationSite original, int newId, Location[] callSites) {
        id = newId;
        this.callSites = callSites;
        allocatedTypeId = original.allocatedTypeId;
        stackSizeOriginal = original.stackSizeOriginal;
        stackSizeExtendedStatic = original.stackSizeExtendedStatic;
    }

    private AllocationSite(AllocationSite original, int newTypeId) {
        id = original.id;
        callSites = original.callSites.clone();
        allocatedTypeId = newTypeId;
        stackSizeOriginal = original.stackSizeOriginal;
        stackSizeExtendedStatic = original.stackSizeExtendedStatic;
    }

    public int getOriginalID() {
        return id;
    }

    public AllocationSite copy(int newAllocatedTypeId) {
        return new AllocationSite(this, newAllocatedTypeId);
    }

    public AllocationSite extendStatic(Location[] additionalCallSites) {
        return new AllocationSite(id, callSites, allocatedTypeId, additionalCallSites);
    }

    public AllocationSite extendDynamic(int newId, Location[] extendedCallSites) {
        final int originalId = this.id;
        return new AllocationSite(this, newId, extendedCallSites) {
            @Override
            public int getOriginalID() {
                return originalId;
            }
        };
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + allocatedTypeId;
        result = prime * result + id;
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        AllocationSite other = (AllocationSite) obj;
        if (allocatedTypeId != other.allocatedTypeId) {
            return false;
        }
        if (id != other.id) {
            return false;
        }
        return true;
    }

    @Override
    public String toString() {
        StringBuilder string = new StringBuilder();
        for (Location location : callSites) {
            string.append(location + "\n");
        }
        return string.toString().trim();
    }
}
