/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.parser;

import at.jku.mevss.anttracks.heap.GarbageCollectionLookup;
import at.jku.mevss.anttracks.heap.io.HeapIndexReader;
import at.jku.mevss.anttracks.heap.io.HeapPosition;
import at.jku.mevss.anttracks.heap.io.MetaDataConfig;
import at.jku.mevss.anttracks.heap.io.MetaDataWriterConfig;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.util.Consts;
import at.jku.mevss.anttracks.util.MutableLong;
import at.jku.mevss.anttracks.util.TraceException;
import javafx.beans.property.BooleanProperty;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.lang.management.ManagementFactory;
import java.nio.ByteBuffer;
import java.util.*;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.function.BiFunction;
import java.util.logging.Level;
import java.util.logging.Logger;

public abstract class TraceParser<W> {

    protected final MetaDataWriterConfig writerConfig;
    public final Symbols symbols;
    protected final long fromTime;
    protected final long toTime;
    public W workspace;

    private final MetaDataConfig readerConfig;
    private final List<TraceParserListener> traceParserlisteners = new CopyOnWriteArrayList<>();
    protected final Logger logger;

    private long objectsAllocated = 0;
    private long extendedStackFramesStatic = 0;
    private long extendedStackFramesDynamic = 0;

    public static boolean CONSISTENCY_CHECK;
    public static final boolean MULTITHREADING = true;

    private volatile Throwable error;

    private List<BiFunction<W, ParsingInfo, TraceParsingEventHandler>> additionalEventHandlerSupplier;

    /**
     * Constructor to parse the whole file and writing meta-data the WriterConfig's directory
     *
     * @param symbols
     * @param writerConfig
     * @throws Exception
     */
    public TraceParser(Symbols symbols, MetaDataWriterConfig writerConfig) throws Exception {
        this.symbols = symbols;
        this.writerConfig = writerConfig;
        this.readerConfig = null;
        this.fromTime = 0;
        this.toTime = 0;

        this.additionalEventHandlerSupplier = new ArrayList<>();

        logger = Logger.getLogger(this.getClass().getSimpleName() + " " + symbols.root);
    }

    /**
     * Parse from a given GC to another given GC.
     * Watch out! This assumes that meta-data already exists!
     *
     * @param symbols
     * @param readerConfig
     * @param fromGc
     * @param toGc
     * @throws Exception
     */
    public TraceParser(Symbols symbols, MetaDataConfig readerConfig, GarbageCollectionLookup fromGc, GarbageCollectionLookup toGc) throws Exception {
        this.symbols = symbols;
        this.writerConfig = null;
        assert readerConfig != null : "GC lookup relies on metadata";
        this.readerConfig = readerConfig;
        try (HeapIndexReader heapIndexReaderFrom = new HeapIndexReader(readerConfig.path, symbols)) {
            this.fromTime = heapIndexReaderFrom.getGCTime(fromGc);
        }
        try (HeapIndexReader heapIndexReaderFrom = new HeapIndexReader(readerConfig.path, symbols)) {
            this.toTime = heapIndexReaderFrom.getGCTime(toGc);
        }

        logger = Logger.getLogger(TraceParser.class.getSimpleName() + " " + symbols.root);

        this.additionalEventHandlerSupplier = new ArrayList<>();
    }

    /**
     * Parse from a given time to another given time.
     * Watch out! This assumes that meta-data already exists!
     *
     * @param symbols
     * @param readerConfig
     * @param fromTime
     * @param toTime
     * @throws Exception
     */
    public TraceParser(Symbols symbols, MetaDataConfig readerConfig, long fromTime, long toTime) throws Exception {
        this.symbols = symbols;
        this.writerConfig = null;
        assert readerConfig != null : "GC lookup relies on metadata";
        this.readerConfig = readerConfig;
        this.fromTime = fromTime;
        this.toTime = toTime;
        logger = Logger.getLogger(TraceParser.class.getSimpleName() + " " + symbols.root);

        this.additionalEventHandlerSupplier = new ArrayList<>();
    }

    public void addTraceParserListener(TraceParserListener listener) {
        traceParserlisteners.add(listener);
    }

    public W parse() throws IOException, TraceException, InterruptedException {
        return parse(null);
    }

    public W parse(BooleanProperty cancellationToken) throws IOException, TraceException, InterruptedException {
        logger.log(Level.INFO, "preparing to parse trace");

        File file = new File(symbols.root + File.separator + symbols.trace);

        TraceScannerFactory factory = null;

        if (readerConfig == null) {
            assert fromTime == 0 : "If no reader config is given, we must start parsing at the beginning";
            logger.log(Level.INFO, "parsing trace from beginning with empty workspace");
            factory = TraceScannerFactory.create(new File(symbols.root), symbols.header, file);
            workspace = generatePlainWorkspace(factory);
        } else {
            HeapIndexReader heapIndexReaderFrom = new HeapIndexReader(readerConfig.path, symbols);
            HeapPosition heapPositionFrom = heapIndexReaderFrom.getRangeFromLastHeapDumpToGivenTime(fromTime);

            long to;
            try (HeapIndexReader heapIndexReaderTo = new HeapIndexReader(readerConfig.path, symbols)) {
                to = heapIndexReaderTo.getRangeFromLastHeapDumpToGivenTime(toTime).toPosition;
            }

            logger.log(Level.INFO, "parsing trace, starting at " + heapPositionFrom.toString());
            logger.log(Level.INFO, "building workspace");
            workspace = generateWorkspaceFromMetaData(heapIndexReaderFrom, heapPositionFrom);
            logger.log(Level.INFO, "workspace built");

            factory = TraceScannerFactory.create(new File(symbols.root), symbols.header, file, heapPositionFrom.fromPosition, to);
        }

        logger.log(Level.INFO, "parsing trace (length = {0}b)", factory.getLength());
        long time = System.currentTimeMillis();

        try {
            for (Scanner s = factory.getNext(); s != null; s = factory.getNext()) {
                try (Scanner scanner = s) {
                    if (scanner.getPosition() == 0) {
                        TraceFileInfo info = TraceFile.readTraceFileInfo(new InputStream() {
                            @Override
                            public int read() throws IOException {
                                return (scanner.getByte()) & 0xFF;
                            }
                        });
                        if (!Arrays.equals(symbols.header, info.getHeader())) {
                            throw new TraceException("Headers do not match!");
                        }
                    }
                    parse(workspace, scanner, factory.getFrom(), factory.getTo(), new ParsingInfo(fromTime, toTime, factory.getLength()), cancellationToken);
                }
            }

            doParseCleanupAfterSuccessfulParse(workspace);
        } finally {
            doRemoveListenersOnCompletion(workspace);
        }

        doWorkspaceCompletion(workspace);

        time = System.currentTimeMillis() - time;
        logger.log(Level.INFO, "parsed trace in {0}s", (1.0 * time / 1000));

        return workspace;
    }

    protected abstract void doWorkspaceCompletion(W workspace) throws TraceException;

    protected abstract void doRemoveListenersOnCompletion(W workspace);

    protected abstract W generateWorkspaceFromMetaData(HeapIndexReader heapIndexReader, HeapPosition heapPosition) throws FileNotFoundException, IOException;

    protected abstract W generatePlainWorkspace(TraceScannerFactory factory) throws IOException;

    private void parse(W workspace, Scanner scanner, long from, long to, ParsingInfo parsingInfo, BooleanProperty cancellationToken)
            throws IOException, TraceException, InterruptedException {
        logger.log(Level.INFO, "parsing trace @ {0}", scanner.getGlobalPosition());

        final Thread self = Thread.currentThread();
        final MutableLong queueSize = new MutableLong(0);
        final BlockingQueue<QueueEntry> chunks = new LinkedBlockingQueue<>();
        final BlockingQueue<ThreadLocalHeap> masterQueue = new LinkedBlockingQueue<>();
        final Map<String, ThreadLocalHeap> threadLocalHeaps = new HashMap<>();

        logger.log(Level.INFO, "starting slaves");
        final List<TraceSlaveParser<W>> slaves = startSlaveThreads(queueSize, masterQueue, workspace, e -> abort(e, self), parsingInfo);

        logger.log(Level.INFO, "starting master");
        final Thread master = new Thread(() -> processQueueEntries(chunks, masterQueue, threadLocalHeaps, slaves, e -> abort(e, self)), "Master");
        master.start();

        try {
            dispatch(scanner, chunks, queueSize, from, to, cancellationToken);
            master.join();

            logger.log(Level.INFO, "waiting for slaves");
            interruptSlaves(masterQueue, slaves);
            joinSlaves(slaves);
            if (cancellationToken != null && cancellationToken.get()) {
                throw new InterruptedException("Parser has been cancelled using cancellation token");
            }
        } catch (InterruptedException ie) {
            master.interrupt();
            interruptSlaves(masterQueue, slaves);
            if (error != null) {
                throw new TraceException(error);
            } else {
                throw ie;
            }
        }

        logger.info(() -> {
            if (!symbols.useCallContext()) {
                return "Call context analysis disabled, no stack statistics available.";
            }
            for (ThreadLocalHeap heap : threadLocalHeaps.values()) {
                objectsAllocated += heap.objectsAllocated;
                extendedStackFramesStatic += heap.extendedStackFramesStatic;
                extendedStackFramesDynamic += heap.extendedStackFramesDynamic;
            }
            String msg = "%,d objects allocated, with %,d more stack frames from static information and %,d more from dynamic context.";
            return String.format(msg, objectsAllocated, extendedStackFramesStatic, extendedStackFramesDynamic);
        });
        if (symbols.useDynamicCallContext()) {
            logger.info(() -> {
                final String msg = "Tried to expand allocation site %,d times, failed %,d times, succeeded %,d times.";
                final long tries = symbols.getAllocationSiteCreationAttempts();
                final long fails = symbols.getAllocationSiteCreationFailures();
                return String.format(msg, tries, fails, tries - fails);
            });
            logger.info(() -> {
                final long exhausted = symbols.getAllocationContextExhausted();
                return String.format("Previous allocations exhausted for %,d allocation sites.", exhausted);
            });
        }
    }

    private void dispatch(Scanner scanner, BlockingQueue<QueueEntry> chunks, MutableLong queueSize, long from, long to, BooleanProperty cancellationToken)
            throws IOException, InterruptedException, TraceException {
        final Thread self = Thread.currentThread();
        logger.log(Level.INFO, "starting IO");
        final int COMPRESSED_LOCATION = (Integer.SIZE - 1);
        final int COMPRESSED_MASK = 1 << COMPRESSED_LOCATION;
        final int SYNC_LOCATION = Integer.SIZE - 3;
        final int SYNC_MASK = 3 << SYNC_LOCATION;
        final int LENGTH_MASK = ~(COMPRESSED_MASK | SYNC_MASK);

        // At this location, the trace file is read buffer by buffer.
        // Each buffer may contain multiple events.
        // Following data is read per buffer:
        // 1. Thread
        // 2. Meta-data
        // - a.) Length
        // - b.) Is Compressed
        // - c.) Sync-Level
        // 3. Position
        // 4. Data (Byte-Buffer)
        // All of this data is stored in a (size-limited) BlockingQueue<QueueEntry> queue, which is
        // processed by the processQueueEntries() method (run by the master).
        for (String thread = scanner.getThread(symbols.heapWordSize); thread != null && error == null && !self.isInterrupted() && (cancellationToken == null ||
                                                                                                                                           !cancellationToken.get());
             thread = scanner
                     .getThread(symbols.heapWordSize)) {
            int metadata = scanner.getWord();
            int length = metadata & LENGTH_MASK;
            boolean isCompressed = (metadata & COMPRESSED_MASK) != 0;
            SyncLevel sync = SyncLevel.parse((metadata & SYNC_MASK) >>> SYNC_LOCATION);
            long position = scanner.getGlobalPosition();
            synchronized (queueSize) {
                queueSize.add(length);
                queueSize.notifyAll();
                while (queueSize.get() > getMaxQueueSize()) {
                    queueSize.wait();
                }
            }
            ByteBuffer buffer = scanner.getBuffer(length);
            chunks.put(new QueueEntry(thread, buffer, position, isCompressed, sync));
            traceParserlisteners.forEach(l -> l.report(from, to, scanner.getGlobalPosition()));
        }

        logger.log(Level.INFO, "IO finished ({0} undispatched chunks ahead) ({1} bytes ahead)", new Object[]{chunks.size(), queueSize.get()});
        chunks.put(QueueEntry.NULL);
    }

    private long getMaxQueueSize() {
        long memory = ManagementFactory.getMemoryMXBean().getHeapMemoryUsage().getMax();
        long size = (long) (memory * 0.05);
        return Math.min(size, 1024 * 50);
    }

    protected abstract void doParseCleanupAfterSuccessfulParse(W workspace) throws TraceException;

    private synchronized void abort(Throwable e, Thread thread) {
        if (error == null) {
            logger.log(Level.SEVERE, "error reported");
            error = e;
            if (thread != null) {
                thread.interrupt();
            }
        }
    }

    // This method runs in a separate thread and takes care of the buffer queue.
    // It checks if the queue contains an buffer, and if so, the entry gets
    // added to the respective ThreadLocalHeap
    private void processQueueEntries(BlockingQueue<QueueEntry> chunks,
                                     BlockingQueue<ThreadLocalHeap> queue,
                                     Map<String, ThreadLocalHeap> threads,
                                     List<TraceSlaveParser<W>> slaves,
                                     ErrorHandler error) {
        logger.log(Level.INFO, "master started");
        final Thread self = Thread.currentThread();
        int buffers = 0;

        while (!self.isInterrupted()) {
            try {
                QueueEntry entry = chunks.take();
                if (entry == QueueEntry.NULL) {
                    // NULL entry gets added at the end of read() -> File
                    // parsing finished
                    waitUntilParked(threads);
                    insertCleanUps(queue, threads);
                    waitUntilParked(threads);
                    self.interrupt();
                } else {
                    processQueueEntry(queue, threads, entry);
                    buffers++;
                }
            } catch (InterruptedException ie) {
                self.interrupt();
            } catch (Throwable e) {
                logger.log(Level.SEVERE, "fatal error occured", e);
                error.report(e);
                self.interrupt();
            }
        }
        logger.log(Level.INFO, "master finished, dispatched {0} chunks", buffers);
    }

    // This method processes a single buffer (QueueEntry)
    // This method takes care of the sync level, and assigns a QueueEntry to the
    // master queue.
    // The master queue then is processes by the slave parser(s).
    private void processQueueEntry(BlockingQueue<ThreadLocalHeap> queue, Map<String, ThreadLocalHeap> threads, QueueEntry chunk) throws InterruptedException, Exception {
        if (chunk.sync == SyncLevel.FULL) {
            waitUntilParked(threads);
            insertCleanUps(queue, threads);
            waitUntilParked(threads);
            for (ThreadLocalHeap heap : threads.values()) {
                objectsAllocated += heap.objectsAllocated;
                extendedStackFramesStatic += heap.extendedStackFramesStatic;
                extendedStackFramesDynamic += heap.extendedStackFramesDynamic;
            }
            for (String thread : threads.keySet().toArray(new String[threads.size()])) {
                ThreadLocalHeap heap = threads.get(thread);
                if (heap.currentLabs.size() == 0) {
                    threads.remove(thread);
                }
            }
        } else if (chunk.sync == SyncLevel.ENSURE_ORDER) {
            waitUntilParked(threads);
        } else {
            assert chunk.sync == SyncLevel.NONE;
        }

        ThreadLocalHeap threadLocalHeap = threads.get(chunk.thread);

        if (threadLocalHeap == null) {
            threadLocalHeap = createNewThreadLocalHeap(queue, threads, chunk.thread);
        }
        addToThreadLocalHeap(queue, chunk, threadLocalHeap);

        if (chunk.sync != SyncLevel.NONE || !MULTITHREADING) {
            waitUntilParked(threadLocalHeap);
        }
    }

    private List<TraceSlaveParser<W>> startSlaveThreads(MutableLong queueSize,
                                                        BlockingQueue<ThreadLocalHeap> queue,
                                                        W workspace,
                                                        ErrorHandler handler,
                                                        ParsingInfo parsingInfo)
            throws IOException {
        return startSlaveThreads(queueSize, queue, workspace, handler, CONSISTENCY_CHECK, parsingInfo);
    }

    protected List<TraceSlaveParser<W>> startSlaveThreads(MutableLong queueSize,
                                                          BlockingQueue<ThreadLocalHeap> masterQueue,
                                                          W workspace,
                                                          ErrorHandler handler,
                                                          boolean check,
                                                          ParsingInfo parsingInfo) throws IOException {
        RelAddrFactory relAddrFactory = new RelAddrFactory(symbols.heapWordSize);

        int slaves = 1;
        if (MULTITHREADING) {
            slaves = Consts.AVAILABLE_PROCESSORS();
        }

        List<TraceSlaveParser<W>> result = new ArrayList<>();
        for (int i = 0; i < slaves; i++) {
            TraceSlaveParser<W> parserThread = new TraceSlaveParser<>(i,
                                                                      queueSize,
                                                                      masterQueue,
                                                                      workspace,
                                                                      relAddrFactory,
                                                                      symbols,
                                                                      check,
                                                                      handler,
                                                                      createMainEventHandler(parsingInfo),
                                                                      additionalEventHandlerSupplier.stream()
                                                                                                    .map(supplier -> supplier.apply(workspace, parsingInfo))
                                                                                                    .toArray(TraceParsingEventHandler[]::new));
            result.add(parserThread);
        }
        return result;
    }

    private void insertCleanUps(BlockingQueue<ThreadLocalHeap> queue, Map<String, ThreadLocalHeap> threads) throws InterruptedException {
        for (String key : threads.keySet()) {
            ThreadLocalHeap threadLocalHeap = threads.get(key);
            synchronized (threadLocalHeap) {
                assert (threadLocalHeap.state == ThreadLocalHeap.STATE_PARKED);
                threadLocalHeap.state = ThreadLocalHeap.STATE_IN_QUEUE_FOR_CLEAN_UP;
                queue.add(threadLocalHeap);
            }
        }
    }

    private void waitUntilParked(Map<String, ThreadLocalHeap> threadMap) throws InterruptedException {
        for (String key : threadMap.keySet()) {
            ThreadLocalHeap threadLocalHeap = threadMap.get(key);
            waitUntilParked(threadLocalHeap);
        }
    }

    private void waitUntilParked(ThreadLocalHeap threadLocalHeap) throws InterruptedException {
        synchronized (threadLocalHeap) {
            while (threadLocalHeap.state != ThreadLocalHeap.STATE_PARKED) {
                threadLocalHeap.wait();
            }
        }

    }

    private ThreadLocalHeap createNewThreadLocalHeap(BlockingQueue<ThreadLocalHeap> queue, Map<String, ThreadLocalHeap> threads, String thread)
            throws InterruptedException {
        ThreadLocalHeap threadLocalHeap = new ThreadLocalHeap(thread, ThreadLocalHeap.STATE_IN_QUEUE);
        threads.put(thread, threadLocalHeap);
        queue.add(threadLocalHeap);
        return threadLocalHeap;
    }

    private void addToThreadLocalHeap(BlockingQueue<ThreadLocalHeap> masterQueue, QueueEntry entry, ThreadLocalHeap threadLocalHeap) throws Exception {
        if (threadLocalHeap == null) {
            throw new TraceException("TLH must already exist.");
        }
        synchronized (threadLocalHeap) {
            threadLocalHeap.queue.add(entry);

            if (threadLocalHeap.state == ThreadLocalHeap.STATE_PARKED) {
                threadLocalHeap.state = ThreadLocalHeap.STATE_IN_QUEUE;
                masterQueue.add(threadLocalHeap);
            }
        }
    }

    private void interruptSlaves(BlockingQueue<ThreadLocalHeap> queue, List<TraceSlaveParser<W>> slaves) {
        for (TraceSlaveParser<W> thread : slaves) {
            thread.interrupt();
        }
    }

    private void joinSlaves(List<TraceSlaveParser<W>> slaves) throws InterruptedException {
        for (TraceSlaveParser<W> thread : slaves) {
            thread.join();
        }
    }

    protected abstract TraceParsingEventHandler createMainEventHandler(ParsingInfo parsingInfo);

    public void addEventHandler(BiFunction<W, ParsingInfo, TraceParsingEventHandler> handlerSupplier) {
        this.additionalEventHandlerSupplier.add(handlerSupplier);
    }
}
