/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.parser;

import at.jku.mevss.anttracks.heap.GarbageCollectionCause;
import at.jku.mevss.anttracks.heap.GarbageCollectionType;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.roots.RootPtr;
import at.jku.mevss.anttracks.heap.space.Space;
import at.jku.mevss.anttracks.heap.space.SpaceMode;
import at.jku.mevss.anttracks.heap.space.SpaceType;
import at.jku.mevss.anttracks.heap.symbols.AllocatedType;
import at.jku.mevss.anttracks.heap.symbols.AllocatedTypes;
import at.jku.mevss.anttracks.heap.symbols.AllocationSite;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.parser.heap.pointer.PtrEvent;
import at.jku.mevss.anttracks.util.Consts;
import at.jku.mevss.anttracks.util.MutableLong;
import at.jku.mevss.anttracks.util.TraceException;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.logging.Level;
import java.util.logging.Logger;

import static at.jku.mevss.anttracks.heap.objects.InstanceMirrorObjectInfo.MIRROR_CLASS;
import static at.jku.mevss.anttracks.heap.roots.RootPtr.RootType;
import static at.jku.mevss.anttracks.heap.symbols.AllocatedTypes.MIRROR_CLASS_NAME;

public class TraceSlaveParser<W> {
    public static final int NULL_PTR = -1;

    // --------------------------------------------------------
    // ----------------- GCInfo -------------------------------
    // --------------------------------------------------------

    protected ParserGCInfo parseGCMeta(int word) throws TraceException {
        GarbageCollectionType gcTypeId = GarbageCollectionType.fromID(recoverValue(word, 1, 1));
        GarbageCollectionCause gcCause = symbols.causes.get(recoverValue(word, 2, 2));

        int flags = recoverValue(word, 3, 3);
        boolean concurrent = (flags & (1 << 1)) != 0;

        int id = getNextWord();
        long time = getNextDoubleWord();

        long base = getNextDoubleWord();
        relAddrFactory.setBase(base);

        return new ParserGCInfo(gcTypeId, gcCause, id, time, concurrent);
    }

    // --------------------------------------------------------
    // ------------ AbstractTraceSlaveParser ------------------
    // --------------------------------------------------------

    // --------------------------------------------------------
    // ------------ Fields ------------------------------------
    // --------------------------------------------------------

    protected final W workspace;
    protected final RelAddrFactory relAddrFactory;
    protected final Symbols symbols;
    protected final Logger logger;

    protected static final int ARRAY_SMALL_LENGTH_INDEX = 3;
    protected static final int ALLOCATION_SITE_INDEX_1 = 1;
    protected static final int ALLOCATION_SITE_INDEX_2 = 2;
    protected static final int ALLOCATION_SITE_INDEX_3 = 3; // big allocSite
    protected static final int EVENT_TYPE_INDEX = 0;

    private final MutableLong size;
    private final BlockingQueue<ThreadLocalHeap> masterQueue;
    private final Decompressor decompressor;
    private final boolean check;

    private final ErrorHandler error;
    private final Thread worker;

    private ByteBuffer buffer;

    private final List<TraceParsingEventHandler> otherParsers;
    private final TraceParsingEventHandler mainParser;

    // --------------------------------------------------------
    // ------------ ctor --------------------------------------
    // --------------------------------------------------------

    public TraceSlaveParser(int id,
                            MutableLong size,
                            BlockingQueue<ThreadLocalHeap> masterQueue,
                            W workspace,
                            RelAddrFactory relAddrFactory,
                            Symbols symbols,
                            boolean test,
                            ErrorHandler error,
                            TraceParsingEventHandler mainParser,
                            TraceParsingEventHandler... otherParsers) {
        this.size = size;
        this.masterQueue = masterQueue;
        this.workspace = workspace;
        this.relAddrFactory = relAddrFactory;
        this.symbols = symbols;
        this.decompressor = new Decompressor();
        this.check = test;
        this.error = error;
        worker = new Thread(this::run, "Slave " + id);
        buffer = null;
        logger = Logger.getLogger(this.getClass().getSimpleName() + " " + id + " " + symbols.root);
        worker.start();

        this.mainParser = mainParser;
        this.otherParsers = Arrays.asList(otherParsers);
    }

    // --------------------------------------------------------
    // ------------ Methods ------------------------------------
    // --------------------------------------------------------

    public void interrupt() {
        worker.interrupt();
    }

    public void join() throws InterruptedException {
        worker.join();
    }

    private void run() {
        logger.log(Level.INFO, "slave started");
        assert Thread.currentThread() == worker;
        int parsedBuffers = 0;

        while (!worker.isInterrupted()) {
            try {
                ThreadLocalHeap threadLocalHeap = masterQueue.take();
                boolean cleanUp;
                synchronized (threadLocalHeap) {
                    if (threadLocalHeap.state == ThreadLocalHeap.STATE_IN_QUEUE) {
                        threadLocalHeap.state = ThreadLocalHeap.STATE_IN_PROCESS;
                        cleanUp = false;
                    } else if (threadLocalHeap.state == ThreadLocalHeap.STATE_IN_QUEUE_FOR_CLEAN_UP) {
                        threadLocalHeap.state = ThreadLocalHeap.STATE_IN_PROCESS;
                        cleanUp = true;
                    } else {
                        errorOnthreadLocalHeapState(threadLocalHeap);
                        return; // to make the compiler happy :-)
                    }
                }

                for (QueueEntry entry = threadLocalHeap.queue.poll(); entry != null; entry = threadLocalHeap.queue.poll()) {
                    buffer = entry.buffer;
                    long start = entry.position;
                    long end = start + buffer.limit();
                    if (entry.isCompressed) {
                        buffer = decompressor.decode(buffer);
                    }
                    if (buffer.order() != ByteOrder.LITTLE_ENDIAN) {
                        buffer.order(ByteOrder.LITTLE_ENDIAN);
                    }

                    while (buffer.position() != buffer.limit()) {
                        evaluateEvent(threadLocalHeap, start, end);
                    }

                    buffer = null; // to avoid memory leak with last buffer
                    parsedBuffers++;
                    synchronized (size) {
                        size.add(-(end - start));
                        size.notifyAll();
                    }
                }

                if (cleanUp) {
                    // Cleanups happen:
                    // 1. When a file has been read (i.e., when multiple trace files are generated)
                    // 2. When an event with SyncLevel = FULL is sent
                    cleanUp(threadLocalHeap);
                }
                parkThreadLocalHeap(threadLocalHeap);
            } catch (InterruptedException ie) {
                worker.interrupt();
            } catch (Throwable e) {
                logger.log(Level.SEVERE, "fatal error", e);
                error.report(e);
                worker.interrupt();
            }

        }

        logger.log(Level.INFO, "slave finished, handled {0} chunks", parsedBuffers);
    }

    private void parkThreadLocalHeap(ThreadLocalHeap threadLocalHeap) {
        synchronized (threadLocalHeap) {
            assert threadLocalHeap.state == ThreadLocalHeap.STATE_IN_PROCESS;
            if (threadLocalHeap.queue.isEmpty()) {
                threadLocalHeap.state = ThreadLocalHeap.STATE_PARKED;
            } else {
                // this happens in rare cases where we did not get another
                // buffer (queue is empty) and we start parking while the master
                // is adding another entry to the queue
                // in this case we end up with a thread local heap that is
                // parked, but has an element in the queue which contains
                // important events that will be missed (like missing foreign
                // filler oops, holes in the heap, non-full LABs)
                threadLocalHeap.state = ThreadLocalHeap.STATE_IN_QUEUE;
                masterQueue.add(threadLocalHeap);
            }
            threadLocalHeap.notifyAll();
        }
    }

    protected void cleanUp(ThreadLocalHeap threadLocalHeap) throws TraceException {
        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doCleanUp(threadLocalHeap);
        }
        mainParser.doCleanUp(threadLocalHeap);
    }

    private void evaluateEvent(ThreadLocalHeap threadLocalHeap, long start, long end) throws TraceException {

        int word = getNextWord();
        long unsignedWord = word & 0x00000000ffffffffL;
        int recovered = recoverValue(unsignedWord, EVENT_TYPE_INDEX, EVENT_TYPE_INDEX);
        int toSpace = recovered & 0x3;

        EventType type = EventType.parse(recovered >> 2);

        switch (type) {
            case TLAB_ALLOC: {
                parseTlabAlloc(word, threadLocalHeap);
                break;
            }
            case PLAB_ALLOC: {
                parsePlabAlloc(word, threadLocalHeap);
                break;
            }
            case OBJ_ALLOC_FAST_IR: {
                parseObjAllocFastIr(type, word, threadLocalHeap);
                break;
            }
            case OBJ_ALLOC_FAST_C1:
            case OBJ_ALLOC_FAST_C2: {
                parseObjAllocFastCi(type, word, threadLocalHeap);
                break;
            }
            case OBJ_ALLOC_NORMAL_IR: {
                parseObjAllocNormalIr(type, word, threadLocalHeap);
                break;
            }
            case OBJ_ALLOC_NORMAL_C1:
            case OBJ_ALLOC_NORMAL_C2: {
                parseObjAllocNormalCi(type, word, threadLocalHeap);
                break;
            }
            case OBJ_ALLOC_SLOW: {
                parseObjAllocSlow(type, word, threadLocalHeap);
                break;
            }
            case OBJ_ALLOC_SLOW_IR:
            case OBJ_ALLOC_SLOW_C1:
            case OBJ_ALLOC_SLOW_C2:
            case OBJ_ALLOC_SLOW_IR_DEVIANT_TYPE:
            case OBJ_ALLOC_SLOW_C1_DEVIANT_TYPE:
            case OBJ_ALLOC_SLOW_C2_DEVIANT_TYPE: {
                parseObjAllocSlowCiIr_Deviant(type, word, threadLocalHeap);
                break;
            }
            case GC_START: {
                parseGCStart(word, start, end, threadLocalHeap);
                break;
            }
            case GC_END: {
                parseGCEnd(word, threadLocalHeap, start, end);
                break;
            }
            case GC_MOVE_FAST:
            case GC_MOVE_FAST_WIDE:
            case GC_MOVE_FAST_NARROW:
            case GC_MOVE_FAST_PTR:
            case GC_MOVE_FAST_WIDE_PTR: {
                parseGCMoveFast_WideOrNarrow_Ptr(type, word, threadLocalHeap, toSpace);
                break;
            }
            case GC_MOVE_SLOW:
            case GC_MOVE_SLOW_PTR: {
                parseGCMoveSlow(type, word, threadLocalHeap);
                break;
            }
            case GC_KEEP_ALIVE:
            case GC_KEEP_ALIVE_PTR: {
                parseGCKeepAlive(type, word, threadLocalHeap);
                break;
            }
            case GC_ROOT_PTR: {
                parseGCRootPtr(type, word, threadLocalHeap);
                break;
            }
            case GC_PTR_EXTENSION:
            case GC_PTR_UPDATE_PREMOVE:
            case GC_PTR_UPDATE_POSTMOVE:
            case GC_PTR_MULTITHREADED: {
                parseGCObjPtr(type, word, threadLocalHeap);
                break;
            }
            case GC_MOVE_REGION: {
                parseGCMoveRegion(type, word, threadLocalHeap);
                break;
            }
            case SYNC_OBJ_NARROW:
            case SYNC_OBJ: {
                parseSyncObj(type, word, threadLocalHeap);
                break;
            }
            case THREAD_ALIVE: {
                parseThreadAlive(type, word);
                break;
            }
            case THREAD_DEATH: {
                parseThreadDeath(type, word);
                break;
            }
            case OBJ_ALLOC_FAST_C2_DEVIANT_TYPE: {
                parseObjAllocFastC2DeviantType(type, word, threadLocalHeap);
                break;
            }
            case NOP:
                break;
            case GC_DEALLOCATION: {
                throw new Error("Not implemented");
            }
            case SPACE_CREATE: {
                parseSpaceCreate(word);
                break;
            }
            case SPACE_ALLOC: {
                parseSpaceAlloc(word);
                break;
            }
            case SPACE_RELEASE: {
                parseSpaceRelease(word);
                break;
            }
            case SPACE_REDEFINE: {
                parseSpaceRedefine(word);
                break;
            }
            case SPACE_DESTROY: {
                parseSpaceDestroy(word);
                break;
            }
            case GC_INFO: {
                parseGCInfo(word);
                break;
            }
            case GC_FAILED: {
                parseGCFailed(word);
                break;
            }
            case GC_INTERRUPT: {
                parseGCInterrupt(word);
                break;
            }
            case GC_CONTINUE: {
                parseGCContinue(word);
                break;
            }
            case GC_TAG: {
                parseGCTag(word);
                break;
            }
            default: {
                throw new TraceException("PARSE ERROR. Invalid trace file!. Unknown EventType=" + type);
            }
        }

        if (symbols.anchors && type.mayBeFollowedByAnchor) {
            int anchor = buffer.getInt();
            if ((anchor & Consts.ANCHOR_MASK) != Consts.ANCHOR_MASK) {
                throw new AssertionError("Anchor missing (" + (start + buffer.position()) + ")");
            }
            EventType check = EventType.parse((anchor & ~Consts.ANCHOR_MASK) >> (8 * 1));
            if (check != type) {
                throw new AssertionError("Anchor inconsistent (" + (start + buffer.position()) + ")");
            }
        }
    }

    private void parseGCTag(int word) throws TraceException {
        int tagTextLength = getNextWord();
        String tagText = getString(tagTextLength);

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseGCTag(tagText);
        }
        mainParser.doParseGCTag(tagText);
    }

    protected void parseSpaceRedefine(int word) throws TraceException {
        int index = getNextWord();
        long startAddr = getNextDoubleWord();
        long size = getNextDoubleWord();
        // boolean postponed = getNextWord() > 0;
        // int source = getNextWord();

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseSpaceRedefine(index, startAddr, size, false, -1);
        }
        mainParser.doParseSpaceRedefine(index, startAddr, size, false, -1);
    }

    protected void parseSpaceDestroy(int word) throws TraceException {
        int firstIndex = getNextWord();
        long nRegions = getNextDoubleWord();

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseSpaceDestroy(firstIndex, nRegions);
        }
        mainParser.doParseSpaceDestroy(firstIndex, nRegions);
    }

    protected void parseSpaceRelease(int word) throws TraceException {
        int index = getNextWord();

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseSpaceRelease(index);
        }
        mainParser.doParseSpaceRelease(index);
    }

    protected void parseSpaceAlloc(int word) throws TraceException {
        SpaceType spaceType = SpaceType.values()[(word >> 18) & 0xFF];
        SpaceMode spaceMode = SpaceMode.values()[(word >> 10) & 0xFF];
        assert spaceType != null;
        assert spaceMode != null;
        int index = getNextWord();

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseSpaceAlloc(index, spaceMode, spaceType);
        }
        mainParser.doParseSpaceAlloc(index, spaceMode, spaceType);
    }

    protected void parseSpaceCreate(int word) throws TraceException {
        // int spaceType = (word >> 22) & 0b1111;
        int index = getNextWord();
        long startAddr = getNextDoubleWord();
        long size = getNextDoubleWord();

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseSpaceCreate(index, startAddr, size);
        }
        mainParser.doParseSpaceCreate(index, startAddr, size);
    }

    protected void parseGCInfo(int word) throws TraceException {
        int index = word & 0x00FFFFFF;
        int id = getNextWord();

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseGCInfo(index, id);
        }
        mainParser.doParseGCInfo(index, id);
    }

    protected void parseGCFailed(int word) throws TraceException {
        int index = word & 0x00FFFFFF;

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseGCFailed(index);
        }
        mainParser.doParseGCFailed(index);
    }

    protected void parseGCStart(int word, long start, long end, ThreadLocalHeap threadLocalHeap) throws TraceException {
        ParserGCInfo info = parseGCMeta(word);

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseGCStart(info, start, end, check);
        }
        mainParser.doParseGCStart(info, start, end, check);

    }

    protected void parseGCEnd(int word, ThreadLocalHeap threadLocalHeap, long start, long end) throws TraceException {
        ParserGCInfo info = parseGCMeta(word);
        int flags = recoverValue(word, 3, 3);
        boolean failed = (flags & 1) != 0;

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseGCEnd(info, threadLocalHeap, start, end, failed, check);
        }
        mainParser.doParseGCEnd(info, threadLocalHeap, start, end, failed, check);
    }

    protected void parseGCInterrupt(int word) throws TraceException {
        int id = getNextWord();
        long address = getNextDoubleWord();

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseGCInterrupt(id, address);
        }
        mainParser.doParseGCInterrupt(id, address);
    }

    protected void parseGCContinue(int word) throws TraceException {
        int id = getNextWord();
        long address = getNextDoubleWord();

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseGCContinue(id, address);
        }
        mainParser.doParseGCContinue(id, address);
    }

    protected void parseGCMoveFast_WideOrNarrow_Ptr(EventType type, int word, ThreadLocalHeap threadLocalHeap, int toSpace) throws TraceException {
        SpaceType toSpaceType = RelAddrFactory.getDefinedSpaceOnly(toSpace);

        long toAddr = -1;
        long fromAddr = -1;

        if (type == EventType.GC_MOVE_FAST_NARROW) {
            int from = (int) (RelAddrFactory.getDefinedAddrOnly(recoverValue(word, 1, 3)) >> 2);
            fromAddr = relAddrFactory.create(from);
        } else if (type == EventType.GC_MOVE_FAST_WIDE || type == EventType.GC_MOVE_FAST_WIDE_PTR) {
            fromAddr = getNextDoubleWord();
        } else if (type == EventType.GC_MOVE_FAST || type == EventType.GC_MOVE_FAST_PTR) {
            fromAddr = relAddrFactory.create(getNextWord());
        } else {
            assert false;
            throw new TraceException("internal error");
        }

        Space fromSpace = getSpace(fromAddr);
        ObjectInfo obj = getObjectIsFiller(fromSpace, fromAddr);
        boolean isFiller = obj == null ? false : true;

        long assignedAddr;

        if (!isFiller) {
            obj = fromSpace.getNonFillerObjectInfo(fromAddr);
        }
        assignedAddr = doGCMove(type, threadLocalHeap, obj, fromAddr, toAddr, toSpaceType);

        if (type == EventType.GC_MOVE_FAST_PTR || type == EventType.GC_MOVE_FAST_WIDE_PTR) {
            assert (getMoveTarget(fromAddr, toAddr, toSpaceType, threadLocalHeap) == assignedAddr);
            long[] ptrs = parsePointers(word, assignedAddr, threadLocalHeap);
            doPtrEvent(type, threadLocalHeap, fromAddr, assignedAddr, ptrs, isFiller);
        }
    }

    protected void parseGCMoveSlow(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        long fromAddr = getNextDoubleWord();
        long toAddr = getNextDoubleWord();

        Space fromSpace = getSpace(fromAddr);
        ObjectInfo obj = getObjectIsFiller(fromSpace, fromAddr);
        boolean isFiller = obj == null ? false : true;

        if (!isFiller) {
            obj = fromSpace.getNonFillerObjectInfo(fromAddr);
        }
        doGCMove(type, threadLocalHeap, obj, fromAddr, toAddr, getSpace(toAddr).getType());

        if (type == EventType.GC_MOVE_SLOW_PTR) {
            long[] ptrs = parsePointers(word, toAddr, threadLocalHeap);
            doPtrEvent(type, threadLocalHeap, fromAddr, toAddr, ptrs, isFiller);
        }
    }

    protected void parseGCKeepAlive(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        long addr = getNextDoubleWord();
        Space space = getSpace(addr);
        ObjectInfo obj = getObjectIsFiller(space, addr);
        boolean isFiller = obj == null ? false : true;

        if (addr == 16113991664l) {
            System.out.println("badboi");
        }

        if (!isFiller) {
            obj = space.getNonFillerObjectInfo(addr);
        }

        doGCMove(type, threadLocalHeap, obj, addr, addr, getSpace(addr).getType());

        if (type == EventType.GC_KEEP_ALIVE_PTR) {
            long[] ptrs = parsePointers(word, addr, threadLocalHeap);
            doPtrEvent(type, threadLocalHeap, addr, addr, ptrs, isFiller);
        }
    }

    protected void parseGCRootPtr(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {

        for (int i = 0; i < RootPtr.MAX_ROOTS_PER_EVENT; i++) {
            long ptr = getNextDoubleWord();
            if (ptr == -1) {    // not a nullptr (those are 0) but the convention for end of root block
                // no more roots in this block
                break;
            } else if (ptr == 0) {
                // NULLPTR
                ptr = NULL_PTR;    // because we use -1 for NULLPTRs
            }

            RootType rootType = RootType.values()[getNextWord()];
            switch (rootType) {
                case CLASS_LOADER_ROOT:
                    String loaderName = getString(getNextWord());
                    for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
                        TraceParsingEventHandler parser = otherParsers.get(parserNr);
                        parser.doParseGCClassLoaderRootPtr(ptr, loaderName);
                    }
                    mainParser.doParseGCClassLoaderRootPtr(ptr, loaderName);
                    break;

                case CLASS_ROOT:
                    int classId = getNextWord();
                    for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
                        TraceParsingEventHandler parser = otherParsers.get(parserNr);
                        parser.doParseGCClassRootPtr(ptr, classId);
                    }
                    mainParser.doParseGCClassRootPtr(ptr, classId);
                    break;

                case STATIC_FIELD_ROOT:
                    classId = getNextWord();
                    int offset = getNextWord();
                    for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
                        TraceParsingEventHandler parser = otherParsers.get(parserNr);
                        parser.doParseGCStaticFieldRootPtr(ptr, classId, offset);
                    }
                    mainParser.doParseGCStaticFieldRootPtr(ptr, classId, offset);
                    break;

                case LOCAL_VARIABLE_ROOT:
                    long threadId = getNextDoubleWord();
                    classId = getNextWord();
                    int methodId = getNextWord();
                    int slot = getNextWord();
                    for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
                        TraceParsingEventHandler parser = otherParsers.get(parserNr);
                        parser.doParseGCLocalVariableRootPtr(ptr, threadId, classId, methodId, slot);
                    }
                    mainParser.doParseGCLocalVariableRootPtr(ptr, threadId, classId, methodId, slot);
                    break;

                case VM_INTERNAL_THREAD_DATA_ROOT:
                    threadId = getNextDoubleWord();
                    for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
                        TraceParsingEventHandler parser = otherParsers.get(parserNr);
                        parser.doParseGCVMInternalThreadDataRootPtr(ptr, threadId);
                    }
                    mainParser.doParseGCVMInternalThreadDataRootPtr(ptr, threadId);
                    break;

                case CODE_BLOB_ROOT:
                    classId = getNextWord();
                    methodId = getNextWord();
                    for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
                        TraceParsingEventHandler parser = otherParsers.get(parserNr);
                        parser.doParseGCCodeBlobRootPtr(ptr, classId, methodId);
                    }
                    mainParser.doParseGCCodeBlobRootPtr(ptr, classId, methodId);
                    break;

                case JNI_LOCAL_ROOT:
                    threadId = getNextDoubleWord();
                    for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
                        TraceParsingEventHandler parser = otherParsers.get(parserNr);
                        parser.doParseGCJNILocalRootPtr(ptr, threadId);
                    }
                    mainParser.doParseGCJNILocalRootPtr(ptr, threadId);
                    break;

                case JNI_GLOBAL_ROOT:
                    boolean weak = getNextWord() != 0;
                    for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
                        TraceParsingEventHandler parser = otherParsers.get(parserNr);
                        parser.doParseGCJNIGlobalRootPtr(ptr, weak);
                    }
                    mainParser.doParseGCJNIGlobalRootPtr(ptr, weak);
                    break;

                // other roots
                case CLASS_LOADER_INTERNAL_ROOT:
                case UNIVERSE_ROOT:
                case SYSTEM_DICTIONARY_ROOT:
                case BUSY_MONITOR_ROOT:
                case INTERNED_STRING:
                case FLAT_PROFILER_ROOT:
                case MANAGEMENT_ROOT:
                case JVMTI_ROOT:
                    for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
                        TraceParsingEventHandler parser = otherParsers.get(parserNr);
                        parser.doParseGCOtherRootPtr(ptr, rootType);
                    }
                    mainParser.doParseGCOtherRootPtr(ptr, rootType);
                    break;

                case DEBUG_ROOT:
                    String vmCall = getString(getNextWord());
                    for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
                        TraceParsingEventHandler parser = otherParsers.get(parserNr);
                        parser.doParseGCDebugRootPtr(ptr, vmCall);
                    }
                    mainParser.doParseGCDebugRootPtr(ptr, vmCall);
                    break;

                default:
                    throw new TraceException(rootType + " is not a valid root type id!");
            }
        }
    }

    protected void parseGCObjPtr(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        long addr = getNextDoubleWord();
        long[] ptrs = parsePointers(word, addr, threadLocalHeap);
        doPtrEvent(type, threadLocalHeap, -1, addr, ptrs, false);
    }

    protected List<ObjectInfo> parseGCMoveRegion(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int numOfObjects = recoverValue(word, 1, 3);
        long fromAddr = getNextDoubleWord();
        long toAddr = getNextDoubleWord();
        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseGCMoveRegion(type, fromAddr, toAddr, getSpace(toAddr).getType(), numOfObjects, threadLocalHeap);
        }
        return mainParser.doParseGCMoveRegion(type, fromAddr, toAddr, getSpace(toAddr).getType(), numOfObjects, threadLocalHeap);
    }

    protected void parseThreadAlive(EventType type, int header) throws TraceException {
        final int MAX_NAME_LENGTH = 12;
        long id = getNextDoubleWord();
        StringBuilder name = new StringBuilder();
        end:
        for (int i = 0; i < MAX_NAME_LENGTH; i++) {
            int word = getNextWord();
            for (int j = 0; j < 4; j++) {
                char c = (char) ((word >> ((3 - j) * 8)) & 0xFF);
                if (c == '\0') {
                    break end;
                }
                name.append(c);
            }
        }

        String finalName = name.toString();

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseThreadAlive(header, id, finalName);
        }
        mainParser.doParseThreadAlive(header, id, finalName);
    }

    protected void parseThreadDeath(EventType type, int word) throws TraceException {
        long id = getNextDoubleWord();

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseThreadDeath(id);
        }
        mainParser.doParseThreadDeath(id);
    }

    protected void parseSyncObj(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int allocationSiteId = recoverValue(word, 1, 3);
        int allocatedTypeId = getNextWord();
        AllocatedType allocatedType = symbols.types.get(allocatedTypeId);

        long fromAddr = getNextDoubleWord();
        long toAddr = type == EventType.SYNC_OBJ ? getNextDoubleWord() : fromAddr;
        int length;
        if (allocatedType.name.startsWith("[")) {
            length = getNextWord();
        } else {
            length = Consts.UNDEFINED_LENGTH;
        }
        int size;
        // TODO: We reworked mirror classes, probably something has to be changed here!
        if (allocatedType.name.equals(MIRROR_CLASS_NAME)) {
            size = getNextWord();
        } else {
            size = 0;
        }

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseSyncObj(type, allocationSiteId, allocatedTypeId, allocatedType, fromAddr, toAddr, getSpace(toAddr).getType(), length, size, threadLocalHeap);
        }
        mainParser.doParseSyncObj(type, allocationSiteId, allocatedTypeId, allocatedType, fromAddr, toAddr, getSpace(toAddr).getType(), length, size, threadLocalHeap);
    }

    protected void parseObjAllocFastCi(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_2);
        // check if first bit is set, if yes --> big allocSite (3 byte), otherwise small allocSite (2 byte)
        if (((allocationSiteId >> 15) & 1) != 0) {
            allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_3);
        }
        AllocationSite allocationSite = symbols.sites.getById(allocationSiteId);
        assert allocationSite != null;
        AllocatedType allocatedType = symbols.types.get(allocationSite.allocatedTypeId);

        boolean isArray = allocatedType.name.startsWith("[");
        int arrayLength = Consts.UNDEFINED_LENGTH;
        if (isArray) {
            arrayLength = recoverArrayLength(word);
        }

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseObjAllocFastCi(type, allocationSiteId, allocationSite, allocatedType, isArray, arrayLength, threadLocalHeap);
        }
        mainParser.doParseObjAllocFastCi(type, allocationSiteId, allocationSite, allocatedType, isArray, arrayLength, threadLocalHeap);
    }

    protected void parseObjAllocFastC2DeviantType(EventType type, int header, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int allocationSiteId = recoverValue(header, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_2);
        // check if first bit is set, if yes --> big allocSite (3 byte), otherwise small allocSite (2 byte)
        if (((allocationSiteId >> 15) & 1) != 0) {
            allocationSiteId = recoverValue(header, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_3);
        }
        AllocatedType allocatedType = symbols.types.get(getNextWord());

        boolean isArray = allocatedType.name.startsWith("[");
        int arrayLength = Consts.UNDEFINED_LENGTH;
        if (isArray) {
            arrayLength = recoverArrayLength(header);
        }

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseObjAllocFastC2DeviantType(type, header, allocationSiteId, allocatedType, isArray, arrayLength, threadLocalHeap);
        }
        mainParser.doParseObjAllocFastC2DeviantType(type, header, allocationSiteId, allocatedType, isArray, arrayLength, threadLocalHeap);
    }

    protected void parseObjAllocFastIr(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_2);
        // check if first bit is set, if yes --> big allocSite (3 byte), otherwise small allocSite (2 byte)
        if (((allocationSiteId >> 15) & 1) != 0) {
            allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_3);
        }

        AllocationSite allocationSite = symbols.sites.getById(allocationSiteId);

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseObjAllocFastIr(type, allocationSiteId, allocationSite, threadLocalHeap);
        }
        mainParser.doParseObjAllocFastIr(type, allocationSiteId, allocationSite, threadLocalHeap);
    }

    protected void parseObjAllocNormalCi(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_2);
        // check if first bit is set, if yes --> big allocSite (3 byte), otherwise small allocSite (2 byte)
        if (((allocationSiteId >> 15) & 1) != 0) {
            allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_3);
        }
        AllocationSite allocationSite = symbols.sites.getById(allocationSiteId);
        AllocatedType allocatedType = symbols.types.get(allocationSite.allocatedTypeId);
        errorOnMirrorClass(allocatedType);

        boolean isArray = allocatedType.name.startsWith("[");
        long addr = getNextDoubleWord();
        int arrayLength = Consts.UNDEFINED_LENGTH;
        if (isArray) {
            arrayLength = recoverArrayLength(word);
        }

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseObjAllocNormalCi(type, allocationSiteId, allocationSite, allocatedType, addr, isArray, arrayLength, threadLocalHeap);
        }
        mainParser.doParseObjAllocNormalCi(type, allocationSiteId, allocationSite, allocatedType, addr, isArray, arrayLength, threadLocalHeap);
    }

    protected void parseObjAllocNormalIr(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {

        int allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_2);
        // check if first bit is set, if yes --> big allocSite (3 byte), otherwise small allocSite (2 byte)
        if (((allocationSiteId >> 15) & 1) != 0) {
            allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_3);
        }
        AllocationSite allocationSite = symbols.sites.getById(allocationSiteId);
        long addr = getNextDoubleWord();

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseObjAllocNormalIr(type, allocationSiteId, allocationSite, addr, threadLocalHeap);
        }
        mainParser.doParseObjAllocNormalIr(type, allocationSiteId, allocationSite, addr, threadLocalHeap);
    }

    protected void parseObjAllocSlowCiIr_Deviant(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_2);

        boolean bigAllocSite = false;
        // check if first bit is set, if yes --> big allocSite (3 byte), otherwise small allocSite (2 byte)
        if (((allocationSiteId >> 15) & 1) != 0) {
            allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_3);
            bigAllocSite = true;
        }

        AllocationSite allocationSite = symbols.sites.getById(allocationSiteId);
        long addr = getNextDoubleWord();

        int arrayLength = 0;
        if (!bigAllocSite) { // all arrays have a small allocSite ID
            arrayLength = recoverArrayLength(word);
        }

        int realAllocatedTypeId = -1;
        if (type == EventType.OBJ_ALLOC_SLOW_C1_DEVIANT_TYPE || type == EventType.OBJ_ALLOC_SLOW_IR_DEVIANT_TYPE || type == EventType.OBJ_ALLOC_SLOW_C2_DEVIANT_TYPE) {
            realAllocatedTypeId = getNextWord();
        }

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseObjAllocSlowCiIr_Deviant(type, allocationSiteId, allocationSite, addr, arrayLength, realAllocatedTypeId, threadLocalHeap);
        }
        mainParser.doParseObjAllocSlowCiIr_Deviant(type, allocationSiteId, allocationSite, addr, arrayLength, realAllocatedTypeId, threadLocalHeap);
    }

    protected void parseObjAllocSlow(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_2);
        boolean bigAllocSite = false;
        // check if first bit is set, if yes --> big allocSite (3 byte), otherwise small allocSite (2 byte)
        if (((allocationSiteId >> 15) & 1) != 0) {
            allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_3);
            bigAllocSite = true;
        }

        AllocationSite allocationSite = symbols.sites.getById(allocationSiteId);

        long addr = getNextDoubleWord();

        int arrayLength = 0;
        if (!bigAllocSite) { // all arrays have a small allocSite ID
            arrayLength = recoverArrayLength(word);
        }

        if (allocationSite.allocatedTypeId == AllocatedType.ALLOCATED_TYPE_IDENTIFIER_UNKNOWN) {
            int id = getNextWord();
            allocationSite = allocationSite.copy(id);
        }

        AllocatedType allocatedType = symbols.types.get(allocationSite.allocatedTypeId);

        boolean isArray = false;
        int size = Consts.UNDEFINED_LENGTH;

        if (allocatedType.name.equals(AllocatedTypes.MIRROR_CLASS_NAME)) {
            size = getNextWord();
            allocatedType.size = size;
            arrayLength = Consts.UNDEFINED_LENGTH;
        } else {
            isArray = allocatedType.name.startsWith("[");
        }

        boolean mayBeFiller = mayBeFiller(allocationSite);

        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseObjAllocSlow(type, allocationSiteId, allocationSite, addr, isArray, arrayLength, size, mayBeFiller, threadLocalHeap);
        }
        mainParser.doParseObjAllocSlow(type, allocationSiteId, allocationSite, addr, isArray, arrayLength, size, mayBeFiller, threadLocalHeap);
    }

    protected void parseTlabAlloc(int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int size = (int) getNextDoubleWord();
        long addr = getNextDoubleWord();
        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParseTlabAlloc(size, addr, threadLocalHeap);
        }
        mainParser.doParseTlabAlloc(size, addr, threadLocalHeap);
    }

    protected void parsePlabAlloc(int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int size = (int) getNextDoubleWord();
        long addr = getNextDoubleWord();
        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doParsePlabAlloc(size, addr, threadLocalHeap);
        }
        mainParser.doParsePlabAlloc(size, addr, threadLocalHeap);
    }

    // --------------------------------------------------------
    // ---------------- Helper methods ------------------------
    // --------------------------------------------------------

    protected int recoverArrayLength(int word) throws TraceException {
        int arrayLength = recoverValue(word, ARRAY_SMALL_LENGTH_INDEX, ARRAY_SMALL_LENGTH_INDEX);
        if (arrayLength == Consts.ARRAY_SIZE_MAX_SMALL) {
            arrayLength = getNextWord();
        }
        return arrayLength;
    }

    protected boolean mayBeFiller(AllocationSite allocationSite) {
        AllocatedType allocatedType = symbols.types.get(allocationSite.allocatedTypeId);

        return (allocatedType.name.equals("[I") || allocatedType.name.equals("Ljava/lang/Object;"));
    }

    protected Space getSpace(long addr) {
        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.getSpace(addr);
        }
        return mainParser.getSpace(addr);
    }

    protected ObjectInfo getObjectIsFiller(Space space, long addr) throws TraceException {
        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.getObjectIsFiller(space, addr);
        }
        return mainParser.getObjectIsFiller(space, addr);
    }

    protected long doGCMove(EventType type, ThreadLocalHeap threadLocalHeap, ObjectInfo obj, long fromAddr, long toAddr, SpaceType toSpaceType) throws TraceException {
        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doGCMove(type, threadLocalHeap, obj, fromAddr, toAddr, toSpaceType);
        }
        return mainParser.doGCMove(type, threadLocalHeap, obj, fromAddr, toAddr, toSpaceType);
    }

    protected long getMoveTarget(long fromAddr, long toAddr, SpaceType toSpaceType, ThreadLocalHeap threadLocalHeap) throws TraceException {
        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.getMoveTarget(fromAddr, toAddr, toSpaceType, threadLocalHeap);
        }
        return mainParser.getMoveTarget(fromAddr, toAddr, toSpaceType, threadLocalHeap);
    }

    protected void doPtrEvent(EventType eventType, ThreadLocalHeap threadLocalHeap, long fromAddr, long toAddr, long[] ptrs, boolean isFiller) throws TraceException {
        for (int parserNr = 0; parserNr < otherParsers.size(); parserNr++) {
            TraceParsingEventHandler parser = otherParsers.get(parserNr);
            parser.doPtrEvent(eventType, threadLocalHeap, fromAddr, toAddr, ptrs, isFiller);
        }
        mainParser.doPtrEvent(eventType, threadLocalHeap, fromAddr, toAddr, ptrs, isFiller);
    }

    // --------------------------------------------------------
    // ----------------- Error methods ------------------------
    // --------------------------------------------------------

    protected void errorOnthreadLocalHeapState(ThreadLocalHeap threadLocalHeap) throws TraceException {
        throw new TraceException("State is " + threadLocalHeap.state + ", but expected " + ThreadLocalHeap.STATE_IN_QUEUE + " or " + ThreadLocalHeap
                .STATE_IN_QUEUE_FOR_CLEAN_UP);
    }

    protected void errorOnMirrorClass(AllocatedType allocatedType) throws TraceException {
        if (allocatedType.name.equals(MIRROR_CLASS)) {
            throw new TraceException("PARSE ERROR. Expected no MIRROR_CLASS here");
        }
    }

    // --------------------------------------------------------
    // ------------------ Word methods ------------------------
    // --------------------------------------------------------

    // just consumes
    protected void getPointers(int word) throws TraceException {

        int ptrKinds = recoverValue(word, 1, 3);
        int i = 0;
        boolean encodingEnd = false;

        while (!encodingEnd && i < PtrEvent.MAX_PTRS_PER_EVENT) {
            int kind = (ptrKinds >> ((PtrEvent.MAX_PTRS_PER_EVENT - i - 1) * 2)) & 0x3;
            if (kind == PtrEvent.ENCODING_RELATIVE_PTR) {
                getNextWord();
            } else if (kind == PtrEvent.ENCODING_ABSOLUTE_PTR) {
                getNextDoubleWord();
            } else if (kind == PtrEvent.ENCODING_END) {
                encodingEnd = true;
            }
            i++;
        }

    }

    // returns absolute pointer addresses
    protected long[] parsePointers(int word, long ref, ThreadLocalHeap localHeap) throws TraceException {

        int ptrKinds = recoverValue(word, 1, 3);
        long lastRef = 0;

        // return null if event contains no pointers
        long[] ptrs = ptrKinds == 0 ? null : new long[PtrEvent.MAX_PTRS_PER_EVENT];

        if (ptrs != null) {
            long addr;
            int i = 0;
            boolean encodingEnd = false;
            while (!encodingEnd && i < PtrEvent.MAX_PTRS_PER_EVENT) {
                addr = NULL_PTR; // NULL PTR

                int kind = (ptrKinds >> ((PtrEvent.MAX_PTRS_PER_EVENT - i - 1) * 2)) & 0x3;
                if (kind == PtrEvent.ENCODING_RELATIVE_PTR) {
                    addr = getNextWord();
                    addr = addr + lastRef;
                    lastRef = addr;
                    addr *= symbols.heapWordSize;
                    addr = ref - addr;
                    ptrs[i] = addr;
                } else if (kind == PtrEvent.ENCODING_ABSOLUTE_PTR) {
                    addr = getNextDoubleWord();
                    assert addr > 0 : "Negative address while parsing absolute pointer";
                    lastRef = (ref - addr) / symbols.heapWordSize;
                    ptrs[i] = addr;
                } else if (kind == PtrEvent.ENCODING_NULL_PTR) {
                    ptrs[i] = NULL_PTR;
                } else if (kind == PtrEvent.ENCODING_END) {
                    encodingEnd = true;
                    ptrs = Arrays.copyOf(ptrs, i);
                } else {
                    assert false : "Unknown pointer encoding type";
                }
                i++;
            }
        }

        return ptrs;

    }

    protected int recoverValue(long word, int from, int to) {
        int start = Consts.HEX_WORD >>> (8 * from);
        int value = start;

        for (int i = from; i < to; i++) {
            value = value | (value >> 8);
        }

        value = (int) ((word & value) >> (8 * (3 - to)));
        return value;
    }

    protected long getNextDoubleWord() throws TraceException {
        final long MASK = (1L << 32) - 1;
        long low = (getNextWord()) & MASK;
        long high = (getNextWord()) & MASK;
        return low | (high << 32);
    }

    protected int getNextWord() throws TraceException {
        if (buffer == null || buffer.position() == buffer.capacity()) {
            throw new TraceException("Buffer is null or reached limit");
        }
        return buffer.getInt();
    }

    protected String getString(int length) throws TraceException {
        StringBuilder ret = new StringBuilder();
        end:
        for (int j = 0; j < length; j += 4) {
            int word1 = getNextWord();
            for (int k = 0; k < 4; k++) {
                char c = (char) ((word1 >> ((3 - k) * 8)) & 0xFF);
                if (c == '\0') {
                    break end;
                }
                ret.append(c);
            }
        }

        return ret.toString();
    }
}
