/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.parser.heap;

import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.GarbageCollectionType;
import at.jku.mevss.anttracks.heap.labs.Lab;
import at.jku.mevss.anttracks.heap.listener.*;
import at.jku.mevss.anttracks.heap.objects.ArrayObjectInfo;
import at.jku.mevss.anttracks.heap.objects.InstanceMirrorObjectInfo;
import at.jku.mevss.anttracks.heap.objects.InstanceObjectInfo;
import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.roots.*;
import at.jku.mevss.anttracks.heap.space.Space;
import at.jku.mevss.anttracks.heap.space.SpaceMode;
import at.jku.mevss.anttracks.heap.space.SpaceRedefinition;
import at.jku.mevss.anttracks.heap.space.SpaceType;
import at.jku.mevss.anttracks.heap.symbols.AllocatedType;
import at.jku.mevss.anttracks.heap.symbols.AllocationSite;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.parser.*;
import at.jku.mevss.anttracks.util.ApplicationStatistics;
import at.jku.mevss.anttracks.util.Consts;
import at.jku.mevss.anttracks.util.TraceException;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;
import java.util.stream.Collectors;

import static at.jku.mevss.anttracks.heap.labs.Lab.UNKNOWN_CAPACITY;

public class HeapBuilder {

    // --------------------------------------------------------
    // ---------------------- Fields --------------------------
    // --------------------------------------------------------

    private final DetailedHeap heap;
    private final Symbols symbols;
    private final boolean useDynamicCallContext;
    private final ParsingInfo parsingInfo;
    private AllocationContext allocContext;
    private ThreadInfo currentThread; // build callstack
    private Logger LOGGER = Logger.getLogger(HeapBuilder.class.getSimpleName());

    // --------------------------------------------------------
    // ------------------- Constructor ------------------------
    // --------------------------------------------------------

    public HeapBuilder(DetailedHeap workspace, Symbols symbols, ParsingInfo parsingInfo) {
        heap = workspace;
        this.symbols = symbols;
        useDynamicCallContext = symbols.useDynamicCallContext();
        if (useDynamicCallContext) {
            allocContext = new AllocationContext();
        }

        this.parsingInfo = parsingInfo;
    }

    // --------------------------------------------------------
    // --------------------- Methods --------------------------
    // --------------------------------------------------------

    // --------------------------------------------------------
    // --------------------- Static ---------------------------
    // --------------------------------------------------------

    public static DetailedHeap constructHeap(Symbols symbols) {
        DetailedHeap heap = new DetailedHeap(symbols);
        PauseTimeLogListener.init(heap);
        SurvivorLogListener.init(heap);
        ObjectSizeLogListener.init(heap);
        ArrayLengthLogListener.init(heap);
        StackTraceLogListener.init(heap);
        TotalAllocationListener.init(heap);
        MultiLabVSSingleLabListener.init(heap);
        RootPointerDistributionListener.init(heap);
        return heap;
    }

    // --------------------------------------------------------
    // --------------------- General ---------------------------
    // --------------------------------------------------------

    public void doCleanUp(ThreadLocalHeap tlh) throws TraceException {
        tlh.retireCurrentLabs(heap);
        heap.insertFillers(tlh);
        tlh.retiredLabs.clear();

        tlh.finishCurrentObjectPointers(heap);
        tlh.copyAndClearMultiThreadedEvents(heap.notYetMatchedMultiThreadedPtrEventsToHandleAtGCEnd);
    }

    public void doParseGCStart(ParserGCInfo info, long start, long end, boolean check) throws TraceException {
        heap.assignLabsIncorrectlyAssumedToBeFillers(false);
        heap.startGC(info, start - Scanner.META_BUFFER_INFO, parsingInfo);

        if (check) {
            heap.validate(symbols.isHeapFragmented);
        }
        executeSpaceRedefinitions();
    }

    private void executeSpaceRedefinitions() {
        List<Space> redefinedSpaces = heap.spaceRedefinitions.stream().map(redef -> redef.space).collect(Collectors.toList());
        Set<Space> spacesWithMultipleRedefinitions = redefinedSpaces.stream().filter(i -> Collections.frequency(redefinedSpaces, i) > 1).collect(Collectors.toSet());

        if (!spacesWithMultipleRedefinitions.isEmpty()) {
            LOGGER.info("Multiple redefinitions exist for at least one space");
            heap.spaceRedefinitions.stream()
                                   .filter(redef -> spacesWithMultipleRedefinitions.contains(redef.space))
                                   .map(SpaceRedefinition::toString)
                                   .forEach(LOGGER::info);
        }

        heap.spaceRedefinitions.forEach(SpaceRedefinition::execute);
        heap.spaceRedefinitions.clear();
    }

    public void doParseGCEnd(ParserGCInfo info, ThreadLocalHeap threadLocalHeap, long start, long end, boolean failed, boolean check) throws TraceException {
        ApplicationStatistics.Measurement m = ApplicationStatistics.getInstance().createMeasurement("GCEnd");

        ApplicationStatistics.Measurement mInner = ApplicationStatistics.getInstance().createMeasurement("fillerHandling");
        heap.assignLabsIncorrectlyAssumedToBeFillers(false);
        mInner.end();

        threadLocalHeap.finishCurrentObjectPointers(heap);

        mInner = ApplicationStatistics.getInstance().createMeasurement("heapAdjusting");
        heap.stopGC(info, failed, end, parsingInfo);
        mInner.end();

        executeSpaceRedefinitions();
        m.end();
    }

    public void doParseGCInterrupt(int id, long address) throws TraceException {
        heap.interruptGC(id, address);
    }

    public void doParseGCContinue(int id, long address) throws TraceException {
        heap.continueGC(id, address);
    }

    // --------------------------------------------------------
    // --------------------- Moving ---------------------------
    // --------------------------------------------------------

    public List<ObjectInfo> doParseGCMoveRegion(EventType type, long fromAddr, long toAddr, SpaceType toSpaceType, int numOfObjects, ThreadLocalHeap threadLocalHeap)
            throws TraceException {
        List<ObjectInfo> movedObjects = new ArrayList<>();

        Space toSpace = heap.getSpace(toAddr);
        Lab regionLab = toSpace.assignRealLab(threadLocalHeap.thread, Lab.KIND_REGION_VIRTUAL, toAddr, UNKNOWN_CAPACITY, symbols.expectPointers);
        threadLocalHeap.currentLabs.put(toSpace.getType(), regionLab);

        for (int i = 0; i < numOfObjects; i++) {
            // if next object is in another space
            if (!toSpace.contains(toAddr)) {
                // retire current lab
                regionLab.variableCapacity();
                regionLab.resetCapacity();
                threadLocalHeap.retireCurrentLab(heap, toSpace.getType());
                // create new lab
                toSpace = heap.getSpace(toAddr);
                regionLab = toSpace.assignRealLab(threadLocalHeap.thread, Lab.KIND_REGION_VIRTUAL, toAddr, UNKNOWN_CAPACITY, symbols.expectPointers);
                threadLocalHeap.currentLabs.put(toSpace.getType(), regionLab);
            }
            ObjectInfo obj = getObjectAndMove(type, fromAddr, toAddr, toSpaceType, threadLocalHeap);
            toAddr += obj.size;
            fromAddr += obj.size;
            movedObjects.add(obj);
        }
        // Make sure size is correctly set for LAB
        regionLab.variableCapacity();
        regionLab.resetCapacity();

        threadLocalHeap.retireCurrentLab(heap, toSpace.getType());
        return movedObjects;
    }

    private ObjectInfo getObjectAndMove(EventType type, long fromAddr, long toAddr, SpaceType toSpaceType, ThreadLocalHeap threadLocalHeap) throws TraceException {
        Space fromSpace = heap.getSpace(fromAddr);
        ObjectInfo obj = fromSpace.checkIfObjectInfoIsFiller(fromAddr, heap.getGC().isFull());
        boolean isFiller = obj != null;

        if (!isFiller) {
            obj = fromSpace.getNonFillerObjectInfo(fromAddr);
        }
        doGCMove(type, threadLocalHeap, obj, fromAddr, toAddr, toSpaceType);

        return obj;
    }

    public Space getSpace(long addr) {
        return heap.getSpace(addr);
    }

    public ObjectInfo getObjectIfFiller(Space space, long addr) throws TraceException {
        return space.checkIfObjectInfoIsFiller(addr, heap.getGC().isFull());
    }

    public long doGCMove(EventType type, ThreadLocalHeap threadLocalHeap, ObjectInfo obj, long fromAddr, long toAddr, SpaceType toSpaceType) throws TraceException {
        /*
        if (toAddr >= 0) {
            Space space = heap.getSpace(toAddr);
            if (space == null) {
                System.err.println("DAMN! Could not find space for toAddress " + String.format("%,d", toAddr));
                System.err.println(heap.toString());
            } else {
                toSpaceType = heap.getSpace(toAddr).getType();
            }
        }
        */

        // if an object pointed by a root was moved, the root DS must be updated
        if (symbols.expectPointers) {
            synchronized (heap.rootPtrs) {
                List<RootPtr> rootsToUpdate = heap.rootPtrs.remove(fromAddr);
                if (rootsToUpdate != null) {
                    for (RootPtr r : rootsToUpdate) {
                        r.setAddr(toAddr);
                    }
                    heap.rootPtrs.put(toAddr, rootsToUpdate);
                }
            }
        }

        return assignToLabOrSpace(threadLocalHeap.thread, threadLocalHeap, toSpaceType, toAddr, obj);
    }

    public ObjectInfo doParseSyncObj(EventType eventType,
                                     int allocationSiteId,
                                     int allocatedTypeId,
                                     AllocatedType allocatedType,
                                     long fromAddr,
                                     long toAddr,
                                     SpaceType toSpaceType,
                                     int length,
                                     int size,
                                     ThreadLocalHeap threadLocalHeap) throws TraceException {
        boolean firstSyncGC = heap.getGCNo() == 0 && (heap.getGC() == GarbageCollectionType.MAJOR_SYNC || heap.getGC() == GarbageCollectionType.MINOR_SYNC);

        ObjectInfo obj = null;
        AllocationSite site = symbols.sites.getById(allocationSiteId);
        AllocatedType type = symbols.types.get(allocatedTypeId);
        if (firstSyncGC) {
            if (size > 0) {
                obj = InstanceMirrorObjectInfo.get(heap.getCache(), threadLocalHeap.instanceMirrorInfo, eventType, site, type, size, symbols);
            } else if (length != Consts.UNDEFINED_LENGTH) {
                obj = ArrayObjectInfo.get(heap.getCache(), threadLocalHeap.arrayInfo, eventType, site, type, length, symbols);
            } else {
                obj = InstanceObjectInfo.get(heap.getCache(), threadLocalHeap.instanceInfo, eventType, site, type, symbols);
            }

            assignToLabOrSpace(threadLocalHeap.thread, threadLocalHeap, heap.getSpace(toAddr).getType(), toAddr, obj);
        } else {
            if (heap.getGC() != GarbageCollectionType.MINOR_SYNC || heap.getSpace(fromAddr).getType() != SpaceType.OLD) {
                obj = getObjectAndMove(eventType, fromAddr, toAddr, toSpaceType, threadLocalHeap);
            }
        }
        return obj;
    }

    public void doPtrEvent(EventType eventType, ThreadLocalHeap threadLocalHeap, long fromAddr, long toAddr, long[] ptrs, boolean isFiller) throws TraceException {
        // Damn, that's a lot of memory consumption!
        // PtrEvent event = new PtrEvent(eventType, fromAddr, toAddr, ptrs, isFiller);

        switch (eventType) {
            case GC_MOVE_SLOW_PTR:
            case GC_MOVE_FAST_WIDE_PTR:
            case GC_MOVE_FAST_PTR:
            case GC_KEEP_ALIVE_PTR:
                // This thread starts sending pointers for a new object, finish previous object and start processing new object

                assert fromAddr >= 0 : "From address must be known for \"move+pointer events\"";
                assert toAddr >= 0 : "To address must be known for \"move+pointer events\"";
                startNewCurrentObjectPointers(threadLocalHeap, eventType, fromAddr, toAddr, ptrs, isFiller);
                break;

            case GC_PTR_EXTENSION:
                // An extension event always has to belong to the object that is currently processed by the current thread
                assert fromAddr == -1 : "From address is not known for dedicated pointer events";
                assert toAddr != -1 : "To address must be known for dedicated pointer events";
                assert ptrs != null : "Dedicated ptr event must have ptrs";
                assert !isFiller : "Filler objects should not have extension pointer events";
                assert threadLocalHeap.isMatchingPtrEvent(toAddr) :
                        String.format("The received extension ptr event does not belong to the currently processed object!\n" +
                                              "Current object: %,d (%s)\n" +
                                              "Receive extension for: %,d (%s)",
                                      threadLocalHeap.currentPtrToAddr,
                                      heap.getObject(threadLocalHeap.currentPtrToAddr),
                                      toAddr,
                                      heap.getObject(toAddr));

                threadLocalHeap.addPtrToCurrentObject(toAddr, ptrs);
                break;

            case GC_PTR_MULTITHREADED:
                // A multithreaded ptr event can be sent from any event for any object
                // The thread local lists will be merged after the GC and processed together
                assert fromAddr == -1 : "From address is not known for dedicated pointer events";
                assert toAddr != -1 : "To address must be known for dedicated pointer events";
                assert ptrs != null : "Dedicated ptr event must have ptrs";
                assert !isFiller : "Filler objects should not have multithreaded pointer events";

                threadLocalHeap.addMultithreadedPtrEvent(heap, toAddr, ptrs);
                break;

            case GC_PTR_UPDATE_PREMOVE:
                // A pointer update event can either belong to the current object
                assert fromAddr == -1 : "From address is not known for dedicated pointer events";
                assert toAddr != -1 : "To address must be known for dedicated pointer events";
                assert ptrs != null : "Dedicated ptr event must have ptrs";
                assert !isFiller : "Filler objects should not have dedicated pointer events";

                if (threadLocalHeap.isMatchingPtrEvent(toAddr)) {
                    threadLocalHeap.addPtrToCurrentObject(toAddr, ptrs);
                } else {
                    // or start a new object pointer handling
                    threadLocalHeap.finishCurrentObjectPointers(heap);
                    startNewCurrentObjectPointers(threadLocalHeap, eventType, toAddr, toAddr, ptrs, isFiller);
                }
                break;

            case GC_PTR_UPDATE_POSTMOVE:
                assert fromAddr == -1 : "From address is not known for dedicated pointer events";
                assert toAddr != -1 : "To address must be known for dedicated pointer events";
                assert ptrs != null : "Dedicated ptr event must have ptrs";
                assert !isFiller : "Filler objects should not have dedicated pointer events";

                if (threadLocalHeap.isMatchingPtrEvent(toAddr)) {
                    threadLocalHeap.addPtrToCurrentObject(toAddr, ptrs);
                } else {
                    startNewCurrentObjectPointers(threadLocalHeap, eventType, toAddr, toAddr, ptrs, isFiller);
                }
                break;

            default:
                throw new TraceException("Cannot handle an event as pointer event that is not a pointer event");
        }
    }

    private void startNewCurrentObjectPointers(ThreadLocalHeap tlh, EventType eventType, long fromAddr, long toAddr, long[] ptrs, boolean isFiller) throws
            TraceException {
        ObjectInfo obj = heap.getObjectInBack(fromAddr);
        int pointerCount = getPointerArrayLength(obj, isFiller);
        int top = ptrs == null ? 0 : ptrs.length;

        if (ptrs == null || ptrs.length < pointerCount) {
            long[] fullPointerArray = new long[pointerCount];
            if (ptrs != null) {
                System.arraycopy(ptrs, 0, fullPointerArray, 0, ptrs.length);
            }
            ptrs = fullPointerArray;
        }

        tlh.finishCurrentObjectPointers(heap);
        if (isFiller) {
            LOGGER.info(String.format("Pointer event for filler object @ %,d. How is it possible that there is a filler object with pointers?!", fromAddr));
            // Not sure why filler objects are moved at all...
            // I hope the following assertion holds:
            assert ptrs == null : "I don't know why filler objects (normally int[]) are moved at all... And now they also have pointers!?";
        }
        tlh.startNewCurrentObjectPointers(eventType, fromAddr, toAddr, ptrs, top, isFiller);
    }

    private int getPointerArrayLength(ObjectInfo obj, boolean isFiller) throws TraceException {
        if (isFiller) {
            // Filler objects do (hopefully) not have pointers, they are int[]
            return 0;
        }
        if (obj == null) {
            throw new TraceException("We cannot determine pointer array length for unknown object");
        }
        assert obj.type != null : "There must not be an object without a type";

        // Arrays
        if (obj instanceof ArrayObjectInfo) {
            return obj.type.isPrimitiveArray() ? 0 : ((ArrayObjectInfo) obj).getLength();
        }
        // Class objects
        if (obj instanceof InstanceMirrorObjectInfo) {
            AllocatedType.MirrorAllocatedType type = (AllocatedType.MirrorAllocatedType) obj.type;
            // All class objects share the same pointers that come from Class.java
            int ptrCount = type.nonStaticReferenceFieldsCount();
            // Also, each class object differs in pointer count because each class object references its class's static fields
            if (type.mirrorKlass != null) {
                AllocatedType.FieldInfo[] staticFields = type.mirrorKlass.getStaticFieldInfos();
                assert staticFields != null : "Every type must have initialized field info";
                for (AllocatedType.FieldInfo fi : staticFields) {
                    if (fi.isInstance() || fi.isArray()) {
                        ptrCount++;
                    }
                }
            }

            return ptrCount;
        }
        // Instance
        return obj.type.nonStaticReferenceFieldsCount();
    }

    // --------------------------------------------------------
    // --------------------- Alloc ----------------------------
    // --------------------------------------------------------

    public ObjectInfo doParseObjAllocFastCi(EventType type,
                                            int allocationSiteId,
                                            AllocationSite allocationSite,
                                            AllocatedType allocatedType,
                                            boolean isArray,
                                            int arrayLength,
                                            ThreadLocalHeap threadLocalHeap) throws TraceException {
        AllocationSite site = amendStackTrace(threadLocalHeap, allocationSite);

        ObjectInfo obj;
        if (isArray) {
            obj = ArrayObjectInfo.get(heap.getCache(), threadLocalHeap.arrayInfo, type, site, symbols.types.get(site.allocatedTypeId), arrayLength, symbols);
        } else {
            obj = InstanceObjectInfo.get(heap.getCache(), threadLocalHeap.instanceInfo, type, site, symbols.types.get(site.allocatedTypeId), symbols);
        }

        long addr = assignToLab(threadLocalHeap.thread, threadLocalHeap.currentLabs.get(SpaceType.EDEN), obj, Consts.UNDEFINED_ADDR);

        threadLocalHeap.recordAllocation(site, useDynamicCallContext);
        return obj;
    }

    public ObjectInfo doParseObjAllocFastC2DeviantType(EventType type,
                                                       int header,
                                                       int allocationSiteId,
                                                       AllocatedType allocatedType,
                                                       boolean isArray,
                                                       int arrayLength,
                                                       ThreadLocalHeap threadLocalHeap) throws TraceException {
        AllocationSite site = amendStackTrace(threadLocalHeap, symbols.sites.getById(allocationSiteId));

        ObjectInfo obj;
        if (isArray) {
            obj = ArrayObjectInfo.get(heap.getCache(), threadLocalHeap.arrayInfo, type, site, symbols.types.get(allocatedType.id), arrayLength, symbols);
        } else {
            obj = InstanceObjectInfo.get(heap.getCache(), threadLocalHeap.instanceInfo, type, site, symbols.types.get(allocatedType.id), symbols);
        }

        long addr = assignToLab(threadLocalHeap.thread, threadLocalHeap.currentLabs.get(SpaceType.EDEN), obj, Consts.UNDEFINED_ADDR);

        threadLocalHeap.recordAllocation(site, useDynamicCallContext);
        return obj;
    }

    public ObjectInfo doParseObjAllocFastIr(EventType type, int allocationSiteId, AllocationSite allocationSite, ThreadLocalHeap threadLocalHeap) throws TraceException {
        AllocationSite site = amendStackTrace(threadLocalHeap, allocationSite);

        ObjectInfo obj = InstanceObjectInfo.get(heap.getCache(), threadLocalHeap.instanceInfo, type, site, symbols.types.get(site.allocatedTypeId), symbols);
        long addr = assignToLab(threadLocalHeap.thread, threadLocalHeap.currentLabs.get(SpaceType.EDEN), obj, Consts.UNDEFINED_ADDR);

        threadLocalHeap.recordAllocation(site, useDynamicCallContext);
        return obj;
    }

    public ObjectInfo doParseObjAllocNormalCi(EventType type,
                                              int allocationSiteId,
                                              AllocationSite allocationSite,
                                              AllocatedType allocatedType,
                                              long addr,
                                              boolean isArray,
                                              int arrayLength,
                                              ThreadLocalHeap threadLocalHeap) throws TraceException {
        AllocationSite site = amendStackTrace(threadLocalHeap, allocationSite);

        ObjectInfo obj;
        if (isArray) {
            obj = ArrayObjectInfo.get(heap.getCache(), threadLocalHeap.arrayInfo, type, site, symbols.types.get(site.allocatedTypeId), arrayLength, symbols);
        } else {
            obj = InstanceObjectInfo.get(heap.getCache(), threadLocalHeap.instanceInfo, type, site, symbols.types.get(site.allocatedTypeId), symbols);
        }

        heap.getSpace(addr).assign(threadLocalHeap.thread, obj, addr, symbols.expectPointers);

        threadLocalHeap.recordAllocation(site, useDynamicCallContext);
        return obj;
    }

    public ObjectInfo doParseObjAllocNormalIr(EventType type, int allocationSiteId, AllocationSite allocationSite, long addr, ThreadLocalHeap threadLocalHeap)
            throws TraceException {
        AllocationSite site = amendStackTrace(threadLocalHeap, allocationSite);

        ObjectInfo obj = InstanceObjectInfo.get(heap.getCache(), threadLocalHeap.instanceInfo, type, site, symbols.types.get(site.allocatedTypeId), symbols);

        heap.getSpace(addr).assign(threadLocalHeap.thread, obj, addr, symbols.expectPointers);

        threadLocalHeap.recordAllocation(site, useDynamicCallContext);
        return obj;
    }

    public ObjectInfo doParseObjAllocSlowCiIr_Deviant(EventType type,
                                                      int allocationSiteId,
                                                      AllocationSite allocationSite,
                                                      long addr,
                                                      int arrayLength,
                                                      int realAllocatedTypeId,
                                                      ThreadLocalHeap threadLocalHeap) throws TraceException {
        AllocationSite site = amendStackTrace(threadLocalHeap, allocationSite);

        SpaceType relAddrSpace = heap.getSpace(addr).getType();

        if (type == EventType.OBJ_ALLOC_SLOW_C1_DEVIANT_TYPE || type == EventType.OBJ_ALLOC_SLOW_IR_DEVIANT_TYPE || type == EventType.OBJ_ALLOC_SLOW_C2_DEVIANT_TYPE) {
            site = amendStackTrace(threadLocalHeap, site.copy(realAllocatedTypeId));
        }
        AllocatedType allocatedType = symbols.types.get(site.allocatedTypeId);

        errorOnMirrorClass(allocatedType);

        boolean isArray = allocatedType.name.startsWith("[");
        ObjectInfo obj;
        if (isArray) {
            obj = ArrayObjectInfo.get(heap.getCache(), threadLocalHeap.arrayInfo, type, site, allocatedType, arrayLength, symbols);
        } else {
            obj = InstanceObjectInfo.get(heap.getCache(), threadLocalHeap.instanceInfo, type, site, allocatedType, symbols);
        }

        assignToLabOrSpace(threadLocalHeap.thread, threadLocalHeap, relAddrSpace, addr, obj);

        threadLocalHeap.recordAllocation(site, useDynamicCallContext);
        return obj;
    }

    public ObjectInfo doParseObjAllocSlow(EventType type,
                                          int allocationSiteId,
                                          AllocationSite allocationSite,
                                          long addr,
                                          boolean isArray,
                                          int arrayLength,
                                          int size,
                                          boolean mayBeFiller,
                                          ThreadLocalHeap threadLocalHeap) throws TraceException {
        AllocationSite site = amendStackTrace(threadLocalHeap, allocationSite);
        Space space = heap.getSpace(addr);
        SpaceType spaceType = space.getType();

        ObjectInfo obj = ObjectInfo.get(type,
                                        site,
                                        symbols.types.get(site.allocatedTypeId),
                                        size,
                                        isArray,
                                        arrayLength,
                                        threadLocalHeap.instanceInfo,
                                        threadLocalHeap.instanceMirrorInfo,
                                        threadLocalHeap.arrayInfo,
                                        heap.getCache(),
                                        symbols);

        assignToLabOrSpace(threadLocalHeap.thread, threadLocalHeap, spaceType, addr, obj, mayBeFiller);

        threadLocalHeap.recordAllocation(site, useDynamicCallContext);
        return obj;
    }

    public void doParseTlabAlloc(int size, long addr, ThreadLocalHeap threadLocalHeap) throws TraceException {
        SpaceType relAddrSpace = heap.getSpace(addr).getType();
        errorOnTLABAlloc(relAddrSpace);
        setLab(threadLocalHeap, size, addr, Lab.KIND_TLAB);
    }

    public void doParsePlabAlloc(int size, long addr, ThreadLocalHeap threadLocalHeap) throws TraceException {
        Space space = heap.getSpace(addr);
        if (space == null) {
            throw new TraceException(String.format("PLAB should be allocated at address %,d, but no space found there.\nCurrent heap: %s", addr, heap.toString()));
        }
        SpaceType relAddrSpace = space.getType();
        errorOnPLABAlloc(relAddrSpace);
        setLab(threadLocalHeap, size > 0 ? size : UNKNOWN_CAPACITY, addr, size > 0 ? Lab.KIND_PLAB : Lab.KIND_REGION_VIRTUAL);
    }

    // --------------------------------------------------------
    // ---------------- Helper methods ------------------------
    // --------------------------------------------------------

    private long assignToLabOrSpace(String thread, ThreadLocalHeap tlh, SpaceType type, long addr, ObjectInfo obj, boolean mayBeFiller) throws TraceException {
        Lab currentLab = tlh.currentLabs.get(type);
        long assignedAddr = Lab.OBJECT_NOT_ASSIGNED;
        if (currentLab != null) {
            if (mayBeFiller && addr + obj.size == currentLab.end()) {
                // object can be for sure handled as filler, method "assign" takes care of everything
                heap.getSpace(addr).assign(thread, true, obj, addr, symbols.expectPointers);
                assignedAddr = addr;
            } else {
                // object is not a filler, current lab exists -> current object must be the next object in the LAB
                assignedAddr = assignToLab(thread, currentLab, obj, addr);
            }
        }
        if (assignedAddr == Lab.OBJECT_NOT_ASSIGNED) {
            heap.getSpace(addr).assign(thread, mayBeFiller, obj, addr, symbols.expectPointers);
        }

        return assignedAddr;
    }

    private long assignToLabOrSpace(String thread, ThreadLocalHeap tlh, SpaceType type, long addr, ObjectInfo obj) throws TraceException {
        long assignedAddr = assignToLab(thread, tlh.currentLabs.get(type), obj, addr);
        if (assignedAddr == Lab.OBJECT_NOT_ASSIGNED) {
            assignedAddr = heap.getSpace(addr).assign(thread, obj, addr, symbols.expectPointers);
        }

        return assignedAddr;
    }

	/*
    private long assignToLabOrSpace(String thread, ThreadLocalHeap tlh, SpaceType type, long addr, ObjectInfoAge objAge)
			throws TraceException {
		long assignedAddr = assignToLab(thread, tlh.currentLabs.get(type), objAge.obj, addr);
		if (assignedAddr == Lab.OBJECT_NOT_ASSIGNED) {
			assignedAddr = heap.getSpace(addr).assign(thread, objAge, addr);
		} else {
			tlh.currentLabs.get(type).setAge(assignedAddr, objAge.age);
		}

		return assignedAddr;
	}
	*/

    private long assignToLab(String thread, Lab currentLab, ObjectInfo obj, long addr) throws TraceException {
        long assignedAddr = Lab.OBJECT_NOT_ASSIGNED;
        if (currentLab != null) {
            assignedAddr = currentLab.tryAllocate(obj, addr);
        } else {
            errorOnUndefinedAddr(addr);
        }

        return assignedAddr;
    }

    private void setLab(ThreadLocalHeap threadLocalHeap, int size, long addr, int kind) throws TraceException {
        SpaceType relAddrSpace = heap.getSpace(addr).getType();
        Lab current = heap.getSpace(addr).assignRealLab(threadLocalHeap.thread, kind, addr, size, symbols.expectPointers);
        threadLocalHeap.retireCurrentLab(heap, relAddrSpace);
        threadLocalHeap.currentLabs.put(relAddrSpace, current);
    }

    // --------------------------------------------------------
    // ---------------- Stack trace extension -----------------
    // --------------------------------------------------------

    /**
     * Try to add missing methods to the stack trace of the specified allocation site using call context information.
     *
     * @param localHeap The {@link ThreadLocalHeap} to work with.
     * @param site      The allocation site.
     * @return A new allocation site with an amended stack trace, or {@code site} if no changes were made.
     */
    private AllocationSite amendStackTrace(ThreadLocalHeap localHeap, AllocationSite site) {
        // We can't do much without allocation context
        /*
         * Problem: if only second allocation decides, lookup stops at first allocation -> quick fix look through all nodes but does that
         * explode the tree?
         */
        if (useDynamicCallContext && symbols.isDynamicallyExtendableSite(site.id) && !localHeap.lastAllocations.isEmpty()) {
            allocContext.assign(site, localHeap.lastAllocations, symbols);
            site = symbols.getOrCreateDynamicAllocationSite(site, allocContext);
        }
        return site;
    }

    // --------------------------------------------------------
    // ----------------- Error methods ------------------------
    // --------------------------------------------------------

    private static void errorOnUndefinedAddr(long addr) throws TraceException {
        if (addr == Consts.UNDEFINED_ADDR) {
            throw new TraceException("#CalculateAddrByTLAB, How can there be an allocation in a LAB without a LAB being allocated in the first place?");
        }
    }

    private static void errorOnTLABAlloc(SpaceType space) {
        // if (space != SpaceType.EDEN) {
        // throw new TraceException("#TLAB allocated in " + space);
        // }
        // Serial GC allocates into survivor sapces in rare cases ...
    }

    private static void errorOnPLABAlloc(SpaceType space) {
        // if (space != SpaceType.SURVIVOR_TO && space != SpaceType.OLD) { throw
        // new TraceException("#PLAB allocated in " + space); }
        // TODO: Discuss if further error handling has to be done here
    }

    private static void errorOnMirrorClass(AllocatedType allocatedType) throws TraceException {
        if (allocatedType.name.equals(InstanceMirrorObjectInfo.MIRROR_CLASS)) {
            throw new TraceException("PARSE ERROR. Expected no MIRROR_CLASS here");
        }
    }

    // --------------------------------------------------------
    // ----------------------- Spaces -------------------------
    // --------------------------------------------------------

    public void doParseSpaceRedefine(int index, long addr, long size, boolean postponed, int source) {
        SpaceRedefinition redef = new SpaceRedefinition(heap, index, addr, size);
        // LOGGER.info(redef.toString());
        if (!redef.hasBeenExecuted()) {
            heap.spaceRedefinitions.add(redef);
        }
    }

    public void doParseSpaceDestroy(int firstIndex, long nRegions) {
        for (int i = 0; i < nRegions; i++) {
            heap.removeSpace(firstIndex + i);
        }
        // LOGGER.info(heap.toString());
    }

    public void doParseSpaceRelease(int index) {
        Space space = heap.getSpace(index);
        space.setMode(null);
        space.setType(null);
        // LOGGER.info(heap.toString());
    }

    public void doParseSpaceAlloc(int index, SpaceMode mode, SpaceType type) {
        Space space = heap.getSpace(index);
        space.setMode(mode);
        space.setType(type);
        // LOGGER.info(heap.toString());
    }

    public void doParseSpaceCreate(int index, long addr, long size) throws TraceException {
        Space space = heap.getSpace(index);
        if (space != null) {
            assert space.getAddress() == addr;
            // assert space.getLength() == size; //this may be valid due to rotation and postponed space events
        } else {
            space = new Space("Space #" + index);
            heap.addSpace(index, space);
            space.resetAddressAndLength(addr, size);
        }
        // LOGGER.info(heap.toString());
    }

    public void doParseGCInfo(int index, int gcId) throws TraceException {
        heap.markSpaceForCollection(index, gcId);
    }

    public void doParseGCFailed(int index) throws TraceException {
        heap.failGC(index);
    }

    // Used in assertions to validate move location
    public long getMoveTarget(long fromAddr, long toAddr, SpaceType toSpaceType, ThreadLocalHeap threadLocalHeap) throws TraceException {
        Space space = heap.getSpace(fromAddr);
        ObjectInfo obj = space.getObjectInfo(fromAddr, heap.getGC().isFull());
        assert obj != null;
        assert (toSpaceType != SpaceType.UNDEFINED);
        assert (toAddr >= 0 ? toSpaceType == heap.getSpace(toAddr).getType() : true);
        // SpaceType toSpace = toAddr < 0 ? toSpaceType :
        // RelAddrFactory.getSpaceFromAbsolute(workspace, toAddr).getType();
        Lab current = threadLocalHeap.currentLabs.get(toSpaceType);
        return current.addr + (current.position() - obj.size);
    }

    protected void doParseGCClassLoaderRootPtr(long ptr, String loaderName) throws TraceException {
        heap.addRoot(new ClassLoaderRoot(ptr, loaderName));
    }

    protected void doParseGCClassRootPtr(long ptr, int classId) throws TraceException {
        heap.addRoot(new ClassRoot(ptr, classId));
    }

    protected void doParseGCStaticFieldRootPtr(long ptr, int classId, int offset) throws TraceException {
        heap.addRoot(new StaticFieldRoot(ptr, classId, offset));
    }

    protected void doParseGCLocalVariableRootPtr(long ptr, long threadId, int classId, int methodId, int slot) throws TraceException {
        LocalVariableRoot localVariableRoot = new LocalVariableRoot(ptr, threadId, classId, methodId, slot);
        heap.addRoot(localVariableRoot);

        ThreadInfo currentThread = heap.threads.get(threadId);

        if (currentThread != null) {
            assert currentThread.isAlive();

            if (currentThread.isNewFrame(classId, methodId)) {
                currentThread.addStackframe(classId, methodId);
            }

            localVariableRoot.setCallStackIndex(currentThread.getStackDepth() - 1);
        }
    }

    protected void doParseGCVMInternalThreadDataRootPtr(long ptr, long threadId) throws TraceException {
        heap.addRoot(new VMInternalThreadDataRoot(ptr, threadId));
    }

    protected void doParseGCCodeBlobRootPtr(long ptr, int classId, int methodId) {
        try {
            heap.addRoot(new CodeBlobRoot(ptr, classId, methodId));
        } catch (TraceException e) {
            e.printStackTrace();
        }
    }

    protected void doParseGCJNILocalRootPtr(long ptr, long threadId) throws TraceException {
        heap.addRoot(new JNILocalRoot(ptr, threadId));
    }

    protected void doParseGCJNIGlobalRootPtr(long ptr, boolean weak) throws TraceException {
        heap.addRoot(new JNIGlobalRoot(ptr, weak));
    }

    protected void doParseGCOtherRootPtr(long ptr, RootPtr.RootType rootType) throws TraceException {
        heap.addRoot(new OtherRoot(ptr, rootType));
    }

    protected void doParseGCDebugRootPtr(long ptr, String vmCall) throws TraceException {
        heap.addRoot(new DebugRoot(ptr, vmCall));
    }

    public void doParseThreadAlive(int header, long id, String name) {
        heap.threads.put(id, new ThreadInfo(id, name, true));
    }

    public void doParseThreadDead(long id) {
        ThreadInfo dyingThread = heap.threads.get(id);
        if (dyingThread == null) {
            // just in case
            heap.threads.put(id, new ThreadInfo(id, "some dead thread", false));
        } else {
            dyingThread.setAlive(false);
        }
    }

    public void addTag(String tagText) {
        heap.addTag(tagText);
    }
}
