/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.parser.main;

import at.jku.mevss.anttracks.heap.io.MetaDataWriterConfig;
import at.jku.mevss.anttracks.heap.labs.Lab;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.parser.TraceParser;
import at.jku.mevss.anttracks.parser.classdefinitions.ClassDefinitionsFile;
import at.jku.mevss.anttracks.parser.heap.HeapTraceParser;
import at.jku.mevss.anttracks.parser.symbols.SymbolsFile;
import at.jku.mevss.anttracks.parser.symbols.SymbolsParser;
import at.jku.mevss.anttracks.util.ApplicationStatistics;
import at.jku.mevss.anttracks.util.Consts;
import at.jku.mevss.anttracks.util.GCReporter;

import java.io.File;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

public class Main {

    public static void main(String argv[]) throws IOException {
        List<File> traceFiles = new ArrayList<>();
        String metaPath = null;
        File features = null;
        boolean reportGc = true;
        int useCallContext = Symbols.CALLCONTEXT_STATIC;

        int metaDumpCount = MetaDataWriterConfig.DEFAULT_DUMP_COUNT;
        boolean metaMajorGCsOnly = MetaDataWriterConfig.DEFAULT_AT_MAJOR_GCS_ONLY;
        boolean check = false;

        for (int index = 0; index < argv.length; index++) {
            System.out.printf("Parameter %d of %d: %s\n", index + 1, argv.length, argv[index]);
            String arg = argv[index];
            if (arg.contains("=")) {
                int splitterIndex = arg.indexOf("=");
                String key = arg.substring(0, splitterIndex);
                String value = arg.substring(splitterIndex + 1);
                System.out.printf("Key: %s, Value: %s\n", key, value);
                if (key.equals("ConsistencyTest")) {
                    check = Boolean.parseBoolean(value);
                    System.out.printf("ConsistencyTest: %s\n", check);
                } else if (key.equals("MetaDataPath")) {
                    metaPath = value;
                    System.out.printf("Meta path: %s\n", metaPath);
                } else if (key.equals("MetaDataDumpCount")) {
                    metaDumpCount = Integer.parseInt(value);
                    System.out.printf("Meta dump count: %s\n", metaDumpCount);
                } else if (key.equals("MetaDataAtMajorGCsOnly")) {
                    metaMajorGCsOnly = Boolean.parseBoolean(value);
                    System.out.printf("Meta data at MajorGCs only: %s\n", metaMajorGCsOnly);
                } else if (key.equals("FeaturesPath")) {
                    features = new File(value);
                    System.out.printf("Feature Path: %s\n", features);
                } else if (key.equals("ReportGC")) {
                    reportGc = Boolean.parseBoolean(value);
                    System.out.printf("Report GC: %s\n", reportGc);
                } else if (key.equals("UseCallContext")) {
                    switch (value.toLowerCase()) {
                        case "none":
                        case "false":
                            useCallContext = Symbols.CALLCONTEXT_NONE;
                            break;
                        case "static":
                        case "fast":
                            useCallContext = Symbols.CALLCONTEXT_STATIC;
                            break;
                        case "full":
                        case "true":
                            useCallContext = Symbols.CALLCONTEXT_FULL;
                            break;
                        default:
                            System.err.println("Unknown value for UseCallContext ignored: " + value);
                            break;
                    }
                    System.out.printf("Call Context: %s\n", useCallContext);
                } else {
                    System.err.println("Did not understand parameter '" + arg + "'");
                    System.exit(1);
                }
            } else if (arg.equals("GenerateMetaData")) {
                metaPath = Consts.ANT_META_DIRECTORY;
            } else {
                traceFiles.add(new File(arg).getAbsoluteFile());
            }
        }

        // affects performance only if LightweightStorage is enabled
        String addrGranularity = System.getProperty("EvalAddrGranularity");
        Lab.ADDR_GRANULARITY = addrGranularity != null ? Integer.parseInt(addrGranularity) : Lab.ADDR_GRANULARITY;
        // benchmark flag to compare parsing vs consuming of ptrs
        String parsePtrs = System.getProperty("ParsePtrs");
        SymbolsParser.parsePtrs = parsePtrs != null ? Boolean.parseBoolean(parsePtrs) : true;

        TraceParser.CONSISTENCY_CHECK = check;

        int errors = 0;
        if (reportGc) {
            GCReporter.getInstance();
        }

        if (traceFiles.size() == 0) {
            System.err.println("No trace file path given!");
        }

        // 1. Logging
        String loggingConfigFileString = "java.util.logging.config.file";
        if (System.getProperty(loggingConfigFileString) == null) {
            System.getProperties().setProperty(loggingConfigFileString, "log.config");
        }
        System.out.println("Log config file: " + System.getProperty(loggingConfigFileString));

        // 2. Statistics output path
        String statisticsPathString = "at.jku.mevss.trace.gui.printStatisticsPath";
        String printStatisticsPath = System.getProperty(statisticsPathString);
        System.out.println("Statistics result file: " + (printStatisticsPath == null ? "null" : printStatisticsPath));

        Runtime.getRuntime().addShutdownHook(new Thread((Runnable) () -> {
            if (printStatisticsPath != null) {
                ApplicationStatistics.getInstance()
                                     .export(new File(printStatisticsPath.replace("%d",
                                                                                  new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss").format(new Date(System.currentTimeMillis
                                                                                          ())))
                                                                         .replace(" ", "")));
            }
            ApplicationStatistics.getInstance().print();
        }, "Shutdown-thread"));

        // 3. Assertions
        boolean assertionsEnabled = false;
        assert assertionsEnabled = true;
        System.out.println("Assertions enabled: " + assertionsEnabled);

        for (File file : traceFiles) {
            try {
                SymbolsParser symbolsParser = new SymbolsParser(SymbolsFile.findSymbolsFileToTrace(file),
                                                                ClassDefinitionsFile.findClassDefinitionsFileToTrace(file),
                                                                features,
                                                                useCallContext);
                Symbols sym = symbolsParser.parse();

                MetaDataWriterConfig config = metaPath != null ? new MetaDataWriterConfig(sym.root + File.separator + metaPath, metaDumpCount, metaMajorGCsOnly) : null;

                HeapTraceParser parser = new HeapTraceParser(sym, config);
                parser.addTraceParserListener(new ProgressReporter(1000 * 60));
                parser.parse();
            } catch (Throwable e) {
                e.printStackTrace(System.err);
                errors++;
            }
        }
        if (errors > 0) {
            System.err.println(errors + " errors found.");
        }
    }

}
