package at.jku.mevss.anttracks.parser.printing;

import at.jku.mevss.anttracks.heap.objects.ObjectInfo;
import at.jku.mevss.anttracks.heap.roots.RootPtr;
import at.jku.mevss.anttracks.heap.space.Space;
import at.jku.mevss.anttracks.heap.space.SpaceMode;
import at.jku.mevss.anttracks.heap.space.SpaceType;
import at.jku.mevss.anttracks.heap.symbols.AllocatedType;
import at.jku.mevss.anttracks.heap.symbols.AllocatedTypes;
import at.jku.mevss.anttracks.heap.symbols.AllocationSite;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.parser.*;
import at.jku.mevss.anttracks.util.Consts;
import at.jku.mevss.anttracks.util.MutableLong;
import at.jku.mevss.anttracks.util.TraceException;

import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.logging.Logger;

import static at.jku.mevss.anttracks.heap.symbols.AllocatedTypes.MIRROR_CLASS_NAME;

public class PrintingTraceSlaveParser extends TraceSlaveParser<Void> {

    private Logger LOGGER = Logger.getLogger(PrintingTraceSlaveParser.class.getSimpleName());

    public PrintingTraceSlaveParser(int id,
                                    MutableLong size,
                                    BlockingQueue<ThreadLocalHeap> masterQueue,
                                    Void workspace,
                                    RelAddrFactory relAddrFactory,
                                    Symbols symbols,
                                    boolean test,
                                    ErrorHandler error) {
        super(id, size, masterQueue, workspace, relAddrFactory, symbols, test, error, null);
    }

    @Override
    protected void cleanUp(ThreadLocalHeap threadLocalHeap) throws TraceException {
        LOGGER.info("CLEAN-UP");
    }

    protected void parseSpaceRedefine(int word) throws TraceException {
        int index = getNextWord();
        long startAddr = getNextDoubleWord();
        long size = getNextDoubleWord();
        boolean postponed = getNextWord() != 0;
        int source = getNextWord();

        LOGGER.info(String.format("Redefine space %d by %s (postponed: %s, source: %d)", index, Thread.currentThread().getName(), postponed, source));
        LOGGER.info(String.format("New range: %,d - %,d, Size: %,d", startAddr, startAddr + size, size));
    }

    protected void parseSpaceDestroy(int word) throws TraceException {
        int firstIndex = getNextWord();
        long nRegions = getNextDoubleWord();

        LOGGER.info("Destroy space " + firstIndex + " to " + (firstIndex + nRegions - 1) + " by " + Thread.currentThread().getName());
    }

    protected void parseSpaceRelease(int word) throws TraceException {
        int index = getNextWord();
        LOGGER.info("Release space " + index + " by " + Thread.currentThread().getName());
    }

    protected void parseSpaceAlloc(int word) throws TraceException {
        SpaceType spaceType = SpaceType.values()[(word >> 18) & 0xFF];
        SpaceMode spaceMode = SpaceMode.values()[(word >> 10) & 0xFF];
        assert spaceType != null;
        assert spaceMode != null;
        int index = getNextWord();

        LOGGER.info("Alloc space " + index + " by " + Thread.currentThread().getName());
    }

    protected void parseSpaceCreate(int word) throws TraceException {
        // int spaceType = (word >> 22) & 0b1111;
        int index = getNextWord();
        long startAddr = getNextDoubleWord();
        long size = getNextDoubleWord();
        LOGGER.info("Create space " + index + " by " + Thread.currentThread().getName());
    }

    protected void parseGCInfo(int word) throws TraceException {
        int index = word & 0x00FFFFFF;
        int id = getNextWord();

    }

    protected void parseGCFailed(int word) throws TraceException {
        int index = word & 0x00FFFFFF;

    }

    protected void parseGCStart(int word, long start, long end, ThreadLocalHeap threadLocalHeap) throws TraceException {
        ParserGCInfo info = parseGCMeta(word);

        LOGGER.info(String.format("Start GC %d (%s, %s)", info.id, info.type, info.cause));
    }

    protected void parseGCEnd(int word, ThreadLocalHeap threadLocalHeap, long start, long end) throws TraceException {
        ParserGCInfo info = parseGCMeta(word);
        int flags = recoverValue(word, 3, 3);
        boolean failed = (flags & 1) != 0;

        LOGGER.info("End GC " + info.id);
    }

    protected void parseGCInterrupt(int word) throws TraceException {
        int id = getNextWord();
        long address = getNextDoubleWord();

    }

    protected void parseGCContinue(int word) throws TraceException {
        int id = getNextWord();
        long address = getNextDoubleWord();

    }

    protected void parseGCMoveFast_WideOrNarrow_Ptr(EventType type, int word, ThreadLocalHeap threadLocalHeap, int toSpace) throws TraceException {
        SpaceType toSpaceType = RelAddrFactory.getDefinedSpaceOnly(toSpace);

        long toAddr = -1;
        long fromAddr = -1;

        if (type == EventType.GC_MOVE_FAST_NARROW) {
            int from = (int) (RelAddrFactory.getDefinedAddrOnly(recoverValue(word, 1, 3)) >> 2);
            fromAddr = relAddrFactory.create(from);
        } else if (type == EventType.GC_MOVE_FAST_WIDE || type == EventType.GC_MOVE_FAST_WIDE_PTR) {
            fromAddr = getNextDoubleWord();
        } else if (type == EventType.GC_MOVE_FAST || type == EventType.GC_MOVE_FAST_PTR) {
            fromAddr = relAddrFactory.create(getNextWord());
        } else {
            assert false;
            throw new TraceException("internal error");
        }

        Space fromSpace = getSpace(fromAddr);
        ObjectInfo obj = getObjectIsFiller(fromSpace, fromAddr);
        boolean isFiller = obj == null ? false : true;

        long assignedAddr;

        /*
        if (!isFiller) {
            obj = fromSpace.getNonFillerObjectInfo(fromAddr);
        }
        assignedAddr = doGCMove(type, threadLocalHeap, obj, fromAddr, toAddr, toSpaceType);
        */
        assignedAddr = -1;

        if (type == EventType.GC_MOVE_FAST_PTR || type == EventType.GC_MOVE_FAST_WIDE_PTR) {
            assert (getMoveTarget(fromAddr, toAddr, toSpaceType, threadLocalHeap) == assignedAddr);
            long[] ptrs = parsePointers(word, assignedAddr, threadLocalHeap);
            doPtrEvent(type, threadLocalHeap, fromAddr, assignedAddr, ptrs, isFiller);
        }
    }

    protected void parseGCMoveSlow(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {

        long fromAddr = getNextDoubleWord();
        long toAddr = getNextDoubleWord();

        Space fromSpace = getSpace(fromAddr);
        ObjectInfo obj = getObjectIsFiller(fromSpace, fromAddr);
        boolean isFiller = obj == null ? false : true;

        /*
        if (!isFiller) {
            obj = fromSpace.getNonFillerObjectInfo(fromAddr);
        }
        doGCMove(type, threadLocalHeap, obj, fromAddr, toAddr, null);
        */

        if (type == EventType.GC_MOVE_SLOW_PTR) {
            long[] ptrs = parsePointers(word, toAddr, threadLocalHeap);
            doPtrEvent(type, threadLocalHeap, fromAddr, toAddr, ptrs, isFiller);
        }
    }

    protected void parseGCKeepAlive(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        long addr = getNextDoubleWord();
        Space space = getSpace(addr);
        ObjectInfo obj = getObjectIsFiller(space, addr);
        boolean isFiller = obj == null ? false : true;

        /*
        if (!isFiller) {
            obj = space.getNonFillerObjectInfo(addr);
        }
        doGCMove(type, threadLocalHeap, obj, addr, addr, null);
        */

        if (type == EventType.GC_KEEP_ALIVE_PTR) {
            long[] ptrs = parsePointers(word, addr, threadLocalHeap);
            doPtrEvent(type, threadLocalHeap, addr, addr, ptrs, isFiller);
        }
    }

    protected void parseGCRootPtr(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {

        for (int i = 0; i < RootPtr.MAX_ROOTS_PER_EVENT; i++) {
            long ptr = getNextDoubleWord();
            if (ptr == -1) {    // not a nullptr (those are 0) but the convention for end of root block
                // no more roots in this block
                break;
            } else if (ptr == 0) {
                // NULLPTR
                ptr = NULL_PTR;    // because we use -1 for NULLPTRs
            }

            RootPtr.RootType rootType = RootPtr.RootType.values()[getNextWord()];
            switch (rootType) {
                case CLASS_LOADER_ROOT:
                    String loaderName = getString(getNextWord());

                    break;

                case CLASS_ROOT:
                    int classId = getNextWord();

                    break;

                case STATIC_FIELD_ROOT:
                    classId = getNextWord();
                    int offset = getNextWord();

                    break;

                case LOCAL_VARIABLE_ROOT:
                    long threadId = getNextDoubleWord();
                    classId = getNextWord();
                    int methodId = getNextWord();
                    int slot = getNextWord();

                    break;

                case VM_INTERNAL_THREAD_DATA_ROOT:
                    threadId = getNextDoubleWord();

                    break;

                case CODE_BLOB_ROOT:
                    classId = getNextWord();
                    methodId = getNextWord();

                    break;

                case JNI_LOCAL_ROOT:
                    threadId = getNextDoubleWord();

                    break;

                case JNI_GLOBAL_ROOT:
                    boolean weak = getNextWord() != 0;

                    break;

                // other roots
                case CLASS_LOADER_INTERNAL_ROOT:
                case UNIVERSE_ROOT:
                case SYSTEM_DICTIONARY_ROOT:
                case BUSY_MONITOR_ROOT:
                case INTERNED_STRING:
                case FLAT_PROFILER_ROOT:
                case MANAGEMENT_ROOT:
                case JVMTI_ROOT:

                    break;

                case DEBUG_ROOT:
                    String vmCall = getString(getNextWord());

                    break;

                default:
                    throw new TraceException(rootType + " is not a valid root type id!");
            }
        }
    }

    protected void parseGCObjPtr(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        long addr = getNextDoubleWord();
        long[] ptrs = parsePointers(word, addr, threadLocalHeap);
        doPtrEvent(type, threadLocalHeap, -1, addr, ptrs, false);
    }

    protected List<ObjectInfo> parseGCMoveRegion(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int numOfObjects = recoverValue(word, 1, 3);
        long fromAddr = getNextDoubleWord();
        long toAddr = getNextDoubleWord();

        return null;
    }

    protected void parseThreadAlive(EventType type, int header) throws TraceException {
        final int MAX_NAME_LENGTH = 12;
        long id = getNextDoubleWord();
        StringBuilder name = new StringBuilder();
        end:
        for (int i = 0; i < MAX_NAME_LENGTH; i++) {
            int word = getNextWord();
            for (int j = 0; j < 4; j++) {
                char c = (char) ((word >> ((3 - j) * 8)) & 0xFF);
                if (c == '\0') {
                    break end;
                }
                name.append(c);
            }
        }

        String finalName = name.toString();

    }

    protected void parseThreadDeath(EventType type, int word) throws TraceException {
        long id = getNextDoubleWord();

    }

    protected void parseSyncObj(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int allocationSiteId = recoverValue(word, 1, 3);
        int allocatedTypeId = getNextWord();
        AllocatedType allocatedType = symbols.types.get(allocatedTypeId);

        long fromAddr = getNextDoubleWord();
        long toAddr = type == EventType.SYNC_OBJ ? getNextDoubleWord() : fromAddr;
        int length;
        if (allocatedType.name.startsWith("[")) {
            length = getNextWord();
        } else {
            length = Consts.UNDEFINED_LENGTH;
        }
        int size;
        // TODO: We reworked mirror classes, probably something has to be changed here!
        if (allocatedType.name.equals(MIRROR_CLASS_NAME)) {
            size = getNextWord();
        } else {
            size = 0;
        }

    }

    protected void parseObjAllocFastCi(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_2);
        // check if first bit is set, if yes --> big allocSite (3 byte), otherwise small allocSite (2 byte)
        if (((allocationSiteId >> 15) & 1) != 0) {
            allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_3);
        }
        AllocationSite allocationSite = symbols.sites.getById(allocationSiteId);
        assert allocationSite != null;
        AllocatedType allocatedType = symbols.types.get(allocationSite.allocatedTypeId);

        boolean isArray = allocatedType.name.startsWith("[");
        int arrayLength = Consts.UNDEFINED_LENGTH;
        if (isArray) {
            arrayLength = recoverArrayLength(word);
        }

    }

    protected void parseObjAllocFastC2DeviantType(EventType type, int header, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int allocationSiteId = recoverValue(header, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_2);
        // check if first bit is set, if yes --> big allocSite (3 byte), otherwise small allocSite (2 byte)
        if (((allocationSiteId >> 15) & 1) != 0) {
            allocationSiteId = recoverValue(header, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_3);
        }
        AllocatedType allocatedType = symbols.types.get(getNextWord());

        boolean isArray = allocatedType.name.startsWith("[");
        int arrayLength = Consts.UNDEFINED_LENGTH;
        if (isArray) {
            arrayLength = recoverArrayLength(header);
        }

    }

    protected void parseObjAllocFastIr(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_2);
        // check if first bit is set, if yes --> big allocSite (3 byte), otherwise small allocSite (2 byte)
        if (((allocationSiteId >> 15) & 1) != 0) {
            allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_3);
        }

        AllocationSite allocationSite = symbols.sites.getById(allocationSiteId);

    }

    protected void parseObjAllocNormalCi(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_2);
        // check if first bit is set, if yes --> big allocSite (3 byte), otherwise small allocSite (2 byte)
        if (((allocationSiteId >> 15) & 1) != 0) {
            allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_3);
        }
        AllocationSite allocationSite = symbols.sites.getById(allocationSiteId);
        AllocatedType allocatedType = symbols.types.get(allocationSite.allocatedTypeId);
        errorOnMirrorClass(allocatedType);

        boolean isArray = allocatedType.name.startsWith("[");
        long addr = getNextDoubleWord();
        int arrayLength = Consts.UNDEFINED_LENGTH;
        if (isArray) {
            arrayLength = recoverArrayLength(word);
        }
    }

    protected void parseObjAllocNormalIr(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {

        int allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_2);
        // check if first bit is set, if yes --> big allocSite (3 byte), otherwise small allocSite (2 byte)
        if (((allocationSiteId >> 15) & 1) != 0) {
            allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_3);
        }
        AllocationSite allocationSite = symbols.sites.getById(allocationSiteId);
        long addr = getNextDoubleWord();

    }

    protected void parseObjAllocSlowCiIr_Deviant(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_2);

        boolean bigAllocSite = false;
        // check if first bit is set, if yes --> big allocSite (3 byte), otherwise small allocSite (2 byte)
        if (((allocationSiteId >> 15) & 1) != 0) {
            allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_3);
            bigAllocSite = true;
        }

        AllocationSite allocationSite = symbols.sites.getById(allocationSiteId);
        long addr = getNextDoubleWord();

        int arrayLength = 0;
        if (!bigAllocSite) { // all arrays have a small allocSite ID
            arrayLength = recoverArrayLength(word);
        }

        int realAllocatedTypeId = -1;
        if (type == EventType.OBJ_ALLOC_SLOW_C1_DEVIANT_TYPE || type == EventType.OBJ_ALLOC_SLOW_IR_DEVIANT_TYPE || type == EventType.OBJ_ALLOC_SLOW_C2_DEVIANT_TYPE) {
            realAllocatedTypeId = getNextWord();
        }

    }

    protected void parseObjAllocSlow(EventType type, int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_2);
        boolean bigAllocSite = false;
        // check if first bit is set, if yes --> big allocSite (3 byte), otherwise small allocSite (2 byte)
        if (((allocationSiteId >> 15) & 1) != 0) {
            allocationSiteId = recoverValue(word, ALLOCATION_SITE_INDEX_1, ALLOCATION_SITE_INDEX_3);
            bigAllocSite = true;
        }

        AllocationSite allocationSite = symbols.sites.getById(allocationSiteId);

        long addr = getNextDoubleWord();

        int arrayLength = 0;
        if (!bigAllocSite) { // all arrays have a small allocSite ID
            arrayLength = recoverArrayLength(word);
        }

        if (allocationSite.allocatedTypeId == AllocatedType.ALLOCATED_TYPE_IDENTIFIER_UNKNOWN) {
            int id = getNextWord();
            allocationSite = allocationSite.copy(id);
        }

        AllocatedType allocatedType = symbols.types.get(allocationSite.allocatedTypeId);

        boolean isArray = false;
        int size = Consts.UNDEFINED_LENGTH;

        if (allocatedType.name.equals(AllocatedTypes.MIRROR_CLASS_NAME)) {
            size = getNextWord();
            allocatedType.size = size;
            arrayLength = Consts.UNDEFINED_LENGTH;
        } else {
            isArray = allocatedType.name.startsWith("[");
        }

        boolean mayBeFiller = mayBeFiller(allocationSite);

    }

    protected void parseTlabAlloc(int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int size = (int) getNextDoubleWord();
        long addr = getNextDoubleWord();

    }

    protected void parsePlabAlloc(int word, ThreadLocalHeap threadLocalHeap) throws TraceException {
        int size = (int) getNextDoubleWord();
        long addr = getNextDoubleWord();

        LOGGER.info(String.format("PLAB allocated by %s (%,d - %,d, Size: %,d)", threadLocalHeap.thread, addr, addr + size, size));
    }

    // --------------------------------------------------------
    // ---------------- Helper methods ------------------------
    // --------------------------------------------------------

    protected Space getSpace(long addr) {
        return null;
    }

    protected ObjectInfo getObjectIsFiller(Space space, long addr) throws TraceException {
        return null;
    }

    protected long doGCMove(EventType type, ThreadLocalHeap threadLocalHeap, ObjectInfo obj, long fromAddr, long toAddr, SpaceType toSpaceType) throws TraceException {
        return -1;
    }

    protected long getMoveTarget(long fromAddr, long toAddr, SpaceType toSpaceType, ThreadLocalHeap threadLocalHeap) throws TraceException {
        return -1;
    }

    protected void doPtrEvent(EventType eventType, ThreadLocalHeap threadLocalHeap, long fromAddr, long toAddr, long[] ptrs, boolean isFiller) throws TraceException {

    }
}
