/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.datastructures.test;

import at.jku.mevss.anttracks.heap.datastructures.cocor.DataStructurePart;
import at.jku.mevss.anttracks.heap.datastructures.cocor.Parser;
import at.jku.mevss.anttracks.heap.datastructures.cocor.PointsToType;
import at.jku.mevss.anttracks.heap.datastructures.cocor.Scanner;
import at.jku.mevss.anttracks.heap.symbols.AllocatedTypes;
import at.jku.mevss.anttracks.util.DataStructureUtil;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

public class DataStructureParserTest {

    Parser parser;

    @Before
    public void setup() {
        parser = new Parser(new Scanner(DataStructureParserTest.class.getResourceAsStream("/datastructure_example.ds")));
        parser.Parse();
        parser.getDataStructureDefinitions().resolveDataStructureParts(new AllocatedTypes());
    }

    // ---------------------------------------------------------------------------
    // Extract package

    @Test
    public void testDataStructureCount() {
        Assert.assertEquals(6, parser.getAllDataStructureParts().size()); // unknown (comes from type resolving, is a placeholder type for null), Test, A, B, C, D
        Assert.assertEquals(1, parser.getHeadDataStructureParts().size()); // Test
        Assert.assertEquals(5, parser.getInternalDataStructureParts().size()); // unknown, A, B, C, D
    }

    @Test
    public void testNamespacePrependedToDataStructeNames() {
        Assert.assertTrue(parser.getAllDataStructureParts().stream().anyMatch(dsp -> dsp.getType().equals("java.util.Test")));
        Assert.assertTrue(parser.getAllDataStructureParts().stream().anyMatch(dsp -> dsp.getType().equals("java.util.A")));
        Assert.assertTrue(parser.getAllDataStructureParts().stream().anyMatch(dsp -> dsp.getType().equals("java.B")));
        Assert.assertTrue(parser.getAllDataStructureParts().stream().anyMatch(dsp -> dsp.getType().equals("C")));
        Assert.assertTrue(parser.getAllDataStructureParts().stream().anyMatch(dsp -> dsp.getType().equals("D")));
    }

    @Test
    public void testNamespacePrependedToPointsToTypes() {
        DataStructurePart javaUtilTest = parser.getHeadDataStructureParts().stream().filter(dsp -> dsp.getType().equals("java.util.Test")).findAny().orElse(null);
        if (javaUtilTest == null) {
            Assert.fail();
        }

        Assert.assertEquals(12, javaUtilTest.getPointsToTypes().size());
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().anyMatch(ptt -> ptt.getName().equals("A")));
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().anyMatch(ptt -> ptt.getName().equals("B")));
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().anyMatch(ptt -> ptt.getName().equals("C")));
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().anyMatch(ptt -> ptt.getName().equals("D")));
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().anyMatch(ptt -> ptt.getName().equals("*Test")));
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().anyMatch(ptt -> ptt.getName().equals("A[]")));
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().anyMatch(ptt -> ptt.getName().equals("*")));
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().anyMatch(ptt -> ptt.getName().equals("java.util.A")));
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().anyMatch(ptt -> ptt.getName().equals("java.util.B")));
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().anyMatch(ptt -> ptt.getName().equals("java.util.C")));
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().anyMatch(ptt -> ptt.getName().equals("java.util.D")));
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().anyMatch(ptt -> ptt.getName().equals("java.util.A[]")));

        Assert.assertFalse(javaUtilTest.getPointsToTypes().stream().anyMatch(ptt -> ptt.getName().equals("java.util.Test")));
        Assert.assertFalse(javaUtilTest.getPointsToTypes().stream().anyMatch(ptt -> ptt.getName().equals("java.util.*Test")));
        Assert.assertFalse(javaUtilTest.getPointsToTypes().stream().anyMatch(ptt -> ptt.getName().equals("java.util.*")));
    }

    @Test
    public void testPointsToTypesFollow() {
        DataStructurePart javaUtilTest = parser.getHeadDataStructureParts().stream().filter(dsp -> dsp.getType().equals("java.util.Test")).findAny().orElse(null);
        if (javaUtilTest == null) {
            Assert.fail();
        }

        Assert.assertEquals(12, javaUtilTest.getPointsToTypes().size());
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("A")).findAny().get().isFollow());
        Assert.assertFalse(javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("B")).findAny().get().isFollow());
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("C")).findAny().get().isFollow());
        Assert.assertFalse(javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("D")).findAny().get().isFollow());
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("*Test")).findAny().get().isFollow());
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("A[]")).findAny().get().isFollow());
        Assert.assertFalse(javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("*")).findAny().get().isFollow());
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("java.util.A")).findAny().get().isFollow());
        Assert.assertFalse(javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("java.util.B")).findAny().get().isFollow());
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("java.util.C")).findAny().get().isFollow());
        Assert.assertFalse(javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("java.util.D")).findAny().get().isFollow());
        Assert.assertTrue(javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("java.util.A[]")).findAny().get().isFollow());
    }

    @Test
    public void testMatchingDataStructurePartsWithoutWildcard() {
        DataStructurePart javaUtilTest = parser.getHeadDataStructureParts().stream().filter(dsp -> dsp.getType().equals("java.util.Test")).findAny().orElse(null);
        if (javaUtilTest == null) {
            Assert.fail();
        }

        PointsToType aInJavaUtilTest = javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("A")).findAny().orElse(null);
        PointsToType bInJavaUtilTest = javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("B")).findAny().orElse(null);
        PointsToType cInJavaUtilTest = javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("C")).findAny().orElse(null);
        PointsToType dInJavaUtilTest = javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("D")).findAny().orElse(null);
        Assert.assertEquals(0, aInJavaUtilTest.getMatchingDataStructureParts().size());
        Assert.assertEquals(0, bInJavaUtilTest.getMatchingDataStructureParts().size());
        Assert.assertEquals(1, cInJavaUtilTest.getMatchingDataStructureParts().size());
        Assert.assertEquals(1, dInJavaUtilTest.getMatchingDataStructureParts().size());

        PointsToType javaUtilAInJavaUtilTest = javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("java.util.A")).findAny().orElse(null);
        PointsToType javaUtilBInJavaUtilTest = javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("java.util.B")).findAny().orElse(null);
        PointsToType javaUtilCInJavaUtilTest = javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("java.util.C")).findAny().orElse(null);
        PointsToType javaUtilDInJavaUtilTest = javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("java.util.D")).findAny().orElse(null);
        Assert.assertEquals(1, javaUtilAInJavaUtilTest.getMatchingDataStructureParts().size());
        Assert.assertEquals(0, javaUtilBInJavaUtilTest.getMatchingDataStructureParts().size());
        Assert.assertEquals(0, javaUtilCInJavaUtilTest.getMatchingDataStructureParts().size());
        Assert.assertEquals(0, javaUtilDInJavaUtilTest.getMatchingDataStructureParts().size());
    }

    @Test
    public void testMatchingDataStructurePartsWithWildcard() {
        DataStructurePart javaUtilTest = parser.getHeadDataStructureParts().stream().filter(dsp -> dsp.getType().equals("java.util.Test")).findAny().orElse(null);
        if (javaUtilTest == null) {
            Assert.fail();
        }

        PointsToType testInJavaUtilTest = javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("*Test")).findAny().orElse(null);
        Assert.assertEquals(1, testInJavaUtilTest.getMatchingDataStructureParts().size());

        PointsToType starInJavaUtilTest = javaUtilTest.getPointsToTypes().stream().filter(ptt -> ptt.getName().equals("*")).findAny().orElse(null);
        Assert.assertEquals(1, starInJavaUtilTest.getMatchingDataStructureParts().size());
        Assert.assertEquals(DataStructureUtil.starDummyDataStructurePart, starInJavaUtilTest.getMatchingDataStructureParts().get(0));
    }

    @Test
    public void testDataStructureResolvingDeepPointsToDataStructureParts() {
        DataStructurePart javaUtilTest = parser.getHeadDataStructureParts().stream().filter(dsp -> dsp.getType().equals("java.util.Test")).findAny().orElse(null);
        if (javaUtilTest == null) {
            Assert.fail();
        }

        Assert.assertEquals(3, javaUtilTest.getDeepPointsToDataStructureParts().size()); // A, C, Test
        Assert.assertTrue(javaUtilTest.getDeepPointsToDataStructureParts().stream().anyMatch(dsp -> dsp.getType().equals("java.util.A")));
        Assert.assertTrue(javaUtilTest.getDeepPointsToDataStructureParts().stream().anyMatch(dsp -> dsp.getType().equals("C")));
        Assert.assertTrue(javaUtilTest.getDeepPointsToDataStructureParts().stream().anyMatch(dsp -> dsp.getType().equals("java.util.Test")));
    }

    @Test
    public void testDataStructureResolvingFlatPointsTo() {
        DataStructurePart javaUtilTest = parser.getHeadDataStructureParts().stream().filter(dsp -> dsp.getType().equals("java.util.Test")).findAny().orElse(null);
        if (javaUtilTest == null) {
            Assert.fail();
        }

        Assert.assertEquals(2, javaUtilTest.getFlatDataStructureParts().size()); // star dummy + D (not B because it is not in the same package and not without package)
        /*
        Assert.assertTrue(javaUtilTest.getFlatDataStructureParts().stream().anyMatch(dsp -> dsp.getType().equals("B")));
        Assert.assertTrue(javaUtilTest.getFlatDataStructureParts().stream().anyMatch(dsp -> dsp.getType().equals("java.util.B")));
        Assert.assertTrue(javaUtilTest.getFlatDataStructureParts().stream().anyMatch(dsp -> dsp.getType().equals("D")));
        Assert.assertTrue(javaUtilTest.getFlatDataStructureParts().stream().anyMatch(dsp -> dsp.getType().equals("java.util.D")));
        Assert.assertTrue(javaUtilTest.getFlatDataStructureParts().stream().anyMatch(dsp -> dsp.getType().equals("*")));
        */
    }

    @Test
    public void testDataStructureResolvingDeepArrays() {
        DataStructurePart javaUtilTest = parser.getHeadDataStructureParts().stream().filter(dsp -> dsp.getType().equals("java.util.Test")).findAny().orElse(null);
        if (javaUtilTest == null) {
            Assert.fail();
        }

        // Current problem: Array DSPs would get generated based on the types in the symbols file.
        // Since we use empty type information in this test yet, we have no array types, and thus, no array DSPs.
        // TODO introduce mocking in JUnit to allow to mock array types
        Assert.assertEquals(0, javaUtilTest.getDeepArrayDataStructureParts().size());

        /*
        Assert.assertEquals(2, javaUtilTest.getDeepArrayDataStructureParts().size()); // A[], java.util.A[]
        Assert.assertTrue(javaUtilTest.getDeepArrayDataStructureParts().stream().anyMatch(dsp -> dsp.getType().equals("A[]")));
        Assert.assertTrue(javaUtilTest.getDeepArrayDataStructureParts().stream().anyMatch(dsp -> dsp.getType().equals("java.util.A[]")));
        */
    }

    @After
    public void teardown() {

    }
}
