/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.anttracks.parser.test;

import at.jku.mevss.anttracks.parser.Scanner;
import at.jku.mevss.anttracks.parser.TraceFile;
import at.jku.mevss.anttracks.parser.TraceScannerFactory;
import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;

import java.io.*;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

public class TraceScannerFactory_Range_Test {

    private static final TraceFile.Access[] FILES = new TraceFile.Access[10];
    private static final File[] OS_FILES = new File[FILES.length];
    private static final long FILE_SIZE = 100_000;
    private static final long FILES_TOTAL_SIZE = FILE_SIZE * FILES.length;

    @BeforeClass
    public static void createFiles() throws IOException {
        for (int i = 0; i < FILES.length; i++) {
            File file = File.createTempFile(TraceScannerFactory_Range_Test.class.getName(), ".tmp");
            file.deleteOnExit();
            try (DataOutputStream out = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(file)))) {
                out.writeInt(toLittleEndian(0xC0FFEE)); // magic prefix
                out.writeInt(toLittleEndian(0)); // header size
                out.writeInt(toLittleEndian(1)); // file type
                out.writeInt(toLittleEndian(i)); // file index
                for (int j = 4; j < FILE_SIZE; j++) { // random data
                    out.writeInt(toLittleEndian(j));
                }
            }
            FILES[i] = createAccess(file);
            OS_FILES[i] = file;
        }
    }

    @AfterClass
    public static void deleteFiles() throws IOException {
        for (TraceFile.Access file : FILES) {
            file.close();
        }
        for (File file : OS_FILES) {
            file.delete();
        }
    }

    private static int toLittleEndian(int value) {
        ByteBuffer buffer = ByteBuffer.allocate(4);
        buffer.order(ByteOrder.BIG_ENDIAN);
        buffer.putInt(value);
        buffer.rewind();
        buffer.order(ByteOrder.LITTLE_ENDIAN);
        return buffer.getInt();
    }

    /*
     * private static TraceFile.Access[] convertFilesToAccess(File... files) { return
     * Arrays.stream(files).map(TraceScannerFactory_Range_Test::createAccess).collect(Collectors.toList()).toArray(new
     * TraceFile.Access[files.length]); }
     */

    private static TraceFile.Access createAccess(File f) {
        try {
            return TraceFile.open(f, new int[0], f)[0];
        } catch (IOException e) {
            throw new AssertionError(e);
        }
    }

    @Test
    public void testNoRange() throws IOException {
        TraceScannerFactory testee = new TraceScannerFactory(FILES);
        int scanners = 0;
        for (Scanner scanner = testee.getNext(); scanner != null; scanner = testee.getNext()) {
            // assertEquals(FILES[scanners], scanner.getFile());
            assertEquals(scanners * FILE_SIZE, scanner.getGlobalPosition() / 4);
            long size = consume(scanner);
            assertEquals(FILE_SIZE, size);
            assertEquals(scanners * FILE_SIZE + FILE_SIZE, scanner.getGlobalPosition() / 4);
            scanners++;
            scanner.close();
        }
        assertEquals(FILES.length, scanners);
    }

    @Test
    public void testSkipFirstFile() throws IOException {
        TraceScannerFactory testee = new TraceScannerFactory(FILES, FILE_SIZE * 4, FILES_TOTAL_SIZE * 4);
        int scanners = 0;
        for (Scanner scanner = testee.getNext(); scanner != null; scanner = testee.getNext()) {
            // assertEquals(FILES[scanners + 1], scanner.getFile());
            assertEquals((scanners + 1) * FILE_SIZE, scanner.getGlobalPosition() / 4);
            long size = consume(scanner);
            assertEquals(FILE_SIZE, size);
            assertEquals((scanners + 1) * FILE_SIZE + FILE_SIZE, scanner.getGlobalPosition() / 4);
            scanners++;
            scanner.close();
        }
        assertEquals(FILES.length - 1, scanners);
    }

    @Test
    public void testSkipLastFile() throws IOException {
        TraceScannerFactory testee = new TraceScannerFactory(FILES, 0, FILES_TOTAL_SIZE * 4 - FILE_SIZE * 4);
        int scanners = 0;
        for (Scanner scanner = testee.getNext(); scanner != null; scanner = testee.getNext()) {
            // assertEquals(FILES[scanners], scanner.getFile());
            assertEquals(scanners * FILE_SIZE, scanner.getGlobalPosition() / 4);
            long size = consume(scanner);
            assertEquals(FILE_SIZE, size);
            assertEquals(scanners * FILE_SIZE + FILE_SIZE, scanner.getGlobalPosition() / 4);
            scanners++;
            scanner.close();
        }
        assertEquals(FILES.length - 1, scanners);
    }

    @Test
    public void testOnlyOneFile() throws IOException {
        final int file = 3;
        long from = FILE_SIZE * 4 * file;
        long to = from + FILE_SIZE * 4;
        TraceScannerFactory testee = new TraceScannerFactory(FILES, from, to);
        int scanners = 0;
        for (Scanner scanner = testee.getNext(); scanner != null; scanner = testee.getNext()) {
            // assertEquals(FILES[scanners + file], scanner.getFile());
            assertEquals((scanners + file) * FILE_SIZE, scanner.getGlobalPosition() / 4);
            long size = consume(scanner);
            assertEquals(FILE_SIZE, size);
            assertEquals((scanners + file) * FILE_SIZE + FILE_SIZE, scanner.getGlobalPosition() / 4);
            scanners++;
            scanner.close();
        }
        assertEquals(1, scanners);
    }

    @Test
    public void testRangeInOneFile() throws IOException {
        final int file = 5;
        long from = FILE_SIZE * 4 * file + FILE_SIZE * 4 / 4;
        long to = from + FILE_SIZE * 4 / 2;
        TraceScannerFactory testee = new TraceScannerFactory(FILES, from, to);
        int scanners = 0;
        for (Scanner scanner = testee.getNext(); scanner != null; scanner = testee.getNext()) {
            // assertEquals(FILES[scanners + file], scanner.getFile());
            assertEquals((scanners + file) * FILE_SIZE + FILE_SIZE / 4, scanner.getGlobalPosition() / 4);
            long size = consume(scanner);
            assertEquals(FILE_SIZE / 2, size);
            assertEquals((scanners + file) * FILE_SIZE + FILE_SIZE / 4 * 3, scanner.getGlobalPosition() / 4);
            scanners++;
            scanner.close();
        }
        assertEquals(1, scanners);
    }

    @Test
    public void testRangeInThreeFiles() throws IOException {
        long from = FILE_SIZE * 4 * 7 + FILE_SIZE * 4 / 2;
        long to = from + FILE_SIZE * 4 * 2;
        TraceScannerFactory testee = new TraceScannerFactory(FILES, from, to);

        Scanner scanner7 = testee.getNext();
        assertNotNull(scanner7);
        // assertEquals(FILES[7], scanner7.getFile());
        assertEquals(7 * FILE_SIZE + FILE_SIZE / 2, scanner7.getGlobalPosition() / 4);
        long scanner7size = consume(scanner7);
        assertEquals(FILE_SIZE * 4 / 2, scanner7size * 4);
        assertEquals(7 * FILE_SIZE + FILE_SIZE, scanner7.getGlobalPosition() / 4);
        scanner7.close();

        Scanner scanner8 = testee.getNext();
        assertNotNull(scanner8);
        // assertEquals(FILES[8], scanner8.getFile());
        assertEquals(8 * FILE_SIZE, scanner8.getGlobalPosition() / 4);
        long scanner8size = consume(scanner8);
        assertEquals(FILE_SIZE * 4, scanner8size * 4);
        assertEquals(8 * FILE_SIZE + FILE_SIZE, scanner8.getGlobalPosition() / 4);
        scanner8.close();

        Scanner scanner9 = testee.getNext();
        assertNotNull(scanner9);
        // assertEquals(FILES[9], scanner9.getFile());
        assertEquals(9 * FILE_SIZE, scanner9.getGlobalPosition() / 4);
        long scanner9size = consume(scanner9);
        assertEquals(FILE_SIZE * 4 / 2, scanner9size * 4);
        assertEquals(9 * FILE_SIZE + FILE_SIZE / 2, scanner9.getGlobalPosition() / 4);
        scanner9.close();

    }

    private static long consume(Scanner scanner) {
        long count = 0;
        try {
            for (; ; ) {
                scanner.getWord();
                count++;
            }
        } catch (IOException e) {
        }
        return count;
    }

}
