/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.scala.trace.parser.heap.classifier

/**
  * Created by neon on 03.10.16.
  */
object Grouping {
    def apply() = new Grouping(null, null, 0, "Overall")
}

class Grouping(parent: Grouping, classifier: Classifier, level: Int, key: Any) {
    private val children: scala.collection.mutable.Map[Any, Grouping] = scala.collection.mutable.Map()

    def getChildren(): List[Grouping] = children.values.toList

    val keys: List[Any] = (if (level == 0) Nil else parent.keys) :+ key
    private val leafs: scala.collection.mutable.Map[Integer, Long] = scala.collection.mutable.Map()

    def apply(address: Long, space: SpaceInfo, objType: AllocatedType, size: Integer, isArray: Boolean, arrayLength: Integer,
              allocationSite: AllocationSite, pointedFrom: Array[Long], pointsTo: Array[Long], eventType: EventType,
              classifiers: Seq[Classifier], n: Long = 1) = classify(address, space, objType, size, isArray, arrayLength, allocationSite, pointedFrom, pointsTo, eventType, classifiers, n = n)

    def ->(key: Any): Option[Grouping] = {
        children.get(key);
    }

    def +->(childData: (Classifier, Any)): Grouping = {
        if (!children.contains(childData._2)) {
            children += childData._2 -> new Grouping(this, childData._1, level + 1, childData._2)
        }
        return children(childData._2)
    }

    def classify(address: Long, space: SpaceInfo, objType: AllocatedType, size: Integer, isArray: Boolean, arrayLength: Integer,
                 allocationSite: AllocationSite, pointedFrom: Array[Long], pointsTo: Array[Long], eventType: EventType,
                 classifiers: Seq[Classifier], conditionalClassifiers: Seq[ConditionedClassifierList] = Seq(), n: Long = 1): Grouping = {
        val classifiersToUse: Seq[Classifier] = conditionalClassifiers.find(_.cond() == keys).getOrElse(classifiers);

        if (classifiersToUse.isEmpty) {
            // Recursion hook: No more classifiers to process
            leafs.get(size) match {
                case Some(prevCount) => leafs += size -> (prevCount + n)
                case None => leafs += size -> n
            }
            return this
        }

        // Invariant: At least one classifier is left to process
        val classifierToUse = classifiersToUse.head;
        val remainingClassifiers = classifiersToUse.tail;

        // Local methods
        def processNextClassifierIn(grouping: Grouping): Grouping = {
            grouping.classify(address, space, objType, size, isArray, arrayLength, allocationSite, pointedFrom, pointsTo, eventType, remainingClassifiers, conditionalClassifiers, n)
        }

        classifierToUse match {
            case multi: Multi[_] => multi.classify(address, space, objType, size, isArray, arrayLength, allocationSite, pointedFrom, pointsTo, eventType).toIterator.foreach(key => processNextClassifierIn(this +-> (multi, key)))
            case hierarchy: Hierarchy[_] =>
                var insertionGrouping = this
                for (key <- hierarchy.classify(address, space, objType, size, isArray, arrayLength, allocationSite, pointedFrom, pointsTo, eventType)) {
                    insertionGrouping = insertionGrouping +-> (hierarchy, key)
                    val condClassifier = conditionalClassifiers
                        .find(_.cond == keys)
                        .map(insertionGrouping.classify(address, space, objType, size, isArray, arrayLength, allocationSite, pointedFrom, pointsTo, eventType, _, conditionalClassifiers, n))
                }
                processNextClassifierIn(insertionGrouping)
            case single => processNextClassifierIn(this +-> (single, single(address, space, objType, size, isArray, arrayLength, allocationSite, pointedFrom, pointsTo, eventType)))
        }
        this
    }

    def summedInfos: Map[Integer, Long] = {
        val allLeafs: mutable.Map[Integer, Long] = leafs.clone()
        children.values.foreach(
            _.summedInfos map tupled { (key, value) => allLeafs += key -> (value + allLeafs.getOrElse[Long](key, 0)) })

        // Used to return immutable data structure
        Map() ++ allLeafs
    }

    def summedAmount: Long = summedInfos.values sum

    def summedBytes: Long = summedInfos map { case (key, value) => key * value } sum // Streaming at its finest

    override def toString: String = {
        val sb: StringBuilder = new StringBuilder
        sb.append("--" * level) // Wow, String multiplication! :D
        if (level > 0) sb.append(" ")
        sb.append(key)
        sb.append(" -> ")
        sb.append(summedAmount)
        sb.append(" (")
        sb.append(summedBytes)
        sb.append(")")
        for (child <- children.values) {
            sb.append("\n")
            sb.append(child.toString)
        }
        sb.toString
    }
}
