package at.jku.mevss.trace.gui.chart.base;

import at.jku.mevss.anttracks.util.Tuple;
import at.jku.mevss.trace.gui.chart.jfreechart.xy.util.AddALotXYSeries;
import javafx.collections.ObservableList;
import javafx.scene.chart.XYChart;
import org.jfree.data.xy.AbstractIntervalXYDataset;
import org.jfree.data.xy.DefaultTableXYDataset;
import org.jfree.data.xy.XYDataItem;
import org.jfree.data.xy.XYSeriesCollection;

import java.util.ArrayList;
import java.util.List;

public interface AxisScaler<DS> {
    public static final String[] OBJECT_UNITS = new String[]{"", "Thousands", "Million", "Billion"};
    public static final long OBJECT_DIVISOR = 1000;
    public static final String[] MEMORY_UNITS = new String[]{"bytes", "Kb", "Mb", "Gb", "Tb", "Pb"};
    public static final long MEMORY_DIVISOR = 1024;
    public static final String[] TIME_UNITS = new String[]{"ms", "sec", "min", "hour", "day", "week"};
    public static final long[] TIME_DIVISORS = new long[]{1000, 60, 60, 24, 7, Integer.MAX_VALUE};

    public static class NoAxisScaler<DS> implements AxisScaler<DS> {
        public long getXUnitDivisor(int level) {
            return 1;
        }

        public long getYUnitDivisor(int level) {
            return 1;
        }

        public String getXUnit(int level) {
            return null;
        }

        public String getYUnit(int level) {
            return null;
        }

        public int maxXUnitDownscalLevel() {
            return 0;
        }

        public int maxYUnitDownscaleLevel() {
            return 0;
        }

        @Override
        public void updateXAxisLabel(String label) {
            // Do not change default label
        }

        @Override
        public void updateYAxisLabel(String label) {
            // Do not change default label
        }

        @Override
        public Tuple<Integer, DS> processDataSetByDownscalingXUnits(DS dataset) {
            return new Tuple<>(0, dataset);
        }

        @Override
        public Tuple<Integer, DS> processDataSetByDownscalingYUnits(DS dataset) {
            return new Tuple<>(0, dataset);
        }
    }

    public abstract static class JFreeChartAxisScaler implements AxisScaler<AbstractIntervalXYDataset> {
        public static class AliveDeadScaler extends JFreeChartAxisScaler {
            ApplicationChartFactory.AliveDeadMemoryConsumptionUnit unit;

            public AliveDeadScaler(AntChartPane<AbstractIntervalXYDataset, ?> chart, ApplicationChartFactory.AliveDeadMemoryConsumptionUnit unit) {
                super(chart);
                this.unit = unit;
            }

            @Override
            public int maxXUnitDownscalLevel() {
                return TIME_UNITS.length;
            }

            @Override
            public long getXUnitDivisor(int level) {
                return TIME_DIVISORS[level];
            }

            @Override
            public String getXUnit(int level) {
                return "Time [" + TIME_UNITS[level] + "]";
            }

            @Override
            public int maxYUnitDownscaleLevel() {
                switch (unit) {
                    case RELATIVE:
                        return 0;
                    case OBJECTS:
                        return OBJECT_UNITS.length;
                    case BYTES:
                        return MEMORY_UNITS.length;
                    default:
                        assert false;
                        return 0;
                }
            }

            @Override
            public long getYUnitDivisor(int level) {
                switch (unit) {
                    case RELATIVE:
                        return 1;
                    case OBJECTS:
                        return OBJECT_DIVISOR;
                    case BYTES:
                        return MEMORY_DIVISOR;
                    default:
                        assert false;
                        return 0;
                }
            }

            @Override
            public String getYUnit(int level) {
                switch (unit) {
                    case RELATIVE:
                        return "Objects [%]";
                    case OBJECTS:
                        return level == 0 ? "Objects" : ("Objects [" + OBJECT_UNITS[level] + "]");
                    case BYTES:
                        return "Memory [" + MEMORY_UNITS[level] + "]";
                    default:
                        assert false;
                        return null;
                }
            }
        }

        public static class GCScaler extends JFreeChartAxisScaler {
            public GCScaler(AntChartPane<AbstractIntervalXYDataset, ?> chart) {
                super(chart);
            }

            @Override
            public long getXUnitDivisor(int level) {
                return TIME_DIVISORS[level];
            }

            @Override
            public long getYUnitDivisor(int level) {
                return TIME_DIVISORS[level];
            }

            @Override
            public String getXUnit(int level) {
                return "Time [" + TIME_UNITS[level] + "]";
            }

            @Override
            public String getYUnit(int level) {
                return "Pause [" + TIME_UNITS[level] + "]";
            }

            @Override
            public int maxXUnitDownscalLevel() {
                return TIME_UNITS.length;
            }

            @Override
            public int maxYUnitDownscaleLevel() {
                return TIME_UNITS.length;
            }
        }

        public static class ObjectsScaler extends JFreeChartAxisScaler {
            public ObjectsScaler(AntChartPane<AbstractIntervalXYDataset, ?> chart) {
                super(chart);
            }

            @Override
            public int maxXUnitDownscalLevel() {
                return TIME_UNITS.length;
            }

            @Override
            public long getXUnitDivisor(int level) {
                return TIME_DIVISORS[level];
            }

            @Override
            public String getXUnit(int level) {
                return "Time [" + TIME_UNITS[level] + "]";
            }

            @Override
            public String getYUnit(int level) {
                return level == 0 ? "Objects" : ("Objects [" + OBJECT_UNITS[level] + "]");
            }

            @Override
            public int maxYUnitDownscaleLevel() {
                return OBJECT_UNITS.length;
            }

            @Override
            public long getYUnitDivisor(int level) {
                return OBJECT_DIVISOR;
            }
        }

        public static class BytesScaler extends JFreeChartAxisScaler {
            public BytesScaler(AntChartPane<AbstractIntervalXYDataset, ?> chart) {
                super(chart);
            }

            @Override
            public int maxYUnitDownscaleLevel() {
                return MEMORY_UNITS.length;
            }

            @Override
            public long getYUnitDivisor(int level) {
                return MEMORY_DIVISOR;
            }

            @Override
            public String getYUnit(int level) {
                return "Memory [" + MEMORY_UNITS[level] + "]";
            }

            @Override
            public int maxXUnitDownscalLevel() {
                return TIME_UNITS.length;
            }

            @Override
            public long getXUnitDivisor(int level) {
                return TIME_DIVISORS[level];
            }

            @Override
            public String getXUnit(int level) {
                return "Time [" + TIME_UNITS[level] + "]";
            }

        }

        public static class NoAxisScaler extends JFreeChartAxisScaler {

            public NoAxisScaler(AntChartPane<AbstractIntervalXYDataset, ?> chart) {
                super(chart);
            }

            public long getXUnitDivisor(int level) {
                return 1;
            }

            public long getYUnitDivisor(int level) {
                return 1;
            }

            public String getXUnit(int level) {
                return null;
            }

            public String getYUnit(int level) {
                return null;
            }

            public int maxXUnitDownscalLevel() {
                return 0;
            }

            public int maxYUnitDownscaleLevel() {
                return 0;
            }

            public void updateXAxisLabel(String label) {
                // Do not change default label
            }

            public void updateYAxisLabel(String label) {
                // Do not change default label
            }
        }

        protected AntChartPane<AbstractIntervalXYDataset, ?> chart;

        public JFreeChartAxisScaler(AntChartPane<AbstractIntervalXYDataset, ?> chart) {
            this.chart = chart;
        }

        @Override
        public void updateXAxisLabel(String label) {
            chart.setXLabel(label);
        }

        @Override
        public void updateYAxisLabel(String label) {
            chart.setYLabel(label);
        }

        @Override
        public Tuple<Integer, AbstractIntervalXYDataset> processDataSetByDownscalingXUnits(AbstractIntervalXYDataset dataset) {
        /*
        // downscaling x units does not work, because for clicks we would have to revert to original, and double arithmetic is a bitch
		// we could use a custom number impl to also save the original
		double maxX = 0;
		for(Series<Number, Number> series : allSeries) {
			for(XYChart.Data<Number, Number> point : series.getData()) {
				if(maxX < point.getXValue().doubleValue()) {
					maxX = point.getXValue().doubleValue();
				}
			}
		}
		int n = 0;
		double totalDivisor = 1;
		long divisor;
		for(int i = 0; i < maxXUnitDownscalLevel() && maxX > (divisor = getXUnitDivisor(n)); i++) {
			maxX = maxX / divisor;
			totalDivisor = totalDivisor * divisor;
			n++;
		}
		for(Series<Number, Number> series : allSeries) {
			for(XYChart.Data<Number, Number> point : series.getData()) {
				point.setXValue(point.getXValue().doubleValue() / totalDivisor);
			}
		}
		return n;
		*/
            return new Tuple<>(0, dataset);
        }

        @Override
        public Tuple<Integer, AbstractIntervalXYDataset> processDataSetByDownscalingYUnits(AbstractIntervalXYDataset dataset) {
            double maxY = chart.getMaxYValue(dataset);

            int currentDivisionLevel = 0;
            double totalDivisor = 1.0;
            long divisor;

            for (currentDivisionLevel = 0; currentDivisionLevel < maxYUnitDownscaleLevel() && maxY > (divisor = getYUnitDivisor(currentDivisionLevel));
                 currentDivisionLevel++) {
                maxY = maxY / divisor;
                totalDivisor = totalDivisor * divisor;
            }

            XYSeriesCollection seriesCollection = null;
            XYSeriesCollection newSeriesCollection = null;
            DefaultTableXYDataset table = null;
            DefaultTableXYDataset newTable = null;

            if (dataset instanceof XYSeriesCollection) {
                seriesCollection = (XYSeriesCollection) dataset;
                newSeriesCollection = new XYSeriesCollection();
            } else if (dataset instanceof DefaultTableXYDataset) {
                table = (DefaultTableXYDataset) dataset;
                newTable = new DefaultTableXYDataset(false);
            } else {
                //throw new Exception("Unsupported dataset type");
            }

            for (int s = 0; s < dataset.getSeriesCount(); s++) {
                List<XYDataItem> newData = new ArrayList<>();
                for (int i = 0; i < dataset.getItemCount(s); i++) {
                    if (seriesCollection != null) {
                        double newY = seriesCollection.getSeries(s).getY(i).doubleValue() / totalDivisor;
                        newData.add(new XYDataItem(seriesCollection.getSeries(s).getX(i).doubleValue(), newY));
                    } else if (table != null) {
                        double newY = table.getSeries(s).getY(i).doubleValue() / totalDivisor;
                        newData.add(new XYDataItem(table.getSeries(s).getX(i).doubleValue(), newY));
                    } else {
                        //throw new Exception("Unsupported dataset type");
                    }
                }
                if (seriesCollection != null) {
                    AddALotXYSeries newSeries = new AddALotXYSeries(dataset.getSeriesKey(s), true);
                    newSeries.setData(newData, false);
                    newSeriesCollection.addSeries(newSeries);
                } else if (table != null) {
                    AddALotXYSeries newSeries = new AddALotXYSeries(dataset.getSeriesKey(s), false);
                    newSeries.setData(newData, false);
                    newTable.addSeries(newSeries);
                } else {
                    //throw new Exception("Unsupported dataset type");
                }
            }

            return new Tuple<>(currentDivisionLevel, (newSeriesCollection != null ? newSeriesCollection : (newTable != null ? newTable : null)));
        }
    }

    public abstract static class JavaFXAxisScaler implements AxisScaler<ObservableList<XYChart.Series<Number, Number>>> {
        public static class AliveDeadScaler extends JavaFXAxisScaler {
            ApplicationChartFactory.AliveDeadMemoryConsumptionUnit unit;

            public AliveDeadScaler(AntChartPane<ObservableList<XYChart.Series<Number, Number>>, ?> chart, ApplicationChartFactory.AliveDeadMemoryConsumptionUnit unit) {
                super(chart);
                this.unit = unit;
            }

            @Override
            public int maxXUnitDownscalLevel() {
                return TIME_UNITS.length;
            }

            @Override
            public long getXUnitDivisor(int level) {
                return TIME_DIVISORS[level];
            }

            @Override
            public String getXUnit(int level) {
                return "Time [" + TIME_UNITS[level] + "]";
            }

            @Override
            public int maxYUnitDownscaleLevel() {
                switch (unit) {
                    case RELATIVE:
                        return 0;
                    case OBJECTS:
                        return OBJECT_UNITS.length;
                    case BYTES:
                        return MEMORY_UNITS.length;
                    default:
                        assert false;
                        return 0;
                }
            }

            @Override
            public long getYUnitDivisor(int level) {
                switch (unit) {
                    case RELATIVE:
                        return 1;
                    case OBJECTS:
                        return OBJECT_DIVISOR;
                    case BYTES:
                        return MEMORY_DIVISOR;
                    default:
                        assert false;
                        return 0;
                }
            }

            @Override
            public String getYUnit(int level) {
                switch (unit) {
                    case RELATIVE:
                        return "Objects [%]";
                    case OBJECTS:
                        return level == 0 ? "Objects" : ("Objects [" + OBJECT_UNITS[level] + "]");
                    case BYTES:
                        return "Memory [" + MEMORY_UNITS[level] + "]";
                    default:
                        assert false;
                        return null;
                }
            }
        }

        public static class GCScaler extends JavaFXAxisScaler {
            public GCScaler(AntChartPane<ObservableList<XYChart.Series<Number, Number>>, ?> chart) {
                super(chart);
            }

            @Override
            public long getXUnitDivisor(int level) {
                return TIME_DIVISORS[level];
            }

            @Override
            public long getYUnitDivisor(int level) {
                return TIME_DIVISORS[level];
            }

            @Override
            public String getXUnit(int level) {
                return "Time [" + TIME_UNITS[level] + "]";
            }

            @Override
            public String getYUnit(int level) {
                return "Pause [" + TIME_UNITS[level] + "]";
            }

            @Override
            public int maxXUnitDownscalLevel() {
                return TIME_UNITS.length;
            }

            @Override
            public int maxYUnitDownscaleLevel() {
                return TIME_UNITS.length;
            }
        }

        public static class ObjectsScaler extends JavaFXAxisScaler {
            public ObjectsScaler(AntChartPane<ObservableList<XYChart.Series<Number, Number>>, ?> chart) {
                super(chart);
            }

            @Override
            public int maxXUnitDownscalLevel() {
                return TIME_UNITS.length;
            }

            @Override
            public long getXUnitDivisor(int level) {
                return TIME_DIVISORS[level];
            }

            @Override
            public String getXUnit(int level) {
                return "Time [" + TIME_UNITS[level] + "]";
            }

            @Override
            public String getYUnit(int level) {
                return level == 0 ? "Objects" : ("Objects [" + OBJECT_UNITS[level] + "]");
            }

            @Override
            public int maxYUnitDownscaleLevel() {
                return OBJECT_UNITS.length;
            }

            @Override
            public long getYUnitDivisor(int level) {
                return OBJECT_DIVISOR;
            }
        }

        public static class BytesScaler extends JavaFXAxisScaler {
            public BytesScaler(AntChartPane<ObservableList<XYChart.Series<Number, Number>>, ?> chart) {
                super(chart);
            }

            @Override
            public int maxYUnitDownscaleLevel() {
                return MEMORY_UNITS.length;
            }

            @Override
            public long getYUnitDivisor(int level) {
                return MEMORY_DIVISOR;
            }

            @Override
            public String getYUnit(int level) {
                return "Memory [" + MEMORY_UNITS[level] + "]";
            }

            @Override
            public int maxXUnitDownscalLevel() {
                return TIME_UNITS.length;
            }

            @Override
            public long getXUnitDivisor(int level) {
                return TIME_DIVISORS[level];
            }

            @Override
            public String getXUnit(int level) {
                return "Time [" + TIME_UNITS[level] + "]";
            }

        }

        public static class NoAxisScaler extends JavaFXAxisScaler {

            public NoAxisScaler(AntChartPane<ObservableList<XYChart.Series<Number, Number>>, ?> chart) {
                super(chart);
            }

            public long getXUnitDivisor(int level) {
                return 1;
            }

            public long getYUnitDivisor(int level) {
                return 1;
            }

            public String getXUnit(int level) {
                return null;
            }

            public String getYUnit(int level) {
                return null;
            }

            public int maxXUnitDownscalLevel() {
                return 0;
            }

            public int maxYUnitDownscaleLevel() {
                return 0;
            }

            public void updateXAxisLabel(String label) {
                // Do not change default label
            }

            public void updateYAxisLabel(String label) {
                // Do not change default label
            }
        }

        protected AntChartPane<ObservableList<XYChart.Series<Number, Number>>, ?> chart;

        public JavaFXAxisScaler(AntChartPane<ObservableList<XYChart.Series<Number, Number>>, ?> chart) {
            this.chart = chart;
        }

        @Override
        public void updateXAxisLabel(String label) {
            chart.setXLabel(label);
        }

        @Override
        public void updateYAxisLabel(String label) {
            chart.setYLabel(label);
        }

        @Override
        public Tuple<Integer, ObservableList<XYChart.Series<Number, Number>>> processDataSetByDownscalingXUnits(ObservableList<XYChart.Series<Number, Number>> dataset) {
        /*
        // downscaling x units does not work, because for clicks we would have to revert to original, and double arithmetic is a bitch
		// we could use a custom number impl to also save the original
		double maxX = 0;
		for(Series<Number, Number> series : allSeries) {
			for(XYChart.Data<Number, Number> point : series.getData()) {
				if(maxX < point.getXValue().doubleValue()) {
					maxX = point.getXValue().doubleValue();
				}
			}
		}
		int n = 0;
		double totalDivisor = 1;
		long divisor;
		for(int i = 0; i < maxXUnitDownscalLevel() && maxX > (divisor = getXUnitDivisor(n)); i++) {
			maxX = maxX / divisor;
			totalDivisor = totalDivisor * divisor;
			n++;
		}
		for(Series<Number, Number> series : allSeries) {
			for(XYChart.Data<Number, Number> point : series.getData()) {
				point.setXValue(point.getXValue().doubleValue() / totalDivisor);
			}
		}
		return n;
		*/
            return new Tuple<>(0, dataset);
        }

        @Override
        public Tuple<Integer, ObservableList<XYChart.Series<Number, Number>>> processDataSetByDownscalingYUnits(ObservableList<XYChart.Series<Number, Number>> dataset) {
            double maxY = 0;
            for (XYChart.Series<Number, Number> series : dataset) {
                for (XYChart.Data<Number, Number> point : series.getData()) {
                    if (maxY < point.getXValue().doubleValue()) {
                        maxY = point.getXValue().doubleValue();
                    }
                }
            }

            int currentDivisionLevel = 0;
            double totalDivisor = 1;
            long divisor;

            for (currentDivisionLevel = 0; currentDivisionLevel < maxYUnitDownscaleLevel() && maxY > (divisor = getYUnitDivisor(currentDivisionLevel));
                 currentDivisionLevel++) {
                maxY = maxY / divisor;
                totalDivisor = totalDivisor * divisor;
                currentDivisionLevel++;
            }

            for (XYChart.Series<Number, Number> series : dataset) {
                for (XYChart.Data<Number, Number> point : series.getData()) {
                    point.setYValue(point.getYValue().doubleValue() / totalDivisor);
                }
            }

            return new Tuple<>(currentDivisionLevel, dataset);
        }
    }

    long getXUnitDivisor(int level);

    long getYUnitDivisor(int level);

    String getXUnit(int level);

    String getYUnit(int level);

    int maxXUnitDownscalLevel();

    int maxYUnitDownscaleLevel();

    void updateXAxisLabel(String label);

    void updateYAxisLabel(String label);

    Tuple<Integer, DS> processDataSetByDownscalingXUnits(DS dataset);

    Tuple<Integer, DS> processDataSetByDownscalingYUnits(DS dataset);

    default DS postProcessDataset(DS dataset) {
        Tuple<Integer, DS> x = processDataSetByDownscalingXUnits(dataset);
        updateXAxisLabel(getXUnit(x.a));
        Tuple<Integer, DS> y = processDataSetByDownscalingYUnits(x.b);
        updateYAxisLabel(getYUnit(y.a));
        return y.b;
    }
}
