/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.chart.fxchart.xy.stackedareapane;

import at.jku.mevss.trace.gui.chart.base.ApplicationChartFactory.MemoryConsumptionUnit;
import at.jku.mevss.trace.gui.chart.base.AxisScaler;
import at.jku.mevss.trace.gui.chart.fxchart.xy.stackedareapane.base.StackedAreaJavaFXChartPane;
import at.jku.mevss.trace.gui.frame.main.tab.heapdiffing.model.DiffingData;
import at.jku.mevss.trace.gui.model.AppInfo;
import at.jku.mevss.trace.gui.model.Size;
import at.jku.mevss.trace.gui.utils.FXMLUtil;
import javafx.collections.ObservableList;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.scene.chart.XYChart;
import javafx.scene.chart.XYChart.Series;

import java.util.Arrays;
import java.util.List;
import java.util.function.Function;

public class DiffingStackedAreaChartPane extends StackedAreaJavaFXChartPane<AppInfo> {
    Series<Number, Number> permSeries;
    Series<Number, Number> diedSeries;
    Series<Number, Number> bornMinSeries;
    Series<Number, Number> bornMaxSeries;
    Series<Number, Number> tempSeries;

    private MemoryConsumptionUnit unit;
    private Function<Size, Long> accessor;
    private List<DiffingData> data;

    public DiffingStackedAreaChartPane() {
        FXMLUtil.load(this, DiffingStackedAreaChartPane.class);
    }

    public void init(MemoryConsumptionUnit unit) {
        this.unit = unit;
        super.init(null);
        switch (unit) {
            case OBJECTS:
                accessor = size -> (long) size.objects;
                break;
            case BYTES:
                accessor = size -> (long) size.bytes;
                break;
        }
        super.chart.setLegendVisible(false);
    }

    public void setData(List<DiffingData> data) throws Exception {
        this.data = data;
        update(null);
    }

    // TODO: Do not always generate the complete data, but only the needed one!
    @Override
    protected ObservableList<XYChart.Series<Number, Number>> updateDataSet(AppInfo appInfo) {
        boolean ea = false;
        assert ea = true;

        double maxX = getMaxXValue(chart.getData());

        DiffingData[] data = this.data.stream().toArray(DiffingData[]::new);
        // if (data.length >= 2 && accessor.apply(data[0].perm) == 0) {
        // data[0] = new DiffingData(data[0].id, data[0].time);
        // data[0].add(DiffingDataInfo.PermObjects, data[1].perm.objects + data[1].died.objects);
        // data[0].add(DiffingDataInfo.PermBytes, data[1].perm.bytes + data[1].died.bytes);
        // }
        DiffingData last = data[data.length - 1];

        // perm is always falling (going to died)
        for (int i = 1; i < data.length; i++) {
            assert accessor.apply(data[i - 1].perm) >= accessor.apply(data[i].perm) : "Perm must not grow";
        }
        // died is always rising (coming from perm)
        for (int i = 1; i < data.length; i++) {
            assert accessor.apply(data[i - 1].died) <= accessor.apply(data[i].died) : "Died must not shrink";
        }
        // difference between perm and died must match
        for (int i = 1; i < data.length; i++) {
            long deltaDied = accessor.apply(data[i].died) - accessor.apply(data[i - 1].died);
            long deltaPerm = accessor.apply(data[i - 1].perm) - accessor.apply(data[i].perm);

            assert deltaDied == deltaPerm : String.format(
                    "Changes in perm and died must match. Perm Before: %d; Died Before: %d; Perm After: %d; Died After: %d; DeltaPerm: " +
                            "%d, DeltaDied: %d;",
                    accessor.apply(data[i - 1].perm),
                    accessor.apply(data[i - 1].died),
                    accessor.apply(data[i].perm),
                    accessor.apply(data[i].died),
                    deltaPerm,
                    deltaDied);
        }

        // born is rising or falling (coming from <allocations>, going to temp)
        // temp is always rising (coming from born)
        for (int i = 1; i < data.length; i++) {
            assert accessor.apply(data[i - 1].temp) <= accessor.apply(data[i].temp) : "Temp must not shrink";
        }

        long[] allocations = new long[data.length];
        long[] deallocations = new long[data.length];
        {
            long last_born = 0, last_temp = 0;
            for (int i = 0; i < allocations.length; i++) {
                DiffingData current = data[i];
                deallocations[i] = accessor.apply(current.temp) - last_temp;
                allocations[i] = (accessor.apply(current.born) + deallocations[i]) - last_born;
                last_born = accessor.apply(current.born);
                last_temp = accessor.apply(current.temp);
            }
        }

        long[] perms = new long[data.length];
        long[] dieds = new long[data.length];
        long[] borns = new long[data.length];
        long[] temps = new long[data.length];
        for (int i = 0; i < data.length; i++) {
            DiffingData d = data[i];
            dieds[i] = accessor.apply(last.died) - accessor.apply(d.died);
            perms[i] = accessor.apply(d.perm) - dieds[i];
            temps[i] = (i == 0) ? 0 : (temps[i - 1] + allocations[i] /* not all are really temp, some are born */ - deallocations[i]);
            borns[i] = accessor.apply(d.perm) + accessor.apply(d.born) - perms[i] - dieds[i] - temps[i]; /* borns will always be zero here */
        }

        /* fix current issure that we do not know when temp objects become born */
        {
            long born_leftovers = accessor.apply(last.born) - borns[borns.length - 1];
            long max = born_leftovers;
            for (int i = data.length - 1; i >= 0 && born_leftovers > 0; i--) {
                max = Math.min(max, temps[i]);
                long correction = Math.min(max, (i + 1 < data.length) ? Math.min(allocations[i + 1], born_leftovers) : 0);
                borns[i] = Math.min(max, (i + 1 < data.length) ? borns[i + 1] - correction : born_leftovers);
                temps[i] -= borns[i];
                born_leftovers -= (i + 1 < data.length ? (borns[i + 1] - borns[i]) : 0);
            }
        }

        if (ea) {
            assert dieds[dieds.length - 1] == 0;
            assert borns[0] == 0;
            assert temps[0] == 0 && temps[temps.length - 1] == 0;
            for (int i = 0; i < data.length; i++) {
                if (i > 0) {
                    assert perms[i - 1] == perms[i] : String.format("perm inconsistent: %d -> %d", perms[i - 1], perms[i]);
                    assert dieds[i - 1] >= dieds[i] : String.format("died inconsistent: %d -> %d", dieds[i - 1], dieds[i]);
                    assert borns[i - 1] <= borns[i] : String.format("born inconsistent: %d -> %d", borns[i - 1], borns[i]);
                    // dunno anything about temps
                }
                assert perms[i] >= 0 && dieds[i] >= 0 && borns[i] >= 0 && temps[i] >= 0 : String.format("negative values: %d, %d, %d, %d",
                                                                                                        perms[i],
                                                                                                        dieds[i],
                                                                                                        borns[i],
                                                                                                        temps[i]);
                assert accessor.apply(data[i].perm) + accessor.apply(data[i].born) == perms[i] + dieds[i] + borns[i] + temps[i] : String.format(
                        "live objects do not add up: %d + %d != %d + %d + %d + %d",
                        accessor.apply(data[i].perm),
                        accessor.apply(data[i].born),
                        perms[i],
                        dieds[i],
                        borns[i],
                        temps[i]);
            }
        }

        long[] borns_max = new long[data.length];
        {
            long max = accessor.apply(last.born);
            for (int i = data.length - 1; i >= 0; i--) {
                max = Math.min(max, temps[i] + borns[i]);
                borns_max[i] = max;
                assert borns_max[i] >= borns[i] : String.format("born max inconsistent: %d, %d, %d, %d", borns[i], borns_max[i], temps[i], max);
            }
            for (int i = 0; i < data.length; i++) {
                borns_max[i] -= borns[i];
                temps[i] = temps[i] - borns_max[i];
            }
        }

        if (ea) {
            assert dieds[dieds.length - 1] == 0;
            assert borns[0] == 0;
            assert borns_max[0] == 0;
            assert temps[0] == 0 && temps[temps.length - 1] == 0;
            for (int i = 0; i < data.length; i++) {
                if (i > 0) {
                    assert perms[i - 1] == perms[i] : String.format("perm inconsistent: %d -> %d", perms[i - 1], perms[i]);
                    assert dieds[i - 1] >= dieds[i] : String.format("died inconsistent: %d -> %d", dieds[i - 1], dieds[i]);
                    assert borns[i - 1] <= borns[i] : String.format("born inconsistent: %d -> %d", borns[i - 1], borns[i]);
                    assert borns[i - 1] + borns_max[i - 1] <= borns[i] + borns_max[i] : String.format("born (max) inconsistent: %d + %d -> %d + %d",
                                                                                                      borns[i - 1],
                                                                                                      borns_max[i - 1],
                                                                                                      borns[i],
                                                                                                      borns_max[i]);
                    // dunno anything about temps
                }
                assert perms[i] >= 0 && dieds[i] >= 0 && borns[i] >= 0 && borns_max[i] >= 0 && temps[i] >= 0;
                assert accessor.apply(data[i].perm) + accessor.apply(data[i].born) == perms[i] + dieds[i] + borns[i] + borns_max[i] + temps[i] : String.format(
                        "%d + %d != %d + %d + %d + %d + %d",
                        accessor.apply(data[i].perm),
                        accessor.apply(data[i].born),
                        perms[i],
                        dieds[i],
                        borns[i],
                        borns_max[i],
                        temps[i]);
            }
        }

        double[] times = Arrays.stream(data).mapToDouble(d -> (double) d.time).toArray();
        for (int i = 1; i < times.length; i++) {
            if (times[i - 1] == times[i]) {
                int start = i - 1;
                int end = i + 1;
                while (end < times.length && times[start] == times[end]) {
                    end++;
                }
                double step = 1.0 / (end - start);
                for (int j = 0; j < end - start; j++) {
                    times[start + j] += j * step;
                }
            } else {
                assert times[i - 1] < times[i];
            }
        }

        for (int i = 0; i < data.length; i++) {
            if (data[i].time > maxX) {
                permSeries.getData().add(new XYChart.Data<Number, Number>(times[i], perms[i]));
                diedSeries.getData().add(new XYChart.Data<Number, Number>(times[i], dieds[i]));
                bornMinSeries.getData().add(new XYChart.Data<Number, Number>(times[i], borns[i]));
                bornMaxSeries.getData().add(new XYChart.Data<Number, Number>(times[i], borns_max[i]));
                tempSeries.getData().add(new XYChart.Data<Number, Number>(times[i], temps[i]));
            }
        }

        return chart.getData();
    }

    @Override
    public AxisScaler<ObservableList<Series<Number, Number>>> getScaler() {
        switch (unit) {
            case OBJECTS:
                return new AxisScaler.JavaFXAxisScaler.ObjectsScaler(this);
            case BYTES:
                return new AxisScaler.JavaFXAxisScaler.BytesScaler(this);
            default:
                LOGGER.warning("No scaler defined for " + unit);
                return new AxisScaler.JavaFXAxisScaler.NoAxisScaler(this);
        }
    }

    @Override
    protected void initializeChartSeries() {
        permSeries = new Series<>();
        permSeries.setName("Perm");
        diedSeries = new Series<>();
        diedSeries.setName("Died");
        bornMinSeries = new Series<>();
        bornMinSeries.setName("Born (min)");
        bornMaxSeries = new Series<>();
        bornMaxSeries.setName("Born (max)");
        tempSeries = new Series<>();
        tempSeries.setName("Temp");
        chart.getData().add(permSeries);
        chart.getData().add(diedSeries);
        chart.getData().add(bornMinSeries);
        chart.getData().add(bornMaxSeries);
        chart.getData().add(tempSeries);
    }

    @Override
    protected void updateDataLabelContent() {

    }

    @Override
    protected EventHandler<ActionEvent> getOnConfigureAction() {
        return null;
    }

}
