/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.classification.classifier;

import at.jku.mevss.anttracks.classification.Classifier;
import at.jku.mevss.anttracks.classification.ClassifierChain;
import at.jku.mevss.anttracks.classification.ClassifierException;
import at.jku.mevss.anttracks.classification.Transformer;
import at.jku.mevss.anttracks.classification.annotations.ClassifierProperty;
import at.jku.mevss.anttracks.classification.enumerations.ClassifierSourceCollection;
import at.jku.mevss.anttracks.classification.enumerations.ClassifierType;
import at.jku.mevss.anttracks.classification.nodes.GroupingNode;
import at.jku.mevss.anttracks.heap.IndexBasedHeap;
import at.jku.mevss.anttracks.heap.symbols.Symbols;
import at.jku.mevss.anttracks.util.ImagePack;
import at.jku.mevss.trace.gui.utils.ImageUtil;

import java.util.function.Supplier;

@at.jku.mevss.anttracks.classification.annotations.Classifier(name = "Points To (Direct)", desc = "This transformer is used to classify the objects an object directly " +
        "points to" +
        ".", example = "(using \"Points to Type\") Points To -> char[] & Integer", type = ClassifierType.ONE, collection = ClassifierSourceCollection.FASTHEAP)
public class PointsToTransformer extends Transformer {

    @ClassifierProperty(overviewLevel = 10)
    protected ClassifierChain classifiers = new ClassifierChain();

    public ClassifierChain getClassifiers() {
        return classifiers;
    }

    public void setClassifiers(ClassifierChain classifiers) {
        this.classifiers = classifiers;
        if (classifiers != null && classifiers.length() > 0) {
            classifiers.getList().forEach(classifier -> classifier.setup(() -> symbols(), () -> fastHeap()));
        }
    }

    private static String NULL_KEY = "null";
    private static final Classifier<String> NULL_CLASSIFIER = new Classifier<String>() {
        @Override
        public String classify() throws Exception {
            return NULL_KEY;
        }
    };
    private static final ClassifierChain NULL_CLASSIFIER_CHAIN = new ClassifierChain(NULL_CLASSIFIER);

    @ClassifierProperty(overviewLevel = 10)
    private boolean showNulls = true;

    public boolean getShowNulls() {
        return showNulls;
    }

    public void setShowNulls(boolean showNulls) {
        this.showNulls = showNulls;
    }

    @Override
    public void setup(Supplier<Symbols> symbolsSupplier, Supplier<IndexBasedHeap> fastHeapSupplier) {
        super.setup(symbolsSupplier, fastHeapSupplier);
        if (classifiers != null && classifiers.length() > 0) {
            classifiers.getList().forEach(classifier -> classifier.setup(symbolsSupplier, fastHeapSupplier));
        }
    }

    @Override
    public ImagePack[] loadIcons() {
        return new ImagePack[]{ImageUtil.getResourceImagePack("Points to", "points_to.png")};
    }

    @Override
    public String title() {
        return Classifier.CLASSIFICATION_MODE == ClassificationMode.LIST ? "Points To (# Objects)" : "Pointers To (# Pointers)";
    }

    @Override
    public GroupingNode classify(GroupingNode transformerRoot) throws ClassifierException {
        int[] pointsTo = pointsToIndices();
        if (pointsTo != null && pointsTo.length > 0) {
            for (int i = 0; i < pointsTo.length; i++) {
                int ptr = pointsTo[i];
                if (ptr < 0) {
                    // Null pointer
                    if (showNulls) {
                        ((GroupingNode.FastHeapGroupingNode) transformerRoot).classify(fastHeap(), ptr, NULL_CLASSIFIER_CHAIN, null);
                        GroupingNode nullNode = transformerRoot.getChild(NULL_KEY);
                        if (nullNode != null) {
                            nullNode.getData().clear();
                        }
                    }
                } else {
                    // Normal pointers
                    ((GroupingNode.FastHeapGroupingNode) transformerRoot).classify(fastHeap(), ptr, classifiers, null);
                }
            }
        }
        return transformerRoot;
    }
}
