/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.classification.classifier.component.selectionpane;

import at.jku.mevss.anttracks.classification.Classifier;
import at.jku.mevss.anttracks.classification.ClassifierFactory;
import at.jku.mevss.anttracks.util.ImagePack;
import at.jku.mevss.trace.gui.classification.classifier.dialog.create.CreateClassifierDialog;
import at.jku.mevss.trace.gui.classification.classifier.dialog.edit.EditClassifierDialog;
import at.jku.mevss.trace.gui.classification.component.selectionpane.ClassificationSelectionPane;
import at.jku.mevss.trace.gui.classification.dialog.properties.ClassificationPropertiesDialog;
import at.jku.mevss.trace.gui.component.clickablepane.ClickablePane;
import at.jku.mevss.trace.gui.model.AvailableClassifierInfo;
import at.jku.mevss.trace.gui.utils.FXMLUtil;
import javafx.fxml.FXML;
import javafx.scene.control.Dialog;
import javafx.scene.layout.HBox;
import javafx.scene.layout.Pane;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public final class ObjectClassifierSelectionPane extends ClassificationSelectionPane<Classifier<?>, ClassifierFactory> {

    @FXML
    private HBox availableSingleCardinalitySubPane;

    @FXML
    private HBox availableMultiCardinalitySubPane;

    @FXML
    private HBox availableTreeCardinalitySubPane;

    @FXML
    private HBox availableMultiTreeCardinalitySubPane;

    private List<Pane> availableSingleCardinalityPaneList;
    private List<Pane> availableMultiCardinalityPaneList;
    private List<Pane> availableTreeCardinalityPaneList;
    private List<Pane> availableMultiTreeCardinalityPaneList;

    public ObjectClassifierSelectionPane() {
        super();
        FXMLUtil.load(this, ObjectClassifierSelectionPane.class);
    }

    @Override
    public void init(AvailableClassifierInfo availableClassifierInfo, int maximumSelect, boolean prohibitDuplicates) {
        availableSingleCardinalityPaneList = new ArrayList<>();
        availableMultiCardinalityPaneList = new ArrayList<>();
        availableTreeCardinalityPaneList = new ArrayList<>();
        availableMultiTreeCardinalityPaneList = new ArrayList<>();
        super.init(availableClassifierInfo, maximumSelect, prohibitDuplicates);
    }

    // SELECTED

    @Override
    protected ImagePack getIcon(Classifier<?> classifier) {
        return classifier.getIcon(null);
    }

    @Override
    protected String getName(Classifier<?> classifier) {
        return classifier.getName();
    }

    @Override
    protected String[] getSelectedDescriptions(Classifier<?> classifier) {
        return Stream.of(classifier.configurableProperties())
                     .filter(prop -> prop.getPresentationName() != null && prop.get() != null)
                     .map(prop -> prop.getPresentationName() + ": " + prop.get())
                     .toArray(String[]::new);
    }

    @Override
    protected String getSelectedToolTip(Classifier<?> classifier) {
        return (classifier.getDesc() + "\n" + (classifier.configurableProperties().length > 0 ? "Right click to configure" : ""));
    }

    @Override
    protected boolean removeSelected(ClassifierFactory y) {
        List<Classifier<?>> remainingSelected = getSelected().stream().filter(selected -> !selected.getName().equals(y.getName())).collect(Collectors.toList());
        if (getSelected().size() != remainingSelected.size()) {
            resetSelected(remainingSelected);
            return true;
        }
        return false;
    }

    @Override
    public boolean selectedRightClick(Classifier<?> classifier) {
        if (ClassificationPropertiesDialog.showDialogForClassifier(classifier, availableClassifierInfo, this)) {
            refreshSelected();
            return true;
        }

        return false;
    }

    // AVAILABLE

    @Override
    protected ImagePack getAvailableIcon(ClassifierFactory fac) {
        return fac.getIcon();
    }

    @Override
    protected String getAvailableName(ClassifierFactory fac) {
        return fac.getName();
    }

    @Override
    protected String[] getAvailableDescriptions(ClassifierFactory fac) {
        return new String[]{"e.g. \"" + fac.getExample() + "\""};
    }

    @Override
    protected String getAvailableToolTip(ClassifierFactory fac) {
        return fac.getDesc() + (fac instanceof ClassifierFactory && fac.isOnTheFlyCompilable() ? "\nRight click to edit" : "");
    }

    @Override
    protected void addAvailablePane(ClassifierFactory fac, ClickablePane pane) {
        switch (fac.getType()) {
            case ONE:
                availableSingleCardinalityPaneList.add(pane);
                availableSingleCardinalitySubPane.getChildren().add(pane);
                break;
            case MANY:
                availableMultiCardinalityPaneList.add(pane);
                availableMultiCardinalitySubPane.getChildren().add(pane);
                break;
            case HIERARCHY:
                availableTreeCardinalityPaneList.add(pane);
                availableTreeCardinalitySubPane.getChildren().add(pane);
                break;
            case MANY_HIERARCHY:
                availableMultiTreeCardinalityPaneList.add(pane);
                availableMultiTreeCardinalitySubPane.getChildren().add(pane);
                break;
        }
    }

    @Override
    protected boolean availableRightClickAvailable(ClassifierFactory fac) {
        if (fac == null) {
            return false;
        }
        if (!(fac instanceof ClassifierFactory)) {
            return false;
        }
        ClassifierFactory f = fac;
        if (!f.isOnTheFlyCompilable()) {
            return false;
        }
        return true;
    }

    @Override
    public List<ClassifierFactory> getAvailableList() {
        return availableClassifierInfo.getAvailableClassifier().stream()
                                      .filter(fac -> !isExcludedFromAvailable(fac.create()))
                                      .collect(Collectors.toList());
    }

    @Override
    protected CreateClassifierDialog getCreateDialog() throws IOException {
        return new CreateClassifierDialog(availableClassifierInfo.getSymbolsSupplier(),
                                          availableClassifierInfo.getFastHeapSupplier(),
                                          availableClassifierInfo.getAvailableClassifier());
    }

    @Override
    protected Dialog<? extends ClassifierFactory> getEditDialog(ClassifierFactory fac) throws IOException {
        if (fac instanceof ClassifierFactory) {
            return new EditClassifierDialog(availableClassifierInfo.getSymbolsSupplier(),
                                            availableClassifierInfo.getFastHeapSupplier(),
                                            fac,
                                            availableClassifierInfo.getAvailableClassifier());
        } else {
            return null;
        }
    }

    @Override
    protected Comparator<ClassifierFactory> comperator() {
        return Comparator.comparing(ClassifierFactory::getName);
    }

    @Override
    protected void clearAvailablePanes() {
        availableSingleCardinalitySubPane.getChildren().removeAll(availableSingleCardinalityPaneList);
        availableSingleCardinalityPaneList.clear();
        availableMultiCardinalitySubPane.getChildren().removeAll(availableMultiCardinalityPaneList);
        availableMultiCardinalityPaneList.clear();
        availableTreeCardinalitySubPane.getChildren().removeAll(availableTreeCardinalityPaneList);
        availableTreeCardinalityPaneList.clear();
        availableMultiTreeCardinalitySubPane.getChildren().removeAll(availableMultiTreeCardinalityPaneList);
        availableMultiTreeCardinalityPaneList.clear();
    }
}