/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.classification.dialog.properties;

import at.jku.mevss.anttracks.classification.Classifier;
import at.jku.mevss.anttracks.classification.ClassifierProperty;
import at.jku.mevss.trace.gui.model.AvailableClassifierInfo;
import at.jku.mevss.trace.gui.utils.NodeUtil;
import at.jku.mevss.trace.gui.utils.NodeUtil.AccessibleNode;
import at.jku.mevss.trace.gui.utils.WindowUtil;
import javafx.collections.ObservableList;
import javafx.geometry.Bounds;
import javafx.geometry.Rectangle2D;
import javafx.scene.Node;
import javafx.scene.control.*;
import javafx.scene.layout.HBox;
import javafx.scene.layout.Pane;
import javafx.scene.layout.VBox;
import javafx.stage.Screen;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.function.Consumer;
import java.util.function.Supplier;
import java.util.stream.Collectors;

public class ClassificationPropertiesDialog extends Dialog<Boolean> {

    private class ClassifierPropertyInfos<T> extends ClassifierProperty<T> {
        private Supplier<T> uiGetter = null;
        private Consumer<T> uiSetter = null;
        private boolean fullyInitialized = true;

        @SuppressWarnings("unchecked")
        public ClassifierPropertyInfos(ClassifierProperty<?> property, Supplier<?> uiGetter, Consumer<?> uiSetter) {
            this(((ClassifierProperty<T>) property).getClassifier(), ((ClassifierProperty<T>) property).getField(), (Supplier<T>) uiGetter, (Consumer<T>) uiSetter);
        }

        @SuppressWarnings("unchecked")
        public ClassifierPropertyInfos(Classifier<?> classifier, Field field, Supplier<T> uiGetter, Consumer<T> uiSetter) {
            super((Classifier<T>) classifier, field);
            this.uiGetter = uiGetter;
            if (uiGetter == null) {
                fullyInitialized = false;
            }
            this.uiSetter = uiSetter;
            if (uiSetter == null) {
                fullyInitialized = false;
            }

            classifierToUi();
        }

        /**
         * Copies the data that is currently visible on the UI into the underlying data
         *
         * @return True if the underlying data changed (i.e., if something was changed on the UI), otherwise false
         */
        public boolean uiToClassifier() {
            if (fullyInitialized) {
                T propertyValue = propertyGetter.get();
                T uiValue = uiGetter.get();
                if ((propertyValue != null && uiValue != null && !propertyValue.equals(uiValue)) ||
                        (propertyValue == null && uiValue != null) ||
                        (propertyValue != null && uiValue == null)) {
                    propertySetter.accept(uiValue);
                    return true;
                }
            }
            return false;
        }

        public void classifierToUi() {
            if (fullyInitialized) {
                uiSetter.accept(propertyGetter.get());
            }
        }
    }

    private final List<ClassifierPropertyInfos<?>> properties = new ArrayList<>();

    public ClassificationPropertiesDialog(Classifier<?> classifier, AvailableClassifierInfo availableClassifierInfo, Node creator) {
        setTitle("Configure properties of classifier \"" + classifier.getName() + "\"");
        ClassificationPropertiesDialogPane dialogPane = new ClassificationPropertiesDialogPane();
        setDialogPane(dialogPane);


        /*
         * Return false if any other button than "apply" has been pressed.
         * Only return true if any of the properties has been changed.
         */
        setResultConverter(result -> {
            if (result == ButtonType.APPLY) {
                return closeWithAccept();
            } else {
                return false;
            }
        });

        List<Label> labels = new ArrayList<>();

        if (classifier.configurableProperties().length == 0) {
            Label headerLabel = new Label("No configurable properties in classifier" + classifier.getName());
            dialogPane.mainPane.getChildren().add(headerLabel);
        } else {
            for (ClassifierProperty<?> property : classifier.configurableProperties()) {
                ClassifierPropertyInfos<?> propertyInfos = null;

                AccessibleNode<?> node = NodeUtil.datatypeToNode(property.getField().getType(), () -> {
                    checkAcceptButtonEnabling();
                }, availableClassifierInfo);
                propertyInfos = new ClassifierPropertyInfos<>(property, node.getter, node.setter);

                if (propertyInfos == null || !propertyInfos.fullyInitialized) {
                    continue;
                }
                properties.add(propertyInfos);

                Pane fieldPane;
                if (node.node instanceof CheckBox) {
                    fieldPane = new HBox();
                    Label textLabel = new Label(propertyInfos.getPresentationName());
                    fieldPane.getChildren().add(node.node);
                    fieldPane.getChildren().add(textLabel);
                } else {
                    fieldPane = new VBox();
                    Label textLabel = new Label(propertyInfos.getPresentationName() + ":");
                    labels.add(textLabel);
                    fieldPane.getChildren().add(textLabel);
                    fieldPane.getChildren().add(node.node);
                }
                dialogPane.mainPane.getChildren().add(fieldPane);
            }
        }

        // position dialog to (always) appear on same screen as main window (and make it fit there)
        Bounds creatorScreenCoords = creator.localToScreen(creator.getBoundsInLocal());
        ObservableList<Screen> screenList = Screen.getScreensForRectangle(creatorScreenCoords.getMinX(), creatorScreenCoords.getMinY(), 1, 1);
        if (screenList.size() >= 1) {
            Rectangle2D currentScreen = screenList.get(0).getBounds();
            dialogPane.setMaxWidth(currentScreen.getWidth() * 0.80);
            dialogPane.setMaxHeight(currentScreen.getHeight() * 0.80);
        }
    }

    private void checkAcceptButtonEnabling() {
        boolean everythingSet = properties.stream()
                                          .filter(property -> property.fullyInitialized)
                                          .allMatch(property -> property.uiGetter != null && property.uiGetter.get() != null);
        getApplyButton().setDisable(!everythingSet);
    }

    private Button getApplyButton() {
        Node node = getDialogPane().lookupButton(ButtonType.APPLY);
        if (node instanceof Button) {
            return (Button) node;
        } else {
            throw new IllegalStateException("Apply Button not found - Dialog may not be initialized correctly");
        }
    }

    /**
     * Called when the dialog is closed with "Accept".
     * Copies the current data from the UI into the underlying data.
     *
     * @return True if any of the underlying data changed (i.e., if some property has been changed), otherwise false.
     */
    private boolean closeWithAccept() {
        boolean anythingChanged = false;
        for (ClassifierPropertyInfos<?> property : properties.stream().filter(property -> property.fullyInitialized).collect(Collectors.toList())) {
            anythingChanged |= property.uiToClassifier();
        }
        return anythingChanged;
    }

    public static boolean showDialogForClassifier(Classifier<?> classifier, AvailableClassifierInfo availableClassifierInfo, Node dialogCreator) {
        if (classifier == null) {
            return false;
        }
        if (classifier.configurableProperties().length == 0) {
            return false;
        }

        ClassificationPropertiesDialog dialog = new ClassificationPropertiesDialog(classifier, availableClassifierInfo, dialogCreator);
        WindowUtil.centerInMainFrame(dialog);
        Optional<Boolean> dialogResult = dialog.showAndWait();

        return dialogResult.isPresent() && dialog.getResult();
    }
}
