/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.frame.main;

import at.jku.mevss.anttracks.heap.MemoryMappedFastHeap;
import at.jku.mevss.anttracks.parser.TraceParser;
import at.jku.mevss.anttracks.util.ApplicationStatistics;
import at.jku.mevss.anttracks.util.GCReporter;
import at.jku.mevss.trace.gui.dialog.preference.Settings;
import at.jku.mevss.trace.gui.model.ClientInfo;
import at.jku.mevss.trace.gui.utils.Consts;
import at.jku.mevss.trace.gui.utils.ImageUtil;
import at.jku.mevss.trace.gui.utils.OperationManager;
import javafx.application.Application;
import javafx.application.Platform;
import javafx.fxml.FXMLLoader;
import javafx.scene.Scene;
import javafx.scene.layout.Pane;
import javafx.stage.Stage;

import javax.swing.*;
import javax.swing.UIManager.LookAndFeelInfo;
import java.awt.*;
import java.io.File;
import java.io.IOException;
import java.lang.management.ManagementFactory;
import java.lang.management.RuntimeMXBean;
import java.nio.file.Files;
import java.text.SimpleDateFormat;
import java.util.Date;

public class GUI extends Application {

    public static ClientInfo CLIENT_INFO = null;

    @Override
    public void start(Stage primaryStage) {
        try {
            try {
                Files.list(MemoryMappedFastHeap.getTempDir()).forEach(f -> {
                    try {
                        Files.delete(f);
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                });
            } catch (IOException e) {
                // Could not open directory, do nothing
            }

            FXMLLoader loader = new FXMLLoader();
            loader.setLocation(GUI.class.getResource("MainFrame.fxml"));

            // Workaround because GTKLookAndFeel is not working in JavaFX
            // application (application hangs up on UIManager.setLookAndFeel())
            SwingUtilities.invokeAndWait(() -> {
                try {
                    for (LookAndFeelInfo info : UIManager.getInstalledLookAndFeels()) {
                        if (info.getClassName().equals("javax.swing.plaf.nimbus.NimbusLookAndFeel")) {
                            UIManager.setLookAndFeel(info.getClassName());
                        }
                    }
                } catch (ClassNotFoundException | InstantiationException | IllegalAccessException | UnsupportedLookAndFeelException e) {
                    e.printStackTrace();
                }
            });

            Pane root = (Pane) loader.load();
            MainFrame mainFrame = loader.getController();
            mainFrame.init();
            CLIENT_INFO = new ClientInfo(this, mainFrame, primaryStage);
            CLIENT_INFO.operationManager = new OperationManager(mainFrame.getStatusPane().getChildren());

            Settings.getInstance().setup(); // just make sure it is created ...

            RuntimeMXBean runtimeMXBean = ManagementFactory.getRuntimeMXBean();
            java.util.List<String> jvmArgs = runtimeMXBean.getInputArguments();
            System.out.println("JVM arguments:");
            for (String arg : jvmArgs) {
                System.out.println(arg);
            }

            // 1. Consistency check
            String consistencyCheckString = "at.jku.mevss.trace.consistencycheck";
            if (System.getProperty(consistencyCheckString) == null) {
                System.setProperty(consistencyCheckString, "false");
            }
            TraceParser.CONSISTENCY_CHECK = Boolean.parseBoolean(System.getProperty(consistencyCheckString));
            CLIENT_INFO.consistencyCheck = TraceParser.CONSISTENCY_CHECK;
            System.out.println("Consistency check: " + System.getProperty(consistencyCheckString));

            // 2. Logging
            String loggingConfigFileString = "java.util.logging.config.file";
            if (System.getProperty(loggingConfigFileString) == null) {
                System.getProperties().setProperty(loggingConfigFileString, "log.config");
            }
            String logFile = System.getProperty(loggingConfigFileString);
            CLIENT_INFO.logSettingsFile = logFile;
            System.out.println("Log config file: " + logFile);

            // 3. Statistics output path
            String statisticsPathString = "at.jku.mevss.trace.gui.printStatisticsPath";
            CLIENT_INFO.printStatisticsPath = System.getProperty(statisticsPathString);
            System.out.println("Statistics result file: " + (CLIENT_INFO.printStatisticsPath == null ? "null" : CLIENT_INFO.printStatisticsPath));

            Runtime.getRuntime().addShutdownHook(new Thread((Runnable) () -> {
                if (CLIENT_INFO.printStatisticsPath != null) {
                    ApplicationStatistics.getInstance()
                                         .export(new File(CLIENT_INFO.printStatisticsPath.replace("%d",
                                                                                                  new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss").format(new Date(System
                                                                                                                                                                     .currentTimeMillis())))
                                                                                         .replace(" ", "")));
                }
                ApplicationStatistics.getInstance().print();

                try {
                    Files.list(MemoryMappedFastHeap.getTempDir()).forEach(f -> {
                        try {
                            Files.delete(f);
                        } catch (IOException e) {
                            e.printStackTrace();
                        }
                    });
                } catch (IOException e) {
                    // Could not open directory, do nothing
                }
            }, "Shutdown-thread"));

            // 4. Assertions
            boolean assertionsEnabled = false;
            assert assertionsEnabled = true;
            CLIENT_INFO.assertionsEnabled = assertionsEnabled;
            System.out.println("Assertions enabled: " + assertionsEnabled);

            GCReporter.getInstance(); // just make sure it is running ...

            Scene scene = new Scene(root);
            primaryStage.setTitle("AntTracks Analyzer");
            primaryStage.setScene(scene);
            primaryStage.getIcons().add(ImageUtil.getIconNode(Consts.ANT_ICON_IMAGE).getImage());
            primaryStage.setMaximized(true);
            // TODO is not necessary if all threads are closed properly
            primaryStage.setOnCloseRequest(e -> {
                Platform.exit();
                System.exit(0);
            });

            primaryStage.show();

            SplashScreen screen = SplashScreen.getSplashScreen();
            if (screen != null && screen.isVisible()) {
                screen.close();
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public static void main(String[] args) {
        launch();
    }
}
