/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.frame.main.component.menubar;

import at.jku.mevss.trace.gui.dialog.about.AboutDialog;
import at.jku.mevss.trace.gui.dialog.preference.PreferencesDialog;
import at.jku.mevss.trace.gui.model.AppInfo;
import at.jku.mevss.trace.gui.model.AppInfo.AliveDeadPanelType;
import at.jku.mevss.trace.gui.utils.AppLoader;
import at.jku.mevss.trace.gui.utils.FXMLUtil;
import at.jku.mevss.trace.gui.utils.OsScript;
import at.jku.mevss.trace.gui.utils.WindowUtil;
import javafx.fxml.FXML;
import javafx.fxml.Initializable;
import javafx.scene.control.*;
import javafx.stage.FileChooser;

import java.io.File;
import java.net.URL;
import java.util.List;
import java.util.ResourceBundle;
import java.util.stream.Collectors;

import static at.jku.mevss.trace.gui.frame.main.GUI.CLIENT_INFO;

public class AntTracksMenuBar extends MenuBar implements Initializable {

    @FXML
    private MenuItem loadMenuItem;
    @FXML
    private MenuItem preprocessMenuItem;
    @FXML
    private MenuItem exitMenuItem;

    @FXML
    private Menu windowMenu;
    @FXML
    private CheckMenuItem extendedChartsMenuItem;
    @FXML
    private MenuItem preferencesMenuItem;

    @FXML
    private Menu applicationMenu;
    @FXML
    private MenuItem redefineMenuItem;
    @FXML
    private ToggleGroup aliveDeadToggleGroup;
    @FXML
    private RadioMenuItem aliveDeadRelativeMenuItem;
    @FXML
    private RadioMenuItem aliveDeadNObjectsMenuItem;
    @FXML
    private RadioMenuItem aliveDeadBytesMenuItem;
    @FXML
    private CheckMenuItem featureChartMenuItem;
    @FXML
    private MenuItem addDataStructureDefinitionFileMenuItem;

    @FXML
    private MenuItem aboutMenuItem;

    public AntTracksMenuBar() {
        FXMLUtil.load(this, AntTracksMenuBar.class);
    }

    @Override
    public void initialize(URL location, ResourceBundle resources) {
        initializeMenus();
    }

    public void setAppsEnabled(boolean enabled) {
        applicationMenu.setDisable(!enabled);
    }

    private void initializeMenus() {
        initializeFileMenu();
        initializeWindowMenu();
        initializeApplicationMenu();
        initializeHelpMenu();
    }

    private void initializeFileMenu() {
        loadMenuItem.setOnAction(ae -> load());
        preprocessMenuItem.setDisable(!OsScript.isSupported());
        preprocessMenuItem.setOnAction(ae -> preprocessDirectory());
        exitMenuItem.setOnAction(ae -> System.exit(0));
    }

    private void initializeWindowMenu() {
        windowMenu.setOnShowing(ae -> extendedChartsMenuItem.setSelected(CLIENT_INFO.isExtendedChartVisibility()));
        extendedChartsMenuItem.setOnAction(ae -> switchExtendedChartVisibility());
        preferencesMenuItem.setOnAction(ae -> showPreferences());
    }

    private void initializeApplicationMenu() {
        applicationMenu.setOnShowing(ae -> {
            AppInfo appInfo = CLIENT_INFO.getCurrentAppInfo();
            redefineMenuItem.setDisable(!appInfo.isParsingCompleted());
            featureChartMenuItem.setDisable(appInfo.getSelectedFeaturesFile() == null);
            featureChartMenuItem.setSelected(appInfo.isShowFeatures());
            switch (appInfo.getAliveDeadPanelType()) {
                case RELATIVE:
                    aliveDeadRelativeMenuItem.setSelected(true);
                    break;
                case N_OBJECTS:
                    aliveDeadNObjectsMenuItem.setSelected(true);
                    break;
                case BYTE:
                    aliveDeadBytesMenuItem.setSelected(true);
                    break;
            }
            addDataStructureDefinitionFileMenuItem.setDisable(!appInfo.isParsingCompleted());
        });
        redefineMenuItem.setDisable(!OsScript.isSupported());
        redefineMenuItem.setOnAction(ae -> reloadApp());
        aliveDeadRelativeMenuItem.setOnAction(ae -> CLIENT_INFO.getCurrentAppInfo().setAliveDeadPanelType(AliveDeadPanelType.RELATIVE));
        aliveDeadNObjectsMenuItem.setOnAction(ae -> CLIENT_INFO.getCurrentAppInfo().setAliveDeadPanelType(AliveDeadPanelType.N_OBJECTS));
        aliveDeadBytesMenuItem.setOnAction(ae -> CLIENT_INFO.getCurrentAppInfo().setAliveDeadPanelType(AliveDeadPanelType.BYTE));
        featureChartMenuItem.setDisable(!OsScript.isSupported());
        featureChartMenuItem.setOnAction(ae -> switchFeatureChartVisibility());
        addDataStructureDefinitionFileMenuItem.setOnAction(ae -> {
            FileChooser chooser = new FileChooser();
            chooser.setTitle("Select data structure definition files...");
            chooser.getExtensionFilters().add(new FileChooser.ExtensionFilter("Data structure definitions (*.ds)", "*.ds"));
            List<File> dsFiles = chooser.showOpenMultipleDialog(CLIENT_INFO.getStage());
            if (dsFiles != null) {
                CLIENT_INFO.getCurrentAppInfo().selectDataStructureDefinitionFiles(dsFiles.stream().map(File::toURI).collect(Collectors.toList()));
            }
        });
    }

    private void initializeHelpMenu() {
        aboutMenuItem.setOnAction(ae -> showAbout());
    }

    private void load() {
        AppLoader.load();
    }

    private void preprocessDirectory() {
        AppLoader.preprocessDirectory();
    }

    private void switchExtendedChartVisibility() {
        CLIENT_INFO.setExtendedChartVisibility(!CLIENT_INFO.isExtendedChartVisibility());
        CLIENT_INFO.getMainFrame().plotAllApps();
    }

    private void showPreferences() {
        PreferencesDialog preferencesDialog = new PreferencesDialog();
        WindowUtil.centerInMainFrame(preferencesDialog);
        preferencesDialog.showAndWait();
    }

    private void reloadApp() {
        AppLoader.reload();
    }

    private void switchFeatureChartVisibility() {
        AppInfo appInfo = CLIENT_INFO.getCurrentAppInfo();
        appInfo.setShowFeatures(!appInfo.isShowFeatures());
    }

    private void showAbout() {
        AboutDialog aboutDialog = new AboutDialog();
        WindowUtil.centerInMainFrame(aboutDialog);
        aboutDialog.showAndWait();
    }
}
