/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.frame.main.tab.heapdiffing;

import at.jku.mevss.anttracks.diff.HeapDiffTraceParser;
import at.jku.mevss.anttracks.diff.Interval.IntervalInfo;
import at.jku.mevss.anttracks.diff.Interval.IntervalListener;
import at.jku.mevss.anttracks.heap.DetailedHeap;
import at.jku.mevss.anttracks.heap.io.HeapWriter;
import at.jku.mevss.anttracks.heap.io.MetaDataReaderConfig;
import at.jku.mevss.anttracks.heap.statistics.MemoryConsumption;
import at.jku.mevss.anttracks.parser.io.BaseFile;
import at.jku.mevss.anttracks.parser.symbols.SymbolsFile;
import at.jku.mevss.anttracks.util.Consts;
import at.jku.mevss.anttracks.util.ThreadUtil;
import at.jku.mevss.trace.gui.frame.main.tab.heapdiffing.model.DiffingData;
import at.jku.mevss.trace.gui.model.DiffStatisticsInfo;
import at.jku.mevss.trace.gui.model.Size;
import at.jku.mevss.trace.gui.utils.AntTask;

import javax.naming.OperationNotSupportedException;
import java.io.*;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.stream.Collectors;

import static at.jku.mevss.anttracks.util.Consts.DIFF_HEAP_EXTENSION;
import static at.jku.mevss.trace.gui.frame.main.GUI.CLIENT_INFO;

public class HeapDiffingTask extends AntTask<IntervalInfo> {
    private class HeapDiffingTableUpdateTask extends AntTask<Void> {

        private IntervalInfo intervalInfo;

        public HeapDiffingTableUpdateTask(IntervalInfo intervalInfo) {
            this.intervalInfo = intervalInfo;
        }

        @Override
        protected Void backgroundWork() throws Exception {
            updateTitle(String.format("Diffing %s from %fs to %fs: Calculating grouping",
                                      diffStatisticsInfo.getDiffInfo().getAppInfo().getAppName(),
                                      (fromTime / 1000.0),
                                      (toTime / 1000.0),
                                      "Parsing trace"));
            if (diffStatisticsInfo.getSelectedClassifiers().length() > 0) {
                diffStatisticsInfo.getDiffInfo().setHeap(intervalInfo.getLastHeap());
                diffStatisticsInfo.setGrouping(intervalInfo.classify(diffStatisticsInfo.getSelectedClassifiers(),
                                                                     diffStatisticsInfo.getSelectedConditionalClassifier(),
                                                                     diffStatisticsInfo.getSelectedFilters()));
                LOGGER.finest("Classification finished:\n" + diffStatisticsInfo.getGrouping());
            }

            return null;
        }

        @Override
        protected void suceeded() {
            heapDiffingTab.updateTable();
            currentTableUpdateTask = null;
        }
    }

    private class HeapDiffingChartUpdateTask extends AntTask<Void> {
        @Override
        protected Void backgroundWork() throws Exception {
            updateTitle(String.format("Diffing %s from %fs to %fs: Update chart",
                                      diffStatisticsInfo.getDiffInfo().getAppInfo().getAppName(),
                                      (fromTime / 1000.0),
                                      (toTime / 1000.0),
                                      "Parsing trace"));
            return null;
        }

        @Override
        protected void suceeded() {
            synchronized (stats) {
                try {
                    heapDiffingTab.updateChart(stats);
                } catch (Exception e) {
                    e.printStackTrace();
                }
                currentChartUpdateTask = null;
            }
        }
    }

    private final DiffStatisticsInfo diffStatisticsInfo;

    private final HeapDiffingTab heapDiffingTab;

    private final List<DiffingData> stats = new ArrayList<>();

    private final long fromTime;
    private final long toTime;

    private HeapDiffingTableUpdateTask currentTableUpdateTask = null;
    private HeapDiffingChartUpdateTask currentChartUpdateTask = null;

    private String fullMetaDataPath;

    public HeapDiffingTask(DiffStatisticsInfo diffStatisticsInfo) {
        this.diffStatisticsInfo = diffStatisticsInfo;

        fromTime = diffStatisticsInfo.getDiffInfo().getTime(0);
        toTime = diffStatisticsInfo.getDiffInfo().getTime(1);

        heapDiffingTab = new HeapDiffingTab();
        heapDiffingTab.init(diffStatisticsInfo, fromTime, toTime);
        heapDiffingTab.setLoadingMode(true);
        CLIENT_INFO.getMainFrame().addTab(heapDiffingTab);

        try {
            fullMetaDataPath = (BaseFile.isPlainFile(SymbolsFile.SYMBOL_FILE_ID, diffStatisticsInfo.getDiffInfo().getAppInfo().getSymbolsFile()) ?
                                diffStatisticsInfo.getDiffInfo().getAppInfo().getSymbolsFile().getParent() :
                                diffStatisticsInfo.getDiffInfo().getAppInfo().getSymbolsFile()).toString() + File.separator + Consts.ANT_META_DIRECTORY;
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    @Override
    protected IntervalInfo backgroundWork() throws Exception {
        LOGGER.log(Level.INFO, "Start parsing trace");
        updateTitle(String.format("Diffing: Diff %s from %fs to %fs",
                                  diffStatisticsInfo.getDiffInfo().getAppInfo().getAppName(),
                                  (fromTime / 1000.0),
                                  (toTime / 1000.0),
                                  "Parsing trace"));
        IntervalInfo heapDiff;

        /*if (new File(fullMetaDataPath + File.separator + toTime + DIFF_HEAP_EXTENSION).exists()) {
            // Heap diff loaded from file, also load stats
            HeapReader heapReader = new HeapReader(fullMetaDataPath + File.separator + toTime + DIFF_HEAP_EXTENSION,
                                                   diffStatisticsInfo.getDiffInfo().getAppInfo().getSymbols());
            DetailedHeap heap = heapReader.read();

            synchronized (stats) {
                loadStatsFromMetaData();
            }

            heapDiff = loadIntervalInfoFromMetaData(heap);
            heapDiff.setLastHeap(heap);

            assert currentChartUpdateTask == null : "There cannot be a running chart update task";
            currentChartUpdateTask = new HeapDiffingChartUpdateTask();
            ThreadUtil.startTask(currentChartUpdateTask);

            assert currentTableUpdateTask == null : "There cannot be a running table update task";
            currentTableUpdateTask = new HeapDiffingTableUpdateTask(heapDiff);
            ThreadUtil.startTask(currentTableUpdateTask);
        } else {*/
        HeapDiffTraceParser parser = new HeapDiffTraceParser(diffStatisticsInfo.getDiffInfo().getAppInfo().getSymbols(),
                                                             new MetaDataReaderConfig(diffStatisticsInfo.getDiffInfo()
                                                                                                        .getAppInfo()
                                                                                                        .getSymbols().root + File.separator + Consts.ANT_META_DIRECTORY),
                                                             fromTime,
                                                             toTime);

        parser.addIntervalListener(new IntervalListener() {

            @Override
            public void gcStartHeapStateAdded(long time, MemoryConsumption perm, MemoryConsumption died, MemoryConsumption born, MemoryConsumption temp) {
                updateProgress(time - fromTime, toTime - fromTime);

                synchronized (stats) {
                    stats.add(new DiffingData("", Size.of(perm), Size.of(died), Size.of(born), Size.of(temp), time));
                }
                if (currentChartUpdateTask == null) {
                    currentChartUpdateTask = new HeapDiffingChartUpdateTask();
                    ThreadUtil.startTask(currentChartUpdateTask);
                }
            }

            @Override
            public void gcEndHeapStateAdded(long time,
                                            MemoryConsumption perm,
                                            MemoryConsumption died,
                                            MemoryConsumption born,
                                            MemoryConsumption temp,
                                            IntervalInfo intervalInfo) {
                updateProgress(time - fromTime, toTime - fromTime);

                synchronized (stats) {
                    stats.add(new DiffingData("", Size.of(perm), Size.of(died), Size.of(born), Size.of(temp), time));
                }
                if (currentChartUpdateTask == null) {
                    currentChartUpdateTask = new HeapDiffingChartUpdateTask();
                    ThreadUtil.startTask(currentChartUpdateTask);
                }

                if (currentTableUpdateTask == null) {
                    currentTableUpdateTask = new HeapDiffingTableUpdateTask(intervalInfo);
                    ThreadUtil.startTask(currentTableUpdateTask);
                }
            }

            @Override
            public void intervalEnded(long time,
                                      DetailedHeap heap,
                                      MemoryConsumption perm,
                                      MemoryConsumption died,
                                      MemoryConsumption born,
                                      MemoryConsumption temp,
                                      IntervalInfo intervalInfo) {
                updateProgress(time - fromTime, toTime - fromTime);

                // activly wait for latest update
                while (currentTableUpdateTask != null) {
                    try {
                        Thread.sleep(10);
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }
                while (currentChartUpdateTask != null) {
                    try {
                        Thread.sleep(10);
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }
                currentTableUpdateTask = new HeapDiffingTableUpdateTask(intervalInfo);
                ThreadUtil.startTask(currentTableUpdateTask);
                currentChartUpdateTask = new HeapDiffingChartUpdateTask();
                ThreadUtil.startTask(currentChartUpdateTask);

                try {
                    ExecutorService executorService = Executors.newSingleThreadExecutor();
                    HeapWriter heapWriter = new HeapWriter(fullMetaDataPath + File.separator + toTime + DIFF_HEAP_EXTENSION);
                    heapWriter.write(heap, executorService);

                    executorService.shutdown();
                    executorService.awaitTermination(Integer.MAX_VALUE, TimeUnit.MILLISECONDS);

                    writeStatsToMetaData();
                    writeIntervalInfoToMetaData(intervalInfo);
                } catch (IOException e) {
                    e.printStackTrace();
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        });
        heapDiff = parser.parse(cancelProperty).getIntervals().get(0).info;
        //}

        LOGGER.log(Level.INFO, "Finished parsing trace, heap diff constructed");
        return heapDiff;
    }

    private void writeIntervalInfoToMetaData(IntervalInfo intervalInfo) {
        try {
            intervalInfo.writeToMetaData(new File(fullMetaDataPath + File.separator + fromTime + "_" + toTime + Consts.DIFF_INTERVAL_EXTENSION));
        } catch (OperationNotSupportedException e) {
            e.printStackTrace();
        }
    }

    private IntervalInfo loadIntervalInfoFromMetaData(DetailedHeap heap) {
        File sourceFile = new File(fullMetaDataPath + File.separator + fromTime + "_" + toTime + at.jku.mevss.anttracks.util.Consts.DIFF_INTERVAL_EXTENSION);
        return IntervalInfo.readFromMetaData(heap, sourceFile);
    }

    private void writeStatsToMetaData() {
        File targetFile = new File(fullMetaDataPath + File.separator + fromTime + "_" + toTime + at.jku.mevss.anttracks.util.Consts.DIFF_STAT_EXTENSION);
        try {
            DataOutputStream dos = new DataOutputStream(new FileOutputStream(targetFile));
            stats.stream().sequential().sorted(Comparator.comparingLong(stat -> stat.time)).forEachOrdered(stat -> stat.writeStatToMetaData(dos));
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }
    }

    private void loadStatsFromMetaData() {
        File sourceFile = new File(fullMetaDataPath + File.separator + fromTime + "_" + toTime + at.jku.mevss.anttracks.util.Consts.DIFF_STAT_EXTENSION);
        try {
            DataInputStream is = new DataInputStream(new FileInputStream(sourceFile));
            stats.addAll(diffStatisticsInfo.getDiffInfo()
                                           .getAppInfo()
                                           .getStatistics()
                                           .stream()
                                           .sequential()
                                           .map(x -> x.info.time)
                                           .filter(x -> x >= fromTime && x <= toTime)
                                           .distinct()
                                           .sorted(Long::compareTo)
                                           .map(time -> DiffingData.loadFromMetaData(is, time))
                                           .collect(Collectors.toList()));
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }
    }

    @Override
    protected void suceeded() {
        diffStatisticsInfo.getDiffInfo().setHeapDiff(getValue());
        heapDiffingTab.setLoadingMode(false);
    }
}
