/*
Copyright (c) 2014, 2015, 2016, 2017, dynatrace and/or its affiliates. All rights reserved.
This file is part of the AntTracks Tool.

This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 only, as
published by the Free Software Foundation.

This code is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
version 2 for more details (a copy is included in the LICENSE file that
accompanied this code).

You should have received a copy of the GNU General Public License
along with with this work.  If not, see <http://www.gnu.org/licenses/>.
*/
package at.jku.mevss.trace.gui.frame.main.tab.heapdiffing.model;

import at.jku.mevss.trace.gui.model.Size;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.math.BigDecimal;

public class DiffingData {
    public enum DiffingDataInfo {
        BeforeObjects,
        PermObjects,
        DiedObjects,
        BornObjects,
        TempObjects,
        AfterObjects,
        BeforeBytes,
        PermBytes,
        DiedBytes,
        BornBytes,
        TempBytes,
        AfterBytes,
        PERM,
        DIED,
        BORN,
        TEMP
    }

    public String id;
    public Size before;
    public Size perm;
    public Size died;
    public Size born;
    public Size temp;
    public Size after;
    public long time;

    public DiffingData(long time) {
        this("", time);
    }

    public DiffingData(String id, long time) {
        this(id, 0, 0, 0, 0, 0, 0, 0, 0, time);
    }

    public DiffingData(String id, Size perm, Size died, Size born, Size temp, long time) {
        this(id, perm.objects, died.objects, born.objects, temp.objects, perm.bytes, died.bytes, born.bytes, temp.bytes, time);
    }

    public DiffingData(String id,
                       double permObjects,
                       double diedObjects,
                       double bornObjects,
                       double tempObjects,
                       double permBytes,
                       double diedBytes,
                       double bornBytes,
                       double tempBytes,
                       long time) {
        this.id = id;
        before = new Size();
        perm = new Size();
        died = new Size();
        born = new Size();
        temp = new Size();
        after = new Size();
        perm.bytes = permBytes;
        died.bytes = diedBytes;
        born.bytes = bornBytes;
        temp.bytes = tempBytes;
        before.bytes = permBytes + diedBytes;
        after.bytes = permBytes + bornBytes;
        perm.objects = permObjects;
        died.objects = diedObjects;
        born.objects = bornObjects;
        temp.objects = tempObjects;
        before.objects = permObjects + diedObjects;
        after.objects = permObjects + bornObjects;
        this.time = time;
    }

    public DiffingData(DiffingData data) {
        id = data.id;
        before = data.before;
        perm = data.perm;
        died = data.died;
        born = data.born;
        temp = data.temp;
        after = data.after;
        time = data.time;
    }

    @Override
    public DiffingData clone() {
        return new DiffingData(id, perm.objects, died.objects, born.objects, temp.objects, perm.bytes, died.bytes, born.bytes, temp.bytes, time);
    }

    public void setId(String id) {
        this.id = id;
    }

    public void add(DiffingDataInfo info, long n) {
        switch (info) {
            case BeforeObjects:
                throw new IllegalArgumentException("BeforeObjects should not be set, is implicitly increased with Perm and Died");
            case PermObjects:
                before.objects += n;
                after.objects += n;
                perm.objects += n;
                break;
            case DiedObjects:
                before.objects += n;
                died.objects += n;
                break;
            case BornObjects:
                after.objects += n;
                born.objects += n;
                break;
            case TempObjects:
                temp.objects += n;
                break;
            case AfterObjects:
                throw new IllegalArgumentException("AfterObjects should not be set, is implicitly increased with Perm and Born");
            case BeforeBytes:
                throw new IllegalArgumentException("BeforeBytes should not be set, is implicitly increased with Perm and Died");
            case PermBytes:
                before.bytes += n;
                after.bytes += n;
                perm.bytes += n;
                break;
            case DiedBytes:
                before.bytes += n;
                died.bytes += n;
                break;
            case BornBytes:
                after.bytes += n;
                born.bytes += n;
                break;
            case TempBytes:
                temp.bytes += n;
                break;
            case AfterBytes:
                throw new IllegalArgumentException("AfterBytes should not be set, is implicitly increased with Perm and Born");
            default:
                throw new IllegalArgumentException("The given parameter is not valid in this context");
        }
    }

    public void add(DiffingDataInfo info, Size n) {
        switch (info) {
            case BeforeObjects:
                throw new IllegalArgumentException("BeforeObjects should not be set, is implicitly increased with Perm and Died");
            case PermObjects:
                before.objects += n.objects;
                after.objects += n.objects;
                perm.objects += n.objects;
                break;
            case DiedObjects:
                before.objects += n.objects;
                died.objects += n.objects;
                break;
            case BornObjects:
                after.objects += n.objects;
                born.objects += n.objects;
                break;
            case TempObjects:
                temp.objects += n.objects;
                break;
            case AfterObjects:
                throw new IllegalArgumentException("AfterObjects should not be set, is implicitly increased with Perm and Born");
            case BeforeBytes:
                throw new IllegalArgumentException("BeforeBytes should not be set, is implicitly increased with Perm and Died");
            case PermBytes:
                before.bytes += n.bytes;
                after.bytes += n.bytes;
                perm.bytes += n.bytes;
                break;
            case DiedBytes:
                before.bytes += n.bytes;
                died.bytes += n.bytes;
                break;
            case BornBytes:
                after.bytes += n.bytes;
                born.bytes += n.bytes;
                break;
            case TempBytes:
                temp.bytes += n.bytes;
                break;
            case AfterBytes:
                throw new IllegalArgumentException("AfterBytes should not be set, is implicitly increased with Perm and Born");
            case PERM:
                before.objects += n.objects;
                after.objects += n.objects;
                perm.objects += n.objects;
                before.bytes += n.bytes;
                after.bytes += n.bytes;
                perm.bytes += n.bytes;
                break;
            case DIED:
                before.objects += n.objects;
                died.objects += n.objects;
                before.bytes += n.bytes;
                died.bytes += n.bytes;
                break;
            case BORN:
                after.objects += n.objects;
                born.objects += n.objects;
                after.bytes += n.bytes;
                born.bytes += n.bytes;
                break;
            case TEMP:
                temp.objects += n.objects;
                temp.bytes += n.bytes;
                break;
            default:
                throw new IllegalArgumentException("The given parameter is not valid in this context");
        }
    }

    public Double getPercentage(DiffingDataInfo info, long total) {
        double reference = 0;
        switch (info) {
            case BeforeObjects:
                reference = before.objects;
                break;
            case PermObjects:
                reference = perm.objects;
                break;
            case DiedObjects:
                reference = died.objects;
                break;
            case BornObjects:
                reference = born.objects;
                break;
            case TempObjects:
                reference = temp.objects;
                break;
            case AfterObjects:
                reference = after.objects;
                break;
            case BeforeBytes:
                reference = before.bytes;
                break;
            case PermBytes:
                reference = perm.bytes;
                break;
            case DiedBytes:
                reference = died.bytes;
                break;
            case BornBytes:
                reference = born.bytes;
                break;
            case TempBytes:
                reference = temp.bytes;
                break;
            case AfterBytes:
                reference = after.bytes;
                break;
            default:
                throw new IllegalArgumentException("The given parameter is not valid in this context");
        }

        return new BigDecimal(reference / total * 100).setScale(1, BigDecimal.ROUND_HALF_UP).doubleValue();
    }

    @Override
    public String toString() {
        return id.toString();
    }

    public void add(DiffingData toAdd) {
        add(DiffingDataInfo.PERM, toAdd.perm);
        add(DiffingDataInfo.DIED, toAdd.died);
        add(DiffingDataInfo.BORN, toAdd.born);
        add(DiffingDataInfo.TEMP, toAdd.temp);
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((id == null) ? 0 : id.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        DiffingData other = (DiffingData) obj;
        if (id == null) {
            if (other.id != null) {
                return false;
            }
        } else if (!id.equals(other.id)) {
            return false;
        }
        return true;
    }

    public void writeStatToMetaData(DataOutputStream dos) {
        try {
            dos.writeUTF(id);
            dos.writeDouble(perm.objects);
            dos.writeDouble(perm.bytes);

            dos.writeDouble(died.objects);
            dos.writeDouble(died.bytes);

            dos.writeDouble(born.objects);
            dos.writeDouble(born.bytes);

            dos.writeDouble(temp.objects);
            dos.writeDouble(temp.bytes);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public static DiffingData loadFromMetaData(DataInputStream dis, long time) {
        try {
            return new DiffingData(dis.readUTF(),
                                   new Size(dis.readDouble(), dis.readDouble()),
                                   new Size(dis.readDouble(), dis.readDouble()),
                                   new Size(dis.readDouble(), dis.readDouble()),
                                   new Size(dis.readDouble(), dis.readDouble()),
                                   time);
        } catch (IOException e) {
            e.printStackTrace();
            return null;
        }
    }
}
