package at.jku.mevss.trace.gui.frame.main.tab.heapstate;

import at.jku.mevss.anttracks.classification.ClassifierChain;
import at.jku.mevss.anttracks.classification.ConditionalClassifierList;
import at.jku.mevss.anttracks.classification.Filter;
import at.jku.mevss.anttracks.heap.IndexBasedHeap;
import at.jku.mevss.anttracks.util.Counter;
import at.jku.mevss.trace.gui.model.AppInfo;
import at.jku.mevss.trace.gui.model.FastHeapInfo;
import at.jku.mevss.trace.gui.task.FastHeapTask;
import javafx.application.Platform;
import javafx.event.Event;
import javafx.event.EventHandler;

import java.util.List;
import java.util.logging.Level;

import static at.jku.mevss.anttracks.heap.FastHeap.NULL_INDEX;
import static at.jku.mevss.trace.gui.frame.main.GUI.CLIENT_INFO;

public class HeapStateTask extends FastHeapTask {
    private final HeapStateTab heapStateTab;
    private final FastHeapInfo detailsInfo;
    private boolean customClassifiers = false;

    public HeapStateTask(AppInfo appInfo, long time) {
        this(appInfo, time, null, null, null);
    }

    public HeapStateTask(AppInfo appInfo, long time, ClassifierChain selectedClassifier,
                         ConditionalClassifierList conditionalClassifier,
                         List<Filter> selectedFilters) {
        super(appInfo, time);
        detailsInfo = new FastHeapInfo(appInfo, time);

        heapStateTab = new HeapStateTab();
        heapStateTab.init(detailsInfo, this,
                          selectedClassifier,
                          conditionalClassifier,
                          selectedFilters);
        if (selectedClassifier != null) {
            customClassifiers = true;
        }
        heapStateTab.getStatistics().setLoadingMode(true);

        CLIENT_INFO.getMainFrame().addTab(heapStateTab);
    }

    @Override
    protected void handleTaskCancelled() {
        // close tab
        Platform.runLater(() -> {
            EventHandler<Event> handler = heapStateTab.getOnClosed();
            if (handler != null) {
                handler.handle(null);
            }

            heapStateTab.getTabPane().getTabs().remove(heapStateTab);
        });
    }

    private void storeHeap() {
        IndexBasedHeap heap = getValue();
        detailsInfo.setHeap(heap);
    }

    private void updateHeapStateTab() {
        heapStateTab.getStatistics().setLoadingMode(false);
        if (!customClassifiers) {
            heapStateTab.setDefaultClassification();
        }
        heapStateTab.getStatistics().applyClassifiers();
    }

    private void calculateHeapMetrics(IndexBasedHeap heap) {
        //this method takes tens of minutes for large heaps and largely contains only redundant / dev-internal metrics
        //I've disabled it for now, maybe guard it with a system property that must be set explicitly on the cmdline / eclipse launch
        // config? (table should
        // also be inivisible if
        // that property is not set)
        //this way released version do not suffer the time penalty for unwanted metrics
        //also, we have traces that crash this method (for some reason), again, releases should not crash for dev metrics

        LOGGER.log(Level.INFO, "Calculate heap metrics");
        updateTitle("Heap State: Calculate heap metric statistics for time " + (detailsInfo.getTime() / 1000.0) + "s");
        // updateMessage("Calculate heap metrics");
        int metricsCount = 5;
        int cur = -1;
        updateProgress(++cur, metricsCount);

        heapStateTab.getStatistics().getHeapMetricsTable().nObjects.valueProperty.set(heap.getObjectCount());
        updateProgress(++cur, metricsCount);

        if (heap.getSymbols().expectPointers) {
            Counter nToPointers = new Counter();
            heap.stream().forEach(objIndex -> {
                int[] toPointers = heap.getToPointers(objIndex);
                nToPointers.add(toPointers != null ? toPointers.length : 0);
            });
            heapStateTab.getStatistics().getHeapMetricsTable().nToPointersMetric.valueProperty.set(nToPointers.get());
            updateProgress(++cur, metricsCount);

            Counter nToPointersWithoutNull = new Counter();
            heap.stream().forEach(objIndex -> {
                int[] toPointers = heap.getToPointers(objIndex);
                if (toPointers != null) {
                    for (int i = 0; i < toPointers.length; i++) {
                        if (toPointers[i] != NULL_INDEX) {
                            nToPointersWithoutNull.inc();
                        }
                    }
                }
            });
            heapStateTab.getStatistics().getHeapMetricsTable().nToPointersWithoutNullMetric.valueProperty.set(nToPointersWithoutNull.get());
            updateProgress(++cur, metricsCount);

            Counter nFromPointers = new Counter();
            heap.stream().forEach(objIndex -> {
                int[] fromPointers = heap.getFromPointers(objIndex);
                nFromPointers.add(fromPointers != null ? fromPointers.length : 0);
            });
            heapStateTab.getStatistics().getHeapMetricsTable().nFromPointersMetric.valueProperty.set(nFromPointers.get());
            updateProgress(++cur, metricsCount);

            heapStateTab.getStatistics().getHeapMetricsTable().nRootPointers.valueProperty.set(heap.getRootPointerBlocking()
                                                                                                   .values()
                                                                                                   .stream()
                                                                                                   .flatMap(List::stream)
                                                                                                   .count());
            updateProgress(++cur, metricsCount);
        }

        LOGGER.log(Level.INFO, "Finished heap metrics");
    }

    @Override
    protected void suceeded() {
        storeHeap();
        updateHeapStateTab();
    }

    @Override
    protected void additionalBackgroundWork(IndexBasedHeap heap) {
        calculateHeapMetrics(heap);
    }
}
